@@ -588,16 +588,21 @@ Flm_ker_gauss(GEN x, ulong p, long deplin)
 }
 
 GEN
-Flm_intersect(GEN x, GEN y, ulong p)
+Flm_intersect_i(GEN x, GEN y, ulong p)
 {
-  pari_sp av = avma;
   long j, lx = lg(x);
   GEN z;
 
   if (lx==1 || lg(y)==1) return cgetg(1,t_MAT);
   z = Flm_ker(shallowconcat(x,y), p);
   for (j=lg(z)-1; j; j--) setlg(gel(z,j),lx);
-  return gerepileupto(av, Flm_mul(x,z,p));
+  return Flm_mul(x,z,p);
+}
+GEN
+Flm_intersect(GEN x, GEN y, ulong p)
+{
+  pari_sp av = avma;
+  return gerepileupto(av, Flm_image(Flm_intersect_i(x, y, p), p));
 }
 
 static GEN
@@ -649,19 +654,11 @@ Flm_ker_i(GEN x, ulong p, long deplin, long inplace) {
 }
 
 GEN
-Flm_ker_sp(GEN x, ulong p, long deplin) {
-  return Flm_ker_i(x, p, deplin, 1);
-}
-
+Flm_ker_sp(GEN x, ulong p, long deplin) { return Flm_ker_i(x, p, deplin, 1); }
 GEN
-Flm_ker(GEN x, ulong p) {
-  return Flm_ker_i(x, p, 0, 0);
-}
-
+Flm_ker(GEN x, ulong p) { return Flm_ker_i(x, p, 0, 0); }
 GEN
-Flm_deplin(GEN x, ulong p) {
-  return Flm_ker_i(x, p, 1, 0);
-}
+Flm_deplin(GEN x, ulong p) { return Flm_ker_i(x, p, 1, 0); }
 
 /* in place, destroy a, SMALL_ULONG(p) is TRUE */
 static ulong
