@@ -264,6 +264,19 @@ FpXQX_red(GEN z, GEN T, GEN p)
   return FpXQX_renormalize(res,l);
 }
 
+GEN
+FpXQXV_red(GEN x, GEN T, GEN p)
+{ pari_APPLY_type(t_VEC, FpXQX_red(gel(x,i), T, p)) }
+
+GEN
+FpXQXT_red(GEN x, GEN T, GEN p)
+{
+  if (typ(x) == t_POL)
+    return FpXQX_red(x, T, p);
+  else
+    pari_APPLY_type(t_VEC, FpXQXT_red(gel(x,i), T, p))
+}
+
 static GEN
 to_intmod(GEN x, GEN p) { retmkintmod(modii(x, p), p); }
 
@@ -397,8 +410,8 @@ FpXQX_mulspec(GEN x, GEN y, GEN T, GEN p, long lx, long ly)
   } else if (ZXX_is_ZX_spec(x,lx))
       return FpXY_FpY_mulspec(y,x,T,p,ly,lx);
   n = get_FpX_degree(T);
-  kx = ZXX_to_Kronecker_spec(x, lx, n);
-  ky = ZXX_to_Kronecker_spec(y, ly, n);
+  kx = RgXX_to_Kronecker_spec(x, lx, n);
+  ky = RgXX_to_Kronecker_spec(y, ly, n);
   z = Kronecker_to_FpXQX(ZX_mul(ky,kx), T, p);
   return gerepileupto(av, z);
 }
@@ -416,7 +429,7 @@ FpXQX_sqr(GEN x, GEN T, GEN p)
   pari_sp av = avma;
   GEN z, kx;
   if (ZXX_is_ZX(x)) return ZX_sqr(x);
-  kx= ZXX_to_Kronecker(x, get_FpX_degree(T));
+  kx= RgXX_to_Kronecker(x, get_FpX_degree(T));
   z = Kronecker_to_FpXQX(ZX_sqr(kx), T, p);
   return gerepileupto(av, z);
 }
@@ -753,8 +766,8 @@ FpXQX_extgcd_halfgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
 GEN
 FpXQX_extgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
 {
+  pari_sp av = avma;
   GEN d;
-  pari_sp ltop=avma;
   if (lgefint(p) == 3)
   {
     GEN Pl, Ql, Tl, Dl;
@@ -773,8 +786,7 @@ FpXQX_extgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
     else
       d = FpXQX_extgcd_basecase(x, y, T, p, ptu, ptv);
   }
-  gerepileall(ltop,ptu?3:2,&d,ptv,ptu);
-  return d;
+  return gc_all(av, ptu?3:2, &d, ptv, ptu);
 }
 
 GEN
@@ -1091,8 +1103,7 @@ FpXQX_divrem(GEN x, GEN S, GEN T, GEN p, GEN *pr)
     GEN q = FpXQX_divrem_Barrett(x,mg,y,T,p,pr);
     if (!q) return gc_NULL(av);
     if (!pr || pr==ONLY_DIVIDES) return gerepilecopy(av, q);
-    gerepileall(av,2,&q,pr);
-    return q;
+    return gc_all(av, 2, &q, pr);
   }
 }
 
@@ -1222,24 +1233,18 @@ static struct bb_ring FpXQX_ring = { _FpXQX_add, _FpXQX_mul, _FpXQX_sqr };
 GEN
 FpXQX_digits(GEN x, GEN B, GEN T, GEN p)
 {
-  pari_sp av = avma;
   long d = degpol(B), n = (lgpol(x)+d-1)/d;
-  GEN z;
   struct _FpXQXQ D;
   D.T = T; D.p = p;
-  z = gen_digits(x, B, n, (void *)&D, &FpXQX_ring, _FpXQX_divrem);
-  return gerepileupto(av, z);
+  return gen_digits(x, B, n, (void *)&D, &FpXQX_ring, _FpXQX_divrem);
 }
 
 GEN
 FpXQXV_FpXQX_fromdigits(GEN x, GEN B, GEN T, GEN p)
 {
-  pari_sp av = avma;
   struct _FpXQXQ D;
-  GEN z;
   D.T = T; D.p = p;
-  z = gen_fromdigits(x,B,(void *)&D, &FpXQX_ring);
-  return gerepileupto(av, z);
+  return gen_fromdigits(x,B,(void *)&D, &FpXQX_ring);
 }
 
 /* Q an FpXY (t_POL with FpX coeffs), evaluate at X = x */
@@ -1754,8 +1759,8 @@ FpXQXn_mul(GEN x, GEN y, long n, GEN T, GEN p)
   if (ZXX_is_ZX(y) && ZXX_is_ZX(x))
     return FpXn_mul(x,y,n,p);
   d = get_FpX_degree(T);
-  kx = ZXX_to_Kronecker(x, d);
-  ky = ZXX_to_Kronecker(y, d);
+  kx = RgXX_to_Kronecker(x, d);
+  ky = RgXX_to_Kronecker(y, d);
   z = Kronecker_to_FpXQX(ZXn_mul(ky,kx,(2*d-1)*n), T, p);
   return gerepileupto(av, z);
 }
@@ -1768,7 +1773,7 @@ FpXQXn_sqr(GEN x, long n, GEN T, GEN p)
   long d;
   if (ZXX_is_ZX(x)) return ZXn_sqr(x, n);
   d = get_FpX_degree(T);
-  kx= ZXX_to_Kronecker(x, d);
+  kx= RgXX_to_Kronecker(x, d);
   z = Kronecker_to_FpXQX(ZXn_sqr(kx, (2*d-1)*n), T, p);
   return gerepileupto(av, z);
 }
@@ -1891,7 +1896,7 @@ FpXQXn_exp(GEN h, long e, GEN T, GEN p)
 }
 
 GEN
-FpXQXn_inv(GEN f, long e, GEN T, GEN p)
+FpXQXn_div(GEN g, GEN f, long e, GEN T, GEN p)
 {
   pari_sp av = avma, av2;
   ulong mask;
@@ -1900,8 +1905,8 @@ FpXQXn_inv(GEN f, long e, GEN T, GEN p)
 
   if (!signe(f)) pari_err_INV("FpXXn_inv",f);
   a = Fq_inv(gel(f,2), T, p);
-  if (e == 1) return scalarpol(a, v);
-  else if (e == 2)
+  if (e == 1 && !g) return scalarpol(a, v);
+  else if (e == 2 && !g)
   {
     GEN b;
     if (degpol(f) <= 0) return scalarpol(a, v);
@@ -1921,8 +1926,17 @@ FpXQXn_inv(GEN f, long e, GEN T, GEN p)
     n<<=1; if (mask & 1) n--;
     mask >>= 1;
     fr = FpXXn_red(f, n);
-    u = FpXQXn_mul(W, FpXQXn_mulhigh(fr, W, n2, n, T, p), n-n2, T, p);
-    W = FpXX_sub(W, FpXX_shift(u, n2), p);
+    if (mask>1 || !g)
+    {
+      u = FpXQXn_mul(W, FpXQXn_mulhigh(fr, W, n2, n, T, p), n-n2, T, p);
+      W = FpXX_sub(W, FpXX_shift(u, n2), p);
+    }
+    else
+    {
+      GEN y = FpXQXn_mul(g, W, n, T, p), yt =  FpXXn_red(y, n-n2);
+      u = FpXQXn_mul(yt, FpXQXn_mulhigh(fr,  W, n2, n, T, p), n-n2, T, p);
+      W = FpXX_sub(y, FpXX_shift(u, n2), p);
+    }
     if (gc_needed(av2,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"FpXQXn_inv, e = %ld", n);
@@ -1931,3 +1945,7 @@ FpXQXn_inv(GEN f, long e, GEN T, GEN p)
   }
   return gerepileupto(av, W);
 }
+
+GEN
+FpXQXn_inv(GEN f, long e, GEN T, GEN p)
+{ return FpXQXn_div(NULL, f, e, T, p); }
