@@ -275,8 +275,7 @@ RgM_Rg_add(GEN x, GEN y)
   GEN z = cgetg(l,t_MAT);
 
   if (l==1) return z;
-  if (l != lgcols(x)) pari_err_OP( "+", x, y);
-  z = cgetg(l,t_MAT);
+  if (l != lgcols(x)) pari_err_OP("+", x, y);
   for (i=1; i<l; i++)
   {
     GEN zi = cgetg(l,t_COL), xi = gel(x,i);
@@ -293,8 +292,7 @@ RgM_Rg_sub(GEN x, GEN y)
   GEN z = cgetg(l,t_MAT);
 
   if (l==1) return z;
-  if (l != lgcols(x)) pari_err_OP( "-", x, y);
-  z = cgetg(l,t_MAT);
+  if (l != lgcols(x)) pari_err_OP("-", x, y);
   for (i=1; i<l; i++)
   {
     GEN zi = cgetg(l,t_COL), xi = gel(x,i);
@@ -587,20 +585,20 @@ RgV_RgM_mul(GEN x, GEN y)
 }
 
 static GEN
-RgM_mul_FpM(GEN x, GEN y, GEN p)
+RgM_mul_FpM_i(GEN x, GEN y, GEN p)
 {
-  pari_sp av = avma;
-  GEN r;
   if (lgefint(p) == 3)
   {
     ulong pp = uel(p, 2);
-    r = Flm_to_ZM_inplace(Flm_mul(RgM_to_Flm(x, pp),
-                                  RgM_to_Flm(y, pp), pp));
+    if (pp==2) return F2m_to_mod(F2m_mul(RgM_to_F2m(x), RgM_to_F2m(y)));
+    if (pp==3) return F3m_to_mod(F3m_mul(RgM_to_F3m(x), RgM_to_F3m(y)));
+    return Flm_to_mod(Flm_mul(RgM_to_Flm(x, pp), RgM_to_Flm(y, pp), pp), pp);
   }
-  else
-    r = FpM_mul(RgM_to_FpM(x, p), RgM_to_FpM(y, p), p);
-  return gerepileupto(av, FpM_to_mod(r, p));
+  return FpM_to_mod(FpM_mul(RgM_to_FpM(x,p), RgM_to_FpM(y,p), p), p);
 }
+static GEN
+RgM_mul_FpM(GEN x, GEN y, GEN p)
+{ pari_sp av = avma; return gerepileupto(av, RgM_mul_FpM_i(x, y, p)); }
 
 static GEN
 RgM_mul_FqM(GEN x, GEN y, GEN pol, GEN p)
@@ -699,10 +697,19 @@ RgM_sqr_fast(GEN x)
 
 #undef code
 
+/* lx, ly > 1 */
+static GEN
+RgM_mul_i(GEN x, GEN y, long lx, long ly)
+{
+  GEN z = cgetg(ly, t_MAT);
+  long j, l = lgcols(x);
+  for (j = 1; j < ly; j++) gel(z,j) = RgM_RgC_mul_i(x, gel(y,j), lx, l);
+  return z;
+}
 GEN
 RgM_mul(GEN x, GEN y)
 {
-  long j, l, lx, ly = lg(y);
+  long lx, ly = lg(y);
   GEN z;
   if (ly == 1) return cgetg(1,t_MAT);
   lx = lg(x);
@@ -710,10 +717,7 @@ RgM_mul(GEN x, GEN y)
   if (lx == 1) return zeromat(0,ly-1);
   z = RgM_mul_fast(x, y);
   if (z) return z;
-  z = cgetg(ly, t_MAT);
-  l = lgcols(x);
-  for (j=1; j<ly; j++) gel(z,j) = RgM_RgC_mul_i(x, gel(y,j), lx, l);
-  return z;
+  return RgM_mul_i(x, y, lx, ly);
 }
 
 GEN
