@@ -15,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_pol
+
 /*******************************************************************/
 /*                                                                 */
 /*                         GENERIC                                 */
@@ -158,12 +160,11 @@ GEN
 RgX_homogenous_evalpow(GEN P, GEN A, GEN B)
 {
   pari_sp av = avma;
-  long d, i, v;
+  long i, d = degpol(P);
   GEN s;
-  if (typ(P)!=t_POL)
-    return mkvec2(P, gen_1);
-  d = degpol(P); v = varn(A);
-  s = scalarpol_shallow(gel(P, d+2), v);
+  if (signe(P)==0) return pol_0(varn(P));
+  s = gel(P, d+2);
+  if (d == 0) return gcopy(s);
   for (i = d-1; i >= 0; i--)
   {
     s = gadd(gmul(s, A), gmul(gel(B,d+1-i), gel(P,i+2)));
@@ -173,8 +174,7 @@ RgX_homogenous_evalpow(GEN P, GEN A, GEN B)
       s = gerepileupto(av, s);
     }
   }
-  s = gerepileupto(av, s);
-  return mkvec2(s, gel(B,d+1));
+  return gerepileupto(av, s);
 }
 
 GEN
@@ -183,7 +183,8 @@ QXQX_homogenous_evalpow(GEN P, GEN A, GEN B, GEN T)
   pari_sp av = avma;
   long i, d = degpol(P), v = varn(A);
   GEN s;
-  if (signe(P)==0) return mkvec2(pol_0(v), pol_1(v));
+  if (signe(P)==0) return pol_0(v);
+  if (d == 0) return scalarpol(gel(P, d+2), v);
   s = scalarpol_shallow(gel(P, d+2), v);
   for (i = d-1; i >= 0; i--)
   {
@@ -195,8 +196,7 @@ QXQX_homogenous_evalpow(GEN P, GEN A, GEN B, GEN T)
       s = gerepileupto(av, s);
     }
   }
-  s = gerepileupto(av, s);
-  return mkvec2(s, gel(B,d+1));
+  return gerepileupto(av, s);
 }
 
 const struct bb_algebra *
@@ -217,10 +217,8 @@ _RgX_divrem(void *E, GEN x, GEN y, GEN *r)
 GEN
 RgX_digits(GEN x, GEN T)
 {
-  pari_sp av = avma;
   long d = degpol(T), n = (lgpol(x)+d-1)/d;
-  GEN z = gen_digits(x,T,n,NULL, &Rg_ring, _RgX_divrem);
-  return gerepileupto(av, z);
+  return gen_digits(x,T,n,NULL, &Rg_ring, _RgX_divrem);
 }
 
 /*******************************************************************/
@@ -396,16 +394,27 @@ ZX_unscale_div(GEN P, GEN h)
   }
   return Q;
 }
+/* P(h*X) / h^k, assuming the result is a ZX */
+GEN
+ZX_unscale_divpow(GEN P, GEN h, long k)
+{
+  long i, j, l = lg(P);
+  GEN H, Q = cgetg(l, t_POL);
+  Q[1] = P[1]; if (l == 2) return Q;
+  H = gpowers(h, maxss(k, l - 3 - k));
+  for (i = 2, j = k+1; j > 1 && i < l; i++)
+    gel(Q, i) = diviiexact(gel(P, i), gel(H, j--));
+  if (i == l) return Q;
+  gel(Q, i) = gel(P, i); i++;
+  for (j = 2; i < l; i++) gel(Q, i) = mulii(gel(P, i), gel(H, j++));
+  return Q;
+}
 
 GEN
-RgXV_unscale(GEN v, GEN h)
+RgXV_unscale(GEN x, GEN h)
 {
-  long i, l;
-  GEN w;
-  if (!h || isint1(h)) return v;
-  w = cgetg_copy(v, &l);
-  for (i=1; i<l; i++) gel(w,i) = RgX_unscale(gel(v,i), h);
-  return w;
+  if (!h || isint1(h)) return x;
+  pari_APPLY_same(RgX_unscale(gel(x,i), h));
 }
 
 /* Return h^degpol(P) P(x / h), not memory clean */
@@ -439,6 +448,16 @@ RgX_deflate(GEN x0, long d)
   return y;
 }
 
+GEN
+RgX_homogenize(GEN P, long v)
+{
+  long i, l, d;
+  GEN Q = cgetg_copy(P, &l);
+  Q[1] = P[1]; d = l-3;
+  for (i = 2; i < l; i++) gel(Q,i) = monomial(gel(P,i), d--, v);
+  return Q;
+}
+
 /* F a t_RFRAC */
 long
 rfrac_deflate_order(GEN F)
@@ -548,6 +567,14 @@ RgX_translate(GEN P, GEN c)
   }
 }
 
+/* P(ax + b) */
+GEN
+RgX_affine(GEN P, GEN a, GEN b)
+{
+  if (signe(b)) P = RgX_translate(P, b);
+  return RgX_unscale(P, a);
+}
+
 /* return lift( P(X + c) ) using Horner, c in R[y]/(T) */
 GEN
 RgXQX_translate(GEN P, GEN c, GEN T)
@@ -736,10 +763,13 @@ Rg_to_RgC(GEN x, long N)
   return (typ(x) == t_POL)? RgX_to_RgC(x,N): scalarcol_shallow(x, N);
 }
 
-/* vector of polynomials (in v) whose coeffs are given by the columns of x */
+/* vector of polynomials (in v) whose coefs are given by the columns of x */
 GEN
 RgM_to_RgXV(GEN x, long v)
 { pari_APPLY_type(t_VEC, RgV_to_RgX(gel(x,i), v)) }
+GEN
+RgM_to_RgXV_reverse(GEN x, long v)
+{ pari_APPLY_type(t_VEC, RgV_to_RgX_reverse(gel(x,i), v)) }
 
 /* matrix whose entries are given by the coeffs of the polynomials in
  * vector v (considered as degree n-1 polynomials) */
@@ -821,6 +851,9 @@ RgXY_degreex(GEN b)
   return deg;
 }
 
+GEN
+RgXY_derivx(GEN x) { pari_APPLY_pol(RgX_deriv(gel(x,i))); }
+
 /* return (x % X^n). Shallow */
 GEN
 RgXn_red_shallow(GEN a, long n)
@@ -940,7 +973,11 @@ RgX_valrem(GEN x, GEN *Z)
   for (i = 2; i < lx; i++)
     if (!isexactzero(gel(x,i))) break;
   /* possible with nonrational zeros */
-  if (i == lx) { *Z = pol_0(varn(x)); return LONG_MAX; }
+  if (i == lx)
+  {
+    *Z = scalarpol_shallow(Rg_get_0(x), varn(x));
+    return LONG_MAX;
+  }
   v = i - 2;
   *Z = RgX_shift_shallow(x, -v);
   return v;
@@ -1002,10 +1039,8 @@ RgX_recipspec_shallow(GEN x, long l, long n)
   long i;
   GEN z = cgetg(n+2,t_POL);
   z[1] = 0; z += 2;
-  for(i=0; i<l; i++)
-    gel(z,n-i-1) = gel(x,i);
-  for(   ; i<n; i++)
-    gel(z, n-i-1) = gen_0;
+  for(i=0; i<l; i++) gel(z,n-i-1) = gel(x,i);
+  for(   ; i<n; i++) gel(z, n-i-1) = gen_0;
   return normalizepol_lg(z-2,n+2);
 }
 
@@ -1029,6 +1064,15 @@ RgX_recip(GEN x)
 /* shallow version */
 GEN
 RgX_recip_shallow(GEN x)
+{
+  long lx, i, j;
+  GEN y = cgetg_copy(x, &lx);
+  y[1] = x[1]; for (i=2,j=lx-1; i<lx; i++,j--) gel(y,i) = gel(x,j);
+  return normalizepol_lg(y,lx);
+}
+
+GEN
+RgX_recip_i(GEN x)
 {
   long lx, i, j;
   GEN y = cgetg_copy(x, &lx);
@@ -1108,7 +1152,7 @@ RgX_Rg_add_shallow(GEN y, GEN x)
   z = cgetg(lz,t_POL); z[1] = y[1];
   gel(z,2) = gadd(gel(y,2),x);
   for(i=3; i<lz; i++) gel(z,i) = gel(y,i);
-  return z = normalizepol_lg(z,lz);
+  return normalizepol_lg(z,lz);
 }
 GEN
 RgX_Rg_sub(GEN y, GEN x)
@@ -1127,7 +1171,7 @@ RgX_Rg_sub(GEN y, GEN x)
   z = cgetg(lz,t_POL); z[1] = y[1];
   gel(z,2) = gsub(gel(y,2),x);
   for(i=3; i<lz; i++) gel(z,i) = gcopy(gel(y,i));
-  return z = normalizepol_lg(z,lz);
+  return normalizepol_lg(z,lz);
 }
 GEN
 Rg_RgX_sub(GEN x, GEN y)
@@ -1138,7 +1182,7 @@ Rg_RgX_sub(GEN x, GEN y)
   z = cgetg(lz,t_POL); z[1] = y[1];
   gel(z,2) = gsub(x, gel(y,2));
   for(i=3; i<lz; i++) gel(z,i) = gneg(gel(y,i));
-  return z = normalizepol_lg(z,lz);
+  return normalizepol_lg(z,lz);
 }
 /*******************************************************************/
 /*                                                                 */
@@ -1360,9 +1404,9 @@ RgXn_mul2(GEN f, GEN g, long n)
   n0 = n>>1; n1 = n-n0;
   RgX_even_odd(f, &fe, &fo);
   RgX_even_odd(g, &ge, &go);
-  l = RgXn_mul(fe,ge,n1);
-  h = RgXn_mul(fo,go,n0);
-  m = RgX_sub(RgXn_mul(RgX_add(fe,fo),RgX_add(ge,go),n0), RgX_add(l,h));
+  l = RgXn_mul2(fe,ge,n1);
+  h = RgXn_mul2(fo,go,n0);
+  m = RgX_sub(RgXn_mul2(RgX_add(fe,fo),RgX_add(ge,go),n0), RgX_add(l,h));
   /* n1-1 <= n0 <= n1, deg l,m <= n1-1, deg h <= n0-1
    * result is t^2 h(t^2) + t m(t^2) + l(t^2) */
   l = RgX_inflate(l,2); /* deg l <= 2n1 - 2 <= n-1 */
@@ -1382,8 +1426,8 @@ RgX_mulhigh_i2(GEN f, GEN g, long n)
   long d = degpol(f)+degpol(g) + 1 - n;
   GEN h;
   if (d <= 2) return RgX_shift_shallow(RgX_mul(f,g), -n);
-  h = RgX_recip_shallow(RgXn_mul(RgX_recip_shallow(f),
-                                 RgX_recip_shallow(g), d));
+  h = RgX_recip_i(RgXn_mul2(RgX_recip_i(f),
+                                  RgX_recip_i(g), d));
   return RgX_shift_shallow(h, d-1-degpol(h)); /* possibly (fg)(0) = 0 */
 }
 
@@ -1394,7 +1438,7 @@ RgX_sqrhigh_i2(GEN f, long n)
   long d = 2*degpol(f)+ 1 - n;
   GEN h;
   if (d <= 2) return RgX_shift_shallow(RgX_sqr(f), -n);
-  h = RgX_recip_shallow(RgXn_sqr(RgX_recip_shallow(f), d));
+  h = RgX_recip_i(RgXn_sqr(RgX_recip_i(f), d));
   return RgX_shift_shallow(h, d-1-degpol(h)); /* possibly (fg)(0) = 0 */
 }
 
@@ -1630,12 +1674,7 @@ RgX_normalize(GEN x)
   gel(z,n) = Rg_get_1(d); return z;
 }
 GEN
-RgX_divs(GEN x, long y) {
-  long i, lx;
-  GEN z = cgetg_copy(x, &lx); z[1] = x[1];
-  for (i=2; i<lx; i++) gel(z,i) = gdivgs(gel(x,i),y);
-  return normalizepol_lg(z, lx);
-}
+RgX_divs(GEN x, long y) { pari_APPLY_pol(gdivgs(gel(x,i),y)); }
 GEN
 RgX_div_by_X_x(GEN a, GEN x, GEN *r)
 {
@@ -1714,8 +1753,7 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
     if (!pr || pr == ONLY_REM || pr == ONLY_DIVIDES)
       return gerepileupto(av, z);
     *pr = FpX_to_mod(*pr, p);
-    gerepileall(av, 2, pr, &z);
-    return z;
+    return gc_all(av, 2, &z, pr);
   }
   switch(typ(y_lead))
   {
@@ -1959,7 +1997,7 @@ rem(GEN c, GEN T)
 int
 ZXQX_dvd(GEN x, GEN y, GEN T)
 {
-  long dx, dy, dz, i, p, T_ismonic;
+  long dx, dy, i, T_ismonic;
   pari_sp av = avma, av2;
   GEN y_lead;
 
@@ -1972,17 +2010,16 @@ ZXQX_dvd(GEN x, GEN y, GEN T)
   dx = degpol(x);
   if (dx < dy) return !signe(x);
   (void)new_chunk(2);
-  x = RgX_recip_shallow(x)+2;
-  y = RgX_recip_shallow(y)+2;
+  x = RgX_recip_i(x)+2;
+  y = RgX_recip_i(y)+2;
   /* pay attention to sparse divisors */
   for (i = 1; i <= dy; i++)
     if (!signe(gel(y,i))) gel(y,i) = NULL;
-  dz = dx-dy; p = dz+1;
   av2 = avma;
   for (;;)
   {
     GEN m, x0 = gel(x,0), y0 = y_lead, cx = content(x0);
-    x0 = gneg(x0); p--;
+    x0 = gneg(x0);
     m = gcdii(cx, y0);
     if (!equali1(m))
     {
@@ -2029,8 +2066,8 @@ RgXQX_pseudorem(GEN x, GEN y, GEN T)
   dx = degpol(x);
   if (dx < dy) return RgX_copy(x);
   (void)new_chunk(2);
-  x = RgX_recip_shallow(x)+2;
-  y = RgX_recip_shallow(y)+2;
+  x = RgX_recip_i(x)+2;
+  y = RgX_recip_i(y)+2;
   /* pay attention to sparse divisors */
   for (i = 1; i <= dy; i++)
     if (isexactzero(gel(y,i))) gel(y,i) = NULL;
@@ -2062,7 +2099,7 @@ RgXQX_pseudorem(GEN x, GEN y, GEN T)
   lx = dx+3; x -= 2;
   x[0] = evaltyp(t_POL) | evallg(lx);
   x[1] = evalsigne(1) | evalvarn(vx);
-  x = RgX_recip_shallow(x);
+  x = RgX_recip_i(x);
   if (p)
   { /* multiply by y[0]^p   [beware dummy vars from FpX_FpXY_resultant] */
     GEN t = y_lead;
@@ -2105,8 +2142,8 @@ RgXQX_pseudodivrem(GEN x, GEN y, GEN T, GEN *ptr)
     *ptr = gerepileupto(av, r); return scalarpol(x_lead, vx);
   }
   (void)new_chunk(2);
-  x = RgX_recip_shallow(x)+2;
-  y = RgX_recip_shallow(y)+2;
+  x = RgX_recip_i(x)+2;
+  y = RgX_recip_i(y)+2;
   /* pay attention to sparse divisors */
   for (i = 1; i <= dy; i++)
     if (isexactzero(gel(y,i))) gel(y,i) = NULL;
@@ -2158,17 +2195,16 @@ RgXQX_pseudodivrem(GEN x, GEN y, GEN T, GEN *ptr)
     lx = dx+3; x -= 2;
     x[0] = evaltyp(t_POL) | evallg(lx);
     x[1] = evalsigne(1) | evalvarn(vx);
-    x = RgX_recip_shallow(x);
+    x = RgX_recip_i(x);
   }
-  z = RgX_recip_shallow(z);
+  z = RgX_recip_i(z);
   r = x;
   if (p)
   {
     GEN c = gel(ypow,p); r = RgX_Rg_mul(r, c);
     if (T && typ(c) == t_POL && varn(c) == varn(T)) r = RgXQX_red(r, T);
   }
-  gerepileall(av, 2, &z, &r);
-  *ptr = r; return z;
+  *ptr = r; return gc_all(av, 2, &z, ptr);
 }
 GEN
 RgX_pseudodivrem(GEN x, GEN y, GEN *ptr)
@@ -2176,41 +2212,20 @@ RgX_pseudodivrem(GEN x, GEN y, GEN *ptr)
 
 GEN
 RgXQX_mul(GEN x, GEN y, GEN T)
-{
-  return RgXQX_red(RgX_mul(x,y), T);
-}
+{ return RgXQX_red(RgX_mul(x,y), T); }
 GEN
-RgX_Rg_mul(GEN y, GEN x) {
-  long i, ly;
-  GEN z = cgetg_copy(y, &ly); z[1] = y[1];
-  if (ly == 2) return z;
-  for (i = 2; i < ly; i++) gel(z,i) = gmul(x,gel(y,i));
-  return normalizepol_lg(z,ly);
-}
+RgX_Rg_mul(GEN x, GEN y) { pari_APPLY_pol(gmul(y, gel(x,i))); }
 GEN
-RgX_muls(GEN y, long x) {
-  long i, ly;
-  GEN z = cgetg_copy(y, &ly); z[1] = y[1];
-  if (ly == 2) return z;
-  for (i = 2; i < ly; i++) gel(z,i) = gmulsg(x,gel(y,i));
-  return normalizepol_lg(z,ly);
-}
+RgX_mul2n(GEN x, long n) { pari_APPLY_pol(gmul2n(gel(x,i), n)); }
 GEN
-RgXQX_RgXQ_mul(GEN x, GEN y, GEN T)
-{
-  return RgXQX_red(RgX_Rg_mul(x,y), T);
-}
+RgX_muls(GEN x, long y) { pari_APPLY_pol(gmulsg(y, gel(x,i))); }
 GEN
-RgXQV_RgXQ_mul(GEN v, GEN x, GEN T)
-{
-  return RgXQV_red(RgV_Rg_mul(v,x), T);
-}
+RgXQX_RgXQ_mul(GEN x, GEN y, GEN T) { return RgXQX_red(RgX_Rg_mul(x,y), T); }
+GEN
+RgXQV_RgXQ_mul(GEN v, GEN x, GEN T) { return RgXQV_red(RgV_Rg_mul(v,x), T); }
 
 GEN
-RgXQX_sqr(GEN x, GEN T)
-{
-  return RgXQX_red(RgX_sqr(x), T);
-}
+RgXQX_sqr(GEN x, GEN T) { return RgXQX_red(RgX_sqr(x), T); }
 
 GEN
 RgXQX_powers(GEN P, long n, GEN T)
@@ -2231,6 +2246,8 @@ _sub(void *data, GEN x, GEN y) { (void)data; return RgX_sub(x, y); }
 static GEN
 _sqr(void *data, GEN x) { return RgXQ_sqr(x, (GEN)data); }
 static GEN
+_pow(void *data, GEN x, GEN n) { return RgXQ_pow(x, n, (GEN)data); }
+static GEN
 _mul(void *data, GEN x, GEN y) { return RgXQ_mul(x,y, (GEN)data); }
 static GEN
 _cmul(void *data, GEN P, long a, GEN x) { (void)data; return RgX_Rg_mul(x,gel(P,a+2)); }
@@ -2367,8 +2384,8 @@ RgXn_sqrhigh(GEN f, long n2, long n)
   return RgX_add(RgX_mulhigh_i(fl, f, n2), RgXn_mul(fh, f, n - n2));
 }
 
-GEN
-RgXn_inv_i(GEN f, long e)
+static GEN
+RgXn_div_gen(GEN g, GEN f, long e)
 {
   pari_sp av;
   ulong mask;
@@ -2377,8 +2394,8 @@ RgXn_inv_i(GEN f, long e)
 
   if (!signe(f)) pari_err_INV("RgXn_inv",f);
   a = ginv(gel(f,2));
-  if (e == 1) return scalarpol(a, v);
-  else if (e == 2)
+  if (e == 1 && !g) return scalarpol(a, v);
+  else if (e == 2 && !g)
   {
     GEN b;
     if (degpol(f) <= 0 || gequal0(b = gel(f,3))) return scalarpol(a, v);
@@ -2396,8 +2413,17 @@ RgXn_inv_i(GEN f, long e)
     n<<=1; if (mask & 1) n--;
     mask >>= 1;
     fr = RgXn_red_shallow(f, n);
-    u = RgXn_mul(W, RgXn_mulhigh(fr, W, n2, n), n-n2);
-    W = RgX_sub(W, RgX_shift_shallow(u, n2));
+    if (mask>1 || !g)
+    {
+      u = RgXn_mul(W, RgXn_mulhigh(fr, W, n2, n), n-n2);
+      W = RgX_sub(W, RgX_shift_shallow(u, n2));
+    }
+    else
+    {
+      GEN y = RgXn_mul(g, W, n), yt =  RgXn_red_shallow(y, n-n2);
+      u = RgXn_mul(yt, RgXn_mulhigh(fr,  W, n2, n), n-n2);
+      W = RgX_sub(y, RgX_shift_shallow(u, n2));
+    }
     if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"RgXn_inv, e = %ld", n);
@@ -2407,10 +2433,35 @@ RgXn_inv_i(GEN f, long e)
   return W;
 }
 
+static GEN
+RgXn_div_FpX(GEN x, GEN y, long e, GEN p)
+{
+  GEN r;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    if (pp == 2)
+      r = F2x_to_ZX(F2xn_div(RgX_to_F2x(x), RgX_to_F2x(y), e));
+    else
+      r = Flx_to_ZX_inplace(Flxn_div(RgX_to_Flx(x, pp), RgX_to_Flx(y, pp), e, pp));
+  }
+  else
+    r = FpXn_div(RgX_to_FpX(x, p), RgX_to_FpX(y, p), e, p);
+  return FpX_to_mod(r, p);
+}
+
+static GEN
+RgXn_div_FpXQX(GEN x, GEN y, long n, GEN pol, GEN p)
+{
+  GEN r, T = RgX_to_FpX(pol, p);
+  if (signe(T) == 0) pari_err_OP("/", x, y);
+  r = FpXQXn_div(RgX_to_FpXQX(x, T, p), RgX_to_FpXQX(y, T, p), n, T, p);
+  return FpXQX_to_mod(r, T, p);
+}
+
 static GEN
 RgXn_inv_FpX(GEN x, long e, GEN p)
 {
-  pari_sp av = avma;
   GEN r;
   if (lgefint(p) == 3)
   {
@@ -2422,17 +2473,16 @@ RgXn_inv_FpX(GEN x, long e, GEN p)
   }
   else
     r = FpXn_inv(RgX_to_FpX(x, p), e, p);
-  return gerepileupto(av, FpX_to_mod(r, p));
+  return FpX_to_mod(r, p);
 }
 
 static GEN
 RgXn_inv_FpXQX(GEN x, long n, GEN pol, GEN p)
 {
-  pari_sp av = avma;
   GEN r, T = RgX_to_FpX(pol, p);
   if (signe(T) == 0) pari_err_OP("/", gen_1, x);
   r = FpXQXn_inv(RgX_to_FpXQX(x, T, p), n, T, p);
-  return gerepileupto(av, FpXQX_to_mod(r, T, p));
+  return FpXQX_to_mod(r, T, p);
 }
 
 #define code(t1,t2) ((t1 << 6) | t2)
@@ -2451,14 +2501,50 @@ RgXn_inv_fast(GEN x, long e)
     default:       return NULL;
   }
 }
+
+static GEN
+RgXn_div_fast(GEN x, GEN y, long e)
+{
+  GEN p, pol;
+  long pa;
+  long t = RgX_type2(x,y,&p,&pol,&pa);
+  switch(t)
+  {
+    case t_INTMOD: return RgXn_div_FpX(x, y, e, p);
+    case code(t_POLMOD, t_INTMOD):
+                   return RgXn_div_FpXQX(x, y, e, pol, p);
+    default:       return NULL;
+  }
+}
 #undef code
 
+GEN
+RgXn_div_i(GEN g, GEN f, long e)
+{
+  GEN h = RgXn_div_fast(g, f, e);
+  if (h) return h;
+  return RgXn_div_gen(g, f, e);
+}
+
+GEN
+RgXn_div(GEN g, GEN f, long e)
+{
+  pari_sp av = avma;
+  return gerepileupto(av, RgXn_div_i(g, f, e));
+}
+
+GEN
+RgXn_inv_i(GEN f, long e)
+{
+  GEN h = RgXn_inv_fast(f, e);
+  if (h) return h;
+  return RgXn_div_gen(NULL, f, e);
+}
+
 GEN
 RgXn_inv(GEN f, long e)
 {
   pari_sp av = avma;
-  GEN h = RgXn_inv_fast(f, e);
-  if (h) return h;
   return gerepileupto(av, RgXn_inv_i(f, e));
 }
 
@@ -2653,11 +2739,19 @@ RgXQ_powers(GEN x, long l, GEN T)
   return gen_powers(x, l, use_sqr, (void *)T,_sqr,_mul,_one);
 }
 
+GEN
+RgXQV_factorback(GEN L, GEN e, GEN T)
+{
+  return gen_factorback(L, e, (void*)T, &_mul, &_pow, &_one);
+}
+
 /* a in K = Q[X]/(T), returns [a^0, ..., a^n] */
 GEN
 QXQ_powers(GEN a, long n, GEN T)
 {
-  GEN den, v = RgXQ_powers(Q_remove_denom(a, &den), n, T);
+  GEN den, v;
+  if (!isint1(leading_coeff(T))) return RgXQ_powers(a, n, T);
+  v = ZXQ_powers(Q_remove_denom(a, &den), n, T);
   /* den*a integral; v[i+1] = (den*a)^i in K */
   if (den)
   { /* restore denominators */
@@ -2697,8 +2791,9 @@ do_QXQ_eval(GEN v, long imin, GEN a, GEN T)
 GEN
 QXV_QXQ_eval(GEN v, GEN a, GEN T)
 { return do_QXQ_eval(v, 1, a, T); }
+
 GEN
-QXX_QXQ_eval(GEN v, GEN a, GEN T)
+QXY_QXQ_evalx(GEN v, GEN a, GEN T)
 { return normalizepol(do_QXQ_eval(v, 2, a, T)); }
 
 GEN
@@ -2720,6 +2815,19 @@ RgXQ_norm(GEN x, GEN T)
   return gerepileupto(av, gdiv(y, gpowgs(L, dx)));
 }
 
+GEN
+RgXQ_trace(GEN x, GEN T)
+{
+  pari_sp av = avma;
+  GEN dT, z;
+  long n;
+  if (degpol(T)==0) return gmulgs(x,0);
+  dT = RgX_deriv(T); n = degpol(dT);
+  z = RgXQ_mul(x, dT, T);
+  if (degpol(z)<n) return gc_const(av, gen_0);
+  return gerepileupto(av, gdiv(gel(z,2+n), gel(T,3+n)));
+}
+
 GEN
 RgX_blocks(GEN P, long n, long m)
 {
@@ -2863,8 +2971,8 @@ RgX_mul_FpXQX(GEN x, GEN y, GEN pol, GEN p)
   GEN T = RgX_to_FpX(pol, p);
   if (signe(T)==0) pari_err_OP("*", x, y);
   dT = degpol(T);
-  kx = ZXX_to_Kronecker(RgX_to_FpXQX(x, T, p), dT);
-  ky = ZXX_to_Kronecker(RgX_to_FpXQX(y, T, p), dT);
+  kx = RgXX_to_Kronecker(RgX_to_FpXQX(x, T, p), dT);
+  ky = RgXX_to_Kronecker(RgX_to_FpXQX(y, T, p), dT);
   r = FpX_mul(kx, ky, p);
   if (signe(r)==0) { set_avma(av); return zero_FpXQX_mod(pol, p, varn(x)); }
   return gerepileupto(av, Kronecker_to_mod(FpX_to_mod(r, p), pol));
@@ -2879,8 +2987,8 @@ RgX_mul_QXQX(GEN x, GEN y, GEN T)
 {
   pari_sp av = avma;
   long dT = degpol(T);
-  GEN r = QX_mul(ZXX_to_Kronecker(RgX_liftred(x, T), dT),
-                 ZXX_to_Kronecker(RgX_liftred(y, T), dT));
+  GEN r = QX_mul(RgXX_to_Kronecker(RgX_liftred(x, T), dT),
+                 RgXX_to_Kronecker(RgX_liftred(y, T), dT));
   return gerepileupto(av, Kronecker_to_mod(r, T));
 }
 
@@ -2908,7 +3016,7 @@ RgX_sqr_FpXQX(GEN x, GEN pol, GEN p)
   GEN kx, r, T = RgX_to_FpX(pol, p);
   if (signe(T)==0) pari_err_OP("*",x,x);
   dT = degpol(T);
-  kx = ZXX_to_Kronecker(RgX_to_FpXQX(x, T, p), dT);
+  kx = RgXX_to_Kronecker(RgX_to_FpXQX(x, T, p), dT);
   r = FpX_sqr(kx, p);
   if (signe(r)==0) { set_avma(av); return zero_FpXQX_mod(pol, p, varn(x)); }
   return gerepileupto(av, Kronecker_to_mod(FpX_to_mod(r, p), pol));
@@ -2919,7 +3027,7 @@ RgX_sqr_QXQX(GEN x, GEN T)
 {
   pari_sp av = avma;
   long dT = degpol(T);
-  GEN r = QX_sqr(ZXX_to_Kronecker(RgX_liftred(x, T), dT));
+  GEN r = QX_sqr(RgXX_to_Kronecker(RgX_liftred(x, T), dT));
   return gerepileupto(av, Kronecker_to_mod(r, T));
 }
 
@@ -3057,6 +3165,14 @@ RgX_rem(GEN x, GEN y)
   return z;
 }
 
+static GEN
+_RgX_mul(void* E, GEN x, GEN y)
+{ (void) E; return RgX_mul(x, y); }
+
+GEN
+RgXV_prod(GEN V)
+{ return gen_product(V, NULL, &_RgX_mul); }
+
 GEN
 RgXn_mul(GEN f, GEN g, long n)
 {
