@@ -21,6 +21,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_mat
+
 /*******************************************************************/
 /*                                                                 */
 /*                               SUM                               */
@@ -700,7 +702,7 @@ genapply(void *E, GEN (*f)(void* E, GEN x), GEN x)
   switch(tx) {
     case t_POL: y = normalizepol(vecapply2(E,f,x)); break;
     case t_SER:
-      y = ser_isexactzero(x)? gcopy(x): normalize(vecapply2(E,f,x));
+      y = ser_isexactzero(x)? gcopy(x): normalizeser(vecapply2(E,f,x));
       break;
     case t_LIST:
       {
@@ -807,9 +809,9 @@ gen_parapply_slice(GEN worker, GEN D, long mmin)
 }
 
 GEN
-gen_parapply(GEN worker, GEN D)
+gen_parapply_percent(GEN worker, GEN D, long percent)
 {
-  long l = lg(D), i, pending = 0;
+  long l = lg(D), i, pending = 0, cnt = 0, lper = -1, lcnt = 0;
   GEN W, V;
   struct pari_mt pt;
 
@@ -824,11 +826,26 @@ gen_parapply(GEN worker, GEN D)
     if (i < l) gel(W,1) = gel(D,i);
     mt_queue_submit(&pt, i, i<l? W: NULL);
     done = mt_queue_get(&pt, &workid, &pending);
-    if (done) gel(V,workid) = done;
+    if (done)
+    {
+      gel(V,workid) = done;
+      if (percent && (++cnt)-lcnt>=percent)
+      {
+        long per = (long)(cnt*100./(l-1));
+        lcnt = cnt;
+        if (per > lper) { err_printf("%ld%% ",per); lper = per; }
+      }
+    }
   }
   mt_queue_end(&pt); return V;
 }
 
+GEN
+gen_parapply(GEN worker, GEN D)
+{
+  return gen_parapply_percent(worker, D, 0);
+}
+
 GEN
 parapply(GEN C, GEN D)
 {
@@ -863,7 +880,7 @@ genfold(void *E, GEN (*f)(void* E, GEN x, GEN y), GEN x)
 GEN
 fold0(GEN f, GEN x)
 {
-  if (typ(f) != t_CLOSURE || closure_arity(f) < 2) pari_err_TYPE("apply",f);
+  if (typ(f) != t_CLOSURE || closure_arity(f) < 2) pari_err_TYPE("fold",f);
   return genfold((void *) f, gp_call2, x);
 }
 /*******************************************************************/
@@ -923,7 +940,7 @@ gtomat(GEN x)
       y = mkmatcopy(x); break;
     case t_MAT:
       y = gcopy(x); break;
-    case t_QFI: case t_QFR: {
+    case t_QFB: {
       GEN b;
       y = cgetg(3,t_MAT); b = gmul2n(gel(x,2),-1);
       gel(y,1) = mkcol2(icopy(gel(x,1)), b);
