@@ -13,14 +13,14 @@ with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 /*********************************************************************/
-/**                                                                 **/
 /**                     ARITHMETIC FUNCTIONS                        **/
 /**                        (second part)                            **/
-/**                                                                 **/
 /*********************************************************************/
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_arith
+
 GEN
 boundfact(GEN n, ulong lim)
 {
@@ -41,23 +41,39 @@ boundfact(GEN n, ulong lim)
 
 /* NOT memory clean */
 GEN
-Z_smoothen(GEN N, GEN L, GEN *pP, GEN *pe)
+Z_lsmoothen(GEN N, GEN L, GEN *pP, GEN *pE)
 {
   long i, j, l = lg(L);
-  GEN e = new_chunk(l), P = new_chunk(l);
+  GEN E = new_chunk(l), P = new_chunk(l);
   for (i = j = 1; i < l; i++)
   {
     ulong p = uel(L,i);
     long v = Z_lvalrem(N, p, &N);
-    if (v) { P[j] = p; e[j] = v; j++; if (is_pm1(N)) { N = NULL; break; } }
+    if (v) { P[j] = p; E[j] = v; j++; if (is_pm1(N)) { N = NULL; break; } }
   }
-  P[0] = evaltyp(t_VECSMALL) | evallg(j); *pP = P;
-  e[0] = evaltyp(t_VECSMALL) | evallg(j); *pe = e; return N;
+  P[0] = evaltyp(t_VECSMALL) | evallg(j); if (pP) *pP = P;
+  E[0] = evaltyp(t_VECSMALL) | evallg(j); if (pE) *pE = E; return N;
+}
+GEN
+Z_smoothen(GEN N, GEN L, GEN *pP, GEN *pE)
+{
+  long i, j, l = lg(L);
+  GEN E = new_chunk(l), P = new_chunk(l);
+  for (i = j = 1; i < l; i++)
+  {
+    GEN p = gel(L,i);
+    long v = Z_pvalrem(N, p, &N);
+    if (v)
+    {
+      gel(P,j) = p; gel(E,j) = utoipos(v); j++;
+     if (is_pm1(N)) { N = NULL; break; }
+    }
+  }
+  P[0] = evaltyp(t_COL) | evallg(j); if (pP) *pP = P;
+  E[0] = evaltyp(t_COL) | evallg(j); if (pE) *pE = E; return N;
 }
 /***********************************************************************/
-/**                                                                   **/
 /**                    SIMPLE FACTORISATIONS                          **/
-/**                                                                   **/
 /***********************************************************************/
 /* Factor n and output [p,e,c] where
  * p, e and c are vecsmall with n = prod{p[i]^e[i]} and c[i] = p[i]^e[i] */
@@ -153,8 +169,8 @@ factor_eulerphi(GEN n)
     }
     if (e > 1) {
       if (B) {
-        gel(B,1) = shallowconcat(gel(B,1), p);
-        gel(B,2) = shallowconcat(gel(B,2), utoipos(e-1));
+        gel(B,1) = vec_append(gel(B,1), p);
+        gel(B,2) = vec_append(gel(B,2), utoipos(e-1));
       } else
         B = to_mat(p, e-1);
     }
@@ -178,9 +194,7 @@ factor_eulerphi(GEN n)
 #endif
 
 /***********************************************************************/
-/**                                                                   **/
 /**         CHECK FACTORIZATION FOR ARITHMETIC FUNCTIONS              **/
-/**                                                                   **/
 /***********************************************************************/
 int
 RgV_is_ZVpos(GEN v)
@@ -221,7 +235,7 @@ RgV_is_ZV0(GEN v)
   return 1;
 }
 
-static int
+int
 RgV_is_prV(GEN v)
 {
   long l = lg(v), i;
@@ -276,8 +290,8 @@ fuse_Z_factor(GEN f, GEN B)
   /* tail / initial segment */
   P2 = vecslice(P, i, l-1); P = vecslice(P, 1, i-1);
   E2 = vecslice(E, i, l-1); E = vecslice(E, 1, i-1);
-  P = shallowconcat(P, mkvec(factorback2(P2,E2)));
-  E = shallowconcat(E, mkvec(gen_1));
+  P = vec_append(P, factorback2(P2,E2));
+  E = vec_append(E, gen_1);
   return mkmat2(P, E);
 }
 
@@ -337,10 +351,8 @@ check_arith_all(GEN n, const char *f) {
 }
 
 /***********************************************************************/
-/**                                                                   **/
 /**                MISCELLANEOUS ARITHMETIC FUNCTIONS                 **/
 /**                (ultimately depend on Z_factor())                  **/
-/**                                                                   **/
 /***********************************************************************/
 /* set P,E from F. Check whether F is an integer and kill "factor" -1 */
 static void
@@ -618,14 +630,11 @@ _mod4(GEN c) {
 
 long
 corediscs(long D, ulong *f)
-{
-  /* D = f^2 dK */
-  long dK = D>=0 ? (long) coreu(D) : -(long) coreu(-(ulong) D);
-  ulong dKmod4 = ((ulong)dK)&3UL;
-  if (dKmod4 == 2 || dKmod4 == 3)
-    dK *= 4;
-  if (f) *f = usqrt((ulong)(D/dK));
-  return D;
+{ /* D = f^2 d */
+  long d = D >= 0? (long)coreu(D) : -(long)coreu(-(ulong)D);
+  if ((((ulong)d)&3UL) != 1) d *= 4;
+  if (f) *f = usqrt((ulong)(D/d));
+  return d;
 }
 
 GEN
@@ -652,6 +661,63 @@ coredisc2(GEN n)
 GEN
 coredisc0(GEN n,long flag) { return flag? coredisc2(n): coredisc(n); }
 
+/* Write x = Df^2, where D = fundamental discriminant,
+ * P^E = factorisation of conductor f */
+GEN
+coredisc2_fact(GEN fa, long s, GEN *pP, GEN *pE)
+{
+  GEN P, E, P0 = gel(fa,1), E0 = gel(fa,2), D = s > 0? gen_1: gen_m1;
+  long l = lg(P0), i, j;
+
+  E = cgetg(l, t_VECSMALL);
+  P = cgetg(l, t_VEC);
+  for (i = j = 1; i < l; i++)
+  {
+    long e = itos(gel(E0,i));
+    GEN p = gel(P0,i);
+    if (odd(e)) D = mulii(D, p);
+    e >>= 1; if (e) { gel(P, j) = p; E[j] = e; j++; }
+  }
+  if (Mod4(D) != 1)
+  {
+    D = shifti(D, 2);
+    if (!--E[1])
+    {
+      P[1] = P[0]; P++;
+      E[1] = E[0]; E++; j--;
+    }
+  }
+  setlg(P,j); *pP = P;
+  setlg(E,j); *pE = E; return D;
+}
+ulong
+coredisc2u_fact(GEN fa, long s, GEN *pP, GEN *pE)
+{
+  GEN P, E, P0 = gel(fa,1), E0 = gel(fa,2);
+  ulong D = 1;
+  long i, j, l = lg(P0);
+
+  E = cgetg(l, t_VECSMALL);
+  P = cgetg(l, t_VECSMALL);
+  for (i = j = 1; i < l; i++)
+  {
+    long e = E0[i], p = P0[i];
+    if (odd(e)) D *= p;
+    e >>= 1; if (e) { P[j] = p; E[j] = e; j++; }
+  }
+  if ((D & 3) != (s > 0? 1: 3))
+  {
+    D *= 4;
+    if (!--E[1])
+    {
+      P[1] = P[0]; P++;
+      E[1] = E[0]; E++; j--;
+    }
+  }
+  setlg(P,j); *pP = P;
+  setlg(E,j); *pE = E; return D;
+}
+
 long
 omegau(ulong n)
 {
@@ -922,6 +988,7 @@ uissquarefree_fact(GEN f)
 {
   GEN E = gel(f,2);
   long i, l = lg(E);
+  if (l == 2) return umael(f,1,1)? E[1] == 1: 0; /* handle factor(0) */
   for (i = 1; i < l; i++)
     if (E[i] > 1) return 0;
   return 1;
@@ -943,14 +1010,12 @@ Z_issquarefree(GEN n)
   return moebius(n)? 1: 0;
 }
 
-static int
-fa_issquarefree(GEN F)
+long
+Z_issquarefree_fact(GEN F)
 {
-  GEN P = gel(F,1), E = gel(F,2);
-  long i, s, l = lg(P);
-  if (l == 1) return 1;
-  s = signe(gel(P,1)); /* = signe(x) */
-  if (!s) return 0;
+  GEN E = gel(F,2);
+  long i, l = lg(E);
+  if (l == 2) return signe(gcoeff(F,1,1))? equali1(gel(E,1)): 0;
   for(i = 1; i < l; i++)
     if (!equali1(gel(E,i))) return 0;
   return 1;
@@ -968,355 +1033,8 @@ issquarefree(GEN x)
       av = avma; d = RgX_gcd(x, RgX_deriv(x));
       return gc_long(av, lg(d)==3);
     case t_VEC:
-    case t_MAT: return fa_issquarefree(check_arith_all(x,"issquarefree"));
+    case t_MAT: return Z_issquarefree_fact(check_arith_all(x,"issquarefree"));
     default: pari_err_TYPE("issquarefree",x);
       return 0; /* LCOV_EXCL_LINE */
   }
 }
-
-/*********************************************************************/
-/**                                                                 **/
-/**                    DIGITS / SUM OF DIGITS                       **/
-/**                                                                 **/
-/*********************************************************************/
-
-/* set v[i] = 1 iff B^i is needed in the digits_dac algorithm */
-static void
-set_vexp(GEN v, long l)
-{
-  long m;
-  if (v[l]) return;
-  v[l] = 1; m = l>>1;
-  set_vexp(v, m);
-  set_vexp(v, l-m);
-}
-
-/* return all needed B^i for DAC algorithm, for lz digits */
-static GEN
-get_vB(GEN T, long lz, void *E, struct bb_ring *r)
-{
-  GEN vB, vexp = const_vecsmall(lz, 0);
-  long i, l = (lz+1) >> 1;
-  vexp[1] = 1;
-  vexp[2] = 1;
-  set_vexp(vexp, lz);
-  vB = zerovec(lz); /* unneeded entries remain = 0 */
-  gel(vB, 1) = T;
-  for (i = 2; i <= l; i++)
-    if (vexp[i])
-    {
-      long j = i >> 1;
-      GEN B2j = r->sqr(E, gel(vB,j));
-      gel(vB,i) = odd(i)? r->mul(E, B2j, T): B2j;
-    }
-  return vB;
-}
-
-static void
-gen_digits_dac(GEN x, GEN vB, long l, GEN *z,
-               void *E, GEN div(void *E, GEN a, GEN b, GEN *r))
-{
-  GEN q, r;
-  long m = l>>1;
-  if (l==1) { *z=x; return; }
-  q = div(E, x, gel(vB,m), &r);
-  gen_digits_dac(r, vB, m, z, E, div);
-  gen_digits_dac(q, vB, l-m, z+m, E, div);
-}
-
-static GEN
-gen_fromdigits_dac(GEN x, GEN vB, long i, long l, void *E,
-                   GEN add(void *E, GEN a, GEN b),
-                   GEN mul(void *E, GEN a, GEN b))
-{
-  GEN a, b;
-  long m = l>>1;
-  if (l==1) return gel(x,i);
-  a = gen_fromdigits_dac(x, vB, i, m, E, add, mul);
-  b = gen_fromdigits_dac(x, vB, i+m, l-m, E, add, mul);
-  return add(E, a, mul(E, b, gel(vB, m)));
-}
-
-static GEN
-gen_digits_i(GEN x, GEN B, long n, void *E, struct bb_ring *r,
-                          GEN (*div)(void *E, GEN x, GEN y, GEN *r))
-{
-  GEN z, vB;
-  if (n==1) retmkvec(gcopy(x));
-  vB = get_vB(B, n, E, r);
-  z = cgetg(n+1, t_VEC);
-  gen_digits_dac(x, vB, n, (GEN*)(z+1), E, div);
-  return z;
-}
-
-GEN
-gen_digits(GEN x, GEN B, long n, void *E, struct bb_ring *r,
-                          GEN (*div)(void *E, GEN x, GEN y, GEN *r))
-{
-  pari_sp av = avma;
-  return gerepilecopy(av, gen_digits_i(x, B, n, E, r, div));
-}
-
-GEN
-gen_fromdigits(GEN x, GEN B, void *E, struct bb_ring *r)
-{
-  pari_sp av = avma;
-  long n = lg(x)-1;
-  GEN vB = get_vB(B, n, E, r);
-  GEN z = gen_fromdigits_dac(x, vB, 1, n, E, r->add, r->mul);
-  return gerepilecopy(av, z);
-}
-
-static GEN
-_addii(void *data /* ignored */, GEN x, GEN y)
-{ (void)data; return addii(x,y); }
-static GEN
-_sqri(void *data /* ignored */, GEN x) { (void)data; return sqri(x); }
-static GEN
-_mulii(void *data /* ignored */, GEN x, GEN y)
- { (void)data; return mulii(x,y); }
-static GEN
-_dvmdii(void *data /* ignored */, GEN x, GEN y, GEN *r)
- { (void)data; return dvmdii(x,y,r); }
-
-static struct bb_ring Z_ring = { _addii, _mulii, _sqri };
-
-static GEN
-check_basis(GEN B)
-{
-  if (!B) return utoipos(10);
-  if (typ(B)!=t_INT) pari_err_TYPE("digits",B);
-  if (abscmpiu(B,2)<0) pari_err_DOMAIN("digits","B","<",gen_2,B);
-  return B;
-}
-
-/* x has l digits in base B, write them to z[0..l-1], vB[i] = B^i */
-static void
-digits_dacsmall(GEN x, GEN vB, long l, ulong* z)
-{
-  pari_sp av = avma;
-  GEN q,r;
-  long m;
-  if (l==1) { *z=itou(x); return; }
-  m=l>>1;
-  q = dvmdii(x, gel(vB,m), &r);
-  digits_dacsmall(q,vB,l-m,z);
-  digits_dacsmall(r,vB,m,z+l-m);
-  set_avma(av);
-}
-
-GEN
-digits(GEN x, GEN B)
-{
-  pari_sp av=avma;
-  long lz;
-  GEN z, vB;
-  if (typ(x)!=t_INT) pari_err_TYPE("digits",x);
-  B = check_basis(B);
-  if (signe(B)<0) pari_err_DOMAIN("digits","B","<",gen_0,B);
-  if (!signe(x))       {set_avma(av); return cgetg(1,t_VEC); }
-  if (abscmpii(x,B)<0) {set_avma(av); retmkvec(absi(x)); }
-  if (Z_ispow2(B))
-  {
-    long k = expi(B);
-    if (k == 1) return binaire(x);
-    if (k < BITS_IN_LONG)
-    {
-      (void)new_chunk(4*(expi(x) + 2)); /* HACK */
-      z = binary_2k_nv(x, k);
-      set_avma(av); return Flv_to_ZV(z);
-    }
-    else
-    {
-      set_avma(av); return binary_2k(x, k);
-    }
-  }
-  x = absi_shallow(x);
-  lz = logint(x,B) + 1;
-  if (lgefint(B)>3)
-  {
-    z = gerepileupto(av, gen_digits_i(x, B, lz, NULL, &Z_ring, _dvmdii));
-    vecreverse_inplace(z); return z;
-  }
-  else
-  {
-    vB = get_vB(B, lz, NULL, &Z_ring);
-    (void)new_chunk(3*lz); /* HACK */
-    z = zero_zv(lz);
-    digits_dacsmall(x,vB,lz,(ulong*)(z+1));
-    set_avma(av); return Flv_to_ZV(z);
-  }
-}
-
-static GEN
-fromdigitsu_dac(GEN x, GEN vB, long i, long l)
-{
-  GEN a, b;
-  long m = l>>1;
-  if (l==1) return utoi(uel(x,i));
-  if (l==2) return addumului(uel(x,i), uel(x,i+1), gel(vB, m));
-  a = fromdigitsu_dac(x, vB, i, m);
-  b = fromdigitsu_dac(x, vB, i+m, l-m);
-  return addii(a, mulii(b, gel(vB, m)));
-}
-
-GEN
-fromdigitsu(GEN x, GEN B)
-{
-  pari_sp av = avma;
-  long n = lg(x)-1;
-  GEN vB, z;
-  if (n==0) return gen_0;
-  vB = get_vB(B, n, NULL, &Z_ring);
-  z = fromdigitsu_dac(x, vB, 1, n);
-  return gerepileuptoint(av, z);
-}
-
-static int
-ZV_in_range(GEN v, GEN B)
-{
-  long i, l = lg(v);
-  for(i=1; i < l; i++)
-  {
-    GEN vi = gel(v, i);
-    if (signe(vi) < 0 || cmpii(vi, B) >= 0)
-      return 0;
-  }
-  return 1;
-}
-
-GEN
-fromdigits(GEN x, GEN B)
-{
-  pari_sp av = avma;
-  if (typ(x)!=t_VEC || !RgV_is_ZV(x)) pari_err_TYPE("fromdigits",x);
-  if (lg(x)==1) return gen_0;
-  B = check_basis(B);
-  if (Z_ispow2(B) && ZV_in_range(x, B))
-    return fromdigits_2k(x, expi(B));
-  x = vecreverse(x);
-  return gerepileuptoint(av, gen_fromdigits(x, B, NULL, &Z_ring));
-}
-
-static const ulong digsum[] ={
-  0,1,2,3,4,5,6,7,8,9,1,2,3,4,5,6,7,8,9,10,2,3,4,5,6,7,8,9,10,11,3,4,5,6,7,8,
-  9,10,11,12,4,5,6,7,8,9,10,11,12,13,5,6,7,8,9,10,11,12,13,14,6,7,8,9,10,11,
-  12,13,14,15,7,8,9,10,11,12,13,14,15,16,8,9,10,11,12,13,14,15,16,17,9,10,11,
-  12,13,14,15,16,17,18,1,2,3,4,5,6,7,8,9,10,2,3,4,5,6,7,8,9,10,11,3,4,5,6,7,8,
-  9,10,11,12,4,5,6,7,8,9,10,11,12,13,5,6,7,8,9,10,11,12,13,14,6,7,8,9,10,11,
-  12,13,14,15,7,8,9,10,11,12,13,14,15,16,8,9,10,11,12,13,14,15,16,17,9,10,11,
-  12,13,14,15,16,17,18,10,11,12,13,14,15,16,17,18,19,2,3,4,5,6,7,8,9,10,11,3,
-  4,5,6,7,8,9,10,11,12,4,5,6,7,8,9,10,11,12,13,5,6,7,8,9,10,11,12,13,14,6,7,8,
-  9,10,11,12,13,14,15,7,8,9,10,11,12,13,14,15,16,8,9,10,11,12,13,14,15,16,17,
-  9,10,11,12,13,14,15,16,17,18,10,11,12,13,14,15,16,17,18,19,11,12,13,14,15,
-  16,17,18,19,20,3,4,5,6,7,8,9,10,11,12,4,5,6,7,8,9,10,11,12,13,5,6,7,8,9,10,
-  11,12,13,14,6,7,8,9,10,11,12,13,14,15,7,8,9,10,11,12,13,14,15,16,8,9,10,11,
-  12,13,14,15,16,17,9,10,11,12,13,14,15,16,17,18,10,11,12,13,14,15,16,17,18,
-  19,11,12,13,14,15,16,17,18,19,20,12,13,14,15,16,17,18,19,20,21,4,5,6,7,8,9,
-  10,11,12,13,5,6,7,8,9,10,11,12,13,14,6,7,8,9,10,11,12,13,14,15,7,8,9,10,11,
-  12,13,14,15,16,8,9,10,11,12,13,14,15,16,17,9,10,11,12,13,14,15,16,17,18,10,
-  11,12,13,14,15,16,17,18,19,11,12,13,14,15,16,17,18,19,20,12,13,14,15,16,17,
-  18,19,20,21,13,14,15,16,17,18,19,20,21,22,5,6,7,8,9,10,11,12,13,14,6,7,8,9,
-  10,11,12,13,14,15,7,8,9,10,11,12,13,14,15,16,8,9,10,11,12,13,14,15,16,17,9,
-  10,11,12,13,14,15,16,17,18,10,11,12,13,14,15,16,17,18,19,11,12,13,14,15,16,
-  17,18,19,20,12,13,14,15,16,17,18,19,20,21,13,14,15,16,17,18,19,20,21,22,14,
-  15,16,17,18,19,20,21,22,23,6,7,8,9,10,11,12,13,14,15,7,8,9,10,11,12,13,14,
-  15,16,8,9,10,11,12,13,14,15,16,17,9,10,11,12,13,14,15,16,17,18,10,11,12,13,
-  14,15,16,17,18,19,11,12,13,14,15,16,17,18,19,20,12,13,14,15,16,17,18,19,20,
-  21,13,14,15,16,17,18,19,20,21,22,14,15,16,17,18,19,20,21,22,23,15,16,17,18,
-  19,20,21,22,23,24,7,8,9,10,11,12,13,14,15,16,8,9,10,11,12,13,14,15,16,17,9,
-  10,11,12,13,14,15,16,17,18,10,11,12,13,14,15,16,17,18,19,11,12,13,14,15,16,
-  17,18,19,20,12,13,14,15,16,17,18,19,20,21,13,14,15,16,17,18,19,20,21,22,14,
-  15,16,17,18,19,20,21,22,23,15,16,17,18,19,20,21,22,23,24,16,17,18,19,20,21,
-  22,23,24,25,8,9,10,11,12,13,14,15,16,17,9,10,11,12,13,14,15,16,17,18,10,11,
-  12,13,14,15,16,17,18,19,11,12,13,14,15,16,17,18,19,20,12,13,14,15,16,17,18,
-  19,20,21,13,14,15,16,17,18,19,20,21,22,14,15,16,17,18,19,20,21,22,23,15,16,
-  17,18,19,20,21,22,23,24,16,17,18,19,20,21,22,23,24,25,17,18,19,20,21,22,23,
-  24,25,26,9,10,11,12,13,14,15,16,17,18,10,11,12,13,14,15,16,17,18,19,11,12,
-  13,14,15,16,17,18,19,20,12,13,14,15,16,17,18,19,20,21,13,14,15,16,17,18,19,
-  20,21,22,14,15,16,17,18,19,20,21,22,23,15,16,17,18,19,20,21,22,23,24,16,17,
-  18,19,20,21,22,23,24,25,17,18,19,20,21,22,23,24,25,26,18,19,20,21,22,23,24,
-  25,26,27
-};
-
-ulong
-sumdigitsu(ulong n)
-{
-  ulong s = 0;
-  while (n) { s += digsum[n % 1000]; n /= 1000; }
-  return s;
-}
-
-/* res=array of 9-digits integers, return sum_{0 <= i < l} sumdigits(res[i]) */
-static ulong
-sumdigits_block(ulong *res, long l)
-{
-  ulong s = sumdigitsu(*--res);
-  while (--l > 0) s += sumdigitsu(*--res);
-  return s;
-}
-
-GEN
-sumdigits(GEN n)
-{
-  const long L = (long)(ULONG_MAX / 81);
-  pari_sp av = avma;
-  ulong *res;
-  long l;
-
-  if (typ(n) != t_INT) pari_err_TYPE("sumdigits", n);
-  switch(lgefint(n))
-  {
-    case 2: return gen_0;
-    case 3: return utoipos(sumdigitsu(n[2]));
-  }
-  res = convi(n, &l);
-  if (l < L)
-  {
-    ulong s = sumdigits_block(res, l);
-    set_avma(av); return utoipos(s);
-  }
-  else /* Huge. Overflows ulong */
-  {
-    GEN S = gen_0;
-    while (l > L)
-    {
-      S = addiu(S, sumdigits_block(res, L));
-      res += L; l -= L;
-    }
-    if (l)
-      S = addiu(S, sumdigits_block(res, l));
-    return gerepileuptoint(av, S);
-  }
-}
-
-GEN
-sumdigits0(GEN x, GEN B)
-{
-  pari_sp av = avma;
-  GEN z;
-  long lz;
-
-  if (!B) return sumdigits(x);
-  if (typ(x) != t_INT) pari_err_TYPE("sumdigits", x);
-  B = check_basis(B);
-  if (Z_ispow2(B))
-  {
-    long k = expi(B);
-    if (k == 1) { set_avma(av); return utoi(hammingweight(x)); }
-    if (k < BITS_IN_LONG)
-    {
-      GEN z = binary_2k_nv(x, k);
-      if (lg(z)-1 > 1L<<(BITS_IN_LONG-k)) /* may overflow */
-        return gerepileuptoint(av, ZV_sum(Flv_to_ZV(z)));
-      set_avma(av); return utoi(zv_sum(z));
-    }
-    return gerepileuptoint(av, ZV_sum(binary_2k(x, k)));
-  }
-  if (!signe(x))       { set_avma(av); return gen_0; }
-  if (abscmpii(x,B)<0) { set_avma(av); return absi(x); }
-  if (absequaliu(B,10))   { set_avma(av); return sumdigits(x); }
-  x = absi_shallow(x); lz = logint(x,B) + 1;
-  z = gen_digits_i(x, B, lz, NULL, &Z_ring, _dvmdii);
-  return gerepileuptoint(av, ZV_sum(z));
-}
