@@ -21,6 +21,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_nf
+
 /*******************************************************************/
 /*                                                                 */
 /*                     IDEAL OPERATIONS                            */
@@ -46,16 +48,17 @@ idealtyp(GEN *ideal, GEN *arch)
   GEN x = *ideal;
   long t,lx,tx = typ(x);
 
-  if (tx!=t_VEC || lg(x)!=3) *arch = NULL;
+  if (tx!=t_VEC || lg(x)!=3) { if (arch) *arch = NULL; }
   else
   {
     GEN a = gel(x,2);
     if (typ(a) == t_MAT && lg(a) != 3)
     { /* allow [;] */
       if (lg(a) != 1) pari_err_TYPE("idealtyp [extended ideal]",x);
-      a = trivial_fact();
+      if (arch) *arch = trivial_fact();
     }
-    *arch = a;
+    else
+      if (arch) *arch = a;
     x = gel(x,1); tx = typ(x);
   }
   switch(tx)
@@ -154,8 +157,7 @@ idealhnf_shallow(GEN nf, GEN x)
       x = ZM_hnfmod(x, ZM_detmult(x));
       return cx? ZM_Q_mul(x,cx): x;
     }
-    case t_QFI:
-    case t_QFR:
+    case t_QFB:
     {
       pari_sp av = avma;
       GEN u, D = nf_get_disc(nf), T = nf_get_pol(nf), f = nf_get_index(nf);
@@ -178,11 +180,12 @@ idealhnf_shallow(GEN nf, GEN x)
     default: return idealhnf_principal(nf, x); /* PRINCIPAL */
   }
 }
+/* true nf */
 GEN
 idealhnf(GEN nf, GEN x)
 {
   pari_sp av = avma;
-  GEN y = idealhnf_shallow(checknf(nf), x);
+  GEN y = idealhnf_shallow(nf, x);
   return (avma == av)? gcopy(y): gerepileupto(av, y);
 }
 
@@ -286,10 +289,11 @@ idealhnf0(GEN nf, GEN a, GEN b)
   long ta, tb;
   pari_sp av;
   GEN x;
+  nf = checknf(nf);
   if (!b) return idealhnf(nf,a);
 
   /* HNF of aZ_K+bZ_K */
-  av = avma; nf = checknf(nf);
+  av = avma;
   a = nf_to_scalar_or_basis(nf,a); ta = typ(a);
   b = nf_to_scalar_or_basis(nf,b); tb = typ(b);
   if (ta == t_COL)
@@ -410,7 +414,7 @@ mat_ideal_two_elt(GEN nf, GEN x)
     const long FB[] = { _evallg(15+1) | evaltyp(t_VECSMALL),
       2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47
     };
-    GEN P, E, a1 = Z_smoothen(xZ, (GEN)FB, &P, &E);
+    GEN P, E, a1 = Z_lsmoothen(xZ, (GEN)FB, &P, &E);
     if (!a1) /* factors completely */
       a = idealapprfact_i(nf, idealfactor(nf,x), 1);
     else if (lg(P) == 1) /* no small factors */
@@ -465,8 +469,7 @@ GEN
 idealtwoelt(GEN nf, GEN x)
 {
   pari_sp av;
-  GEN z;
-  long tx = idealtyp(&x,&z);
+  long tx = idealtyp(&x, NULL);
   nf = checknf(nf);
   if (tx == id_MAT) return mat_ideal_two_elt(nf,x);
   if (tx == id_PRIME) return mkvec2copy(gel(x,1), gel(x,2));
@@ -704,10 +707,8 @@ Q_nffactor(GEN nf, GEN y, ulong lim)
     gel(P,i) = idealprimedec(nf, gel(P,i));
     gel(E,i) = prV_e_muls(gel(P,i), itos(gel(E,i)));
   }
-  settyp(P,t_VEC); P = shallowconcat1(P);
-  settyp(E,t_VEC); E = shallowconcat1(E);
-  gel(f,1) = P; settyp(P, t_COL);
-  gel(f,2) = E; return f;
+  P = shallowconcat1(P); gel(f,1) = P; settyp(P, t_COL);
+  E = shallowconcat1(E); gel(f,2) = E; return f;
 }
 
 GEN
@@ -741,7 +742,7 @@ idealfactor_limit(GEN nf, GEN x, ulong lim)
 {
   pari_sp av = avma;
   GEN fa, y;
-  long tx = idealtyp(&x,&y);
+  long tx = idealtyp(&x, NULL);
 
   if (tx == id_PRIME)
   {
@@ -793,26 +794,31 @@ static int
 ramified_root_simple(GEN nf, long n, GEN P, GEN v)
 {
   long i, l = lg(v);
-  for (i = 1; i < l; i++) if (v[i])
+  for (i = 1; i < l; i++)
   {
-    GEN vpr = idealprimedec(nf, gel(P,i));
-    long lpr = lg(vpr), j;
-    for (j = 1; j < lpr; j++)
+    long w = v[i] % n;
+    if (w)
     {
-      long e = pr_get_e(gel(vpr,j));
-      if ((e * v[i]) % n) return 0;
+      GEN vpr = idealprimedec(nf, gel(P,i));
+      long lpr = lg(vpr), j;
+      for (j = 1; j < lpr; j++)
+      {
+        long e = pr_get_e(gel(vpr,j));
+        if ((e * w) % n) return 0;
+      }
     }
   }
   return 1;
 }
-/* true nf; A is assumed to be the n-th power of an integral ideal,
- * return its n-th root; n > 1 */
+/* true nf, n > 1, A a non-zero integral ideal; check whether A is the n-th
+ * power of an ideal and set *pB to its n-th root if so */
 static long
 idealsqrtn_int(GEN nf, GEN A, long n, GEN *pB)
 {
   GEN C, root;
   long i, l;
 
+  if (typ(A) == t_MAT && ZM_isscalar(A, NULL)) A = gcoeff(A,1,1);
   if (typ(A) == t_INT) /* > 0 */
   {
     GEN P = nf_get_ramified_primes(nf), v, q;
@@ -927,8 +933,8 @@ long
 idealval(GEN nf, GEN A, GEN P)
 {
   pari_sp av = avma;
-  GEN a, p, cA;
-  long vcA, v, Zval, tx = idealtyp(&A,&a);
+  GEN p, cA;
+  long vcA, v, Zval, tx = idealtyp(&A, NULL);
 
   if (tx == id_PRINCIPAL) return nfval(nf,A,P);
   checkprid(P);
@@ -964,8 +970,8 @@ idealadd(GEN nf, GEN x, GEN y)
   long tx, ty;
   GEN z, a, dx, dy, dz;
 
-  tx = idealtyp(&x,&z);
-  ty = idealtyp(&y,&z); nf = checknf(nf);
+  tx = idealtyp(&x, NULL);
+  ty = idealtyp(&y, NULL); nf = checknf(nf);
   if (tx != id_MAT) x = idealhnf_shallow(nf,x);
   if (ty != id_MAT) y = idealhnf_shallow(nf,y);
   if (lg(x) == 1) return gerepilecopy(av,y);
@@ -996,8 +1002,8 @@ static GEN
 _idealaddtoone(GEN nf, GEN x, GEN y, long red)
 {
   GEN a;
-  long tx = idealtyp(&x, &a/*junk*/);
-  long ty = idealtyp(&y, &a/*junk*/);
+  long tx = idealtyp(&x, NULL);
+  long ty = idealtyp(&y, NULL);
   long ea;
   if (tx != id_MAT) x = idealhnf_shallow(nf, x);
   if (ty != id_MAT) y = idealhnf_shallow(nf, y);
@@ -1906,10 +1912,9 @@ GEN
 idealnorm(GEN nf, GEN x)
 {
   pari_sp av;
-  GEN y;
   long tx;
 
-  switch(idealtyp(&x,&y))
+  switch(idealtyp(&x, NULL))
   {
     case id_PRIME: return pr_norm(x);
     case id_MAT: return RgM_det_triangular(x);
@@ -1929,7 +1934,7 @@ idealdown(GEN nf, GEN x)
 {
   pari_sp av = avma;
   GEN y, c;
-  switch(idealtyp(&x,&y))
+  switch(idealtyp(&x, NULL))
   {
     case id_PRIME: return icopy(pr_get_p(x));
     case id_MAT: return gcopy(gcoeff(x,1,1));
@@ -1989,13 +1994,12 @@ idealismaximal_i(GEN nf, GEN x)
 {
   GEN L, p, pr, c;
   long i, l;
-  switch(idealtyp(&x,&c))
+  switch(idealtyp(&x, NULL))
   {
     case id_PRIME: return x;
     case id_MAT: return idealismaximal_mat(nf, x);
   }
   /* id_PRINCIPAL */
-  nf = checknf(nf);
   x = nf_to_scalar_or_basis(nf, x);
   switch(typ(x))
   {
@@ -2111,8 +2115,8 @@ GEN
 idealnumden(GEN nf, GEN x)
 {
   pari_sp av = avma;
-  GEN x0, ax, c, d, A, B, J;
-  long tx = idealtyp(&x,&ax);
+  GEN x0, c, d, A, B, J;
+  long tx = idealtyp(&x, NULL);
   nf = checknf(nf);
   switch (tx)
   {
@@ -2129,7 +2133,7 @@ idealnumden(GEN nf, GEN x)
       }
       /* t_COL */
       x = Q_remove_denom(x, &d);
-      if (!d) return gerepilecopy(av, mkvec2(idealhnf(nf, x), gen_1));
+      if (!d) return gerepilecopy(av, mkvec2(idealhnf_shallow(nf, x), gen_1));
       mx = zk_multable(nf, x);
       xZ = zkmultable_capZ(mx);
       x = ZM_hnfmodid(mx, xZ); /* principal ideal (x) */
@@ -2210,15 +2214,13 @@ idealpowprime(GEN nf, GEN pr, GEN n, GEN *pc)
   return mkvec2(q, gen);
 }
 
-/* x * pr^n. Assume x in HNF or scalar (possibly nonintegral) */
+/* True nf. x * pr^n. Assume x in HNF or scalar (possibly nonintegral) */
 GEN
 idealmulpowprime(GEN nf, GEN x, GEN pr, GEN n)
 {
   GEN c, cx, y;
-  long N;
+  long N = nf_get_degree(nf);
 
-  nf = checknf(nf);
-  N = nf_get_degree(nf);
   if (!signe(n)) return typ(x) == t_MAT? x: scalarmat_shallow(x, N);
 
   /* inert, special cased for efficiency */
@@ -2307,9 +2309,8 @@ idealpow(GEN nf, GEN x, GEN n)
   av = avma;
   x = gerepileupto(av, idealpow_aux(checknf(nf), x, tx, n));
   if (!ax) return x;
-  ax = ext_pow(nf, ax, n);
   gel(res,1) = x;
-  gel(res,2) = ax;
+  gel(res,2) = ext_pow(nf, ax, n);
   return res;
 }
 
@@ -2578,7 +2579,7 @@ GEN
 idealred0(GEN nf, GEN I, GEN vdir)
 {
   pari_sp av = avma;
-  GEN G, aI, IZ, J, y, yZ, my, c1 = NULL;
+  GEN G, aI, IZ, J, y, yZ, my, yi, c1 = NULL;
   long N;
 
   nf = checknf(nf);
@@ -2623,23 +2624,81 @@ idealred0(GEN nf, GEN I, GEN vdir)
   my = zk_multable(nf, y);
   I = ZM_Z_divexact(ZM_mul(my, I), IZ); /* y I / (I\cap Z), integral */
   c1 = mul_content(c1, IZ);
-  my = ZM_gauss(my, col_ei(N,1)); /* y^-1 */
-  yZ = Q_denom(my); /* (y) \cap Z */
+  if (equali1(c1)) c1 = NULL; /* can be simplified with IZ */
+  yi = ZM_gauss(my, col_ei(N,1)); /* y^-1 */
+  yZ = Q_denom(yi); /* (y) \cap Z */
   I = hnfmodid(I, yZ);
   if (!aI) return gerepileupto(av, I);
-  c1 = RgC_Rg_mul(my, c1);
+  if (typ(aI) == t_MAT) /* yi is not integral and usually larger than y */
+    aI = famat_div(aI, y);
+  else if (c1)
+    c1 = RgC_Rg_mul(yi, c1);
 END:
   if (c1) aI = ext_mul(nf, aI,c1);
   return gerepilecopy(av, mkvec2(I, aI));
 }
 
+/* I integral ZM (not HNF), G ZM, rounded Cholesky form of a weighted
+ * T2 matrix. Reduce I wrt G */
+GEN
+idealpseudored(GEN I, GEN G)
+{ return ZM_mul(I, ZM_lll(ZM_mul(G, I), 0.99, LLL_IM)); }
+
+/* Same I, G; m in I with T2(m) small */
+GEN
+idealpseudomin(GEN I, GEN G)
+{
+  GEN u = ZM_lll(ZM_mul(G, I), 0.99, LLL_IM);
+  return ZM_ZC_mul(I, gel(u,1));
+}
+/* Same I,G; irrational m in I with T2(m) small */
+GEN
+idealpseudomin_nonscalar(GEN I, GEN G)
+{
+  GEN u = ZM_lll(ZM_mul(G, I), 0.99, LLL_IM);
+  GEN m = ZM_ZC_mul(I, gel(u,1));
+  if (ZV_isscalar(m) && lg(u) > 2) m = ZM_ZC_mul(I, gel(u,2));
+  return m;
+}
+/* Same I,G; t_VEC of irrational m in I with T2(m) small */
+GEN
+idealpseudominvec(GEN I, GEN G)
+{
+  long i, j, k, n = lg(I)-1;
+  GEN x, L, b = idealpseudored(I, G);
+  L = cgetg(1 + (n*(n+1))/2, t_VEC);
+  for (i = k = 1; i <= n; i++)
+  {
+    x = gel(b,i);
+    if (!ZV_isscalar(x)) gel(L,k++) = x;
+  }
+  for (i = 2; i <= n; i++)
+  {
+    long J = minss(i, 4);
+    for (j = 1; j < J; j++)
+    {
+      x = ZC_add(gel(b,i),gel(b,j));
+      if (!ZV_isscalar(x)) gel(L,k++) = x;
+    }
+  }
+  setlg(L,k); return L;
+}
+
+GEN
+idealred_elt(GEN nf, GEN I)
+{
+  pari_sp av = avma;
+  GEN u = idealpseudomin(I, nf_get_roundG(nf));
+  return gerepileupto(av, u);
+}
+
 GEN
 idealmin(GEN nf, GEN x, GEN vdir)
 {
   pari_sp av = avma;
   GEN y, dx;
   nf = checknf(nf);
-  switch( idealtyp(&x,&y) )
+  switch( idealtyp(&x, NULL) )
   {
     case id_PRINCIPAL: return gcopy(x);
     case id_PRIME: x = pr_hnf(nf,x); break;
@@ -2965,51 +3024,54 @@ prV_lcm_capZ(GEN L)
   }
   return F;
 }
+/* v vector of prid. Return underlying list of rational primes */
+GEN
+prV_primes(GEN v)
+{
+  long i, l = lg(v);
+  GEN w = cgetg(l,t_VEC);
+  for (i=1; i<l; i++) gel(w,i) = pr_get_p(gel(v,i));
+  return ZV_sort_uniq(w);
+}
 
 /* Given a prime ideal factorization with possibly zero or negative
  * exponents, gives b such that v_p(b) = v_p(x) for all prime ideals pr | x
  * and v_pr(b) >= 0 for all other pr.
  * For optimal performance, all [anti-]uniformizers should be precomputed,
- * but no support for this yet.
- *
- * If nored, do not reduce result.
- * No garbage collecting */
+ * but no support for this yet. If nored, do not reduce result. */
 static GEN
 idealapprfact_i(GEN nf, GEN x, int nored)
 {
-  GEN z, d, L, e, e2, F;
+  GEN d = NULL, z, L, e, e2, F;
   long i, r;
-  int flagden;
+  int hasden = 0;
 
   nf = checknf(nf);
   L = gel(x,1);
   e = gel(x,2);
   F = prV_lcm_capZ(L);
-  flagden = 0;
   z = NULL; r = lg(e);
   for (i = 1; i < r; i++)
   {
     long s = signe(gel(e,i));
     GEN pi, q;
     if (!s) continue;
-    if (s < 0) flagden = 1;
+    if (s < 0) hasden = 1;
     pi = pr_uniformizer(gel(L,i), F);
     q = nfpow(nf, pi, gel(e,i));
     z = z? nfmul(nf, z, q): q;
   }
   if (!z) return gen_1;
-  if (nored || typ(z) != t_COL) return z;
-  e2 = cgetg(r, t_VEC);
-  for (i=1; i<r; i++) gel(e2,i) = addiu(gel(e,i), 1);
-  x = factorbackprime(nf, L,e2);
-  if (flagden) /* denominator */
+  if (hasden) /* denominator */
   {
     z = Q_remove_denom(z, &d);
     d = diviiexact(d, Z_ppo(d, F));
-    x = RgM_Rg_mul(x, d);
   }
-  else
-    d = NULL;
+  if (nored || typ(z) != t_COL) return d? gdiv(z, d): z;
+  e2 = cgetg(r, t_VEC);
+  for (i = 1; i < r; i++) gel(e2,i) = addiu(gel(e,i), 1);
+  x = factorbackprime(nf, L, e2);
+  if (d) x = RgM_Rg_mul(x, d);
   z = ZC_reducemodlll(z, x);
   return d? RgC_Rg_div(z,d): z;
 }
