@@ -20,6 +20,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_rnf
+
 /* eq is an rnfeq; must return a t_POL */
 GEN
 eltreltoabs(GEN eq, GEN x)
@@ -215,7 +217,7 @@ rnf_zkabs(GEN rnf)
     if (!isint1(D)) pari_err_TYPE("rnf_zkabs (old style rnf)", rnf);
   }
   v = shallowconcat(nf_get_ramified_primes(rnf_get_nf(rnf)), v);
-  return mkvec3(T, RgM_to_RgXV(M, varn(T)), ZV_sort_uniq(v));
+  return mkvec3(T, RgM_to_RgXV(M, varn(T)), ZV_sort_uniq_shallow(v));
 }
 
 static GEN
@@ -660,7 +662,7 @@ rnfidealdown(GEN rnf,GEN x)
       {
         GEN z, proj = obj_check(rnf,rnf_MAPS), ZK = gel(proj,1);
         long i, lz, l;
-        x = idealhnf(NF,x);
+        x = idealhnf_shallow(NF,x);
         if (lg(x) == 1) { set_avma(av); return cgetg(1,t_MAT); }
         z = ZM_lll(shallowconcat(ZK,x), 0.99, LLL_KER);
         lz = lg(z); l = lg(ZK);
@@ -688,7 +690,7 @@ rnfidealup(GEN rnf,GEN x)
   n = rnf_get_degree(rnf);
   bas = rnf_get_zk(rnf); bas2 = gel(bas,2);
 
-  (void)idealtyp(&x, &I); /* I is junk */
+  (void)idealtyp(&x, NULL);
   x = Q_remove_denom(x, &dx);
   x2 = idealtwoelt(nf,x);
   I = cgetg(n+1,t_VEC);
@@ -723,7 +725,7 @@ rnfidealup0(GEN rnf,GEN x, long flag)
   proj = obj_check(rnf,rnf_MAPS);
   NF = obj_check(rnf,rnf_NFABS);
 
-  (void)idealtyp(&x, &d); /* d is junk */
+  (void)idealtyp(&x, NULL);
   x2 = idealtwoelt(nf,x);
   x2 = Q_remove_denom(x2,&d);
   if (typ(gel(x2,2)) == t_COL) gel(x2,2) = ZM_ZC_mul(gel(proj,1),gel(x2,2));
@@ -1209,11 +1211,10 @@ rnfpolred(GEN nf, GEN pol, long prec)
 {
   pari_sp av = avma;
   long i, j, n, v = varn(pol);
-  GEN id, w, I, O, bnf, nfpol;
+  GEN id, w, I, O, nfpol, bnf = checkbnf_i(nf);
 
   if (typ(pol)!=t_POL) pari_err_TYPE("rnfpolred",pol);
-  bnf = nf; nf = checknf(bnf);
-  bnf = (nf == bnf)? NULL: checkbnf(bnf);
+  nf = bnf? bnf_get_nf(bnf): checknf(nf);
   if (degpol(pol) <= 1) { w = cgetg(2, t_VEC); gel(w,1) = pol_x(v); return w; }
   nfpol = nf_get_pol(nf);
 
@@ -1272,9 +1273,8 @@ GEN
 nfreduce(GEN nf, GEN x, GEN I)
 {
   pari_sp av = avma;
-  GEN aI;
   x = nf_to_scalar_or_basis(checknf(nf), x);
-  if (idealtyp(&I,&aI) != id_MAT || lg(I)==1) pari_err_TYPE("nfreduce",I);
+  if (idealtyp(&I, NULL) != id_MAT || lg(I)==1) pari_err_TYPE("nfreduce",I);
   if (typ(x) != t_COL) x = scalarcol( gmod(x, gcoeff(I,1,1)), lg(I)-1 );
   else x = reducemodinvertible(x, I);
   return gerepileupto(av, x);
