@@ -15,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_ellisogeny
+
 /* Return 1 if the point Q is a Weierstrass (2-torsion) point of the
  * curve E, return 0 otherwise */
 static long
@@ -46,6 +48,17 @@ get_isog_vars(GEN phi, long *vx, long *vy)
 /* x must be nonzero */
 INLINE long _degree(GEN x) { return typ(x)==t_POL ? degpol(x): 0; }
 
+static GEN
+RgH_eval(GEN P, GEN A, GEN B)
+{
+  if (typ(P)==t_POL)
+  {
+    if (signe(P)==0) return mkvec2(P, gen_1);
+    return mkvec2(RgX_homogenous_evalpow(P, A, B), gel(B,degpol(P)+1));
+  } else
+    return mkvec2(P, gen_1);
+}
+
 /* Given isogenies F:E' -> E and G:E'' -> E', return the composite
  * isogeny F o G:E'' -> E */
 static GEN
@@ -59,14 +72,15 @@ ellcompisog(GEN F, GEN G)
   checkellisog(G);
   get_isog_vars(F, &vx, &vy);
   Gh = gel(G,3); Gh2 = gsqr(Gh); Gh3 = gmul(Gh, Gh2);
-  F0 = polcoef(gel(F,2), 0, vy); F1 = polcoef(gel(F,2), 1, vy);
+  F0 = polcoef_i(gel(F,2), 0, vy);
+  F1 = polcoef_i(gel(F,2), 1, vy);
   d = maxss(maxss(degpol(gel(F,1)),_degree(gel(F,3))),
             maxss(_degree(F0),_degree(F1)));
   Gp = gpowers(Gh2, d);
-  f  = RgX_homogenous_evalpow(gel(F,1), gel(G,1), Gp);
-  g0 = RgX_homogenous_evalpow(F0, gel(G,1), Gp);
-  g1 = RgX_homogenous_evalpow(F1, gel(G,1), Gp);
-  h =  RgX_homogenous_evalpow(gel(F,3), gel(G,1), Gp);
+  f  = RgH_eval(gel(F,1), gel(G,1), Gp);
+  g0 = RgH_eval(F0, gel(G,1), Gp);
+  g1 = RgH_eval(F1, gel(G,1), Gp);
+  h =  RgH_eval(gel(F,3), gel(G,1), Gp);
   K = gmul(gel(h,1), Gh);
   K = RgX_normalize(RgX_div(K, RgX_gcd(K,RgX_deriv(K))));
   K2 = gsqr(K); K3 = gmul(K, K2);
@@ -100,6 +114,13 @@ divy(GEN P0, GEN P1, GEN Q, GEN T, long vy)
   return P2;
 }
 
+static GEN
+QXQH_eval(GEN P, GEN A, GEN B, GEN T)
+{
+  if (signe(P)==0) return mkvec2(P, pol_1(varn(P)));
+  return mkvec2(QXQX_homogenous_evalpow(P, A, B, T), gel(B,degpol(P)+1));
+}
+
 static GEN
 ellnfcompisog(GEN nf, GEN F, GEN G)
 {
@@ -114,21 +135,20 @@ ellnfcompisog(GEN nf, GEN F, GEN G)
   if (!nf) return ellcompisog(F, G);
   T = nf_get_pol(nf);
   timer_start(&ti);
-  checkellisog(F);
-  checkellisog(G);
-  F = liftpol(F); G = liftpol(G);
+  checkellisog(F); F = liftpol_shallow(F);
+  checkellisog(G); G = liftpol_shallow(G);
   get_isog_vars(F, &vx, &vy);
   Gh = to_RgX(gel(G,3),vx); Gh2 = QXQX_sqr(Gh, T); Gh3 = QXQX_mul(Gh, Gh2, T);
-  F0 = to_RgX(polcoef(gel(F,2), 0, vy), vx);
-  F1 = to_RgX(polcoef(gel(F,2), 1, vy), vx);
-  G0 = to_RgX(polcoef(gel(G,2), 0, vy), vx);
-  G1 = to_RgX(polcoef(gel(G,2), 1, vy), vx);
+  F0 = to_RgX(polcoef_i(gel(F,2), 0, vy), vx);
+  F1 = to_RgX(polcoef_i(gel(F,2), 1, vy), vx);
+  G0 = to_RgX(polcoef_i(gel(G,2), 0, vy), vx);
+  G1 = to_RgX(polcoef_i(gel(G,2), 1, vy), vx);
   d = maxss(maxss(degpol(gel(F,1)),degpol(gel(F,3))),maxss(degpol(F0),degpol(F1)));
   Gp = QXQX_powers(Gh2, d, T);
-  f  = QXQX_homogenous_evalpow(to_RgX(gel(F,1),vx), gel(G,1), Gp, T);
-  g0 = QXQX_homogenous_evalpow(F0, to_RgX(gel(G,1),vx), Gp, T);
-  g1 = QXQX_homogenous_evalpow(F1, to_RgX(gel(G,1),vx), Gp, T);
-  h  = QXQX_homogenous_evalpow(to_RgX(gel(F,3),vx), gel(G,1), Gp, T);
+  f  = QXQH_eval(to_RgX(gel(F,1),vx), gel(G,1), Gp, T);
+  g0 = QXQH_eval(F0, to_RgX(gel(G,1),vx), Gp, T);
+  g1 = QXQH_eval(F1, to_RgX(gel(G,1),vx), Gp, T);
+  h  = QXQH_eval(to_RgX(gel(F,3),vx), gel(G,1), Gp, T);
   K = Q_remove_denom(QXQX_mul(to_RgX(gel(h,1),vx), Gh, T), NULL);
   K = RgXQX_div(K, nfgcd(K, RgX_deriv(K), T, NULL), T);
   K = RgX_normalize(K);
@@ -254,7 +274,6 @@ isogeny_from_kernel_point(GEN E, GEN P, int only_image, long vx, long vy)
   pari_sp av = avma;
   GEN isog, EE, f, g, h, h2, h3;
   GEN Q = P, t = gen_0, w = gen_0;
-  long c;
   if (!oncurve(E,P))
     pari_err_DOMAIN("isogeny_from_kernel_point", "point", "not on", E, P);
   if (ell_is_inf(P))
@@ -263,7 +282,7 @@ isogeny_from_kernel_point(GEN E, GEN P, int only_image, long vx, long vy)
     return mkvec2(E, isog_identity(vx,vy));
   }
 
-  isog = NULL; c = 1;
+  isog = NULL;
   for (;;)
   {
     GEN tQ, xQ = gel(Q,1), uQ = ec_2divpol_evalx(E, xQ);
@@ -281,7 +300,6 @@ isogeny_from_kernel_point(GEN E, GEN P, int only_image, long vx, long vy)
     if (stop) break;
 
     Q = elladd(E, P, Q);
-    ++c;
     /* IF x([c]P) = x([c-1]P) THEN [c]P = -[c-1]P and [2c-1]P = 0 */
     if (gequal(gel(Q,1), xQ)) break;
     if (gc_needed(av,1))
@@ -511,7 +529,7 @@ non_two_torsion_ordinate_char2(GEN E, GEN h, GEN x, GEN y)
   m = degpol(h);
   p1 = RgX_coeff(h, m-1); /* first power sum */
 
-  t1 = gmul(gadd(gmul(a1, p1), gmulgs(a3, m)), RgX_mul(h,h2));
+  t1 = gmul(gadd(gmul(a1, p1), gmulgu(a3, m)), RgX_mul(h,h2));
 
   t2 = gmul(a1, gadd(gmul(a1, gadd(y, psi2)), RgX_add(RgX_Rg_add(RgX_sqr(x), a4), t)));
   t2 = gmul(t2, gmul(dh, h2));
@@ -660,8 +678,8 @@ static GEN
 isogeny_a4a6(GEN E)
 {
   GEN a1 = ell_get_a1(E), a3 = ell_get_a3(E), b2 = ell_get_b2(E);
-  retmkvec3(deg1pol(gen_1, gdivgs(b2, 12), 0),
-            deg1pol(gdivgs(a1,2), deg1pol(gen_1, gdivgs(a3,2), 1), 0),
+  retmkvec3(deg1pol(gen_1, gdivgu(b2, 12), 0),
+            deg1pol(gdivgu(a1,2), deg1pol(gen_1, gdivgu(a3,2), 1), 0),
             pol_1(0));
 }
 
@@ -671,7 +689,7 @@ invisogeny_a4a6(GEN E)
   GEN a1 = ell_get_a1(E), a3 = ell_get_a3(E), b2 = ell_get_b2(E);
   retmkvec3(deg1pol(gen_1, gdivgs(b2, -12), 0),
             deg1pol(gdivgs(a1,-2),
-              deg1pol(gen_1, gadd(gdivgs(a3,-2), gdivgs(gmul(b2,a1), 24)), 1), 0),
+              deg1pol(gen_1, gadd(gdivgs(a3,-2), gdivgu(gmul(b2,a1), 24)), 1), 0),
             pol_1(0));
 }
 
@@ -717,7 +735,7 @@ static GEN
 a4a6_divpol(GEN a4, GEN a6, long n)
 {
   if (n == 2) return mkpoln(4, gen_1, gen_0, a4, a6);
-  return mkpoln(5, utoi(3), gen_0, gmulgs(a4,6) , gmulgs(a6,12),
+  return mkpoln(5, utoi(3), gen_0, gmulgu(a4,6) , gmulgu(a6,12),
                    gneg(gsqr(a4)));
 }
 
@@ -750,7 +768,7 @@ static GEN
 corr(GEN c4, GEN c6)
 {
   GEN c62 = gmul2n(c6, 1);
-  return gadd(gdiv(gsqr(c4), c62), gdiv(c62, gmulgs(c4,3)));
+  return gadd(gdiv(gsqr(c4), c62), gdiv(c62, gmulgu(c4,3)));
 }
 
 static GEN
@@ -760,25 +778,25 @@ elkies98(GEN a4, GEN a6, long l, GEN s, GEN a4t, GEN a6t)
   long i, n, d;
   d = l == 2 ? 1 : l>>1;
   C = cgetg(d+1, t_VEC);
-  gel(C, 1) = gdivgs(gsub(a4, a4t), 5);
+  gel(C, 1) = gdivgu(gsub(a4, a4t), 5);
   if (d >= 2)
-    gel(C, 2) = gdivgs(gsub(a6, a6t), 7);
+    gel(C, 2) = gdivgu(gsub(a6, a6t), 7);
   if (d >= 3)
-    gel(C, 3) = gdivgs(gsub(gsqr(gel(C, 1)), gmul(a4, gel(C, 1))), 3);
+    gel(C, 3) = gdivgu(gsub(gsqr(gel(C, 1)), gmul(a4, gel(C, 1))), 3);
   for (n = 3; n < d; ++n)
   {
     GEN s = gen_0;
     for (i = 1; i < n; i++)
       s = gadd(s, gmul(gel(C, i), gel(C, n-i)));
-    gel(C, n+1) = gdivgs(gsub(gsub(gmulsg(3, s), gmul(gmulsg((2*n-1)*(n-1), a4), gel(C, n-1))), gmul(gmulsg((2*n-2)*(n-2), a6), gel(C, n-2))), (n-1)*(2*n+5));
+    gel(C, n+1) = gdivgu(gsub(gsub(gmulsg(3, s), gmul(gmulsg((2*n-1)*(n-1), a4), gel(C, n-1))), gmul(gmulsg((2*n-2)*(n-2), a6), gel(C, n-2))), (n-1)*(2*n+5));
   }
   P = cgetg(d+2, t_VEC);
   gel(P, 1 + 0) = stoi(d);
   gel(P, 1 + 1) = s;
   if (d >= 2)
-    gel(P, 1 + 2) = gdivgs(gsub(gel(C, 1), gmulgs(gmulsg(2, a4), d)), 6);
+    gel(P, 1 + 2) = gdivgu(gsub(gel(C, 1), gmulgu(a4, 2*d)), 6);
   for (n = 2; n < d; ++n)
-    gel(P, 1 + n+1) = gdivgs(gsub(gsub(gel(C, n), gmul(gmulsg(4*n-2, a4), gel(P, 1+n-1))), gmul(gmulsg(4*n-4, a6), gel(P, 1+n-2))), 4*n+2);
+    gel(P, 1 + n+1) = gdivgu(gsub(gsub(gel(C, n), gmul(gmulsg(4*n-2, a4), gel(P, 1+n-1))), gmul(gmulsg(4*n-4, a6), gel(P, 1+n-2))), 4*n+2);
   S = cgetg(d+3, t_POL);
   S[1] = evalsigne(1) | evalvarn(0);
   gel(S, 2 + d - 0) = gen_1;
@@ -819,23 +837,17 @@ ellisog_by_jt(GEN c4, GEN c6, GEN jt, GEN jtp, GEN s0, long n, long flag)
   }
 }
 
-/*
-Based on
-RENE SCHOOF
-Counting points on elliptic curves over finite fields
-Journal de Theorie des Nombres de Bordeaux,
-tome 7, no 1 (1995), p. 219-254.
-<http://www.numdam.org/item?id=JTNB_1995__7_1_219_0>
-*/
-
+/* RENE SCHOOF, Counting points on elliptic curves over finite fields,
+ * Journal de Theorie des Nombres de Bordeaux, tome 7, no 1 (1995), p. 219-254.
+ * http://www.numdam.org/item?id=JTNB_1995__7_1_219_0 */
 static GEN
 ellisog_by_j(GEN e, GEN jt, long n, GEN P, long flag)
 {
   pari_sp av = avma;
   GEN c4  = gel(e,1), c6 = gel(e, 2), j = gel(e, 3);
-  GEN Px = deriv(P, 0), Py = deriv(P, 1);
+  GEN Px = RgX_deriv(P), Py = RgXY_derivx(P);
   GEN Pxj = RgXY_eval(Px, j, jt), Pyj = RgXY_eval(Py, j, jt);
-  GEN Pxx  = deriv(Px, 0), Pxy = deriv(Py, 0), Pyy = deriv(Py, 1);
+  GEN Pxx  = RgX_deriv(Px), Pxy = RgX_deriv(Py), Pyy = RgXY_derivx(Py);
   GEN Pxxj = RgXY_eval(Pxx,j,jt);
   GEN Pxyj = RgXY_eval(Pxy,j,jt);
   GEN Pyyj = RgXY_eval(Pyy,j,jt);
@@ -859,8 +871,7 @@ ellisograph_iso(GEN nf, GEN e, ulong p, GEN P, GEN oj, long flag)
   R = nfroots(nf,oj ? RgX_div_by_X_x(Pj, oj, NULL):Pj);
   r = lg(R);
   V = cgetg(r, t_VEC);
-  for (i=1; i < r; i++)
-    gel(V, i) = ellisog_by_j(e, gel(R, i), p, P, flag);
+  for (i=1; i < r; i++) gel(V, i) = ellisog_by_j(e, gel(R, i), p, P, flag);
   return V;
 }
 
@@ -905,13 +916,13 @@ isograph_p(GEN nf, GEN e, ulong p, GEN P, long flag)
 }
 
 static GEN
-get_polmodular(ulong p)
-{ return p > 3 ? polmodular_ZXX(p,0,0,1): NULL; }
+get_polmodular(ulong p, long vx, long vy)
+{ return p > 3 ? polmodular_ZXX(p,0,vx,vy): NULL; }
 static GEN
 ellisograph_p(GEN nf, GEN E, ulong p, long flag)
 {
   GEN e = ellisograph_a4a6(E, flag);
-  GEN P = get_polmodular(p);
+  GEN P = get_polmodular(p, 0, 1);
   return isograph_p(nf, e, p, P, flag);
 }
 
@@ -920,8 +931,7 @@ etree_nbnodes(GEN T)
 {
   GEN F = gel(T,2);
   long n = 1, i, l = lg(F);
-  for (i = 1; i < l; i++)
-    n += etree_nbnodes(gel(F, i));
+  for (i = 1; i < l; i++) n += etree_nbnodes(gel(F, i));
   return n;
 }
 
@@ -989,8 +999,7 @@ etree_distmat(GEN T)
 {
   long i, n = etree_nbnodes(T);
   GEN M = cgetg(n+1, t_MAT);
-  for(i = 1; i <= n; i++)
-    gel(M,i) = cgetg(n+1, t_VECSMALL);
+  for(i = 1; i <= n; i++) gel(M,i) = cgetg(n+1, t_VECSMALL);
   (void)etree_distmatr(T, M, 1);
   return M;
 }
@@ -1013,7 +1022,8 @@ static GEN
 isomatdbl(GEN nf, GEN L, GEN M, ulong p, GEN T2, long flag)
 {
   long i, j, n = lg(L) -1;
-  GEN P = get_polmodular(p), V = cgetg(2*n+1, t_VEC), N = cgetg(2*n+1, t_MAT);
+  GEN P = get_polmodular(p, 0, 1);
+  GEN V = cgetg(2*n+1, t_VEC), N = cgetg(2*n+1, t_MAT);
   for (i=1; i <= n; i++)
   {
     GEN F, E, e = gel(L,i);
@@ -1120,15 +1130,10 @@ mkisomatdbl(ulong p, GEN T, ulong p2, GEN T2, long flag)
   return isomatdbl(NULL, gel(v,1), gel(v,2), p2, T2, flag);
 }
 
-/*
-See
-M.A Kenku
-On the number of Q-isomorphism classes of elliptic curves in each Q-isogeny class
-Journal of Number Theory
-Volume 15, Issue 2, October 1982, Pages 199-202
-http://www.sciencedirect.com/science/article/pii/0022314X82900257
-*/
-
+/*See M.A Kenku, On the number of Q-isomorphism classes of elliptic curves in
+ * each Q-isogeny class, Journal of Number Theory Volume 15, Issue 2,
+ * October 1982, pp 199-202,
+ * http://www.sciencedirect.com/science/article/pii/0022314X82900257 */
 enum { _2 = 1, _3 = 2, _5 = 4, _7 = 8, _13 = 16 };
 static ulong
 ellQ_goodl(GEN E)
@@ -1145,7 +1150,7 @@ ellQ_goodl(GEN E)
     if (umodiu(disc,p)==0) i--;
     else
     {
-      long t = ellap_CM_fast(E, p, CM), D = t*t-4*p;
+      long t = ellap_CM_fast(E, p, CM), D = t*t - 4*p;
       if (t%2) mask &= ~_2;
       if ((mask & _3) && kross(D,3)==-1)  mask &= ~_3;
       if ((mask & _5) && kross(D,5)==-1)  mask &= ~_5;
@@ -1163,22 +1168,22 @@ ellQ_goodl_l(GEN E, long l)
   long i;
   GEN disc = ell_get_disc(E);
   pari_sp av = avma;
-  u_forprime_init(&T, 17UL,ULONG_MAX);
-  for(i=1; i<=20; i++)
+  u_forprime_init(&T, 17UL, ULONG_MAX);
+  for (i=1; i<=20; i++)
   {
     ulong p = u_forprime_next(&T);
     if (umodiu(disc,p)==0) { i--; continue; }
     else
     {
-      long t = itos(ellap(E, utoi(p)));
+      long t = itos(ellap(E, utoipos(p)));
       if (l==2)
       {
-        if (t%2==1) return 0;
+        if (odd(t)) return 0;
       }
       else
       {
-        long D = t*t-4*p;
-        if (kross(D,l)==-1) return 0;
+        long D = t*t - 4*p;
+        if (kross(D,l) == -1) return 0;
       }
       set_avma(av);
     }
@@ -1190,17 +1195,15 @@ static ulong
 ellnf_goodl_l(GEN E, GEN v)
 {
   forprime_t T;
-  long i;
-  GEN nf = ellnf_get_nf(E);
-  GEN disc = ell_get_disc(E);
-  long lv = lg(v);
+  GEN nf = ellnf_get_nf(E), disc = ell_get_disc(E);
+  long i, lv = lg(v);
   ulong w = 0UL;
   pari_sp av = avma;
   u_forprime_init(&T, 17UL,ULONG_MAX);
-  for(i=1; i<=20; i++)
+  for (i = 1; i <= 20; i++)
   {
     ulong p = u_forprime_next(&T);
-    GEN pr = idealprimedec(nf, utoi(p));
+    GEN pr = idealprimedec(nf, utoipos(p));
     long j, k, lv = lg(v), g = lg(pr)-1;
     for (j=1; j<=g; j++)
     {
@@ -1214,11 +1217,11 @@ ellnf_goodl_l(GEN E, GEN v)
           long l = v[k];
           if (l==2)
           {
-            if (t%2==1) w |= 1<<(k-1);
+            if (odd(t)) w |= 1<<(k-1);
           }
           else
           {
-            GEN D = subii(sqrs(t),shifti(pr_norm(prj),2));
+            GEN D = subii(sqrs(t), shifti(pr_norm(prj),2));
             if (krois(D,l)==-1) w |= 1<<(k-1);
           }
         }
@@ -1231,18 +1234,7 @@ ellnf_goodl_l(GEN E, GEN v)
 
 static GEN
 ellnf_charpoly(GEN E, GEN pr)
-{
-  return deg2pol_shallow(gen_1, negi(ellap(E,pr)), pr_norm(pr), 0);
-}
-
-static GEN
-RgX_homogenize(GEN P, long v)
-{
-  GEN Q = leafcopy(P);
-  long i, l = lg(P), d = degpol(P);
-  for (i = 2; i < l; i++) gel(Q,i) = monomial(gel(Q,i), d--, v);
-  return Q;
-}
+{ return deg2pol_shallow(gen_1, negi(ellap(E,pr)), pr_norm(pr), 0); }
 
 static GEN
 starlaw(GEN p, GEN q)
@@ -1254,58 +1246,46 @@ starlaw(GEN p, GEN q)
 static GEN
 startor(GEN p, long r)
 {
-  GEN xr = pol_xn(r, 0);
-  GEN psir = gsub(xr, gen_1);
+  GEN xr = pol_xn(r, 0), psir = gsub(xr, gen_1);
   return gsubstpol(starlaw(p, psir),xr,pol_x(0));
 }
 
 static GEN
 ellnf_get_degree(GEN E, GEN p)
 {
-  GEN nf = ellnf_get_nf(E);
-  long d = nf_get_degree(nf);
-  GEN dec = idealprimedec(nf, p);
-  long i, l = lg(dec), k;
+  GEN nf = ellnf_get_nf(E), dec = idealprimedec(nf, p);
+  long d = nf_get_degree(nf), l = lg(dec), i, k;
   GEN R, starl = deg1pol_shallow(gen_1, gen_m1, 0);
-  for(i=1; i < l; i++)
+  for (i=1; i < l; i++)
   {
-    GEN pr = gel(dec,i);
-    GEN q = ellnf_charpoly(E, pr);
+    GEN pr = gel(dec,i), q = ellnf_charpoly(E, pr);
     starl = starlaw(starl, startor(q, 12*pr_get_e(pr)));
   }
-  R = p;
-  for(k=0; 2*k<=d; k++)
-    R = mulii(R, poleval(starl,powiu(p,12*k)));
+  for (k = 0, R = p; 2*k <= d; k++) R = mulii(R, poleval(starl, powiu(p,12*k)));
   return R;
 }
 
-/*
-Based on a GP script by Nicolas Billerey itself
-based on Th\'eor\`emes 2.4 and 2.8 of the following article:
-N. Billerey, Crit\`eres d'irr\'eductibilit\'e pour les
-repr\'esentations des courbes elliptiques,
-Int. J. Number Theory 7 (2011), no. 4, 1001-1032.
-*/
-
+/* Based on a GP script by Nicolas Billerey and Theorems 2.4 and 2.8 of
+ * N. Billerey, Criteres d'irreductibilite pour les representations des
+ * courbes elliptiques, Int. J. Number Theory 7 (2011), no. 4, 1001-1032. */
 static GEN
 ellnf_prime_degree(GEN E)
 {
   forprime_t T;
   long i;
-  GEN nf = ellnf_get_nf(E);
-  GEN disc = ell_get_disc(E);
-  GEN P, B = gen_0, rB;
-  GEN nor = nfnorm(nf, disc);
-  GEN bad = mulii(mulii(denom_i(nor),numer_i(nor)),nf_get_disc(nf));
+  GEN nf = ellnf_get_nf(E), N = nfnorm(nf, ell_get_disc(E)), B = gen_0, P;
+  GEN bad = mulii(typ(N) == t_FRAC? mulii(gel(N,1),gel(N,2)): N,
+                  nf_get_disc(nf));
   u_forprime_init(&T, 5UL,ULONG_MAX);
-  for(i=1; i<=20; i++)
+  for (i = 1; i <= 20; i++)
   {
     ulong p = u_forprime_next(&T);
-    if (dvdiu(bad, p)) {i--; continue;}
-    B = gcdii(B, ellnf_get_degree(E, utoi(p)));
-    if (Z_issquareall(B,&rB)) B=rB;
+    GEN r;
+    if (dvdiu(bad, p)) { i--; continue; }
+    B = gcdii(B, ellnf_get_degree(E, utoipos(p)));
+    if (Z_issquareall(B, &r)) B = r;
   }
-  if (signe(B)==0) pari_err_DOMAIN("ellisomat", "E","has",strtoGENstr("CM"),E);
+  if (!signe(B)) pari_err_DOMAIN("ellisomat", "E","has",strtoGENstr("CM"),E);
   P = vec_to_vecsmall(gel(Z_factor(B),1));
   return shallowextract(P, utoi(ellnf_goodl_l(E, P)));
 }
@@ -1415,12 +1395,10 @@ ellnf_isocrv(GEN nf, GEN E, GEN v, GEN PE, long flag)
 }
 
 static long
-nfispower(GEN nf, long d, GEN a, GEN b)
+nfispower_quo(GEN nf, long d, GEN a, GEN b)
 {
-  GEN N;
   if (gequal(a,b)) return 1;
-  N = nfroots(nf, gsub(monomial(b, d, 0), monomial(a,0,0)));
-  return lg(N) > 1;
+  return nfispower(nf, nfdiv(nf, a, b), d, NULL);
 }
 
 static long
@@ -1429,10 +1407,10 @@ isomat_eq(GEN nf, GEN e1, GEN e2)
   if (gequal(e1,e2)) return 1;
   if (!gequal(gel(e1,3), gel(e2,3))) return 0;
   if (gequal0(gel(e1,3)))
-    return nfispower(nf,6,gel(e1,2),gel(e2,2));
+    return nfispower_quo(nf,6,gel(e1,2),gel(e2,2));
   if (gequalgs(gel(e1,3),1728))
-    return nfispower(nf,4,gel(e1,1),gel(e2,1));
-  return nfispower(nf,2,gmul(gel(e1,1),gel(e2,2)),gmul(gel(e1,2),gel(e2,1)));
+    return nfispower_quo(nf,4,gel(e1,1),gel(e2,1));
+  return nfispower_quo(nf,2,gmul(gel(e1,1),gel(e2,2)),gmul(gel(e1,2),gel(e2,1)));
 }
 
 static long
@@ -1455,11 +1433,11 @@ isomat_perm(GEN nf, GEN E, GEN L)
 }
 
 static GEN
-ellnf_modpoly(GEN v)
+ellnf_modpoly(GEN v, long vx, long vy)
 {
   long i, l = lg(v);
   GEN P = cgetg(l, t_VEC);
-  for(i = 1; i < l; i++) gel(P, i) = get_polmodular(v[i]);
+  for(i = 1; i < l; i++) gel(P, i) = get_polmodular(v[i],vx,vy);
   return P;
 }
 
@@ -1468,7 +1446,9 @@ ellnf_isomat(GEN E, long flag)
 {
   GEN nf = ellnf_get_nf(E);
   GEN v = ellnf_prime_degree(E);
-  GEN P = ellnf_modpoly(v);
+  long vy = fetch_var_higher();
+  long vx = fetch_var_higher();
+  GEN P = ellnf_modpoly(v,vx,vy);
   GEN LM = ellnf_isocrv(nf, E, v, P, flag), L = gel(LM,1), M = gel(LM,2);
   long i, l = lg(L);
   GEN R = cgetg(l, t_MAT);
@@ -1482,6 +1462,7 @@ ellnf_isomat(GEN E, long flag)
     GEN r = isomat_perm(nf, L, LLi);
     gel(R,i) = vecpermute(Mi, r);
   }
+  delete_var(); delete_var();
   return mkvec2(L, R);
 }
 
@@ -1506,7 +1487,7 @@ ellisomat(GEN E, long p, long flag)
   GEN r = NULL, nf = NULL;
   long good = 1;
   if (flag < 0 || flag > 1) pari_err_FLAG("ellisomat");
-  if (p < 0) pari_err_PRIME("ellisomat", utoi(p));
+  if (p < 0) pari_err_PRIME("ellisomat", stoi(p));
   if (p == 1) { flag = 1; p = 0; } /* for backward compatibility */
   checkell(E);
   switch(ell_get_type(E))
@@ -1517,6 +1498,10 @@ ellisomat(GEN E, long p, long flag)
     case t_ELL_NF:
       if (p) good = ellnf_goodl_l(E, mkvecsmall(p));
       nf = ellnf_get_nf(E);
+      if (nf_get_varn(nf) <= 0)
+        pari_err_PRIORITY("ellisomat", nf_get_pol(nf), "<=", 0);
+      if (flag==0 && nf_get_varn(nf) <= 1)
+        pari_err_PRIORITY("ellisomat", nf_get_pol(nf), "<=", 1);
       break;
     default: pari_err_TYPE("ellisomat",E);
   }
@@ -1596,7 +1581,7 @@ ellweilcurve(GEN E, GEN *ms)
   for (i = 1; i < l; i++) obj_free(gel(vE,i));
   vE = mkvec2(vE, vL);
   if (!ms) return gerepilecopy(av, vE);
-  *ms = Wx; gerepileall(av, 2, &vE, ms); return vE;
+  *ms = Wx; return gc_all(av, 2, &vE, ms);
 }
 
 GEN
