@@ -88,17 +88,16 @@ tors(GEN e, long k, GEN p, GEN q, GEN v)
   return r;
 }
 
-/* Finds a multiplicative upper bound for #E_tor; assume integral model */
+/* Finds a multiplicative upper bound for #E_tor (p-Sylow if p != 0);
+ * assume integral model */
 static long
 torsbound(GEN e, ulong p)
 {
   GEN D = ell_get_disc(e);
   pari_sp av = avma, av2;
-  long m, b, bold, nb;
+  long m, b, bold, nb, CM = ellQ_get_CM(e);
   forprime_t S;
-  long CM = ellQ_get_CM(e);
-  nb = expi(D) >> 3;
-  /* nb = number of primes to try ~ 1 prime every 8 bits in D */
+  nb = expi(D) >> 3; /* number of primes to try ~ 1 prime every 8 bits in D */
   switch (p)
   {
     case 0: b = 5040; break;
@@ -162,8 +161,9 @@ t2points(GEN E, GEN *f2)
   return v;
 }
 
+/* psylow = 0 or prime (return p-Sylow subgroup) */
 static GEN
-elltors_divpol(GEN E, long psylow)
+ellQtors(GEN E, long psylow)
 {
   GEN T2 = NULL, p, P, Q, v;
   long v2, r2, B;
@@ -185,7 +185,7 @@ elltors_divpol(GEN E, long psylow)
       default: r2 = 2; v2--; break; /* 3 */
     }
     if (v2) gel(p,2) = gel(T2,1);
-    /* f = f_2 */
+    /* f = f2 */
     if (v2 > 1) { gel(p,4) = tpoint(E,4, &f); if (!gel(p,4)) v2 = 1; }
     /* if (v2>1) now f = f4 */
     if (v2 > 2) { gel(p,8) = tpoint(E,8, &f); if (!gel(p,8)) v2 = 2; }
@@ -257,6 +257,7 @@ primedec_deg1(GEN K, GEN p)
 }
 
 /* Bound for the elementary divisors of the torsion group of elliptic curve
+ * (p-Sylow if psylow = p is not 0)
  * Reduce the curve modulo some small good primes */
 static GEN
 nftorsbound(GEN E, ulong psylow)
@@ -316,108 +317,129 @@ nftorsbound(GEN E, ulong psylow)
   if (psylow)
   {
     B1 = powuu(psylow, Z_lval(B1, psylow));
-    B2 = powuu(psylow, Z_lval(B1, psylow));
+    B2 = powuu(psylow, Z_lval(B2, psylow));
   }
   return mkvec2(B1,B2);
 }
 
-/* Checks whether the point P is divisible by n in E(K), where xn is
- * [phi_n, psi_n^2]
- * If true, returns a point Q such that nQ = P or -P. Else, returns NULL */
+/* Checks whether the point P != oo is divisible by n in E(K), where xn is
+ * [phi_n, psi_n^2]. If so, returns a point Q such that nQ = P or -P.
+ * Else, returns NULL */
 static GEN
 ellnfis_divisible_by(GEN E, GEN K, GEN P, GEN xn)
 {
-  GEN r, x = gel(P,1);
-  long i, l;
-  if (ell_is_inf(P)) return P;
-  r = nfroots(K, RgX_sub(RgX_Rg_mul(gel(xn,2), x), gel(xn,1)));
-  l = lg(r);
-  for(i=1; i<l; i++)
+  GEN R = nfroots(K, RgX_sub(RgX_Rg_mul(gel(xn,2), gel(P,1)), gel(xn,1)));
+  long i, l = lg(R);
+  for (i = 1; i < l; i++)
   {
-    GEN a = gel(r,i), y = ellordinate(E,a,0);
-    if (lg(y) != 1) return mkvec2(a, gel(y,1));
+    GEN x = gel(R,i), y = ellordinate(E,x,0);
+    if (lg(y) != 1) return mkvec2(x, gel(y,1));
   }
   return NULL;
 }
 
-/* w a variable number of highest priority */
-static long
-ellisdivisible_i(GEN E, GEN P, GEN n, long w, GEN *pQ)
+/* Q is not the point at infinity; do we have Q = [n]Q' on curve C/K ?
+ * return P such that [n] P = \pm Q or NULL if none exist */
+static GEN
+ellnfis_divisible_by_i(GEN C, GEN K, GEN Q, GEN n, long v)
 {
-  GEN xP, R, K = NULL, N = NULL;
-  long i, l;
-  checkell(E);
+  GEN xn;
+  if (!isprimepower(absi_shallow(n), NULL))
+  {
+    GEN f = absZ_factor(n), P = gel(f,1), E = gel(f,2);
+    long i, l = lg(P);
+    for (i = 1; i < l; i++)
+    {
+      GEN q = powii(gel(P,i), gel(E,i));
+      xn = ellxn(C, itou(q), v);
+      if (!(Q = ellnfis_divisible_by(C, K, Q, xn))) return 0;
+    }
+    return Q;
+  }
+  xn = ellxn(C, itou(n), v);
+  return ellnfis_divisible_by(C, K, Q, xn);
+}
+
+static GEN ellorder_nf(GEN E, GEN P, GEN B);
+long
+ellisdivisible(GEN E, GEN Q, GEN n, GEN *pQ)
+{
+  pari_sp av = avma;
+  GEN P = NULL, N = NULL, K = NULL;
+  long v;
+
+  checkell(E); checkellpt(Q);
   switch(ell_get_type(E))
   {
     case t_ELL_Q: break;
     case t_ELL_NF: K = ellnf_get_nf(E); break;
     default: pari_err_TYPE("ellisdivisible",E);
   }
-  checkellpt(P);
+  if (ell_is_inf(Q))
+  {
+    if (pQ) *pQ = ellinf();
+    return 1;
+  }
   switch(typ(n))
   {
     case t_INT:
-      N = n;
-      if (!isprime(absi_shallow(n)))
-      {
-        GEN f = absZ_factor(n), LP = gel(f,1), LE = gel(f,2);
-        l = lg(LP);
-        for (i = 1; i < l; i++)
-        {
-          long j, e = itos(gel(LE,i));
-          GEN xp = ellxn(E,itos(gel(LP,i)), w);
-          for (j = 1; j <= e; j++)
-            if (!ellisdivisible(E, P, xp, &P)) return 0;
-        }
-        if (pQ) *pQ = signe(n) < 0? ellneg(E, P): P;
-        return 1;
+      if (!signe(n)) return 0;
+      if (is_bigint(n))
+      { /* only possible if P is torsion */
+        GEN x, o, ex = cyc_get_expo(abgrp_get_cyc(elltors(E)));
+        if (equali1(ex)) return gc_long(av, 0);
+        o = K? ellorder_nf(E, Q, ex): ellorder(E, Q, NULL);
+        if (isintzero(o)) return gc_long(av, 0);
+        x = Z_ppo(n, o);
+        Q = ellmul(E, Q, Fp_inv(x, o));
+        n = diviiexact(n, x); /* all primes dividing n divide o = ord(Q) | ex */
+        if (!dvdii(diviiexact(ex,o), n)) return gc_long(av, 0);
+        if (is_bigint(n)) pari_err_IMPL("ellisdivisible for huge torsion");
+        if (!ellisdivisible(E, Q, n, pQ)) return gc_long(av, 0);
+        if (!pQ) return gc_long(av, 1);
+        *pQ = gerepilecopy(av, *pQ); return 1;
       }
-      n = ellxn(E, itou(n), w);
-      break;
+      if (!K)
+      { /* could use elltors instead of a multiple ? */
+        ulong nn = itou(n), n2 = u_ppo(nn, torsbound(E, 0));
+        if (n2 > 1)
+        { /* n2 coprime to torsion */
+          if (!(Q = ellQ_isdivisible(E, Q, n2))) return 0;
+          if (signe(n) < 0) Q = ellneg(E, Q);
+          if (nn == n2)
+          {
+            if (pQ) *pQ = Q;
+            return 1;
+          }
+          /* we may have changed n into |n/n2| (and Q in -Q if n < 0) */
+          n = utoipos(nn/n2);
+        }
+      }
+      v = fetch_var_higher();
+      P = ellnfis_divisible_by_i(E, K, Q, n, v);
+      delete_var(); N = n; break;
     case t_VEC:
-      if (lg(n) == 3 && typ(gel(n,1)) == t_POL && typ(gel(n,2)) == t_POL) break;
+      if (lg(n) == 3 && typ(gel(n,1)) == t_POL && typ(gel(n,2)) == t_POL)
+      { /* ellxn */
+        long d, d2 = degpol(gel(n,1));
+        if (d2 < 0) return gc_long(av, 0);
+        if (!uissquareall(d2,(ulong*)&d)) pari_err_TYPE("ellisdivisible",n);
+        P = ellnfis_divisible_by(E, K, Q, n);
+        N = utoi(d); break;
+      } /* fall through */
     default:
       pari_err_TYPE("ellisdivisible",n);
       break;
   }
-  if (ell_is_inf(P)) { if (pQ) *pQ = ellinf(); return 1; }
-  if (!N)
-  {
-    long d, d2 = degpol(gel(n,1));
-    if (d2 < 0)
-      N = gen_0;
-    else
-    {
-      if (!uissquareall(d2,(ulong*)&d)) pari_err_TYPE("ellisdivisible",n);
-      N = stoi(d);
-    }
-  }
-  if (!signe(N)) return 0;
-  xP = gel(P,1);
-  R = nfroots(K, RgX_sub(RgX_Rg_mul(gel(n,2), xP), gel(n,1)));
-  l = lg(R);
-  for(i = 1; i < l; i++)
-  {
-    GEN Q,y, x = gel(R,i), a = ellordinate(E,x,0);
-    if (lg(a) == 1) continue;
-    y = gel(a,1);
-    Q = mkvec2(x,y);
-    if (!gequal(P,ellmul(E,Q,N))) Q = ellneg(E,Q); /* nQ = -P */
-    if (pQ) *pQ = Q;
-    return 1;
-  }
-  return 0;
-}
-long
-ellisdivisible(GEN E, GEN P, GEN n, GEN *pQ)
-{
-  pari_sp av = avma;
-  long w = fetch_var_higher(), t = ellisdivisible_i(E, P, n, w, pQ);
-  delete_var();
-  if (!t) return gc_long(av, 0);
-  if (pQ) *pQ = gerepilecopy(av, *pQ); else set_avma(av);
-  return 1;
+  if (!P) return gc_long(av, 0);
+  if (!pQ) return gc_long(av, 1);
+  if (gequal(Q, ellmul(E, P, N)))
+    P = gerepilecopy(av, P);
+  else
+    P = gerepileupto(av, ellneg(E, P));
+  *pQ = P; return 1;
 }
+
 /* 2-torsion point of abscissa x */
 static GEN
 tor2(GEN E, GEN x) { return mkvec2(x, gmul2n(gneg(ec_h_evalx(E,x)), -1)); }
@@ -444,9 +466,9 @@ ellnftorsprimary(GEN E, long p, long N1, long N2, long v)
   long n1, n2;
 
   /* compute E[p] = < P1 > or < P1, P2 > */
-  P1 = P2 = ellinf();
   X = nfroots(K, elldivpol(E,p,v));
-  if(lg(X) == 1) return ptor0();
+  if (lg(X) == 1) return ptor0();
+  P2 = ellinf();
   if (p==2)
   {
     P1 = tor2(E, gel(X,1));
@@ -591,7 +613,7 @@ elltors(GEN e)
   checkell(e);
   switch(ell_get_type(e))
   {
-    case t_ELL_Q:  t = elltors_divpol(e, 0); break;
+    case t_ELL_Q:  t = ellQtors(e, 0); break;
     case t_ELL_NF: t = ellnftors(e, 0); break;
     case t_ELL_Fp:
     case t_ELL_Fq: return ellgroup0(e,NULL,1);
@@ -611,9 +633,9 @@ elltors_psylow(GEN e, ulong p)
   checkell(e);
   switch(ell_get_type(e))
   {
-    case t_ELL_Q:  t = elltors_divpol(e, p); break;
+    case t_ELL_Q:  t = ellQtors(e, p); break;
     case t_ELL_NF: t = ellnftors(e, p); break;
-    default: pari_err_TYPE("elltorspsylow",e);
+    default: pari_err_TYPE("elltors_psylow",e);
   }
   return gerepilecopy(av, t);
 }
@@ -667,11 +689,11 @@ ellorder_Q(GEN E, GEN P)
   }
   return gc_long(av,k);
 }
-/* E a t_ELL_NF */
+/* E a t_ELL_NF, B a multiplicative bound for exponent(E_tor) or NULL */
 static GEN
-ellorder_nf(GEN E, GEN P)
+ellorder_nf(GEN E, GEN P, GEN B)
 {
-  GEN K = ellnf_get_nf(E), B;
+  GEN K = ellnf_get_nf(E);
   pari_sp av = avma;
   GEN dx, dy, d4, d6, D, ND, Ep, Pp, Q, gp, modpr, pr, T, k;
   forprime_t S;
@@ -679,7 +701,7 @@ ellorder_nf(GEN E, GEN P)
   if (ell_is_inf(P)) return gen_1;
   if (gequal(P, ellneg(E,P))) return gen_2;
 
-  B = gel(nftorsbound(E, 0), 1);
+  if (!B) B = gel(nftorsbound(E, 0), 1);
   dx = Q_denom(gel(P,1));
   dy = Q_denom(gel(P,2));
   d4 = Q_denom(ell_get_c4(E));
@@ -738,7 +760,7 @@ ellorder(GEN E, GEN P, GEN o)
       if (lg(E)==1) pari_err_IMPL("ellorder for curve with singular reduction");
     }
   }
-  if (ell_get_type(E)==t_ELL_NF) return ellorder_nf(E, P);
+  if (ell_get_type(E)==t_ELL_NF) return ellorder_nf(E, P, NULL);
   checkell_Fq(E);
   fg = ellff_get_field(E);
   if (!o) o = ellff_get_o(E);
