@@ -20,6 +20,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_hyperell
+
 /* Implementation of Kedlaya Algorithm for counting point on hyperelliptic
 curves by Bill Allombert based on a GP script by Bernadette Perrin-Riou.
 
@@ -135,7 +137,7 @@ get_UV(GEN *U, GEN *V, GEN T, ulong p, long e)
   GEN R = polresultantext(T, dT);
   long v = varn(T);
   if (dvdiu(gel(R,3),p)) is_sing(T, p);
-  d = Fp_inv(gel(R,3), q);
+  d = Zp_inv(gel(R,3), utoi(p), e);
   *U = FpX_Fp_mul(FpX_red(to_ZX(gel(R,1),v),q),d,q);
   *V = FpX_Fp_mul(FpX_red(to_ZX(gel(R,2),v),q),d,q);
 }
@@ -310,7 +312,7 @@ ZlX_hyperellpadicfrobenius(GEN H, ulong p, long n)
   {
     pari_sp av2 = avma;
     GEN M, D;
-    D = diff_red(s, monomial(utoi(p),p*i-1,1),(k*p-1)>>1, Q, pN1);
+    D = diff_red(s, monomial(utoipos(p),p*i-1,1),(k*p-1)>>1, Q, pN1);
     if (DEBUGLEVEL>1) timer_printf(&ti,"red");
     M = ZpXXQ_frob(D, U, V, (k-1)>>1, Q, p, N + 1);
     if (DEBUGLEVEL>1) timer_printf(&ti,"frob");
@@ -324,7 +326,7 @@ hyperellpadicfrobenius(GEN H, ulong p, long n)
 {
   pari_sp av = avma;
   GEN M = ZlX_hyperellpadicfrobenius(H, p, n);
-  GEN q = zeropadic(utoi(p),n);
+  GEN q = zeropadic(utoipos(p),n);
   return gerepileupto(av, ZM_to_padic(M, q));
 }
 
@@ -358,7 +360,7 @@ ZpXQXXQ_sqr(GEN x, GEN S, GEN T, GEN q, GEN p, long e)
   pari_sp av = avma;
   GEN z, kx;
   long n = degpol(S);
-  kx = ZXX_to_Kronecker(x, n);
+  kx = RgXX_to_Kronecker(x, n);
   z = Kronecker_to_ZXX(FpXQX_sqr(kx, T, q), n, varn(S));
   return gerepileupto(av, ZpXQXXQ_red(z, S, T, q, p, e));
 }
@@ -369,8 +371,8 @@ ZpXQXXQ_mul(GEN x, GEN y, GEN S, GEN T, GEN q, GEN p, long e)
   pari_sp av = avma;
   GEN z, kx, ky;
   long n = degpol(S);
-  kx = ZXX_to_Kronecker(x, n);
-  ky = ZXX_to_Kronecker(y, n);
+  kx = RgXX_to_Kronecker(x, n);
+  ky = RgXX_to_Kronecker(y, n);
   z = Kronecker_to_ZXX(FpXQX_mul(ky, kx, T, q), n, varn(S));
   return gerepileupto(av, ZpXQXXQ_red(z, S, T, q, p, e));
 }
@@ -405,7 +407,7 @@ ZpXQXXQ_invsqrt(GEN F, GEN S, GEN T, ulong p, long e)
   ulong mask;
   long v = varn(F), n=1;
   pari_timer ti;
-  GEN a = pol_1(v), pp = utoi(p);
+  GEN a = pol_1(v), pp = utoipos(p);
   if (DEBUGLEVEL>1) timer_start(&ti);
   if (e <= 1) return gerepilecopy(av, a);
   mask = quadratic_prec_mask(e);
@@ -444,7 +446,7 @@ ZpXQXXQ_frob(GEN F, GEN U, GEN V, long k, GEN S, GEN T, ulong p, long e)
 {
   pari_sp av = avma, av2;
   long i, pr = degpol(F), dS = degpol(S), v = varn(T);
-  GEN q = powuu(p,e), pp = utoi(p);
+  GEN q = powuu(p,e), pp = utoipos(p);
   GEN Sp = RgX_deriv(S), Sp1 = RgX_shift_shallow(Sp, 1);
   GEN M = gel(F,pr+2), R;
   av2 = avma;
@@ -452,7 +454,7 @@ ZpXQXXQ_frob(GEN F, GEN U, GEN V, long k, GEN S, GEN T, ulong p, long e)
   {
     GEN A, B, H, Bc;
     ulong v, r;
-    H = ZpXQX_divrem(FpXQX_mul(V, M, T, q), S, T, q, utoi(p), e, &B);
+    H = ZpXQX_divrem(FpXQX_mul(V, M, T, q), S, T, q, utoipos(p), e, &B);
     A = FpXX_add(FpXQX_mul(U, M, T, q), FpXQX_mul(H, Sp, T, q),q);
     v = u_lvalrem(2*i+1,p,&r);
     Bc = RgX_deriv(B);
@@ -522,13 +524,12 @@ Fq_diff_red(GEN s, GEN A, long m, GEN S, GEN T, GEN q, GEN p, long e)
 static void
 Fq_get_UV(GEN *U, GEN *V, GEN S, GEN T, ulong p, long e)
 {
-  GEN q = powuu(p, e), d;
-  GEN dS = RgX_deriv(S);
-  GEN R  = polresultantext(S, dS), C;
+  GEN q = powuu(p, e), pp = utoipos(p), d;
+  GEN dS = RgX_deriv(S), R  = polresultantext(S, dS), C;
   long v = varn(S);
-  if (signe(FpX_red(to_ZX(gel(R,3),v),utoi(p)))==0) is_sing(S, p);
+  if (signe(FpX_red(to_ZX(gel(R,3),v), pp))==0) is_sing(S, p);
   C = FpXQ_red(to_ZX(gel(R, 3),v), T, q);
-  d = ZpXQ_inv(C, T, utoi(p), e);
+  d = ZpXQ_inv(C, T, pp, e);
   *U = FpXQX_FpXQ_mul(FpXQX_red(to_ZX(gel(R,1),v),T,q),d,T,q);
   *V = FpXQX_FpXQ_mul(FpXQX_red(to_ZX(gel(R,2),v),T,q),d,T,q);
 }
@@ -561,11 +562,9 @@ ZlXQX_hyperellpadicfrobenius(GEN H, GEN T, ulong p, long n)
   if (typ(H) != t_POL) pari_err_TYPE("hyperellpadicfrobenius",H);
   if (p == 2) is_sing(H, 2);
   d = degpol(H);
-  if (d <= 0)
-    pari_err_CONSTPOL("hyperellpadicfrobenius");
-  if (n < 1)
-    pari_err_DOMAIN("hyperellpadicfrobenius","n","<", gen_1, utoi(n));
-  k = get_basis(p, d); pp = utoi(p);
+  if (d <= 0) pari_err_CONSTPOL("hyperellpadicfrobenius");
+  if (n < 1) pari_err_DOMAIN("hyperellpadicfrobenius","n","<", gen_1, utoi(n));
+  k = get_basis(p, d); pp = utoipos(p);
   N = n + ulogint(2*n, p) + 1;
   q = powuu(p,n); N1 = N+1;
   pN1 = powuu(p,N1); T = FpX_get_red(T, pN1);
@@ -600,9 +599,9 @@ GEN
 nfhyperellpadicfrobenius(GEN H, GEN T, ulong p, long n)
 {
   pari_sp av = avma;
+  GEN pp = utoipos(p), q = zeropadic(pp, n);
   GEN M = ZlXQX_hyperellpadicfrobenius(lift_shallow(H),T,p,n);
-  GEN MM = ZpXQM_prodFrobenius(M, T, utoi(p), n);
-  GEN q = zeropadic(utoi(p),n);
+  GEN MM = ZpXQM_prodFrobenius(M, T, pp, n);
   GEN m = gmul(ZXM_to_padic(MM, q), gmodulo(gen_1, T));
   return gerepileupto(av, m);
 }
@@ -725,19 +724,28 @@ hyperell_Weil_bound(GEN q, ulong g, GEN p)
   return gc_long(av, logint(w,p) + 1);
 }
 
+/* return 4P + Q^2 */
+static GEN
+check_hyperell(GEN PQ)
+{
+  GEN H;
+  if (is_vec_t(typ(PQ)) && lg(PQ)==3)
+    H = gadd(gsqr(gel(PQ, 2)), gmul2n(gel(PQ, 1), 2));
+  else
+    H = gmul2n(PQ, 2);
+  return typ(H) == t_POL? H: NULL;
+}
+
 GEN
 hyperellcharpoly(GEN PQ)
 {
   pari_sp av = avma;
-  GEN H, M, R, T=NULL, pp=NULL, q;
+  GEN M, R, T=NULL, pp=NULL, q;
   long d, n, eps = 0;
   ulong p;
-  if (is_vec_t(typ(PQ)) && lg(PQ)==3)
-    H = gadd(gsqr(gel(PQ, 2)), gmul2n(gel(PQ, 1), 2));
-  else
-    H = PQ;
-  if (typ(H)!=t_POL || !RgX_is_FpXQX(H, &T, &pp) || !pp)
-    pari_err_TYPE("hyperellcharpoly",H);
+  GEN H = check_hyperell(PQ);
+  if (!H || !RgX_is_FpXQX(H, &T, &pp) || !pp)
+    pari_err_TYPE("hyperellcharpoly", PQ);
   p = itou(pp);
   if (!T)
   {
@@ -802,37 +810,525 @@ hyperellcharpoly(GEN PQ)
   return gerepilecopy(av, R);
 }
 
-GEN
-hyperell_redsl2(GEN Q)
+int
+hyperellisoncurve(GEN W, GEN P)
 {
   pari_sp av = avma;
-  long d = degpol(Q), v = varn(Q);
-  GEN den, P = Q_primitive_part(Q, &den);
+  long res;
+  GEN x, y;
+  if (typ(P)!=t_VEC || lg(P)!=3) pari_err_TYPE("hyperellisoncurve",P);
+  x = gel(P,1); y = gel(P,2);
+  if (typ(W)==t_POL)
+    res = gequal(gsqr(y), poleval(W,x));
+  else
+  {
+    if (typ(W)!=t_VEC || lg(W)!=3) pari_err_TYPE("hyperellisoncurve",W);
+    res = gequal(gmul(y, gadd(y,poleval(gel(W,2), x))), poleval(gel(W,1), x));
+  }
+  return gc_int(av, res);
+}
+
+GEN
+hyperelldisc(GEN PQ)
+{
+  pari_sp av = avma;
+  GEN D, H = check_hyperell(PQ);
+  long d, g;
+  if (!H || signe(H)==0) pari_err_TYPE("hyperelldisc",PQ);
+  d = degpol(H); g = ((d+1)>>1)-1;
+  D = gmul2n(RgX_disc(H),-4*(g+1));
+  if (odd(d)) D = gmul(D, gsqr(leading_coeff(H)));
+  return gerepileupto(av, D);
+}
+
+static long
+get_ep(GEN W)
+{
+  GEN P = gel(W,1), Q = gel(W,2);
+  if (signe(Q)==0) return ZX_lval(P,2);
+  return minss(ZX_lval(P,2), ZX_lval(Q,2));
+}
+
+static GEN
+algo51(GEN W, GEN M)
+{
+  GEN P = gel(W,1), Q = gel(W,2);
+  for(;;)
+  {
+    long vP = ZX_lval(P,2);
+    long vQ = signe(Q) ? ZX_lval(Q,2): vP+1;
+    long r;
+    /* 1 */
+    if (vQ==0) break;
+    /* 2 */
+    if (vP==0)
+    {
+      GEN H, H1;
+      /* a */
+      RgX_even_odd(FpX_red(P,gen_2),&H, &H1);
+      if (signe(H1)) break;
+      /* b */
+      P = ZX_add(P, ZX_mul(H, ZX_sub(Q, H)));
+      Q = ZX_sub(Q, ZX_mulu(H, 2));
+      vP = ZX_lval(P,2);
+      vQ = signe(Q) ? ZX_lval(Q,2): vP+1;
+    }
+    /* 2c */
+    if (vP==1) break;
+    /* 2d */
+    r = minss(2*vQ, vP)>>1;
+    gel(M,1) = shifti(gel(M,1), r);
+    P = ZX_shifti(P, -2*r);
+    Q = ZX_shifti(Q, -r);
+  }
+  return mkvec2(P,Q);
+}
+
+static GEN
+algo52(GEN W, GEN c, long *pt_lambda)
+{
+  long lambda;
+  GEN P = gel(W,1), Q = gel(W,2);
+  for(;;)
+  {
+    GEN H, H1;
+    /* 1 */
+    GEN Pc = ZX_affine(P,gen_2,c), Qc = ZX_affine(Q,gen_2,c);
+    long mP = ZX_lval(Pc,2), mQ = signe(Qc) ? ZX_lval(Qc,2): mP+1;
+    /* 2 */
+    if (2*mQ <= mP) { lambda = 2*mQ; break; }
+    /* 3 */
+    if (mP%2 == 1) { lambda = mP; break; }
+    /* 4 */
+    RgX_even_odd(FpX_red(ZX_shifti(Pc, -mP),gen_2),&H, &H1);
+    if (signe(H1)) { lambda = mP; break; }
+    /* 5 */
+     P = ZX_add(P, ZX_mul(H, ZX_sub(Q, H)));
+     Q = ZX_sub(Q, ZX_mulu(H, 2));
+  }
+  *pt_lambda = lambda;
+  return mkvec2(P,Q);
+}
+
+static GEN
+algo541(GEN F, GEN p, long ep, long g)
+{
+  GEN Fe = FpX_red(ep ? ZX_Z_divexact(F,p): F, p);
+  return FpX_roots_mult(Fe, g+2-ep, p);
+}
+
+static long
+test53(long lambda, long ep, long g)
+{
+  return (lambda <= g+1) || (g%2 && lambda<g+3 && ep==1);
+}
+
+static long
+test55(GEN W, long ep, long g)
+{
+  GEN P = gel(W,1), Q = gel(W,2);
+  GEN Pe = FpX_red(ep ? ZX_shifti(P,-1): P, gen_2);
+  GEN Qe = FpX_red(ep ? ZX_shifti(Q,-1): Q, gen_2);
+  if (ep==0)
+  {
+    if (signe(Qe)!=0) return ZX_val(Qe) >= (g + 3)>>1;
+    else return ZX_val(FpX_deriv(Pe, gen_2)) >= g+1;
+  }
+  else
+    return ZX_val(Qe) >= (g+1)>>1 && ZX_val(Pe) >= g + 1;
+}
+
+static GEN
+hyperell_reverse(GEN W, long g)
+{
+  return mkvec2(RgXn_recip_shallow(gel(W,1),2*g+3),
+                RgXn_recip_shallow(gel(W,2),g+2));
+}
+
+static GEN
+algo56(GEN W, long g)
+{
+  long ep;
+  GEN M = mkvec2(gen_1, matid(2)), Woo;
+  W = algo51(W, M);
+  Woo = hyperell_reverse(W, g);
+  ep = get_ep(Woo);
+  if (test55(Woo,ep,g))
+  {
+    long lambda;
+    Woo = algo52(Woo, gen_0, &lambda);
+    if (!test53(lambda,ep,g))
+    {
+      long r = lambda>>1;
+      gel(M,1) = shifti(gel(M,1), r);
+      gel(M,2) = ZM2_mul(gel(M,2), mkmat22(gen_0, gen_1, gen_2, gen_0));
+      W = mkvec2(ZX_shifti(ZX_unscale(gel(Woo,1), gen_2), -2*r),
+                 ZX_shifti(ZX_unscale(gel(Woo,2), gen_2), -r));
+    }
+  }
+  for(;;)
+  {
+    long j, ep = get_ep(W);
+    for (j = 0; j < 2; j++)
+    {
+      long lambda;
+      GEN c = utoi(j);
+      GEN Pc = ZX_affine(gel(W,1), gen_2, c);
+      GEN Qc = ZX_affine(gel(W,2), gen_2, c);
+      if (test55(mkvec2(Pc, Qc), ep, g))
+      {
+        GEN Wc = algo52(W, c, &lambda);
+        if (!test53(lambda,ep,g))
+        {
+          long r = lambda>>1;
+          gel(M,1) = shifti(gel(M,1), r);
+          gel(M,2) = ZM2_mul(gel(M,2), mkmat22(gen_2, c, gen_0, gen_1));
+          W = mkvec2(ZX_shifti(ZX_affine(gel(Wc,1), gen_2,c), -2*r),
+                     ZX_shifti(ZX_affine(gel(Wc,2), gen_2,c), -r));
+          break;
+        }
+      }
+    }
+    if (j==2) break;
+  }
+  return mkvec2(W, M);
+}
+
+/* return the (degree 2) apolar invariant (the nth transvectant of P and P) */
+static GEN
+ZX_apolar(GEN P, long n)
+{
+  pari_sp av = avma;
+  long d = degpol(P), i;
+  GEN s = gen_0, g = cgetg(n+2,t_VEC);
+  gel(g,1) = gen_1;
+  for (i = 1; i <= n; i++) gel(g,i+1) = muliu(gel(g,i),i); /* g[i+1] = i! */
+  for (i = n-d; i <= d; i++)
+  {
+     GEN a = mulii(mulii(gel(g,i+1),gel(g,n-i+1)),
+                   mulii(gel(P,i+2),gel(P,n-i+2)));
+     s = odd(i)? subii(s, a): addii(s, a);
+  }
+  return gerepileuptoint(av,s);
+}
+
+static GEN
+algo57(GEN F, long g, GEN pr)
+{
+  long i, l;
+  GEN D, C = content(F);
+  GEN e = gel(core2(shifti(C,-vali(C))),2);
+  GEN M = mkvec2(e, matid(2));
+  long minvd = (2*g+1)>>(odd(g) ? 4:2);
+  F = ZX_Z_divexact(F, sqri(e));
+  D = absi(hyperelldisc(F));
+  if (!pr)
+  {
+    GEN A = gcdii(D, ZX_apolar(F, 2*g+2));
+    pr = gel(factor(shifti(A, -vali(A))),1);
+  }
+  l = lg(pr);
+  for (i = 1; i < l; i++)
+  {
+    long ep;
+    GEN p = gel(pr, i), ps2 = shifti(p,-1), Fe;
+    if (equaliu(p,2) || Z_pval(D,p) < minvd) continue;
+    ep = ZX_pval(F,p);
+    Fe = FpX_red(ep ? ZX_Z_divexact(F,p): F, p);
+    if (degpol(Fe) < g+1+ep)
+    {
+      GEN Fi = ZX_unscale(RgXn_recip_shallow(F,2*g+3), p);
+      long lambda = ZX_pval(Fi,p);
+      if (!test53(lambda,ep,g))
+      {
+        GEN ppr = powiu(p,lambda>>1);
+        F = ZX_Z_divexact(Fi,sqri(ppr));
+        gel(M,1) = mulii(gel(M,1), ppr);
+        gel(M,2) = ZM2_mul(gel(M,2), mkmat22(gen_0,gen_1,p,gen_0));
+      }
+    }
+    for(;;)
+    {
+      long ep = ZX_pval(F,p);
+      GEN R = algo541(F, p, ep, g);
+      long j, lR = lg(R);
+      for (j = 1; j<lR; j++)
+      {
+        GEN c = Fp_center(gel(R,j), p, ps2);
+        GEN Fi = ZX_affine(F,p,c);
+        long lambda = ZX_pval(Fi,p);
+        if (!test53(lambda,ep,g))
+        {
+          GEN ppr = powiu(p,lambda>>1);
+          F = ZX_Z_divexact(Fi, sqri(ppr));
+          gel(M,1) = mulii(gel(M,1), ppr);
+          gel(M,2) = ZM2_mul(gel(M,2), mkmat22(p,c,gen_0,gen_1));
+          break;
+        }
+      }
+      if (j==lR) break;
+    }
+  }
+  return mkvec2(F, M);
+}
+
+static GEN
+RgX_RgM2_eval(GEN P, GEN A, GEN Bp, long d)
+{
+  if (signe(P)==0)
+    return P;
+  else
+  {
+    long dP = degpol(P);
+    GEN R = RgX_homogenous_evalpow(P, A, Bp);
+    if (d > dP)
+      R = gmul(R, gel(Bp,1+d-dP));
+    return R;
+  }
+}
+
+static GEN
+minimalmodel_merge(GEN W2, GEN Modd, long g, long v)
+{
+  GEN P = gel(W2,1), Q = gel(W2,2);
+  GEN e = gel(Modd,1), M = gel(Modd,2);
+  GEN A = deg1pol_shallow(gcoeff(M,1,1), gcoeff(M,1,2), v);
+  GEN B = deg1pol_shallow(gcoeff(M,2,1), gcoeff(M,2,2), v);
+  GEN Bp = gpowers(B, 2*g+2);
+  long f = mod4(e)==1 ? 1: -1;
+  GEN m = shifti(f > 0 ? subui(1,e): addui(1,e), -2);
+  GEN  m24 = subii(shifti(m,1), shifti(sqri(m),2));
+  P = RgX_RgM2_eval(P, A, Bp, 2*g+2);
+  Q = RgX_RgM2_eval(Q, A, Bp, g+1);
+  P = ZX_Z_divexact(ZX_add(P, ZX_Z_mul(ZX_sqr(Q), m24)),sqri(e));
+  if (f < 0) Q = ZX_neg(Q);
+  return mkvec2(P,Q);
+}
+
+static GEN
+hyperell_redQ(GEN W)
+{
+  GEN P = gel(W,1), Q = gel(W,2);
+  GEN Pr, Qr = FpX_red(Q, gen_2);
+  Pr = ZX_add(P, ZX_shifti(ZX_mul(ZX_sub(Q, Qr),ZX_add(Q, Qr)),-2));
+  return mkvec2(Pr, Qr);
+}
+
+static GEN
+minimalmodel_getH(GEN W, GEN Qn, GEN e, GEN M, long g, long v)
+{
+  GEN Q = gel(W,2);
+  GEN A = deg1pol_shallow(gcoeff(M,1,1), gcoeff(M,1,2), v);
+  GEN B = deg1pol_shallow(gcoeff(M,2,1), gcoeff(M,2,2), v);
+  GEN Bp = gpowers(B, g+1);
+  return ZX_shifti(ZX_sub(ZX_Z_mul(Qn,e),RgX_RgM2_eval(Q, A, Bp, g+1)), -1);
+}
+
+static void
+check_hyperell_Q(const char *fun, GEN *pW, GEN *pF)
+{
+  GEN W = *pW, F = check_hyperell(W);
+  long v, d;
+  if (!F || !signe(F) || !RgX_is_ZX(F)) pari_err_TYPE(fun, W);
+  if (!signe(ZX_disc(F))) pari_err_DOMAIN(fun,"disc(W)","==",gen_0,W);
+  v = varn(F); d = degpol(F);
+  if (typ(W)==t_POL) W = mkvec2(W, pol_0(v));
+  else
+  {
+    GEN P = gel(W, 1), Q = gel(W, 2);
+    long g = ((d+1) >> 1) - 1;
+    if (typ(P)!=t_POL) P = scalarpol_shallow(P, v);
+    if (typ(Q)!=t_POL) Q = scalarpol_shallow(Q, v);
+    if (!RgX_is_ZX(P) || !RgX_is_ZX(Q)) pari_err_TYPE(fun,W);
+    if (degpol(P) > 2*g+2) pari_err_DOMAIN(fun, "deg(P)", ">", utoi(2*g+2), P);
+    if (degpol(Q) > g+1) pari_err_DOMAIN(fun, "deg(Q)", ">", utoi(g+1), Q);
+    W = mkvec2(P, Q);
+  }
+  if (d < 3) pari_err_DOMAIN(fun, "genus", "=", gen_0, gen_0);
+  *pW = W; *pF = F;
+}
+
+GEN
+hyperellminimalmodel(GEN W, GEN *pM, GEN pr)
+{
+  pari_sp av = avma;
+  GEN Wr, F, WM2, F2, W2, M2, Modd, Wf, ef, Mf, Hf;
+  long d, g, v;
+  check_hyperell_Q("hyperellminimalmodel",&W, &F);
+  d = degpol(F); g = ((d+1)>>1)-1; v = varn(F);
+  Wr = hyperell_redQ(W);
+  if (!pr || RgV_isin(pr, gen_2))
+  {
+    WM2 = algo56(Wr,g); W2 = gel(WM2, 1); M2 = gel(WM2, 2);
+    F2 = check_hyperell(W2);
+  }
+  else
+  {
+    W2 = Wr; F2 = F; M2 = mkvec2(gen_1, matid(2));
+  }
+  Modd = gel(algo57(F2, g, pr), 2);
+  Wf = hyperell_redQ(minimalmodel_merge(W2, Modd, g, v));
+  if (!pM) return gerepilecopy(av, Wf);
+  ef = mulii(gel(M2,1), gel(Modd,1));
+  Mf = ZM2_mul(gel(M2,2), gel(Modd,2));
+  Hf = minimalmodel_getH(W, gel(Wf,2), ef, Mf, g, v);
+  *pM =  mkvec3(ef, Mf, Hf);
+  return gc_all(av, 2, &Wf, pM);
+}
+
+GEN
+hyperellminimaldisc(GEN W, GEN pr)
+{
+  pari_sp av = avma;
+  GEN C = hyperellminimalmodel(W, NULL, pr);
+  return gerepileuptoint(av, hyperelldisc(C));
+}
+
+static GEN
+redqfbsplit(GEN a, GEN b, GEN c, GEN d)
+{
+  GEN p = subii(d,b), q = shifti(a,1);
+  GEN U, Q, u, v, w = bezout(p, q, &u, &v);
+
+  if (!equali1(w)) { p = diviiexact(p, w); q = diviiexact(q, w); }
+  U = mkmat22(p, negi(v), q, u);
+  Q = qfb_apply_ZM(mkvec3(a,b,c), U);
+  b = gel(Q, 2); c = gel(Q,3);
+  if (signe(b) < 0) gel(U,2) = mkcol2(v, negi(u));
+  gel(U,2) = ZC_lincomb(gen_1, truedivii(negi(c), d), gel(U,2), gel(U,1));
+  return U;
+}
+
+static GEN
+polreduce(GEN P, GEN M)
+{
+  long v = varn(P), dP = degpol(P), d = odd(dP) ? dP+1: dP;
+  GEN A = deg1pol_shallow(gcoeff(M,1,1), gcoeff(M,1,2), v);
+  GEN B = deg1pol_shallow(gcoeff(M,2,1), gcoeff(M,2,2), v);
+  return RgX_RgM2_eval(P, A, gpowers(B, d), d);
+}
+
+/* assume deg(P) > 2 */
+static GEN
+red_Cremona_Stoll(GEN P, GEN *pM)
+{
+  GEN q1, q2, q3, M, R;
+  long i, prec = nbits2prec(2*gexpo(P)) + 1, d = degpol(P);
+  GEN dP = ZX_deriv(P), r = QX_complex_roots(P, prec);
+  q1 = gen_0; q2 = gen_0; q3 = gen_0;
+  for (i = 1; i <= d; i++)
+  {
+    GEN ri = gel(r,i);
+    GEN s = ginv(gabs(RgX_cxeval(dP,ri,NULL), prec));
+    if (d!=4) s = gpow(s, gdivgs(gen_2,d-2), prec);
+    q1 = gadd(q1, s);
+    q2 = gsub(q2, gmul(real_i(ri), s));
+    q3 = gadd(q3, gmul(gnorm(ri), s));
+  }
+  M = lllgram(mkmat22(q1,q2,q2,q3));
+  if (lg(M) != 3) M = matid(2);
+  R = polreduce(P, M);
+  *pM = M;
+  return R;
+}
+
+/* assume deg(P) > 2 */
+GEN
+ZX_hyperellred(GEN P, GEN *pM)
+{
+  pari_sp av = avma;
+  long d = degpol(P);
+  GEN q1, q2, q3, D, vD;
   GEN a = gel(P,d+2), b = gel(P,d+1), c = gel(P, d);
-  GEN q1, q2, q3, D, M, A, B, R;
-  q1 = mulis(sqri(a), d);
+  GEN M, R, M2;
+
+  q1 = muliu(sqri(a), d);
   q2 = shifti(mulii(a,b), 1);
-  q3 = subii(sqri(b),shifti(mulii(a,c), 1));
+  q3 = subii(sqri(b), shifti(mulii(a,c), 1));
   D = gcdii(gcdii(q1, q2), q3);
   if (!equali1(D))
   {
-    q1 = diviiexact(q1,D);
-    q2 = diviiexact(q2,D);
-    q3 = diviiexact(q3,D);
+    q1 = diviiexact(q1, D);
+    q2 = diviiexact(q2, D);
+    q3 = diviiexact(q3, D);
   }
   D = qfb_disc3(q1, q2, q3);
   if (!signe(D))
-    M = mkmat22(gen_1,truedivii(negi(q2),shifti(q1,1)),gen_0,gen_1);
-  else if (issquare(D))
-    M = mkmat22(gen_1,divii(negi(q2),shifti(q1,1)),gen_0,gen_1);
+    M = mkmat22(gen_1, truedivii(negi(q2),shifti(q1,1)), gen_0, gen_1);
+  else if (issquareall(D,&vD))
+    M = redqfbsplit(q1, q2, q3, vD);
+  else
+    M = gel(qfbredsl2(mkqfb(q1,q2,q3,D), NULL), 2);
+  R = red_Cremona_Stoll(polreduce(P, M), &M2);
+  if (pM) *pM = gmul(M, M2);
+  return gc_all(av, pM ? 2: 1, &R, pM);
+}
+
+GEN
+hyperellred(GEN W, GEN *pM)
+{
+  pari_sp av = avma;
+  long g, d, v;
+  GEN F, M, Wf, Hf;
+  check_hyperell_Q("hyperellred", &W, &F);
+  d = degpol(F); g = ((d+1)>>1)-1; v = varn(F);
+  (void) ZX_hyperellred(F, &M);
+  Wf = hyperell_redQ(minimalmodel_merge(W, mkvec2(gen_1, M), g, v));
+  Hf = minimalmodel_getH(W, gel(Wf,2), gen_1, M, g, v);
+  if (pM) *pM = mkvec3(gen_1, M, Hf);
+  return gc_all(av, pM ? 2: 1, &Wf, pM);
+}
+
+static void
+check_hyperell_Rg(const char *fun, GEN *pW, GEN *pF)
+{
+  GEN W = *pW, F = check_hyperell(W);
+  long v = varn(F);
+  if (!F || signe(F)==0)
+    pari_err_TYPE(fun, W);
+  if (typ(W)==t_POL) W = mkvec2(W, pol_0(v));
   else
   {
-    GEN Q = signe(D) < 0 ? qfi(q1,q2,q3): qfr(q1,q2,q3,real_0(DEFAULTPREC));
-    M = gel(qfbredsl2(Q, NULL),2);
+    GEN P = gel(W, 1), Q = gel(W, 2);
+    long g = ((degpol(F)+1)>>1)-1;
+    if( typ(P)!=t_POL) P = scalarpol(P, v);
+    if( typ(Q)!=t_POL) Q = scalarpol(Q, v);
+    if (degpol(P) > 2*g+2)
+      pari_err_DOMAIN(fun, "poldegree(P)", ">", utoi(2*g+2), P);
+    if (degpol(Q) > g+1)
+      pari_err_DOMAIN(fun, "poldegree(Q)", ">", utoi(g+1), Q);
+
+    W = mkvec2(P, Q);
   }
-  A = deg1pol(gcoeff(M,1,1),gcoeff(M,1,2),v);
-  B = gpowers(deg1pol(gcoeff(M,2,1),gcoeff(M,2,2),v), d);
-  R = gel(RgX_homogenous_evalpow(P, A, B),1);
-  if (den) R = gmul(R, den);
-  return gerepilecopy(av,mkvec2(R,M));
+  if (pF) *pF = F;
+  *pW = W;
+}
+
+static void
+check_hyperell_vc(const char *fun, GEN C, long v, GEN *e, GEN *M, GEN *H)
+{
+  if (typ(C) != t_VEC || lg(C) != 4) pari_err_TYPE(fun,C);
+  *e = gel(C,1); *M = gel(C,2); *H = gel(C,3);
+  if (typ(*M) != t_MAT || lg(*M) != 3 || lgcols(*M) != 3) pari_err_TYPE(fun,C);
+  if (typ(*H)!=t_POL || varn(*H)!=v) *H = scalarpol_shallow(*H,v);
+}
+
+GEN
+hyperellchangecurve(GEN W, GEN C)
+{
+  pari_sp av = avma;
+  GEN F, P, Q, A, B, Bp, e, M, H;
+  long d, g, v;
+  check_hyperell_Rg("hyperellchangecurve",&W,&F);
+  P = gel(W,1); Q = gel(W,2);
+  d = degpol(F); g = ((d+1)>>1)-1; v = varn(F);
+  check_hyperell_vc("hyperellchangecurve", C, v, &e, &M, &H);
+  A = deg1pol_shallow(gcoeff(M,1,1), gcoeff(M,1,2), v);
+  B = deg1pol_shallow(gcoeff(M,2,1), gcoeff(M,2,2), v);
+  Bp = gpowers(B, 2*g+2);
+  P = RgX_RgM2_eval(P, A, Bp, 2*g+2);
+  Q = RgX_RgM2_eval(Q, A, Bp, g+1);
+  P = RgX_Rg_div(RgX_sub(P, RgX_mul(H,RgX_add(Q,H))), gsqr(e));
+  Q = RgX_Rg_div(RgX_add(Q, RgX_mul2n(H,1)), e);
+  return gerepilecopy(av, mkvec2(P,Q));
 }
