@@ -21,6 +21,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_lfun
+
 /*******************************************************************/
 /*  Accessors                                                      */
 /*******************************************************************/
@@ -168,7 +170,7 @@ antwist(GEN an, GEN Vga, long prec)
     if (typ(an) == t_VECSMALL)
       for (i = 1; i < l; i++) gel(b,i) = mulss(an[i], i);
     else
-      for (i = 1; i < l; i++) gel(b,i) = gmulgs(gel(an,i), i);
+      for (i = 1; i < l; i++) gel(b,i) = gmulgu(gel(an,i), i);
   }
   else
   {
@@ -252,16 +254,17 @@ isnegint(GEN s)
   if (signe(r) <= 0 && gequal(s, r)) return -itos(r);
   return -1;
 }
+/* if s = a + O(x^n), a <= 0 integer, replace by a + b*x^n + O(x^(n+1)) */
 static GEN
-serextendifnegint(GEN s, GEN s0, long *ext)
+serextendifnegint(GEN s, GEN b, long *ext)
 {
   if (!signe(s) || (serisscalar(s) && isnegint(gel(s,2)) >= 0))
-  { /* a + O(x^n) => a + x^n + O(x^(n+1)) */
+  {
     long l = lg(s);
     GEN t = cgetg(l+1, t_SER);
-    gel(t, l) = s0; while (--l > 1) gel(t,l) = gel(s,l);
+    gel(t, l) = b; while (--l > 1) gel(t,l) = gel(s,l);
     if (gequal0(gel(t,2))) gel(t,2) = gen_0;
-    t[1] = s[1]; s = normalize(t); *ext = 1;
+    t[1] = s[1]; s = normalizeser(t); *ext = 1;
   }
   return s;
 }
@@ -339,12 +342,13 @@ gammafactor(GEN Vga)
     GEN a = gel(Vga,i), r = gmul2n(real_i(a), -1);
     long q = itos(gfloor(r)); /* [Re a/2] */
     r = gmul2n(gsubgs(r, q), 1);
-    gel(F,i) = gadd(r, imag_i(a)); /* 2{Re a/2} + Im a/2 */
+    gel(F,i) = gequal0(imag_i(a)) ? r : mkcomplex(r, imag_i(a)); /* 2{Re a/2} + I*(Im a) */
     if (q) R = gmul(R, gammafrac(gel(F,i), q));
   }
   F = vec_reduce(F, &E); l = lg(E);
   v = cgetg(l, t_VEC);
-  for (i = 1; i < l; i++) gel(v,i) = mkvec2(gfrac(gel(F,i)), stoi(E[i]));
+  for (i = 1; i < l; i++)
+      gel(v,i) = mkvec2(gsub(gel(F,i),gfloor(real_i(gel(F,i)))), stoi(E[i]));
   gen_sort_inplace(v, (void*)cmp_universal, cmp_nodata, &P);
   a = cgetg(l, t_VEC); e = cgetg(l, t_VECSMALL);
   b = cgetg(l, t_VEC); f = cgetg(l, t_VECSMALL);
@@ -421,7 +425,11 @@ gammaordinary(GEN Vga, GEN s)
   {
     GEN z = gadd(s, gel(Vga,i));
     long e;
-    if (gsigne(z) <= 0) { (void)grndtoi(z, &e); if (e < -10) return 0; }
+    if (gexpo(imag_i(z)) < -10)
+    {
+      z = real_i(z);
+      if (gsigne(z) <= 0) { (void)grndtoi(z, &e); if (e < -10) return 0; }
+    }
   }
   return 1;
 }
@@ -469,7 +477,7 @@ lfunthetacost(GEN ldata, GEN tdom, long m, long bitprec)
   pari_sp av = avma;
   GEN Vga = ldata_get_gammavec(ldata);
   long d = lg(Vga)-1;
-  long k1 = ldata_get_k1_dbl(ldata);
+  double k1 = maxdd(ldata_get_k1_dbl(ldata), 0.);
   double c = d/2., a, A, B, logC, al, rho, T;
   double N = gtodouble(ldata_get_conductor(ldata));
 
@@ -534,11 +542,11 @@ fracgammadegree(GEN FVga)
 static GEN
 rtoR(GEN a, GEN r, GEN FVga, GEN N, long prec)
 {
-  long v = lg(r)-2;
-  GEN as = deg1ser_shallow(gen_1, a, varn(r), v);
-  GEN Na = gpow(N, gdivgs(as, 2), prec);
-  long d = fracgammadegree(FVga), ext;
-  if (d) as = sertoser(as, v+d); /* make up for a possible loss of accuracy */
+  long v = lg(r)-2, d = fracgammadegree(FVga), ext;
+  GEN Na, as = deg1ser_shallow(gen_1, a, varn(r), v);
+  Na = gpow(N, gdivgu(as, 2), prec);
+  /* make up for a possible loss of accuracy */
+  if (d) as = deg1ser_shallow(gen_1, a, varn(r), v + d);
   return gmul(gmul(r, Na), gammafactproduct(FVga, as, &ext, prec));
 }
 
@@ -648,6 +656,16 @@ static GEN
 lfunrtoR(GEN ldata, long prec)
 { return lfunrtoR_eno(ldata, ldata_get_rootno(ldata), prec); }
 
+static long
+prec_fix(long prec)
+{
+#ifndef LONG_IS_64BIT
+  /* make sure that default accuracy is the same on 32/64bit */
+  if (odd(prec)) prec += EXTRAPRECWORD;
+#endif
+  return prec;
+}
+
 /* thetainit using {an: n <= L}; if (m = 0 && easytheta), an2 is an * n^al */
 static GEN
 lfunthetainit0(GEN ldata, GEN tdom, GEN an2, long m,
@@ -664,23 +682,25 @@ lfunthetainit0(GEN ldata, GEN tdom, GEN an2, long m,
     get_cone_fuzz(tdom, &r, &a);
     tdom = mkvec2(dbltor(r), a? dbltor(a): gen_0);
   }
+  prec += maxss(EXTRAPRECWORD, nbits2extraprec(extrabit));
   tech = mkvecn(7, an2,K,R, stoi(bitprec), stoi(m), tdom,
-                   gsqrt(ginv(N), prec + EXTRAPRECWORD));
+                   gsqrt(ginv(N), prec_fix(prec)));
   return mkvec3(mkvecsmall(t_LDESC_THETA), ldata, tech);
 }
 
 /* tdom: 1) positive real number r, t real, t >= r; or
  *       2) [r,a], describing the cone |t| >= r, |arg(t)| <= a */
 static GEN
-lfunthetainit_i(GEN data, GEN tdom, long m, long bitprec)
+lfunthetainit_i(GEN data, GEN tdom, long m, long bit)
 {
   GEN ldata = lfunmisc_to_ldata_shallow(data);
-  long L = lfunthetacost(ldata, tdom, m, bitprec), prec = nbits2prec(bitprec);
+  long b = 32, L = lfunthetacost(ldata, tdom, m, bit), prec = nbits2prec(bit);
   GEN ldatan = ldata_newprec(ldata, prec);
   GEN an = ldata_vecan(ldata_get_an(ldatan), L, prec);
   GEN Vga = ldata_get_gammavec(ldatan);
   if (m == 0 && Vgaeasytheta(Vga)) an = antwist(an, Vga, prec);
-  return lfunthetainit0(ldatan, tdom, an, m, bitprec, 32);
+  if (typ(an) != t_VECSMALL) b = maxss(b, gexpo(an));
+  return lfunthetainit0(ldatan, tdom, an, m, bit, b);
 }
 
 GEN
@@ -757,7 +777,7 @@ mkvroots(long d, long lim, long prec)
         return v;
     }
   }
-  return vecpowug(lim, gdivgs(gen_2,d), prec);
+  return vecpowug(lim, gdivgu(gen_2,d), prec);
 }
 
 GEN
@@ -861,13 +881,15 @@ lfuntheta(GEN data, GEN t, long m, long bitprec)
   pari_sp ltop = avma;
   long limt, d;
   GEN isqN, vecan, Vga, ldata, theta, thetainit, S;
-  long n, prec = nbits2prec(bitprec);
-  t = gprec_w(t, prec);
+  long n, prec;
+
   theta = lfunthetacheckinit(data, t, m, bitprec);
   ldata = linit_get_ldata(theta);
   thetainit = linit_get_tech(theta);
   vecan = theta_get_an(thetainit);
   isqN = theta_get_isqrtN(thetainit);
+  prec = maxss(realprec(isqN), nbits2prec(bitprec));
+  t = gprec_w(t, prec);
   limt = lg(vecan)-1;
   if (theta == data)
     limt = minss(limt, lfunthetacost(ldata, t, m, bitprec));
@@ -892,7 +914,7 @@ lfuntheta(GEN data, GEN t, long m, long bitprec)
     GEN K = theta_get_K(thetainit);
     GEN vroots = mkvroots(d, limt, prec);
     pari_sp av;
-    t = gpow(t, gdivgs(gen_2,d), prec);
+    t = gpow(t, gdivgu(gen_2,d), prec);
     S = gen_0; av = avma;
     for (n = 1; n <= limt; ++n)
     {
@@ -919,7 +941,7 @@ struct lfunp {
 };
 
 static void
-lfunparams(GEN ldata, long der, long bitprec, struct lfunp *S)
+lfunp_set(GEN ldata, long der, long bitprec, struct lfunp *S)
 {
   const long derprec = (der > 1)? dbllog2(mpfact(der)): 0; /* log2(der!) */
   GEN Vga, N, L, k;
@@ -950,7 +972,7 @@ lfunparams(GEN ldata, long der, long bitprec, struct lfunp *S)
   hd = M_LN2/S->m0;
 
   logC = d2*M_LN2 - log(d2)/2;
-  k1 = ldata_get_k1_dbl(ldata);
+  k1 = maxdd(ldata_get_k1_dbl(ldata), 0.);
   S->k1 = k1; /* assume |a_n| << n^k1 with small implied constant */
   A = gammavec_expo(d, suma);
 
@@ -1163,7 +1185,7 @@ lfuninit_ab(GEN theta, GEN h, struct lfunp *S)
      * N.B. we use the 'rt' variant and pass (n exp(mh)/sqrt(N))^(2/d).
      * Speedup: if n' = 2n and m' = m - m0 >= 0; then k[m,n] = k[m',n']. */
     vroots = mkvroots(S->d, S->nmax, prec); /* vroots[n] = n^(2/d) */
-    d2 = gdivgs(gen_2, S->d);
+    d2 = gdivgu(gen_2, S->d);
     peh2d = gpowers0(gexp(gmul(d2,h), prec), M, gpow(isqN, d2, prec));
     m0 = S->m0; /* peh2d[m+1] = (exp(mh)/sqrt(N))^(2/d) */
     worker = snm_closure(is_entry("_lfuninit_worker"),
@@ -1265,7 +1287,7 @@ lfuninit_make(long t, GEN ldata, GEN tech, GEN domain)
   GEN Vga = ldata_get_gammavec(ldata);
   long d = lg(Vga)-1;
   GEN w2 = gen_1, k2 = gmul2n(ldata_get_k(ldata), -1);
-  GEN expot = gdivgs(gadd(gmulsg(d, gsubgs(k2, 1)), sumVga(Vga)), 4);
+  GEN expot = gdivgu(gadd(gmulsg(d, gsubgs(k2, 1)), sumVga(Vga)), 4);
   if (typ(ldata_get_dual(ldata))==t_INT)
   {
     GEN eno = ldata_get_rootno(ldata);
@@ -1360,15 +1382,18 @@ lfun_init_theta(GEN ldata, GEN eno, struct lfunp *S)
 }
 
 GEN
-lfuncost(GEN L, GEN dom, long der, long bitprec)
+lfuncost(GEN L, GEN dom, long der, long bit)
 {
   pari_sp av = avma;
   GEN ldata = lfunmisc_to_ldata_shallow(L);
-  GEN k = ldata_get_k(ldata);
+  GEN w, k = ldata_get_k(ldata);
   struct lfunp S;
 
   parse_dom(gtodouble(k), dom, &S);
-  lfunparams(ldata, der, bitprec, &S);
+  lfunp_set(ldata, der, bit, &S);
+  w = ldata_get_rootno(ldata);
+  if (isintzero(w)) /* for lfunrootres */
+    S.nmax = maxss(S.nmax, lfunthetacost(ldata, dbltor(sqrt(0.5)), 0, bit+1));
   set_avma(av); return mkvecsmall2(S.nmax, S.Dmax);
 }
 GEN
@@ -1420,11 +1445,11 @@ lfuninit(GEN lmisc, GEN dom, long der, long bitprec)
   if (ldata_get_type(ldata)==t_LFUN_NF)
   {
     GEN T = gel(ldata_get_an(ldata), 2);
-    return lfunzetakinit(T, dom, der, 0, bitprec);
+    return lfunzetakinit(T, dom, der, bitprec);
   }
   k = ldata_get_k(ldata);
   parse_dom(gtodouble(k), dom, &S);
-  lfunparams(ldata, der, bitprec, &S);
+  lfunp_set(ldata, der, bitprec, &S);
   ldata = ldata_newprec(ldata, nbits2prec(S.Dmax));
   r = ldata_get_residue(ldata);
   /* Note: all guesses should already have been performed (thetainit more
@@ -1679,7 +1704,7 @@ static GEN
 lfununext(GEN L)
 {
   setlg(L, maxss(lg(L)-1, valp(L)? 2: 3));
-  return normalize(L);
+  return normalizeser(L);
 }
 static GEN
 lfununextvec(GEN x)
@@ -1709,7 +1734,7 @@ lfun_OK(GEN linit, GEN s, GEN sdom, long bitprec)
   }
   gas = gammafactproduct(FVga, ss, &ext, prec);
   N = ldata_get_conductor(linit_get_ldata(linit));
-  res = gdiv(S, gmul(gpow(N, gdivgs(ss, 2), prec), gas));
+  res = gdiv(S, gmul(gpow(N, gdivgu(ss, 2), prec), gas));
   if (typ(s) != t_SER && is_ser(res)) res = lfunservec(res);
   else if (ext) res = lfununextvec(res);
   return gprec_w(res, prec);
@@ -1767,7 +1792,7 @@ sersplit1(GEN s, GEN *head)
   y = cgetg(l-1, t_SER); y[1] = s[1];
   setvalp(y, 1);
   for (i=3; i < l; i++) gel(y,i-1) = gel(s,i);
-  return normalize(y);
+  return normalizeser(y);
 }
 
 /* order of pole of Lambda at s (0 if regular point) */
@@ -1829,6 +1854,7 @@ lfunderiv(GEN lmisc, long m, GEN s, long flag, long bitprec)
   {
     long ex = lfunlambdaord(linit, s);
     /* HACK: pretend lfuninit was done to right accuracy */
+    if (gequal0(s)) s = gen_0;
     s = deg1ser_shallow(gen_1, s, 0, m+1+ex);
   }
   res = flag ? lfunlambda_OK(linit, s, dom, bitprec):
@@ -1908,7 +1934,7 @@ theta_pole_contrib(GEN R, long v, GEN L)
   GEN s = mysercoeff(R,-v);
   long i;
   for (i = v-1; i >= 1; i--)
-    s = gadd(mysercoeff(R,-i), gdivgs(gmul(s,L), i));
+    s = gadd(mysercoeff(R,-i), gdivgu(gmul(s,L), i));
   return s;
 }
 /* subtract successively rather than adding everything then subtracting.
@@ -1952,7 +1978,7 @@ lfuncheckfeq_i(GEN theta, GEN thetad, GEN t0, GEN t0i, long bitprec)
       { /* inefficient since theta not needed; no need to optimize for this
            (artificial) query [e.g. lfuncheckfeq(t_POL)] */
         GEN T = gel(ldata_get_an(ldata), 2);
-        GEN L = lfunzetakinit(T,zerovec(3),0,0,bitprec);
+        GEN L = lfunzetakinit(T,zerovec(3),0,bitprec);
         return lfuncheckfeq(L,t0,bitprec);
       }
       v = lfunrootres(theta, bitprec);
@@ -1991,7 +2017,7 @@ lfuncheckfeq(GEN lmisc, GEN t0, long bitprec)
   av = avma;
   if (!t0)
   { /* ~Pi/3 + I/7, some random complex number */
-    t0 = mkcomplex(sstoQ(355,339), sstoQ(1,7));
+    t0 = mkcomplex(uutoQ(355,339), uutoQ(1,7));
     t0i = ginv(t0);
   }
   else if (gcmpgs(gnorm(t0), 1) < 0) { t0i = t0; t0 = ginv(t0); }
@@ -2008,11 +2034,13 @@ lfuncheckfeq(GEN lmisc, GEN t0, long bitprec)
 /*******************************************************************/
 /* round root number to \pm 1 if close to integer. */
 static GEN
-ropm1(GEN eno, long prec)
+ropm1(GEN w, long prec)
 {
   long e;
-  GEN r = grndtoi(eno, &e);
-  return (e < -prec2nbits(prec)/2)? r: eno;
+  GEN r;
+  if (typ(w) == t_INT) return w;
+  r = grndtoi(w, &e);
+  return (e < -prec2nbits(prec)/2)? r: w;
 }
 
 /* theta for t=1/sqrt(2) and t2==2t simultaneously, saving 25% of the work.
@@ -2046,7 +2074,7 @@ lfunthetaspec(GEN linit, long bitprec, GEN *pv, GEN *pv2)
    * theta(t) / theta(2t) without assuming t = 1/sqrt(2) */
   K = theta_get_K(thetainit);
   vroots = mkvroots(d, L, prec);
-  t = gpow(t, gdivgs(gen_2, d), prec); /* rt variant: t->t^(2/d) */
+  t = gpow(t, gdivgu(gen_2, d), prec); /* rt variant: t->t^(2/d) */
   /* v = \sum_{n <= L, n odd} a_n K(nt) */
   for (v = gen_0, n = 1; n <= L; n+=2)
   {
@@ -2079,7 +2107,7 @@ static GEN
 Rtor(GEN a, GEN R, GEN ldata, long prec)
 {
   GEN FVga = gammafactor(ldata_get_gammavec(ldata));
-  GEN Na = gpow(ldata_get_conductor(ldata), gdivgs(a,2), prec);
+  GEN Na = gpow(ldata_get_conductor(ldata), gdivgu(a,2), prec);
   long ext;
   return gdiv(R, gmul(Na, gammafactproduct(FVga, a, &ext, prec)));
 }
@@ -2171,29 +2199,27 @@ lfunrootres(GEN data, long bitprec)
     R = lfuntheta(linit, gen_1, 0, bitprec);
   else
   {
+    GEN p2k = gpow(gen_2,k,prec);
     lfunthetaspec(linit, bitprec, &v2, &v);
     if (gequal(gmulsg(2, be), k)) pari_err_IMPL("pole at k/2 in lfunrootres");
     if (gequal(be, k))
     {
-      GEN p2k = gpow(gen_2,k,prec);
       a = conj_i(gsub(gmul(p2k, v), v2));
       b = subiu(p2k, 1);
       e = gmul(gsqrt(p2k, prec), gsub(v2, v));
     }
     else
     {
-      GEN p2k = gpow(gen_2,k,prec);
       GEN tk2 = gsqrt(p2k, prec);
       GEN tbe = gpow(gen_2, be, prec);
-      GEN tkbe = gpow(gen_2, gdivgs(gsub(k, be), 2), prec);
+      GEN tkbe = gpow(gen_2, gdivgu(gsub(k, be), 2), prec);
       a = conj_i(gsub(gmul(tbe, v), v2));
       b = gsub(gdiv(tbe, tkbe), tkbe);
       e = gsub(gmul(gdiv(tbe, tk2), v2), gmul(tk2, v));
     }
-    if (!isintzero(w)) R = gdiv(gsub(e, gmul(a, w)), b);
-    else
+    if (isintzero(w))
     { /* Now residue unknown, r = [[be,0]], and w unknown. */
-      GEN t0  = mkfrac(stoi(11),stoi(10));
+      GEN t0  = mkfrac(utoi(11),utoi(10));
       GEN th1 = lfuntheta(linit, t0,  0, bitprec);
       GEN th2 = lfuntheta(linit, ginv(t0), 0, bitprec);
       GEN tbe = gpow(t0, gmulsg(2, be), prec);
@@ -2204,12 +2230,13 @@ lfunrootres(GEN data, long bitprec)
       GEN f = gsub(gmul(gdiv(tbe, tk2), th2), gmul(tk2, th1));
       GEN D = gsub(gmul(a, d), gmul(b, c));
       w = gdiv(gsub(gmul(d, e), gmul(b, f)), D);
-      R = gdiv(gsub(gmul(a, f), gmul(c, e)), D);
     }
+    w = ropm1(w, prec);
+    R = gdiv(gsub(e, gmul(a, w)), b);
   }
   r = normalize_simple_pole(Rtor(be, R, ldata, prec), be);
   R = lfunrtoR_i(ldata, r, w, prec);
-  return gerepilecopy(ltop, mkvec3(r, R, ropm1(w, prec)));
+  return gerepilecopy(ltop, mkvec3(r, R, w));
 }
 
 /*******************************************************************/
@@ -2345,7 +2372,7 @@ lfunzeros(GEN ldata, GEN lim, long divz, long bitprec)
   cN = gdiv(ldata_get_conductor(ldata), gpowgs(Pi2n(-1, prec), d));
   cN = gexpo(cN) >= 0? gaddsg(d, gmulsg(2, glog(cN, prec))): utoi(d);
   pi2 = Pi2n(1, prec);
-  pi2div = gdivgs(pi2, labs(divz));
+  pi2div = gdivgu(pi2, labs(divz));
   s1 = gsigne(h1);
   s2 = gsigne(h2);
   w = cgetg(100+1, t_VEC); c = 1; ct = 0; T = NULL;
@@ -2492,7 +2519,7 @@ parse_maxcond(GEN maxN)
   else if (typ(maxN) == t_VEC)
   {
     if (!RgV_is_ZV(maxN)) pari_err_TYPE("lfunconductor",maxN);
-    return ZV_sort(maxN);
+    return ZV_sort_shallow(maxN);
   }
   else
     M = maxN;
@@ -2515,12 +2542,12 @@ lfunconductor(GEN data, GEN maxcond, long flag, long bitprec)
     if (lg(M) == 1) { set_avma(av); return cgetg(1,t_VEC); }
     eval = NULL; tdom = dbltor(0.7);
   }
-  else if (!r) { eval = wrap1; tdom = sstoQ(10,11); }
+  else if (!r) { eval = wrap1; tdom = uutoQ(10,11); }
   else
   {
     if (typ(r) == t_VEC && lg(r) > 2)
       pari_err_IMPL("multiple poles in lfunconductor");
-    eval = wrap2; tdom = sstoQ(11,13);
+    eval = wrap2; tdom = uutoQ(11,13);
   }
   if (eval) bitprec += bitprec/2;
   prec = nbits2prec(bitprec);
@@ -2542,7 +2569,7 @@ lfunconductor(GEN data, GEN maxcond, long flag, long bitprec)
   if (!eval)
   {
     long i, besti = 0, beste = -10, l = lg(M);
-    t0 = sstoQ(11,10); t0i = sstoQ(10,11);
+    t0 = uutoQ(11,10); t0i = uutoQ(10,11);
     for (i = 1; i < l; i++)
     {
       pari_sp av2 = avma;
