@@ -21,6 +21,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_lfun
+
 static GEN
 tag(GEN x, long t) { return mkvec2(mkvecsmall(t), x); }
 
@@ -31,14 +33,20 @@ is_tagged(GEN v)
   GEN T = gel(v,1);
   return (typ(T)==t_VEC && lg(T)==3 && typ(gel(T,1))==t_VECSMALL);
 }
+/* rough check */
+static long
+is_ldata(GEN L)
+{
+  long l = lg(L);
+  return typ(L) == t_VEC && (l == 7 || l == 8);
+}
+/* thorough check */
 static void
 checkldata(GEN ldata)
 {
   GEN vga, w, N;
 #if 0 /* assumed already checked and true */
-  long l = lg(ldata);
-  if (typ(ldata)!=t_VEC || l < 7 || l > 8 || !is_tagged(ldata))
-    pari_err_TYPE("checkldata", ldata);
+  if (!is_ldata(ldata) || !is_tagged(ldata)) pari_err_TYPE("checkldata", ldata);
 #endif
   vga = ldata_get_gammavec(ldata);
   if (typ(vga) != t_VEC) pari_err_TYPE("checkldata [gammavec]",vga);
@@ -53,35 +61,70 @@ checkldata(GEN ldata)
   if (typ(N) != t_INT) pari_err_TYPE("checkldata [conductor]",N);
 }
 
+/* tag as t_LFUN_GENERIC */
+static void
+lfuncreate_tag(GEN L)
+{
+  if (is_tagged(L)) return;
+  gel(L,1) = tag(gel(L,1), t_LFUN_GENERIC);
+  if (typ(gel(L,2)) != t_INT) gel(L,2) = tag(gel(L,2), t_LFUN_GENERIC);
+}
+
+/* shallow */
+static GEN
+closure2ldata(GEN C, long prec)
+{
+  GEN L = closure_callgen0prec(C, prec);
+  if (is_ldata(L)) { checkldata(L); lfuncreate_tag(L); }
+  else L = lfunmisc_to_ldata_shallow(L);
+  return L;
+}
+
 /* data may be either an object (polynomial, elliptic curve, etc...)
  * or a description vector [an,sd,Vga,k,conductor,rootno,{poles}]. */
 GEN
 lfuncreate(GEN data)
 {
-  long lx = lg(data);
-  if (typ(data)==t_VEC && (lx == 7 || lx == 8))
+  if (is_ldata(data))
   {
-    GEN ldata = gcopy(data);
-    if (!is_tagged(data))
-    { /* tag first component as t_LFUN_GENERIC */
-      gel(ldata, 1) = tag(gel(ldata,1), t_LFUN_GENERIC);
-      if (typ(gel(ldata, 2))!=t_INT)
-        gel(ldata, 2) = tag(gel(ldata,2), t_LFUN_GENERIC);
-    }
-    checkldata(ldata); return ldata;
-  } else if (typ(data)==t_CLOSURE && closure_arity(data)==0)
+    GEN L = gcopy(data);
+    lfuncreate_tag(L); checkldata(L); return L;
+  }
+  if (typ(data) == t_CLOSURE && closure_arity(data)==0)
   {
     pari_sp av = avma;
-    GEN ldata = lfuncreate(closure_callgen0prec(data, DEFAULTPREC));
-    gel(ldata,1) = tag(data, t_LFUN_CLOSURE0);
-    return gerepilecopy(av, ldata);
+    GEN L = closure2ldata(data, DEFAULTPREC);
+    gel(L,1) = tag(data, t_LFUN_CLOSURE0); return gerepilecopy(av, L);
   }
   return lfunmisc_to_ldata(data);
 }
 
+GEN
+lfunparams(GEN L, long prec)
+{
+  pari_sp av = avma;
+  GEN k, N, v;
+  long p;
+
+  if (!is_ldata(L) || !is_tagged(L)) L = lfunmisc_to_ldata_shallow(L);
+  N = ldata_get_conductor(L);
+  k = ldata_get_k(L);
+  v = ldata_get_gammavec(L);
+  p = gprecision(v);
+  if (p > prec) v = gprec_wtrunc(v, prec);
+  else if (p < prec)
+  {
+    GEN van = ldata_get_an(L), an = gel(van,2);
+    long t = mael(van,1,1);
+    if (t == t_LFUN_CLOSURE0) L = closure2ldata(an, prec);
+  }
+  return gerepilecopy(av, mkvec3(N, k, v));
+}
+
 /********************************************************************/
 /**                     Simple constructors                        **/
 /********************************************************************/
+static GEN ldata_eulerf(GEN van, GEN p, long prec);
 
 static GEN
 vecan_conj(GEN an, long n, long prec)
@@ -90,6 +133,13 @@ vecan_conj(GEN an, long n, long prec)
   return typ(p1) == t_VEC? conj_i(p1): p1;
 }
 
+static GEN
+eulerf_conj(GEN an, GEN p, long prec)
+{
+  GEN p1 = ldata_eulerf(gel(an,1), p, prec);
+  return conj_i(p1);
+}
+
 static GEN
 vecan_mul(GEN an, long n, long prec)
 {
@@ -100,6 +150,14 @@ vecan_mul(GEN an, long n, long prec)
   return dirmul(p1, p2);
 }
 
+static GEN
+eulerf_mul(GEN an, GEN p, long prec)
+{
+  GEN p1 = ldata_eulerf(gel(an,1), p, prec);
+  GEN p2 = ldata_eulerf(gel(an,2), p, prec);
+  return gmul(p1, p2);
+}
+
 static GEN
 lfunconvol(GEN a1, GEN a2)
 { return tag(mkvec2(a1, a2), t_LFUN_MUL); }
@@ -114,6 +172,14 @@ vecan_div(GEN an, long n, long prec)
   return dirdiv(p1, p2);
 }
 
+static GEN
+eulerf_div(GEN an, GEN p, long prec)
+{
+  GEN p1 = ldata_eulerf(gel(an,1), p, prec);
+  GEN p2 = ldata_eulerf(gel(an,2), p, prec);
+  return gdiv(p1, p2);
+}
+
 static GEN
 lfunconvolinv(GEN a1, GEN a2)
 { return tag(mkvec2(a1,a2), t_LFUN_DIV); }
@@ -152,6 +218,17 @@ vecan_twist(GEN an, long n, long prec)
   return V;
 }
 
+static GEN
+eulerf_twist(GEN an, GEN p, long prec)
+{
+  GEN p1 = ldata_eulerf(gel(an,1), p, prec);
+  GEN p2 = ginv(ldata_eulerf(gel(an,2), p, prec));
+  if (typ(p2)!=t_POL || degpol(p2)==0)
+    return poleval(p1,pol_0(0));
+  if (degpol(p2)!=1) pari_err_IMPL("lfuneuler");
+  return poleval(p1,monomial(gneg(gel(p2,3)),1,0));
+}
+
 static GEN
 vecan_shift(GEN an, long n, long prec)
 {
@@ -167,7 +244,7 @@ vecan_shift(GEN an, long n, long prec)
       for(i = 1; i <= n ; i++)
       {
         GEN gi = gel(p1, i);
-        gel(V, i) = gequal0(gi)? gi: gmulgs(gi, i);
+        gel(V, i) = gequal0(gi)? gi: gmulgu(gi, i);
       }
     else
       for(i = 1; i <= n ; i++)
@@ -185,39 +262,92 @@ vecan_shift(GEN an, long n, long prec)
   return V;
 }
 
+static GEN
+eulerf_shift(GEN an, GEN p, long prec)
+{
+  GEN p1 = ldata_eulerf(gel(an,1), p, prec);
+  GEN s = gel(an,2);
+  return gsubst(p1, 0, monomial(gpow(p, s, prec), 1, 0));
+}
+
+static GEN
+eulerf_hgm(GEN an, GEN p)
+{
+  GEN H = gel(an,1), t = gel(an,2);
+  if (typ(t)==t_VEC && lg(t)==3)
+  {
+    GEN L = gel(t,2);
+    long i, l = lg(L);
+    t = gel(t,1);
+    for (i = 1; i < l; i++) /* wild primes */
+      if (equalii(p, gmael(L, i, 1))) break;
+    if (i<l) return gmael(L,i,2);
+  }
+  return ginv(hgmeulerfactor(H, t, itos(p), NULL));
+}
+
 static GEN
 deg1ser_shallow(GEN a1, GEN a0, long e)
 { return RgX_to_ser(deg1pol_shallow(a1, a0, 0), e+2); }
+/* lfunrtopoles without sort */
 static GEN
-residuetopoles(GEN r, GEN k)
+rtopoles(GEN r)
 {
-  if (!is_vec_t(typ(r))) r = mkvec(mkvec2(k, r));
-  return lfunrtopoles(r);
+  long j, l = lg(r);
+  GEN v = cgetg(l, t_VEC);
+  for (j = 1; j < l; j++)
+  {
+    GEN rj = gel(r,j), a = gel(rj,1);
+    gel(v,j) = a;
+  }
+  return v;
 }
+/* re = polar part; overestimate when re = gen_0 (unknown) */
+static long
+orderpole(GEN re) { return typ(re) == t_SER? -valp(re): 1; }
 static GEN
 lfunmulpoles(GEN ldata1, GEN ldata2, long bitprec)
 {
-  GEN k = ldata_get_k(ldata1);
+  GEN r, k = ldata_get_k(ldata1), b1 = NULL, b2 = NULL;
   GEN r1 = ldata_get_residue(ldata1);
-  GEN r2 = ldata_get_residue(ldata2), r;
-  long l, j;
+  GEN r2 = ldata_get_residue(ldata2);
+  long i, j, l, L = 0;
 
   if (!r1 && !r2) return NULL;
-  if (r1) r1 = residuetopoles(r1, k);
-  if (r2) r2 = residuetopoles(r2, k);
-  r = r1? (r2? setunion_i(r1, r2): r1): r2;
-  l = lg(r); if (l == 1) return NULL;
-  for (j = 1; j < l; j++)
+  if (r1 && !is_vec_t(typ(r1))) r1 = mkvec(mkvec2(k, r1));
+  if (r2 && !is_vec_t(typ(r2))) r2 = mkvec(mkvec2(k, r2));
+  if (r1) { b1 = rtopoles(r1); L += lg(b1); }
+  if (r2) { b2 = rtopoles(r2); L += lg(b2); }
+  r = cgetg(L, t_VEC); j = 1;
+  if (b1)
   {
-    GEN be = gel(r,j), bx = deg1ser_shallow(gen_1, be, 2);
-    GEN z1 = lfun(ldata1,bx,bitprec), z2 = lfun(ldata2,bx,bitprec);
-    long e = valp(z1) + valp(z2);
-    GEN b = deg1ser_shallow(gen_1, be, 2-e);
-    z1 = lfun(ldata1,b,bitprec);
-    z2 = lfun(ldata2,b,bitprec);
-    gel(r,j) = mkvec2(be, gmul(z1, z2));
+    l = lg(b1);
+    for (i = 1; i < l; i++)
+    {
+      GEN z, z1, z2, be = gmael(r1,i,1);
+      long n, v = orderpole(gmael(r1,i,2));
+      if (b2 && (n = RgV_isin(b2, be))) v += orderpole(gmael(r2,n,2));
+      z = deg1ser_shallow(gen_1, be, 2 + v);
+      z1 = lfun(ldata1,z,bitprec);
+      z2 = lfun(ldata2,z,bitprec);
+      gel(r,j++) = mkvec2(be, gmul(z1, z2));
+    }
   }
-  return r;
+  if (b2)
+  {
+    long l = lg(b2);
+    for (i = 1; i < l; i++)
+    {
+      GEN z, z1, z2, be = gmael(r2,i,1);
+      long n, v = orderpole(gmael(r2,i,2));
+      if (b1 && (n = RgV_isin(b1, be))) continue; /* done already */
+      z = deg1ser_shallow(gen_1, be, 2 + v);
+      z1 = lfun(ldata1,z,bitprec);
+      z2 = lfun(ldata2,z,bitprec);
+      gel(r,j++) = mkvec2(be, gmul(z1, z2));
+    }
+  }
+  setlg(r, j); return r;
 }
 
 static GEN
@@ -252,31 +382,56 @@ static GEN
 lfundivpoles(GEN ldata1, GEN ldata2, long bitprec)
 {
   long i, j, l;
-  GEN k  = ldata_get_k(ldata1);
+  GEN be2, k  = ldata_get_k(ldata1);
   GEN r1 = ldata_get_residue(ldata1);
   GEN r2 = ldata_get_residue(ldata2), r;
 
-  if (r1 && typ(r1) != t_VEC) r1 = mkvec(mkvec2(k, r1));
-  if (r2 && typ(r2) != t_VEC) r2 = mkvec(mkvec2(k, r2));
+  if (r1 && !is_vec_t(typ(r1))) r1 = mkvec(mkvec2(k, r1));
+  if (r2 && !is_vec_t(typ(r2))) r2 = mkvec(mkvec2(k, r2));
   if (!r1) return NULL;
-  r1 = lfunrtopoles(r1);
   l = lg(r1); r = cgetg(l, t_VEC);
+  be2 = r2? rtopoles(r2): NULL;
   for (i = j = 1; j < l; j++)
   {
-    GEN be = gel(r1,j);
-    GEN z = gdiv(lfun(ldata1,be,bitprec), lfun(ldata2,be,bitprec));
+    GEN z, v = gel(r1,j), be = gel(v,1), s1 = gel(v,2);
+    long n;
+    if (be2 && (n = RgV_isin(be2, be)))
+    {
+      GEN s2 = gmael(r2,n,2); /* s1,s2: polar parts */
+      if (orderpole(s1) == orderpole(s2)) continue;
+    }
+    z = gdiv(lfun(ldata1,be,bitprec), lfun(ldata2,be,bitprec));
     if (valp(z) < 0) gel(r,i++) = mkvec2(be, z);
   }
   if (i == 1) return NULL;
   setlg(r, i); return r;
 }
 
+static GEN
+lfunvgasub(GEN v01, GEN v2)
+{
+  GEN v1 = shallowcopy(v01), v;
+  long l1 = lg(v1), l2 = lg(v2), j1, j2, j;
+  for (j2 = 1; j2 < l2; j2++)
+  {
+    for (j1 = 1; j1 < l1; j1++)
+      if (gel(v1,j1) && gequal(gel(v1,j1), gel(v2,j2)))
+      {
+        gel(v1,j1) = NULL; break;
+      }
+    if (j1 == l1) pari_err_OP("lfunvgasub", v1, v2);
+  }
+  v = cgetg(l1-l2+1, t_VEC);
+  for (j1 = j = 1; j1 < l1; j1++)
+    if (gel(v1, j1)) gel(v,j++) = gel(v1,j1);
+  return v;
+}
+
 GEN
 lfundiv(GEN ldata1, GEN ldata2, long bitprec)
 {
   pari_sp ltop = avma;
-  GEN k, r, N, v, v1, v2, eno, a1a2, b1b2, LD, eno2;
-  long j, j1, j2, l1, l2;
+  GEN k, r, N, v, eno, a1a2, b1b2, LD, eno2;
   long prec = nbits2prec(bitprec);
   ldata1 = ldata_newprec(lfunmisc_to_ldata_shallow(ldata1), prec);
   ldata2 = ldata_newprec(lfunmisc_to_ldata_shallow(ldata2), prec);
@@ -290,22 +445,7 @@ lfundiv(GEN ldata1, GEN ldata2, long bitprec)
   b1b2 = lfuncombdual(lfunconvolinv, ldata1, ldata2);
   eno2 = ldata_get_rootno(ldata2);
   eno = isintzero(eno2)? gen_0: gdiv(ldata_get_rootno(ldata1), eno2);
-  v1 = shallowcopy(ldata_get_gammavec(ldata1));
-  v2 = ldata_get_gammavec(ldata2);
-  l1 = lg(v1); l2 = lg(v2);
-  for (j2 = 1; j2 < l2; j2++)
-  {
-    for (j1 = 1; j1 < l1; j1++)
-      if (gel(v1,j1) && gequal(gel(v1,j1), gel(v2,j2)))
-      {
-        gel(v1,j1) = NULL; break;
-      }
-    if (j1 == l1) pari_err_OP("lfundiv [Vga]",ldata1, ldata2);
-  }
-  v = cgetg(l1-l2+1, t_VEC);
-  for (j1 = j = 1; j1 < l1; j1++)
-    if (gel(v1, j1)) gel(v,j++) = gel(v1,j1);
-
+  v = lfunvgasub(ldata_get_gammavec(ldata1), ldata_get_gammavec(ldata2));
   LD = mkvecn(7, a1a2, b1b2, v, k, N, eno, r);
   if (!r) setlg(LD,7);
   return gerepilecopy(ltop, LD);
@@ -547,6 +687,38 @@ vecan_closure(GEN a, long L, long prec)
   pop_localprec(); return a;
 }
 
+static GEN
+eulerf_closure(GEN a, GEN p, long prec)
+{
+  long ta = typ(a);
+  GEN Sbad = NULL, f;
+
+  if (ta == t_VEC)
+  {
+    long l = lg(a);
+    if (l == 1) pari_err_TYPE("vecan_closure", a);
+    ta = typ(gel(a,1));
+    /* regular vector, return it */
+    if (ta != t_CLOSURE) return NULL;
+    if (l != 3) pari_err_TYPE("vecan_closure", a);
+    Sbad = gel(a,2);
+    if (typ(Sbad) != t_VEC) pari_err_TYPE("vecan_closure", a);
+    a = gel(a,1);
+  }
+  else if (ta != t_CLOSURE) pari_err_TYPE("vecan_closure", a);
+  push_localprec(prec);
+  switch(closure_arity(a))
+  {
+    case 2:
+      f = closure_callgen2(a, p, mkoo()); break;
+    case 1:
+      f = NULL; break;
+    default:
+      f = NULL; pari_err_TYPE("vecan_closure", a);
+  }
+  pop_localprec(); return f;
+}
+
 /*****************************************************************/
 /*  L-series of Dirichlet characters.                            */
 /*****************************************************************/
@@ -705,6 +877,18 @@ vecan_chiZ(GEN an, long n, long prec)
   return v;
 }
 
+static GEN
+eulerf_chiZ(GEN an, GEN p, long prec)
+{
+  GEN G = gel(an,1);
+  GEN nchi = gel(an,2), gord = gel(nchi,1), chi = gel(nchi,2);
+  long multichi= (lg(chi) > 1 && is_vec_t(typ(gel(chi,1))));
+  GEN z = rootsof1_cx(gord, prec);
+  GEN N = znstar_get_N(G);
+  GEN ch = dvdii(N,p) ? gen_0: chigenevalvec(znconreylog(G, p), nchi, z, prec, multichi);
+  return mkrfrac(gen_1, deg1pol_shallow(gneg(ch), gen_1,0));
+}
+
 static GEN
 vecan_chigen(GEN an, long n, long prec)
 {
@@ -771,7 +955,45 @@ vecan_chigen(GEN an, long n, long prec)
   return v;
 }
 
-static GEN lfunzetak_i(GEN T);
+static GEN
+eulerf_chigen(GEN an, GEN p, long prec)
+{
+  GEN bnr = gel(an,1), nf = bnr_get_nf(bnr);
+  GEN nchi = gel(an,2), gord = gel(nchi,1), chi = gel(nchi,2), z;
+  GEN N = gel(bnr_get_mod(bnr), 1), NZ = gcoeff(N,1,1), f;
+  long multichi= (lg(chi) > 1 && is_vec_t(typ(gel(chi,1))));
+
+  z = rootsof1_cx(gord, prec);
+  if (nf_get_degree(nf) == 1)
+  {
+    GEN ch;
+    if (dvdii(NZ,p)) ch = gen_0;
+    else
+    {
+      ch = chigenevalvec(isprincipalray(bnr,p), nchi, z, prec, multichi);
+      if (typ(ch)==t_VEC) return NULL;
+    }
+    f = deg1pol_shallow(gneg(ch), gen_1, 0);
+  }
+  else
+  {
+    int check = dvdii(NZ,p);
+    GEN L = idealprimedec(nf, p);
+    long j, lL = lg(L);
+    f = pol_1(0);
+    for (j = 1; j < lL; j++)
+    {
+      GEN pr = gel(L, j), ch;
+      if (check && idealval(nf, N, pr)) ch = gen_0;
+      else
+      ch = chigenevalvec(isprincipalray(bnr,pr), nchi, z, prec, multichi);
+      if (typ(ch)==t_VEC) return NULL;
+      f = gmul(f, gsub(gen_1, monomial(ch, pr_get_f(pr), 0)));
+    }
+  }
+  return mkrfrac(gen_1,f);
+}
+
 static GEN
 vec01(long r1, long r2)
 {
@@ -781,6 +1003,26 @@ vec01(long r1, long r2)
   for (     ; i <= d;  i++) gel(v,i) = gen_1;
   return v;
 }
+
+/* true nf or t_POL */
+static GEN
+lfunzetak_i(GEN T)
+{
+  GEN Vga, N;
+  long r1, r2;
+  if (typ(T) == t_POL)
+  {
+    T = nfinit(T, DEFAULTPREC);
+    if (lg(T) == 3) T = gel(T,1); /* [nf,change of var] */
+  }
+  nf_get_sign(T,&r1,&r2); Vga = vec01(r1+r2,r2);
+  N = absi_shallow(nf_get_disc(T));
+  return mkvecn(7, tag(T,t_LFUN_NF), gen_0, Vga, gen_1, N, gen_1, gen_0);
+}
+static GEN
+lfunzetak(GEN T)
+{ pari_sp av = avma; return gerepilecopy(av, lfunzetak_i(T)); }
+
 /* v = vector of normalized characters of order dividing o; renormalize
  * so that all have same apparent order o */
 static GEN
@@ -909,7 +1151,7 @@ lfunchigen(GEN bnr, GEN CHI)
   if (!nchi)
   {
     if (equali1(NN)) { set_avma(av); return lfunzeta(); }
-    if (ZV_equal0(CHI)) return gerepilecopy(av, lfunzetak_i(bnr_get_bnf(bnr)));
+    if (ZV_equal0(CHI)) return gerepilecopy(av, lfunzetak_i(bnr_get_nf(bnr)));
     nchi = char_normalize(CHI, cyc_normalize(bnr_get_cyc(bnr)));
   }
   real = abscmpiu(gel(nchi,1), 2) <= 0;
@@ -942,36 +1184,6 @@ chigenkerfind(GEN bnr, GEN H, GEN *pcnj)
   setlg(res, k); return res;
 }
 
-static GEN
-lfunzetak_i(GEN T)
-{
-  GEN Vga, N, nf, bnf = checkbnf_i(T), r = gen_0/*unknown*/;
-  long r1, r2;
-
-  if (bnf)
-    nf = bnf_get_nf(bnf);
-  else
-  {
-    nf = checknf_i(T);
-    if (!nf) nf = T = nfinit(T, DEFAULTPREC);
-  }
-  nf_get_sign(nf,&r1,&r2);
-  N = absi_shallow(nf_get_disc(nf));
-  if (bnf)
-  {
-    GEN h = bnf_get_no(bnf);
-    GEN R = bnf_get_reg(bnf);
-    long prec = nf_get_prec(nf);
-    r = gdiv(gmul(mulir(shifti(h, r1+r2), powru(mppi(prec), r2)), R),
-             mulur(bnf_get_tuN(bnf), gsqrt(N, prec)));
-  }
-  Vga = vec01(r1+r2,r2);
-  return mkvecn(7, tag(T,t_LFUN_NF), gen_0, Vga, gen_1, N, gen_1, r);
-}
-static GEN
-lfunzetak(GEN T)
-{ pari_sp ltop = avma; return gerepilecopy(ltop, lfunzetak_i(T)); }
-
 /* bnf = NULL: base field = Q */
 GEN
 lfunabelianrelinit(GEN nfabs, GEN bnf, GEN polrel, GEN dom, long der, long bitprec)
@@ -1006,6 +1218,7 @@ lfunabelianrelinit(GEN nfabs, GEN bnf, GEN polrel, GEN dom, long der, long bitpr
 /*****************************************************************/
 /*                 Dedekind zeta functions                       */
 /*****************************************************************/
+/* true nf */
 static GEN
 dirzetak0(GEN nf, ulong N)
 {
@@ -1059,6 +1272,21 @@ dirzetak0(GEN nf, ulong N)
   pari_free(c2); return c;
 }
 
+static GEN
+eulerf_zetak(GEN nf, GEN p)
+{
+  GEN vect, T = nf_get_pol(nf), index = nf_get_index(nf), f = pol_1(0);
+  long i, l;
+  if (dvdii(index, p)) /* p does not divide index */
+    vect = idealprimedec_degrees(nf,p);
+  else
+    vect = gel(FpX_degfact(T,p),1);
+  l = lg(vect);
+  for (i = 1; i < l; i++)
+    f = gmul(f, gsub(gen_1, monomial(gen_1, vect[i], 0)));
+  return gcopy(mkrfrac(gen_1, f));
+}
+
 GEN
 dirzetak(GEN nf, GEN b)
 {
@@ -1093,14 +1321,6 @@ lfunproduct(GEN ldata, GEN linit1, GEN linit2, GEN domain)
   return lfuninit_make(t_LDESC_PRODUCT, ldata, M3, domain);
 }
 
-static GEN
-lfunzetakinit_raw(GEN T, GEN dom, long der, long bitprec)
-{
-  pari_sp ltop = avma;
-  GEN ldata = lfunzetak_i(T);
-  return gerepileupto(ltop, lfuninit(ldata, dom, der, bitprec));
-}
-
 static GEN
 lfunzetakinit_quotient(GEN nf, GEN polk, GEN dom, long der, long bitprec)
 {
@@ -1110,7 +1330,7 @@ lfunzetakinit_quotient(GEN nf, GEN polk, GEN dom, long der, long bitprec)
 
   nf_get_sign(nf,&r1,&r2);
   nfk = nfinit(polk, nbits2prec(bitprec));
-  Lk = lfunzetakinit(nfk, dom, der, 0, bitprec); /* zeta_k */
+  Lk = lfunzetakinit(nfk, dom, der, bitprec); /* zeta_k */
   nf_get_sign(nfk,&r1k,&r2k);
   Vga = vec01((r1+r2) - (r1k+r2k), r2-r2k);
   N = absi_shallow(diviiexact(nf_get_disc(nf), nf_get_disc(nfk)));
@@ -1134,10 +1354,10 @@ lfunzetakinit_Galois(GEN nf, GEN gal, GEN dom, long der, long bitprec)
   else return lfunzetakinit_artin(nf, gal, dom, der, bitprec);
 }
 
+/* true nf */
 GEN
-lfunzetakinit(GEN NF, GEN dom, long der, long flag, long bitprec)
+lfunzetakinit(GEN nf, GEN dom, long der, long bitprec)
 {
-  GEN nf = checknf(NF);
   GEN G, nfs, sbg;
   long lf, d = nf_get_degree(nf);
   if (d == 1) return lfunzetainit(dom, der, bitprec);
@@ -1146,8 +1366,6 @@ lfunzetakinit(GEN NF, GEN dom, long der, long flag, long bitprec)
     return lfunzetakinit_Galois(nf, G, dom, der, bitprec);
   nfs = nfsubfields(nf, 0); lf = lg(nfs)-1;
   sbg = gmael(nfs,lf-1,1);
-  if (flag && d > 4*degpol(sbg))
-    return lfunzetakinit_raw(nf, dom, der, bitprec);
   return lfunzetakinit_quotient(nf, sbg, dom, der, bitprec);
 }
 
@@ -1288,6 +1506,16 @@ direllsympow_worker(GEN P, ulong X, GEN E, ulong m)
   return gerepilecopy(av, mkvec2(P,W));
 }
 
+static GEN
+eulerf_bad(GEN bad, GEN p)
+{
+  long i, l = lg(bad);
+  for (i = 1; i < l; i++)
+    if (equalii(gmael(bad,i,1), p))
+      return gmael(bad,i,2);
+  return NULL;
+}
+
 static GEN
 vecan_ellsympow(GEN an, long n)
 {
@@ -1296,6 +1524,15 @@ vecan_ellsympow(GEN an, long n)
   return pardireuler(worker, gen_2, nn, nn, bad);
 }
 
+static GEN
+eulerf_ellsympow(GEN an, GEN p)
+{
+  GEN crvm = gel(an,1), bad = gel(an,2), E = gel(crvm,1);
+  GEN f = eulerf_bad(bad, p);
+  if (f) return f;
+  retmkrfrac(gen_1,ellsympow_abelian(p, ellap(E, p), itos(gel(crvm,2)), 1));
+}
+
 static long
 ellsympow_betam(long o, long m)
 {
@@ -1507,11 +1744,10 @@ ellsympow_goodred2(GEN E, GEN F, GEN p, long m, long vN, long *cnd, long *w)
 }
 
 static GEN
-ellminimaldotwist(GEN E, GEN *pD, long prec)
+ellminimaldotwist(GEN E, GEN *pD)
 {
   GEN D = ellminimaltwistcond(E), Et = elltwist(E, D), Etmin;
   if (pD) *pD = D;
-  Et = ellinit(Et, NULL, prec);
   Etmin = ellminimalmodel(Et, NULL);
   obj_free(Et); return Etmin;
 }
@@ -1538,7 +1774,7 @@ lfunellsympow(GEN e, ulong m)
   if (ugcd(umodiu(N,6), 6) == 1)
     et = NULL;
   else
-    et = ellminimaldotwist(e, NULL, DEFAULTPREC);
+    et = ellminimaldotwist(e, NULL);
   B = gen_1;
   bad = cgetg(l, t_VEC);
   for (i=1; i<l; i++)
@@ -1763,7 +1999,7 @@ GEN
 lfunellmfpeters(GEN E, long bitprec)
 {
   pari_sp ltop = avma;
-  GEN D, Et = ellminimaldotwist(E, &D, nbits2prec(bitprec));
+  GEN D, Et = ellminimaldotwist(E, &D);
   GEN nor = lfunellmfpetersmintwist(Et, bitprec);
   GEN nor2 = gmul(nor, elldiscfix(E, Et, D));
   obj_free(Et); return gerepileupto(ltop, nor2);
@@ -1856,6 +2092,16 @@ vecan_genus2(GEN an, long L)
   return pardireuler(worker, gen_2, stoi(L), NULL, bad);
 }
 
+static GEN
+eulerf_genus2(GEN an, GEN p)
+{
+  GEN Q = gel(an,1), bad = gel(an, 2);
+  GEN f = eulerf_bad(bad, p);
+  if (f) return f;
+  f = RgX_recip(hyperellcharpoly(gmul(Q,gmodulo(gen_1, p))));
+  return mkrfrac(gen_1,f);
+}
+
 static GEN
 genus2_redmodel(GEN P, GEN p)
 {
@@ -2022,27 +2268,6 @@ F2xqX_quad_nbroots(GEN b, GEN c, GEN T)
     return 1;
 }
 
-static GEN
-genus2_redmodel2(GEN P)
-{
-  GEN Q = pol_0(varn(P));
-  GEN P2 = ZX_to_F2x(P);
-  while (F2x_issquare(P2))
-  {
-    GEN H = F2x_to_ZX(F2x_sqrt(P2));
-    GEN P1 = ZX_sub(P, ZX_sqr(H));
-    GEN Q1 = ZX_add(Q, ZX_mulu(H, 2));
-    if ((signe(P1)==0 ||  ZX_lval(P1,2)>=2)
-     && (signe(Q1)==0 ||  ZX_lval(Q1,2)>=1))
-    {
-      P = ZX_shifti(P1, -2);
-      Q = ZX_shifti(Q1, -1);
-      P2= ZX_to_F2x(P);
-    } else break;
-  }
-  return mkvec2(P,Q);
-}
-
 static GEN
 genus2_eulerfact2(GEN PQ)
 {
@@ -2092,9 +2317,8 @@ lfungenus2(GEN G)
   pari_sp ltop = avma;
   GEN Ldata;
   GEN gr = genus2red(G, NULL);
-  GEN N  = gel(gr, 1), M = gel(gr, 2), Q = gel(gr, 3), L = gel(gr, 4);
-  GEN PQ = genus2_redmodel2(Q);
-  GEN e;
+  GEN N  = gel(gr, 1), M = gel(gr, 2), PQ = gel(gr, 3), L = gel(gr, 4);
+  GEN e, F = gadd(gsqr(gel(PQ, 2)), gmul2n(gel(PQ, 1), 2));
   long i, lL = lg(L), ram2;
   ram2 = absequaliu(gmael(M,1,1),2);
   if (ram2 && equalis(gmael(M,2,1),-1))
@@ -2106,9 +2330,9 @@ lfungenus2(GEN G)
   {
     GEN Li = gel(L, i);
     GEN p = gel(Li, 1);
-    gel(e, ram2 ? i: i+1) = mkvec2(p, genus2_eulerfact(Q,p));
+    gel(e, ram2 ? i: i+1) = mkvec2(p, genus2_eulerfact(F,p));
   }
-  Ldata = mkvecn(6, tag(mkvec2(Q,e), t_LFUN_GENUS2),
+  Ldata = mkvecn(6, tag(mkvec2(F,e), t_LFUN_GENUS2),
       gen_0, mkvec4(gen_0, gen_0, gen_1, gen_1), gen_2, N, gen_0);
   return gerepilecopy(ltop, Ldata);
 }
@@ -2165,10 +2389,11 @@ eta_product_ZXn(GEN eta, long L)
 static GEN
 vecan_eta(GEN an, long L)
 {
-  GEN v = RgX_to_RgC(eta_product_ZXn(an, L), L);
-  settyp(v, t_VEC); return v;
+  long v = itos(gel(an, 3));
+  GEN t = eta_product_ZXn(an, L - v);
+  if (v) t = RgX_shift_shallow(t, v);
+  return RgX_to_RgV(t, L);
 }
-
 /* return 1 if cuspidal, 0 if holomorphic, -1 otherwise */
 static int
 etacuspidal(GEN N, GEN k, GEN B, GEN R, GEN NB)
@@ -2271,10 +2496,9 @@ lfunetaquo(GEN eta0)
   if (!etaquotype(&eta, &N, &k, NULL, &v, &sd, &cusp))
     pari_err_TYPE("lfunetaquo", eta0);
   if (!cusp) pari_err_IMPL("noncuspidal eta quotient");
-  if (v != 1) pari_err_IMPL("valuation != 1");
   if (!sd) pari_err_IMPL("non self-dual eta quotient");
   if (typ(k) != t_INT) pari_err_TYPE("lfunetaquo [nonintegral weight]", eta0);
-  BR = mkvec2(ZV_to_zv(gel(eta,1)), ZV_to_zv(gel(eta,2)));
+  BR = mkvec3(ZV_to_zv(gel(eta,1)), ZV_to_zv(gel(eta,2)), stoi(v - 1));
   Ldata = mkvecn(6, tag(BR,t_LFUN_ETA), gen_0, mkvec2(gen_0,gen_1), k,N, gen_1);
   return gerepilecopy(ltop, Ldata);
 }
@@ -2308,7 +2532,7 @@ lfunqf(GEN M, long prec)
   if (typ(M) != t_MAT) pari_err_TYPE("lfunqf", M);
   if (!RgM_is_ZM(M))   pari_err_TYPE("lfunqf [not integral]", M);
   n = lg(M)-1;
-  k = sstoQ(n,2);
+  k = uutoQ(n,2);
   M = Q_primpart(M);
   Mi = ZM_inv(M, &d); /* d M^(-1) */
   if (!qfiseven(M)) { M = gmul2n(M, 1); d = shifti(d,1); }
@@ -2400,7 +2624,7 @@ artin_dim(GEN ind, GEN ch)
   GEN s = gen_0;
   for(i=1; i<=d; i++)
     s = gadd(s, gel(ch, gel(elts,i)[1]));
-  return gtos(gdivgs(cyclotoi(s), d));
+  return gtos(gdivgu(cyclotoi(s), d));
 }
 
 static GEN
@@ -2410,7 +2634,7 @@ artin_ind(GEN elts, GEN ch, GEN p)
   GEN s = gen_0;
   for(i=1; i<=d; i++)
     s = gadd(s, gel(ch, gmul(gel(elts,i),p)[1]));
-  return gdivgs(s, d);
+  return gdivgu(s, d);
 }
 
 static GEN
@@ -2436,7 +2660,7 @@ artin_ram(GEN nf, GEN gal, GEN aut, GEN pr, GEN ramg, GEN ch, long d)
   setvarn(V,0); return gerepileupto(av, ginv(V));
 }
 
-/* [Artin conductor, vec of [p, Lp]] */
+/* N true nf; [Artin conductor, vec of [p, Lp]] */
 static GEN
 artin_badprimes(GEN N, GEN G, GEN aut, GEN ch)
 {
@@ -2520,6 +2744,7 @@ idealfrobenius_hard(GEN nf, GEN gal, GEN aut, GEN pr)
   return NULL; /* LCOV_EXCL_LINE */
 }
 
+/* true nf */
 static GEN
 dirartin(GEN nf, GEN G, GEN V, GEN aut, GEN p, long n)
 {
@@ -2537,7 +2762,7 @@ dirartin(GEN nf, GEN G, GEN V, GEN aut, GEN p, long n)
     pr = idealprimedec_galois(nf,p);
     frob = idealfrobenius_hard(nf, G, aut, pr);
   }
-  set_avma(av); return RgXn_inv(gel(V, frob[1]), n);
+  set_avma(av); return n ? RgXn_inv(gel(V, frob[1]), n): gel(V, frob[1]);
 }
 
 GEN
@@ -2567,6 +2792,19 @@ vecan_artin(GEN an, long L, long prec)
   return A;
 }
 
+static GEN
+eulerf_artin(GEN an, GEN p, long prec)
+{
+  GEN nf = gel(an,1), G = gel(an,2), V = gel(an,3), aut = gel(an,4);
+  GEN Sbad = gel(an,5);
+  long n = itos(gel(an,6)), isreal = lg(an)<8 ? 0: !itos(gel(an,7));
+  GEN f = eulerf_bad(Sbad, p);
+  if (!f) f = mkrfrac(gen_1,dirartin(nf, G, V, aut, p, 0));
+  f = gsubst(liftpol(f),1, rootsof1u_cx(n, prec));
+  if (isreal) f = real_i(f);
+  return f;
+}
+
 static GEN
 char_expand(GEN conj, GEN ch)
 {
@@ -2685,7 +2923,8 @@ lfunzetakinit_artin(GEN nf, GEN gal, GEN dom, long der, long bitprec)
 /**                    High-level Constructors                     **/
 /********************************************************************/
 enum { t_LFUNMISC_POL, t_LFUNMISC_CHIQUAD, t_LFUNMISC_CHICONREY,
-       t_LFUNMISC_CHIGEN, t_LFUNMISC_ELLINIT, t_LFUNMISC_ETAQUO };
+       t_LFUNMISC_CHIGEN, t_LFUNMISC_ELLINIT, t_LFUNMISC_ETAQUO,
+       t_LFUNMISC_GCHAR };
 static long
 lfundatatype(GEN data)
 {
@@ -2695,27 +2934,29 @@ lfundatatype(GEN data)
     case t_INTMOD: return t_LFUNMISC_CHICONREY;
     case t_POL: return t_LFUNMISC_POL;
     case t_VEC:
+    {
+      long l = lg(data);
       if (checknf_i(data)) return t_LFUNMISC_POL;
-      switch(lg(data))
-      {
-        case 17: return t_LFUNMISC_ELLINIT;
-        case 3:  return t_LFUNMISC_CHIGEN;
-      }
+      if (l == 17) return t_LFUNMISC_ELLINIT;
+      if (l == 3 && typ(gel(data,1)) == t_VEC)
+        return is_gchar_group(gel(data,1))? t_LFUNMISC_GCHAR
+                                          : t_LFUNMISC_CHIGEN;
       break;
+    }
   }
   return -1;
 }
 static GEN
 lfunmisc_to_ldata_i(GEN ldata, long shallow)
 {
-  long lx;
+  GEN x;
   if (is_linit(ldata)) ldata = linit_get_ldata(ldata);
-  lx = lg(ldata);
-  if (typ(ldata)==t_VEC && (lx == 7 || lx == 8) && is_tagged(ldata))
+  if (is_ldata(ldata) && is_tagged(ldata))
   {
     if (!shallow) ldata = gcopy(ldata);
     checkldata(ldata); return ldata;
   }
+  x = checknf_i(ldata); if (x) return lfunzetak(x);
   switch (lfundatatype(ldata))
   {
     case t_LFUNMISC_POL: return lfunzetak(ldata);
@@ -2735,7 +2976,7 @@ lfunmisc_to_ldata_i(GEN ldata, long shallow)
       }
     }
     break;
-
+    case t_LFUNMISC_GCHAR: return lfungchar(gel(ldata,1), gel(ldata,2));
     case t_LFUNMISC_ELLINIT: return lfunell(ldata);
   }
   if (shallow != 2) pari_err_TYPE("lfunmisc_to_ldata",ldata);
@@ -2789,6 +3030,7 @@ ldata_vecan(GEN van, long L, long prec)
     case t_LFUN_KRONECKER: an = vecan_Kronecker(an, L); break;
     case t_LFUN_CHIZ: an = vecan_chiZ(an, L, prec); break;
     case t_LFUN_CHIGEN: an = vecan_chigen(an, L, prec); break;
+    case t_LFUN_HECKE: an = vecan_gchar(an, L, prec); break;
     case t_LFUN_ARTIN: an = vecan_artin(an, L, prec); break;
     case t_LFUN_ETA: an = vecan_eta(an, L); break;
     case t_LFUN_QF: an = vecan_qf(an, L); break;
@@ -2797,6 +3039,8 @@ ldata_vecan(GEN van, long L, long prec)
     case t_LFUN_CONJ: an = vecan_conj(an, L, prec); break;
     case t_LFUN_SYMPOW_ELL: an = vecan_ellsympow(an, L); break;
     case t_LFUN_GENUS2: an = vecan_genus2(an, L); break;
+    case t_LFUN_HGM:
+      an = hgmcoefs(gel(an,1), gel(an,2), L); break;
     case t_LFUN_MFCLOS:
     {
       GEN F = gel(an,1), E = gel(an,2), c = gel(an,3);
@@ -2814,24 +3058,70 @@ ldata_vecan(GEN van, long L, long prec)
   return an;
 }
 
-/* shallow function */
+/* shallow */
 GEN
 ldata_newprec(GEN ldata, long prec)
 {
-  GEN van = ldata_get_an(ldata);
-  GEN an = gel(van, 2);
+  GEN van = ldata_get_an(ldata), an = gel(van, 2);
   long t = mael(van,1,1);
   switch (t)
   {
-    case t_LFUN_CLOSURE0:
-      return lfuncreate(closure_callgen0prec(an, prec));
+    case t_LFUN_CLOSURE0: return closure2ldata(an, prec);
+    case t_LFUN_HECKE:
+    {
+      GEN gc = gel(an, 1), chiw = gel(an, 2);
+      gc = gcharnewprec(gc, prec);
+      return gchari_lfun(gc, chiw, gen_0); /* chi in internal format */
+    }
     case t_LFUN_QF:
     {
       GEN eno = ldata_get_rootno(ldata);
-      if (typ(eno)==t_REAL && realprec(eno) < prec)
-        return lfunqf(an, prec);
+      if (typ(eno)==t_REAL && realprec(eno) < prec) return lfunqf(an, prec);
       break;
     }
   }
   return ldata;
 }
+
+GEN
+ldata_eulerf(GEN van, GEN p, long prec)
+{
+  GEN an = gel(van, 2), f = gen_0;
+  long t = mael(van,1,1);
+  switch (t)
+  {
+    case t_LFUN_GENERIC:
+      f = eulerf_closure(an, p, prec); break;
+    case t_LFUN_CLOSURE0:
+      pari_err_BUG("ldata_vecan: please call ldata_newprec");/*LCOV_EXCL_LINE*/
+    case t_LFUN_ZETA: f = mkrfrac(gen_1,deg1pol(gen_m1, gen_1,0)); break;
+    case t_LFUN_NF:  f = eulerf_zetak(an, p); break;
+    case t_LFUN_ELL: f = elleulerf(an, p); break;
+    case t_LFUN_KRONECKER:
+      f = mkrfrac(gen_1, deg1pol_shallow(stoi(-kronecker(an, p)), gen_1, 0)); break;
+    case t_LFUN_CHIZ: f = eulerf_chiZ(an, p, prec); break;
+    case t_LFUN_CHIGEN: f = eulerf_chigen(an, p, prec); break;
+    case t_LFUN_HECKE: f = eulerf_gchar(an, p, prec); break;
+    case t_LFUN_ARTIN: f = eulerf_artin(an, p, prec); break;
+    case t_LFUN_DIV: f = eulerf_div(an, p, prec); break;
+    case t_LFUN_MUL: f = eulerf_mul(an, p, prec); break;
+    case t_LFUN_CONJ: f = eulerf_conj(an, p, prec); break;
+    case t_LFUN_SYMPOW_ELL: f = eulerf_ellsympow(an, p); break;
+    case t_LFUN_GENUS2: f = eulerf_genus2(an, p); break;
+    case t_LFUN_TWIST: f = eulerf_twist(an, p, prec); break;
+    case t_LFUN_SHIFT: f = eulerf_shift(an, p, prec); break;
+    case t_LFUN_HGM: f = eulerf_hgm(an, p); break;
+    default: f = NULL; break;
+  }
+  if (!f) pari_err_DOMAIN("lfuneuler", "L", "Euler product", strtoGENstr("unknown"), an);
+  return f;
+}
+
+GEN
+lfuneuler(GEN ldata, GEN p, long prec)
+{
+  pari_sp av = avma;
+  if (typ(p)!=t_INT || signe(p)<=0) pari_err_TYPE("lfuneuler", p);
+  ldata = ldata_newprec(lfunmisc_to_ldata_shallow(ldata), prec);
+  return gerepilecopy(av, ldata_eulerf(ldata_get_an(ldata), p, prec));
+}
