@@ -15,59 +15,16 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_gammamellininv
+
 /*******************************************************************/
 /*               Computation of inverse Mellin                     */
 /*               transforms of gamma products.                     */
 /*******************************************************************/
-#ifndef M_E
-#define M_E 2.7182818284590452354
-#endif
-
 /* Handle complex Vga whose sum is real */
 static GEN
 sumVga(GEN Vga) { return real_i(vecsum(Vga)); }
 
-/* rough approximation to W0(a > -1/e), < 1% relative error */
-double
-dbllambertW0(double a)
-{
-  if (a < -0.2583)
-  {
-    const double c2 = -1./3, c3 = 11./72, c4 = -43./540, c5 = 769./17280;
-    double p = sqrt(2*(M_E*a+1));
-    if (a < -0.3243) return -1+p*(1+p*(c2+p*c3));
-    return -1+p*(1+p*(c2+p*(c3+p*(c4+p*c5))));
-  }
-  else
-  {
-    double Wd = log(1.+a);
-    Wd *= (1.-log(Wd/a))/(1.+Wd);
-    if (a < 0.6482 && a > -0.1838) return Wd;
-    return Wd*(1.-log(Wd/a))/(1.+Wd);
-  }
-}
-
-/* rough approximation to W_{-1}(0 > a > -1/e), < 1% relative error */
-double
-dbllambertW_1(double a)
-{
-  if (a < -0.2464)
-  {
-    const double c2 = -1./3, c3 = 11./72, c4 = -43./540, c5 = 769./17280;
-    double p = -sqrt(2*(M_E*a+1));
-    if (a < -0.3243) return -1+p*(1+p*(c2+p*c3));
-    return -1+p*(1+p*(c2+p*(c3+p*(c4+p*c5))));
-  }
-  else
-  {
-    double Wd;
-    a = -a; Wd = -log(a);
-    Wd *= (1.-log(Wd/a))/(1.-Wd);
-    if (a < 0.0056) return -Wd;
-    return -Wd*(1.-log(Wd/a))/(1.-Wd);
-  }
-}
-
 /* ac != 0 */
 static double
 lemma526_i(double ac, double c, double t, double B)
@@ -267,23 +224,15 @@ Kderivsmallinit(GEN ldata, GEN Vga, long m, long bit)
   return mkvec5(L, RgV_neg(M), mat, mkvecsmall(prec2), piA);
 }
 
-/* Evaluate a vector considered as a polynomial using Horner. Unstable!
- * If ui != NULL, ui = 1/u, evaluate P(1/u)*u^(deg P): useful for |u|>1 */
+/* Evaluate a vector considered as a polynomial using Horner. */
 static GEN
-evalvec(GEN vec, long lim, GEN u, GEN ui)
+evalvec(GEN vec, long N, GEN u)
 {
-  pari_sp ltop = avma;
   GEN S = gen_0;
   long n;
-  lim = minss(lim, lg(vec)-1);
-  if (!ui)
-    for (n = lim; n >= 1; --n) S = gmul(u, gadd(gel(vec,n), S));
-  else
-  {
-    for (n = 1; n <= lim; ++n) S = gmul(ui, gadd(gel(vec,n), S));
-    S = gmul(gpowgs(u, n), S);
-  }
-  return gerepileupto(ltop, S);
+  N = minss(N, lg(vec)-1);
+  for (n = N; n >= 1; n--) S = gmul(u, gadd(gel(vec,n), S));
+  return S;
 }
 
 /* gammamellininvinit accessors */
@@ -311,7 +260,7 @@ static GEN
 Kderivsmall(GEN K, GEN x, GEN x2d, long bitprec)
 {
   GEN VS = GMi_get_VS(K), L = gel(VS,1), M = gel(VS,2), mat = gel(VS,3);
-  GEN d2, Lx, x2, x2i, S, pi, piA = gel(VS,5);
+  GEN d2, Lx, x2, S, pi, piA = gel(VS,5);
   long prec = gel(VS,4)[1], d = GMi_get_degree(K);
   long j, k, limn, N = lg(L)-1;
   double xd, Wd, Ed = M_LN2*bitprec / d;
@@ -331,14 +280,13 @@ Kderivsmall(GEN K, GEN x, GEN x2d, long bitprec)
   /* at this stage, x has been replaced by pi^(d/2) x */
   x2 = gsqr(x);
   Lx = gpowers(gneg(glog(x,prec)), vecsmall_max(L));
-  x2i = (gcmp(gnorml2(x2), gen_1) <= 0)? NULL: ginv(x2);
   S = gen_0;
   for (j = 1; j <= N; ++j)
   {
     long lj = L[j];
     GEN s = gen_0;
     for (k = 1; k <= lj; k++)
-      s = gadd(s, gmul(gel(Lx,k), evalvec(gmael(mat,j,k), limn, x2, x2i)));
+      s = gadd(s, gmul(gel(Lx,k), evalvec(gmael(mat,j,k), limn, x2)));
     S = gadd(S, gmul(gpow(x, gel(M,j), prec), s));
   }
   if (!gequal0(piA)) S = gmul(S, piA);
@@ -369,10 +317,8 @@ Kderivlarge_optim(GEN K, long abs, GEN t2d,GEN gcd, long *pbitprec, long *pnlim)
   double a, rtd, E = M_LN2*bitprec;
 
   rtd = (typ(t2d) == t_COMPLEX)? gtodouble(gel(t2d,1)): td;
-
   /* A2/2 = A, log(td) = (2/d)*log t */
-  a = d*gtodouble(A2)*log2(td)/2 - (M_PI/M_LN2)*d*rtd + log2(cd); /*log2 K(t)~a*/
-
+  a = d*gtodouble(A2)*log2(td)/2 - (M_PI/M_LN2)*d*rtd + log2(cd);/*log2 K(t)~a*/
   /* if bitprec <= 0, caller should return K(t) ~ 0 */
   bitprec += 64;
   if (abs)
@@ -401,31 +347,26 @@ Kderivlarge(GEN K, GEN t, GEN t2d, long bitprec0)
   prec = nbits2prec(bitprec);
   t2d = gtofp(t2d, prec);
   if (t)
-    tdA = gpow(t, gdivgs(A2,d), prec);
+    tdA = gpow(t, gdivgu(A2,d), prec);
   else
-    tdA = gpow(t2d, gdivgs(A2,2), prec);
+    tdA = gpow(t2d, gdivgu(A2,2), prec);
   tdA = gmul(cd, tdA);
 
   pi = mppi(prec);
   z = gmul(pi, t2d);
   P = gmul(tdA, gexp(gmulsg(-d, z), prec));
   if (m) P = gmul(P, gpowgs(mulsr(-2, pi), m));
-  M = gel(Ms,1);
-  status = itos(gel(Ms,2));
+  M = gel(Ms,1); status = itos(gel(Ms,2));
   if (status == 2)
-  {
-    if (lg(M) == 2) /* shortcut: continued fraction is constant */
-      S = gel(M,1);
-    else
-      S = poleval(RgV_to_RgX(M, 0), ginv(z));
-  }
+    S = (lg(M) == 2)? gel(M,1) /* constant continued fraction */
+                    : poleval(RgV_to_RgX(M, 0), ginv(z));
   else
   {
     S = contfraceval_inv(M, z, nlim/2);
     if (DEBUGLEVEL>3)
     {
       GEN S0 = contfraceval_inv(M, z, nlim/2 + 1);
-      long e = gexpo(gmul(P, gabs(gsub(S,S0),0)));
+      long e = gexpo(gmul(P, gsub(S,S0)));
       if (-e < bitprec0)
         err_printf("Kderivlarge: e = %ld, bit = %ld\n",e,bitprec0);
     }
@@ -443,7 +384,7 @@ vp(long p, long c, GEN SMd, GEN sh)
   long m, j, k;
 
   gel(ve,1) = gen_1; gel(ve,2) = utoipos(c);
-  for (j = 2; j <= p; j++) gel(ve,j+1) = gdivgs(gmulgs(gel(ve,j), c), j);
+  for (j = 2; j <= p; j++) gel(ve,j+1) = gdivgs(gmulgu(gel(ve,j), c), j);
   s = gel(SMd, 1);
   for (m = 1; m <= p; m++)
   {
@@ -514,14 +455,14 @@ get_SMd(GEN Vga)
  * If status = 2, the asymptotic expansion is finite so return only
  * the necessary number of terms nlim <= nlimmax + d. */
 static GEN
-Klargeinit(GEN Vga, long nlimmax, long *status)
+Klargeinit(GEN Vga, long nlimmax, long *status, long prec)
 {
   long d = lg(Vga) - 1, p, n, cnt;
   GEN M, SMd, se, vsinh, vd;
 
   if (Vgaeasytheta(Vga)) { *status = 2; return mkvec(gen_1); }
   /* d >= 2 */
-  *status = 0;
+  *status = 0; prec += prec >> 1;
   SMd = get_SMd(Vga);
   se = gsinh(RgX_to_ser(pol_x(0), d+2), 0); setvalp(se,0);
   se = gdeflate(se, 0, 2); /* se(x^2) = sinh(x)/x */
@@ -538,6 +479,7 @@ Klargeinit(GEN Vga, long nlimmax, long *status)
       GEN z = vp(p, 2*n-1-p, gel(SMd, p-1), gel(vsinh, d-p+1));
       s = gadd(s, gmul(gdiv(z, gel(vd, p+1)), gel(M, n+1-p)));
     }
+    if (prec && !isinexact(s)) s = gtofp(s, prec);
     gel(M,n) = s = gerepileupto(av, gdivgs(s, 1-n));
     if (gequal0(s))
     {
@@ -567,23 +509,23 @@ stripzeros(GEN M)
  * nlimmax. If status = 2, the asymptotic expansion is finite so return only
  * the necessary number of terms nlim <= nlimmax + d. */
 static GEN
-gammamellininvasymp_i(GEN Vga, long nlimmax, long m, long *status)
+gammamellininvasymp_i(GEN Vga, long nlimmax, long m, long *status, long prec)
 {
   GEN M, A, Aadd;
   long d, i, nlim, n;
 
-  M = Klargeinit(Vga, nlimmax, status);
+  M = Klargeinit(Vga, nlimmax, status, prec);
   if (!m) return M;
   d = lg(Vga)-1;
   /* half the exponent of t in asymptotic expansion. */
-  A = gdivgs(gaddsg(1-d, sumVga(Vga)), 2*d);
+  A = gdivgu(gaddsg(1-d, sumVga(Vga)), 2*d);
   if (*status == 2) M = shallowconcat(M, zerovec(m));
   nlim = lg(M)-1;
   Aadd = sstoQ(2-d, 2*d); /* (1/d) - (1/2) */
   for (i = 1; i <= m; i++, A = gadd(A,Aadd))
     for (n = nlim-1; n >= 1; --n)
       gel(M, n+1) = gsub(gel(M, n+1),
-                         gmul(gel(M, n), gsub(A, sstoQ(n-1, d))));
+                         gmul(gel(M, n), gsub(A, uutoQ(n-1, d))));
   stripzeros(M); return M;
 }
 static GEN
@@ -602,7 +544,7 @@ gammamellininvasymp(GEN Vga, long nlim, long m)
   Vga = get_Vga(Vga, &ldata);
   if (!is_vec_t(typ(Vga)) || lg(Vga) == 1)
     pari_err_TYPE("gammamellininvasymp",Vga);
-  return gerepilecopy(av, gammamellininvasymp_i(Vga, nlim, m, &status));
+  return gerepilecopy(av, gammamellininvasymp_i(Vga, nlim, m, &status, 0));
 }
 
 /* Does the continued fraction of the asymptotic expansion M at oo of inverse
@@ -640,9 +582,9 @@ gammamellininvinit(GEN Vga, long m, long bitprec)
   if (!is_vec_t(typ(Vga)) || !d) pari_err_TYPE("gammamellininvinit",Vga);
   nlimmax = ceil(E * log2(1+M_PI*tmax) * C2 / get_D(d));
   A2 = gaddsg(m*(2-d) + 1-d, sumVga(Vga));
-  cd = (d <= 2)? gen_2: gsqrt(gdivgs(int2n(d+1), d), nbits2prec(bitprec));
+  cd = (d <= 2)? gen_2: gsqrt(gdivgu(int2n(d+1), d), nbits2prec(bitprec));
   /* if in Klarge, we have |t| > tmax = E/C2, thus nlim < E*C2/D. */
-  M = gammamellininvasymp_i(RgV_gtofp(Vga, prec), nlimmax, m, &status);
+  M = gammamellininvasymp_i(Vga, nlimmax, m, &status, prec);
   if (status == 2)
   {
     tmax = -1.; /* only use Klarge */
@@ -672,14 +614,17 @@ gammamellininvinit(GEN Vga, long m, long bitprec)
  * initialization data. Use Taylor expansion at 0 for |s2d| < tmax, and
  * asymptotic expansion at oo otherwise. WARNING: assume that accuracy
  * has been increased according to tmax by the CALLING program. */
-GEN
-gammamellininvrt(GEN K, GEN s2d, long bitprec)
+static GEN
+gammamellininvrt_i(GEN K, GEN s, GEN s2d, long bit)
 {
-  if (dblmodulus(s2d) < GMi_get_tmax(K, bitprec))
-    return Kderivsmall(K, NULL, s2d, bitprec);
+  if (dblmodulus(s2d) < GMi_get_tmax(K, bit))
+    return Kderivsmall(K, s, s2d, bit);
   else
-    return Kderivlarge(K, NULL, s2d, bitprec);
+    return Kderivlarge(K, s, s2d, bit);
 }
+GEN
+gammamellininvrt(GEN K, GEN s2d, long bit)
+{ return gammamellininvrt_i(K, NULL, s2d, bit); }
 
 /* Compute inverse Mellin at s. K from gammamellininv OR a Vga, in which
  * case the initialization data is computed. */
@@ -687,16 +632,12 @@ GEN
 gammamellininv(GEN K, GEN s, long m, long bitprec)
 {
   pari_sp av = avma;
-  GEN z, s2d;
+  GEN s2d;
   long d;
 
   if (!is_vec_t(typ(K)) || lg(K) != 6 || !is_vec_t(typ(GMi_get_Vga(K))))
     K = gammamellininvinit(K, m, bitprec);
   d = GMi_get_degree(K);
-  s2d = gpow(s, gdivgs(gen_2, d), nbits2prec(bitprec));
-  if (dblmodulus(s2d) < GMi_get_tmax(K, bitprec))
-    z = Kderivsmall(K, s, s2d, bitprec);
-  else
-    z = Kderivlarge(K, s, s2d, bitprec);
-  return gerepileupto(av, z);
+  s2d = gpow(s, gdivgu(gen_2, d), nbits2prec(bitprec));
+  return gerepileupto(av, gammamellininvrt_i(K, s, s2d, bitprec));
 }
