@@ -280,8 +280,8 @@ vecvecsmall_indexsort(GEN x)
 { return gen_indexsort(x, (void*)&vecsmall_lexcmp, cmp_nodata); }
 
 long
-vecvecsmall_search(GEN x, GEN y, long flag)
-{ return gen_search(x,y,flag,(void*)vecsmall_prefixcmp, cmp_nodata); }
+vecvecsmall_search(GEN x, GEN y)
+{ return gen_search(x,y,(void*)vecsmall_prefixcmp, cmp_nodata); }
 
 /* assume x non empty */
 long
@@ -882,11 +882,12 @@ tableset_find_index(GEN tbl, GEN set)
 
 GEN
 trivialgroup(void) { retmkvec2(cgetg(1,t_VEC), cgetg(1,t_VECSMALL)); }
+
 /*Cyclic group generated by g of order s*/
 GEN
 cyclicgroup(GEN g, long s)
-{ retmkvec2(mkvec( vecsmall_copy(g) ),
-            mkvecsmall(s)); }
+{ retmkvec2(mkvec( vecsmall_copy(g) ), mkvecsmall(s)); }
+
 /*Return the group generated by g1,g2 of relative orders s1,s2*/
 GEN
 dicyclicgroup(GEN g1, GEN g2, long s1, long s2)
@@ -1380,9 +1381,8 @@ groupelts_conjclasses(GEN elts, long *pnbcl)
       if (j != i)
       {
         GEN h = perm_conj(gel(elts,j), g);
-        long i2 = gen_search(elts,h,0,(void*)&vecsmall_lexcmp,&cmp_nodata);
-        c[i2] = cl;
-        set_avma(av);
+        long i2 = gen_search(elts,h,(void*)&vecsmall_lexcmp,&cmp_nodata);
+        c[i2] = cl; set_avma(av);
       }
   }
   if (pnbcl) *pnbcl = cl;
@@ -1510,43 +1510,46 @@ groupelts_cyclic_subgroups(GEN G)
 {
   pari_sp av = avma;
   long i, j, n = lg(G)-1;
-  GEN elts = zero_F2v(n+1), f;
-  GEN ord = cgetg(n+1, t_VECSMALL);
-  GEN V = cgetg(n+1, t_VEC);
+  GEN elts, f, gen, ord;
+  if (n==1) return cgetg(1,t_VEC);
+  elts = zero_F2v(lg(gel(G,1))-1);
+  gen = cgetg(n+1, t_VECSMALL);
+  ord = cgetg(n+1, t_VECSMALL);
   for (i=1, j=1; i<=n; i++)
   {
     long k = 1, o, c = 0;
     GEN p = gel(G, i);
     if (F2v_coeff(elts, p[1])) continue;
     o = perm_orderu(p);
-    ord[j] = o;
-    gel(V,j++) = p;
+    gen[j] = i; ord[j] = o; j++;
     do
     {
       if (cgcd(o, ++c)==1) F2v_set(elts, p[k]);
       k = p[k];
     } while (k!=1);
   }
+  setlg(gen, j);
   setlg(ord, j);
-  setlg(V, j);
   f = vecsmall_indexsort(ord);
-  return gerepilecopy(av, mkvec2(vecpermute(V, f), vecpermute(ord, f)));
+  return gerepilecopy(av, mkvec2(vecpermute(gen, f), vecpermute(ord, f)));
 }
 
 GEN
 groupelts_to_group(GEN G)
 {
   pari_sp av = avma;
-  GEN L = groupelts_cyclic_subgroups(G);
-  GEN cyc = gel(L,1), ord = gel(L,2);
-  long i, l = lg(cyc), n = lg(G)-1;
+  GEN L, cyc, ord;
+  long i, l, n = lg(G)-1;
+  if (n==1) return trivialgroup();
+  L = groupelts_cyclic_subgroups(G);
+  cyc = gel(L,1); ord = gel(L,2);
+  l = lg(cyc);
   for (i = l-1; i >= 2; i--)
   {
-    GEN p = gel(cyc,i);
+    GEN p = gel(G,cyc[i]);
     long o = ord[i];
-    GEN H;
-    if (o == n) { set_avma(av); return cyclicgroup(p, o); }
-    H = cyclicgroup(p, o);
+    GEN H = cyclicgroup(p, o);
+    if (o == n) return gerepileupto(av, H);
     if (groupelts_subgroup_isnormal(G, H))
     {
       GEN C = groupelts_quotient(G, H);
@@ -1558,19 +1561,242 @@ groupelts_to_group(GEN G)
   }
   if (n==12 && l==9 && ord[2]==2 && ord[3]==2 && ord[5]==3)
     return gerepilecopy(av,
-      mkvec2(mkvec3(gel(cyc,2), gel(cyc,3), gel(cyc,5)), mkvecsmall3(2,2,3)));
+      mkvec2(mkvec3(gel(G,cyc[2]), gel(G,cyc[3]), gel(G,cyc[5])), mkvecsmall3(2,2,3)));
   if (n==24 && l==18 && ord[11]==3 && ord[15]==4 && ord[16]==4)
   {
-    GEN t21 = perm_sqr(gel(cyc,15));
-    GEN t22 = perm_sqr(gel(cyc,16));
+    GEN t21 = perm_sqr(gel(G,cyc[15]));
+    GEN t22 = perm_sqr(gel(G,cyc[16]));
+    GEN s = perm_mul(t22, gel(G,cyc[15]));
     return gerepilecopy(av,
-      mkvec2(mkvec4(t21,t22, gel(cyc,11), gel(cyc,15)), mkvecsmall4(2,2,3,2)));
+      mkvec2(mkvec4(t21,t22, gel(G,cyc[11]), s), mkvecsmall4(2,2,3,2)));
   }
   if (n==36 && l==24 && ord[11]==3 && ord[15]==4)
   {
-    GEN t1 = gel(cyc,11), t3 = gel(cyc,15);
+    GEN t1 = gel(G,cyc[11]), t3 = gel(G,cyc[15]);
     return gerepilecopy(av,
-      mkvec2(mkvec3(t1, perm_conj(t3, t1), t3), mkvecsmall3(3,3,4)));
+      mkvec2(mkvec3(perm_conj(t3, t1), t1, t3), mkvecsmall3(3,3,4)));
   }
   return gc_NULL(av);
 }
+
+static GEN
+subg_get_gen(GEN subg) {  return gel(subg, 1); }
+
+static GEN
+subg_get_set(GEN subg) {  return gel(subg, 2); }
+
+static GEN
+groupelt_subg_normalize(GEN elt, GEN subg, GEN cyc)
+{
+  GEN gen = subg_get_gen(subg), set =  subg_get_set(subg);
+  long i, j, u, n = lg(elt)-1, lgen = lg(gen);
+  GEN b = F2v_copy(cyc), res = zero_F2v(n);
+  for(i = 1; i <= n; i++)
+  {
+    GEN g;
+    if (!F2v_coeff(b, i)) continue;
+    g = gel(elt,i);
+    for(u=1; u<=n; u++)
+      if (g[u]==1) break;
+    for(j=1; j<lgen; j++)
+    {
+      GEN h = gel(elt,gen[j]);
+      if (!F2v_coeff(set,g[h[u]])) break;
+    }
+    if (j < lgen) continue;
+    F2v_set(res,i);
+    for(j=1; j <= n; j++)
+      if (F2v_coeff(set, j))
+        F2v_clear(b,g[gel(elt,j)[1]]);
+  }
+  return res;
+}
+
+static GEN
+triv_subg(GEN elt)
+{
+  GEN v = cgetg(3, t_VEC);
+  gel(v,1) = cgetg(1,t_VECSMALL);
+  gel(v,2) = zero_F2v(lg(elt)-1);
+  F2v_set(gel(v,2),1);
+  return v;
+}
+
+static GEN
+subg_extend(GEN U, long e, long o, GEN elt)
+{
+  long i, j, n = lg(elt)-1;
+  GEN g = gel(elt, e);
+  GEN gen = vecsmall_append(subg_get_gen(U), e);
+  GEN set = subg_get_set(U);
+  GEN Vset = zv_copy(set);
+  for(i = 1; i <= n; i++)
+    if (F2v_coeff(set, i))
+    {
+      long h = gel(elt, i)[1];
+      for(j = 1; j < o; j++)
+      {
+        h = g[h];
+        F2v_set(Vset, h);
+      }
+    }
+  return mkvec2(gen, Vset);
+}
+
+static GEN
+cyclic_subg(long e, long o, GEN elt)
+{
+  long j, n = lg(elt)-1, h = 1;
+  GEN g = gel(elt, e);
+  GEN gen = mkvecsmall(e);
+  GEN set = zero_F2v(n);
+  F2v_set(set,1);
+  for(j = 1; j < o; j++)
+  {
+    h = g[h];
+    F2v_set(set, h);
+  }
+  return mkvec2(gen, set);
+}
+
+static GEN
+groupelts_to_regular(GEN elt)
+{
+  long i, j, n = lg(elt)-1;
+  GEN V = cgetg(n+1,t_VEC);
+  for (i=1; i<=n; i++)
+  {
+    pari_sp av = avma;
+    GEN g = gel(elt, i);
+    GEN W = cgetg(n+1,t_VEC);
+    for(j=1; j<=n; j++)
+      gel(W,j) = perm_mul(g, gel(elt,j));
+    gel(V, i) = gerepileuptoleaf(av,vecvecsmall_indexsort(W));
+  }
+  vecvecsmall_sort_inplace(V, NULL);
+  return V;
+}
+
+static long
+groupelts_pow(GEN elt, long j, long n)
+{
+  GEN g = gel(elt,j);
+  long i, h = 1;
+  for (i=1; i<=n; i++)
+    h = g[h];
+  return h;
+}
+
+static GEN
+groupelts_cyclic_primepow(GEN elt, GEN *pt_pr, GEN *pt_po)
+{
+  GEN R = groupelts_cyclic_subgroups(elt);
+  GEN gen = gel(R,1), ord = gel(R,2);
+  long i, n = lg(elt)-1, l = lg(gen);
+  GEN set = zero_F2v(n);
+  GEN pr  = zero_Flv(n);
+  GEN po  = zero_Flv(n);
+  for (i = 1; i < l; i++)
+  {
+    long h = gen[i];
+    ulong p;
+    if (uisprimepower(ord[i], &p))
+    {
+      F2v_set(set, h);
+      uel(pr,h) = p;
+      po[h] = groupelts_pow(elt, h, p);
+    }
+  }
+  *pt_pr = pr; *pt_po = po;
+  return set;
+}
+
+static GEN
+all_cyclic_subg(GEN pr, GEN po, GEN elt)
+{
+  long i, n = lg(pr)-1, m = 0, k = 1;
+  GEN W;
+  for (i=1; i <= n; i++)
+    m += po[i]==1;
+  W = cgetg(m+1, t_VEC);
+  for (i=1; i <= n; i++)
+    if (po[i]==1)
+      gel(W, k++) = cyclic_subg(i, pr[i], elt);
+  return W;
+}
+
+static GEN
+groupelts_subgroups_raw(GEN elts)
+{
+  pari_sp av = avma;
+  GEN elt = groupelts_to_regular(elts);
+  GEN pr, po, cyc = groupelts_cyclic_primepow(elt, &pr, &po);
+  long n = lg(elt)-1;
+  long i, j, nS = 1;
+  GEN S, L;
+  S = cgetg(1+bigomegau(n)+1, t_VEC);
+  gel(S, nS++) = mkvec(triv_subg(elt));
+  gel(S, nS++) = L = all_cyclic_subg(pr, po, elt);
+  if (DEBUGLEVEL) err_printf("subgroups: level %ld: %ld\n",nS,lg(L)-1);
+  while (lg(L) > 1)
+  {
+    pari_sp av2 = avma;
+    long nW = 1, lL = lg(L);
+    long ng = n;
+    GEN W = cgetg(1+ng, t_VEC);
+    for (i=1; i<lL; i++)
+    {
+      GEN U = gel(L, i), set = subg_get_set(U);
+      GEN G = groupelt_subg_normalize(elt, U, cyc);
+      for (j=1; j<nW; j++)
+      {
+        GEN Wj = subg_get_set(gel(W, j));
+        if (F2v_subset(set, Wj))
+          F2v_negimply_inplace(G, Wj);
+      }
+      for (j=1; j<=n; j++)
+        if(F2v_coeff(G,j))
+        {
+          long p = pr[j];
+          if (F2v_coeff(set, j)) continue;
+          if (F2v_coeff(set, po[j]))
+          {
+            GEN U2 = subg_extend(U, j, p, elt);
+            F2v_negimply_inplace(G, subg_get_set(U2));
+            if (nW > ng) { ng<<=1; W = vec_lengthen(W, ng); }
+            gel(W, nW++) = U2;
+          }
+        }
+    }
+    setlg(W, nW);
+    if (DEBUGLEVEL) err_printf("subgroups: level %ld: %ld\n",nS,nW-1);
+    L = W;
+    if (nW > 1) gel(S, nS++) = L = gerepilecopy(av2, W);
+  }
+  setlg(S, nS);
+  return gerepilecopy(av, shallowconcat1(S));
+}
+
+static GEN
+set_groupelts(GEN S, GEN x)
+{
+  long i, n = F2v_hamming(x), k=1, m = x[1];
+  GEN v = cgetg(n+1, t_VEC);
+  for (i=1; i<=m; i++)
+    if (F2v_coeff(x,i))
+      gel(v,k++) = gel(S,i);
+  return v;
+}
+
+static GEN
+subg_to_elts(GEN S, GEN x)
+{ pari_APPLY_type(t_VEC, set_groupelts(S, gmael(x,i,2))); }
+
+GEN
+groupelts_solvablesubgroups(GEN G)
+{
+  pari_sp av = avma;
+  GEN S = vecvecsmall_sort(checkgroupelts(G));
+  GEN L = groupelts_subgroups_raw(S);
+  return gerepilecopy(av, subg_to_elts(S, L));
+}
