@@ -21,6 +21,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_pol
+
 /************************************************************************
  **                                                                    **
  **                      Ring membership                               **
@@ -834,7 +836,7 @@ FqV_roots_to_pol(GEN V, GEN T, GEN p, long v)
   {
     ulong pp = p[2];
     GEN Tl = ZX_to_Flx(T, pp);
-    GEN Vl = FqV_to_FlxV(V, T, p);
+    GEN Vl = ZXC_to_FlxC(V, pp, get_Flx_var(Tl));
     Tl = FlxqV_roots_to_pol(Vl, Tl, pp, v);
     return gerepileupto(ltop, FlxX_to_ZXX(Tl));
   }
@@ -885,28 +887,6 @@ FqC_FqV_mul(GEN x, GEN y, GEN T, GEN p)
   return z;
 }
 
-GEN
-FqV_to_FlxV(GEN x, GEN T, GEN pp)
-{
-  long vT = evalvarn(get_FpX_var(T));
-  ulong p = pp[2];
-  pari_APPLY_type(t_VEC, typ(gel(x,i))==t_INT?  Z_to_Flx(gel(x,i), p, vT)
-                                             : ZX_to_Flx(gel(x,i), p))
-}
-
-GEN
-FqC_to_FlxC(GEN x, GEN T, GEN pp)
-{
-  long vT = evalvarn(get_FpX_var(T));
-  ulong p = pp[2];
-  pari_APPLY_type(t_COL, typ(gel(x,i))==t_INT?  Z_to_Flx(gel(x,i), p, vT)
-                                             : ZX_to_Flx(gel(x,i), p))
-}
-
-GEN
-FqM_to_FlxM(GEN x, GEN T, GEN p)
-{ pari_APPLY_same(FqC_to_FlxC(gel(x,i), T, p)) }
-
 GEN
 FpXC_center(GEN x, GEN p, GEN pov2)
 { pari_APPLY_type(t_COL, FpX_center(gel(x,i), p, pov2)) }
@@ -1336,34 +1316,62 @@ ZX_norml1(GEN x)
   s = gel(x, l-1); /* != 0 */
   for (i = l-2; i > 1; i--) {
     GEN xi = gel(x,i);
-    if (!signe(x)) continue;
+    if (!signe(xi)) continue;
     s = addii_sign(s,1, xi,1);
   }
   return s;
 }
-
+/* x >= 0, y != 0, return x + |y| */
 static GEN
-L2_bound(GEN nf, GEN den, GEN *pt_roots)
+addii_abs(GEN x, GEN y)
 {
-  GEN M, L, prep, T = nf_get_pol(nf), tozk = nf_get_invzk(nf);
-  long prec = ZM_max_lg(tozk) + ZX_max_lg(T) + nbits2prec(degpol(T));
-  (void)initgaloisborne(nf, den? den: gen_1, prec, &L, &prep, NULL);
-  M = vandermondeinverse(L, RgX_gtofp(T,prec), den, prep);
-  *pt_roots = L;
-  return RgM_fpnorml2(RgM_mul(tozk,M), DEFAULTPREC);
+  if (!signe(x)) return absi_shallow(y);
+  return addii_sign(x,1, y,1);
 }
 
-/* Interpolate at roots of 1 and use Hadamard bound for univariate resultant:
- *   bound = N_2(A)^degpol B N_2(B)^degpol(A),  where
- *     N_2(A) = sqrt(sum (N_1(Ai))^2)
- * Return e such that Res(A, B) < 2^e */
+/* x a ZX, return sum_{i >= k} |x[i]| binomial(i, k) */
 static GEN
-RgX_RgXY_ResBound(GEN A, GEN B, long prec)
+ZX_norml1_1(GEN x, long k)
 {
-  pari_sp av = avma, av2;
-  GEN a = gen_0, b = gen_0, bnd;
-  long i , lA = lg(A), lB = lg(B);
-  for (i=2; i<lA; i++)
+  long i, d = degpol(x);
+  GEN s, C; /* = binomial(i, k) */
+
+  if (!d || k > d) return gen_0;
+  s = absi_shallow(gel(x, k+2)); /* may be 0 */
+  C = gen_1;
+  for (i = k+1; i <= d; i++) {
+    GEN xi = gel(x,i+2);
+    if (k) C = diviuexact(muliu(C, i), i-k);
+    if (signe(xi)) s = addii_abs(s, mulii(C, xi));
+  }
+  return s;
+}
+/* x has non-negative real coefficients */
+static GEN
+RgX_norml1_1(GEN x, long k)
+{
+  long i, d = degpol(x);
+  GEN s, C; /* = binomial(i, k) */
+
+  if (!d || k > d) return gen_0;
+  s = gel(x, k+2); /* may be 0 */
+  C = gen_1;
+  for (i = k+1; i <= d; i++) {
+    GEN xi = gel(x,i+2);
+    if (k) C = diviuexact(muliu(C, i), i-k);
+    if (!gequal0(xi)) s = gadd(s, gmul(C, xi));
+  }
+  return s;
+}
+
+/* N_2(A)^2 */
+static GEN
+sqrN2(GEN A, long prec)
+{
+  pari_sp av = avma;
+  long i, l = lg(A);
+  GEN a = gen_0;
+  for (i = 2; i < l; i++)
   {
     a = gadd(a, gabs(gnorm(gel(A,i)), prec));
     if (gc_needed(av,1))
@@ -1372,34 +1380,65 @@ RgX_RgXY_ResBound(GEN A, GEN B, long prec)
       a = gerepileupto(av, a);
     }
   }
-  av2 = avma;
+  return a;
+}
+/* Interpolate at roots of 1 and use Hadamard bound for univariate resultant:
+ *   bound = N_2(A)^degpol B N_2(B)^degpol(A),  where
+ *     N_2(A) = sqrt(sum (N_1(Ai))^2)
+ * Return e such that Res(A, B) < 2^e */
+static GEN
+RgX_RgXY_ResBound(GEN A, GEN B, long prec)
+{
+  pari_sp av = avma;
+  GEN b = gen_0, bnd;
+  long i, lB = lg(B);
   for (i=2; i<lB; i++)
   {
     GEN t = gel(B,i);
     if (typ(t) == t_POL) t = gnorml1(t, prec);
     b = gadd(b, gabs(gsqr(t), prec));
     if (gc_needed(av,1))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"RgX_RgXY_ResBound i = %ld",i);
+      b = gerepileupto(av, b);
+    }
+  }
+  bnd = gsqrt(gmul(gpowgs(sqrN2(A,prec), degpol(B)),
+                   gpowgs(b, degpol(A))), prec);
+  return gerepileupto(av, bnd);
+}
+/* A,B in C[X] return RgX_RgXY_ResBound(A, B(x+y)) */
+static GEN
+RgX_RgXY_ResBound_1(GEN A, GEN B, long prec)
+{
+  pari_sp av = avma, av2;
+  GEN b = gen_0, bnd;
+  long i, lB = lg(B);
+  B = shallowcopy(B);
+  for (i=2; i<lB; i++) gel(B,i) = gabs(gel(B,i), prec);
+  av2 = avma;
+  for (i=2; i<lB; i++)
+  {
+    b = gadd(b, gsqr(RgX_norml1_1(B, i-2)));
+    if (gc_needed(av2,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"RgX_RgXY_ResBound i = %ld",i);
       b = gerepileupto(av2, b);
     }
   }
-  bnd = gsqrt(gmul(gpowgs(a, degpol(B)), gpowgs(b, degpol(A))), prec);
+  bnd = gsqrt(gmul(gpowgs(sqrN2(A,prec), degpol(B)),
+                   gpowgs(b, degpol(A))), prec);
   return gerepileupto(av, bnd);
 }
 
-/* Interpolate at roots of 1 and use Hadamard bound for univariate resultant:
- *   bound = N_2(A)^degpol B N_2(B)^degpol(A),  where
- *     N_2(A) = sqrt(sum (N_1(Ai))^2)
- * Return e such that Res(A, B) < 2^e */
-ulong
-ZX_ZXY_ResBound(GEN A, GEN B, GEN dB)
+/* log2 N_2(A)^2 */
+static double
+log2N2(GEN A)
 {
   pari_sp av = avma;
-  GEN a = gen_0, b = gen_0;
-  long i , lA = lg(A), lB = lg(B);
-  double loga, logb;
-  for (i=2; i<lA; i++)
+  long i, l = lg(A);
+  GEN a = gen_0;
+  for (i=2; i < l; i++)
   {
     a = addii(a, sqri(gel(A,i)));
     if (gc_needed(av,1))
@@ -1408,7 +1447,19 @@ ZX_ZXY_ResBound(GEN A, GEN B, GEN dB)
       a = gerepileupto(av, a);
     }
   }
-  loga = dbllog2(a); set_avma(av);
+  return gc_double(av, dbllog2(a));
+}
+/* Interpolate at roots of 1 and use Hadamard bound for univariate resultant:
+ *   bound = N_2(A)^degpol B N_2(B)^degpol(A),  where
+ *     N_2(A) = sqrt(sum (N_1(Ai))^2)
+ * Return e such that Res(A, B) < 2^e */
+ulong
+ZX_ZXY_ResBound(GEN A, GEN B, GEN dB)
+{
+  pari_sp av = avma;
+  GEN b = gen_0;
+  long i, lB = lg(B);
+  double logb;
   for (i=2; i<lB; i++)
   {
     GEN t = gel(B,i);
@@ -1421,7 +1472,26 @@ ZX_ZXY_ResBound(GEN A, GEN B, GEN dB)
     }
   }
   logb = dbllog2(b); if (dB) logb -= 2 * dbllog2(dB);
-  i = (long)((degpol(B) * loga + degpol(A) * logb) / 2);
+  i = (long)((degpol(B) * log2N2(A) + degpol(A) * logb) / 2);
+  return gc_ulong(av, (i <= 0)? 1: 1 + (ulong)i);
+}
+/* A,B ZX. Return ZX_ZXY_ResBound(A(x), B(x+y)) */
+static ulong
+ZX_ZXY_ResBound_1(GEN A, GEN B)
+{
+  pari_sp av = avma;
+  GEN b = gen_0;
+  long i, lB = lg(B);
+  for (i=2; i<lB; i++)
+  {
+    b = addii(b, sqri(ZX_norml1_1(B, i-2)));
+    if (gc_needed(av,1))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"ZX_ZXY_ResBound i = %ld",i);
+      b = gerepileupto(av, b);
+    }
+  }
+  i = (long)((degpol(B) * log2N2(A) + degpol(A) * dbllog2(b)) / 2);
   return gc_ulong(av, (i <= 0)? 1: 1 + (ulong)i);
 }
 /* special case B = A' */
@@ -1451,12 +1521,12 @@ ZX_discbound(GEN A)
 
 /* return Res(a(Y), b(n,Y)) over Fp. la = leading_coeff(a) [for efficiency] */
 static ulong
-Flx_FlxY_eval_resultant(GEN a, GEN b, ulong n, ulong p, ulong la)
+Flx_FlxY_eval_resultant(GEN a, GEN b, ulong n, ulong p, ulong pi, ulong la)
 {
-  GEN ev = FlxY_evalx(b, n, p);
+  GEN ev = FlxY_evalx_pre(b, n, p, pi);
   long drop = lg(b) - lg(ev);
-  ulong r = Flx_resultant(a, ev, p);
-  if (drop && la != 1) r = Fl_mul(r, Fl_powu(la, drop,p),p);
+  ulong r = Flx_resultant_pre(a, ev, p, pi);
+  if (drop && la != 1) r = Fl_mul(r, Fl_powu_pre(la, drop, p, pi), p);
   return r;
 }
 static GEN
@@ -1475,6 +1545,7 @@ static GEN
 Flx_FlxY_resultant_polint(GEN a, GEN b, ulong p, long dres, long sx)
 {
   long i;
+  ulong pi = SMALL_ULONG(p)? 0: get_Fl_red(p);
   ulong n, la = Flx_lead(a);
   GEN  x = cgetg(dres+2, t_VECSMALL);
   GEN  y = cgetg(dres+2, t_VECSMALL);
@@ -1482,35 +1553,35 @@ Flx_FlxY_resultant_polint(GEN a, GEN b, ulong p, long dres, long sx)
   * P_{-n}(-X), where P_i is Lagrange polynomial: P_i(j) = delta_{i,j} */
   for (i=0,n = 1; i < dres; n++)
   {
-    x[++i] = n;   y[i] = Flx_FlxY_eval_resultant(a,b, x[i], p,la);
-    x[++i] = p-n; y[i] = Flx_FlxY_eval_resultant(a,b, x[i], p,la);
+    x[++i] = n;   y[i] = Flx_FlxY_eval_resultant(a,b, x[i], p,pi,la);
+    x[++i] = p-n; y[i] = Flx_FlxY_eval_resultant(a,b, x[i], p,pi,la);
   }
   if (i == dres)
   {
-    x[++i] = 0;   y[i] = Flx_FlxY_eval_resultant(a,b, x[i], p,la);
+    x[++i] = 0;   y[i] = Flx_FlxY_eval_resultant(a,b, x[i], p,pi,la);
   }
   return Flv_polint(x,y, p, sx);
 }
 
 static GEN
-FlxX_pseudorem(GEN x, GEN y, ulong p)
+FlxX_pseudorem(GEN x, GEN y, ulong p, ulong pi)
 {
   long vx = varn(x), dx, dy, dz, i, lx, dp;
   pari_sp av = avma, av2;
 
   if (!signe(y)) pari_err_INV("FlxX_pseudorem",y);
   (void)new_chunk(2);
-  dx=degpol(x); x = RgX_recip_shallow(x)+2;
-  dy=degpol(y); y = RgX_recip_shallow(y)+2; dz=dx-dy; dp = dz+1;
+  dx=degpol(x); x = RgX_recip_i(x)+2;
+  dy=degpol(y); y = RgX_recip_i(y)+2; dz=dx-dy; dp = dz+1;
   av2 = avma;
   for (;;)
   {
     gel(x,0) = Flx_neg(gel(x,0), p); dp--;
     for (i=1; i<=dy; i++)
-      gel(x,i) = Flx_add( Flx_mul(gel(y,0), gel(x,i), p),
-                              Flx_mul(gel(x,0), gel(y,i), p), p );
+      gel(x,i) = Flx_add( Flx_mul_pre(gel(y,0), gel(x,i), p, pi),
+                          Flx_mul_pre(gel(x,0), gel(y,i), p, pi), p );
     for (   ; i<=dx; i++)
-      gel(x,i) = Flx_mul(gel(y,0), gel(x,i), p);
+      gel(x,i) = Flx_mul_pre(gel(y,0), gel(x,i), p, pi);
     do { x++; dx--; } while (dx >= 0 && lg(gel(x,0))==2);
     if (dx < dy) break;
     if (gc_needed(av2,1))
@@ -1523,12 +1594,11 @@ FlxX_pseudorem(GEN x, GEN y, ulong p)
   lx = dx+3; x -= 2;
   x[0]=evaltyp(t_POL) | evallg(lx);
   x[1]=evalsigne(1) | evalvarn(vx);
-  x = RgX_recip_shallow(x);
+  x = RgX_recip_i(x);
   if (dp)
   { /* multiply by y[0]^dp   [beware dummy vars from FpX_FpXY_resultant] */
-    GEN t = Flx_powu(gel(y,0), dp, p);
-    for (i=2; i<lx; i++)
-      gel(x,i) = Flx_mul(gel(x,i), t, p);
+    GEN t = Flx_powu_pre(gel(y,0), dp, p, pi);
+    for (i=2; i<lx; i++) gel(x,i) = Flx_mul_pre(gel(x,i), t, p, pi);
   }
   return gerepilecopy(av, x);
 }
@@ -1538,22 +1608,24 @@ GEN
 FlxX_resultant(GEN u, GEN v, ulong p, long sx)
 {
   pari_sp av = avma, av2;
-  long degq,dx,dy,du,dv,dr,signh;
-  GEN z,g,h,r,p1;
+  long degq, dx, dy, du, dv, dr, signh;
+  ulong pi;
+  GEN z, g, h, r, p1;
 
-  dx=degpol(u); dy=degpol(v); signh=1;
+  dx = degpol(u); dy = degpol(v); signh = 1;
   if (dx < dy)
   {
     swap(u,v); lswap(dx,dy);
     if (both_odd(dx, dy)) signh = -signh;
   }
   if (dy < 0) return zero_Flx(sx);
-  if (dy==0) return gerepileupto(av, Flx_powu(gel(v,2),dx,p));
+  pi = SMALL_ULONG(p)? 0: get_Fl_red(p);
+  if (dy==0) return gerepileupto(av, Flx_powu_pre(gel(v,2),dx,p,pi));
 
   g = h = pol1_Flx(sx); av2 = avma;
   for(;;)
   {
-    r = FlxX_pseudorem(u,v,p); dr = lg(r);
+    r = FlxX_pseudorem(u,v,p,pi); dr = lg(r);
     if (dr == 2) { set_avma(av); return zero_Flx(sx); }
     du = degpol(u); dv = degpol(v); degq = du-dv;
     u = v; p1 = g; g = leading_coeff(u);
@@ -1561,10 +1633,11 @@ FlxX_resultant(GEN u, GEN v, ulong p, long sx)
     {
       case 0: break;
       case 1:
-        p1 = Flx_mul(h,p1, p); h = g; break;
+        p1 = Flx_mul_pre(h,p1, p, pi); h = g; break;
       default:
-        p1 = Flx_mul(Flx_powu(h,degq,p), p1, p);
-        h = Flx_div(Flx_powu(g,degq,p), Flx_powu(h,degq-1,p), p);
+        p1 = Flx_mul_pre(Flx_powu_pre(h,degq,p,pi), p1, p, pi);
+        h = Flx_div_pre(Flx_powu_pre(g,degq,p,pi),
+                        Flx_powu_pre(h,degq-1,p,pi), p, pi);
     }
     if (both_odd(du,dv)) signh = -signh;
     v = FlxY_Flx_div(r, p1, p);
@@ -1576,7 +1649,8 @@ FlxX_resultant(GEN u, GEN v, ulong p, long sx)
     }
   }
   z = gel(v,2);
-  if (dv > 1) z = Flx_div(Flx_powu(z,dv,p), Flx_powu(h,dv-1,p), p);
+  if (dv > 1) z = Flx_div_pre(Flx_powu_pre(z,dv,p,pi),
+                              Flx_powu_pre(h,dv-1,p,pi), p, pi);
   if (signh < 0) z = Flx_neg(z,p);
   return gerepileupto(av, z);
 }
@@ -1718,17 +1792,6 @@ FpXV_direct_compositum(GEN V, GEN p)
 
 /* 0, 1, -1, 2, -2, ... */
 #define next_lambda(a) (a>0 ? -a : 1-a)
-GEN
-FpX_compositum(GEN a, GEN b, GEN p)
-{
-  long k, v = fetch_var_higher();
-  for (k = 1;; k = next_lambda(k))
-  {
-    GEN x = deg1pol_shallow(gen_1, gmulsg(k, pol_x(v)), 0); /* x + k y */
-    GEN C = FpX_FpXY_resultant(a, poleval(b,x),p);
-    if (FpX_is_squarefree(C, p)) { (void)delete_var(); return C; }
-  }
-}
 
 /* Assume A in Z[Y], B in Q[Y][X], B squarefree in (Q[Y]/(A))[X] and
  * Res_Y(A, B) in Z[X]. Find a small lambda (start from *lambda, use
@@ -1753,8 +1816,7 @@ ZX_ZXY_resultant_LERS(GEN A, GEN B0, long *plambda, GEN *LERS)
     H = gsubst(B, vY, gdiv(gneg(a0),a1));
    if (!equali1(a1)) H = RgX_Rg_mul(H, powiu(a1, poldegree(B,vY)));
     *LERS = mkvec2(scalarpol_shallow(a0,vX), scalarpol_shallow(a1,vX));
-    gerepileall(av, 2, &H, LERS);
-    return H;
+    return gc_all(av, 2, &H, LERS);
   }
 
   dglist = Hp = H0p = H1p = C0 = C1 = NULL; /* gcc -Wall */
@@ -1871,8 +1933,7 @@ END:
   if (DEBUGLEVEL>5) err_printf(" done\n");
   setvarn(H, vX); (void)delete_var();
   *LERS = mkvec2(H0,H1);
-  gerepileall(av, 2, &H, LERS);
-  *plambda = lambda; return H;
+  *plambda = lambda; return gc_all(av, 2, &H, LERS);
 }
 
 GEN
@@ -1990,9 +2051,7 @@ ZX_resultant_slice(GEN A, GEN B, GEN dB, GEN P, GEN *mod)
     H[i] = ZX_resultant_prime(a, b, dB, degA, degB, p);
   }
   H = ZV_chinese_tree(H, P, T, ZV_chinesetree(P,T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -2179,9 +2238,7 @@ ZXQX_resultant_slice(GEN A, GEN B, GEN U, GEN dB, GEN P, GEN *mod)
   }
   if (redo) T = ZV_producttree(P);
   H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P, T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -2195,20 +2252,22 @@ ZXQX_resultant_worker(GEN P, GEN A, GEN B, GEN T, GEN dB)
 }
 
 static ulong
-ZXQX_resultant_bound(GEN nf, GEN A, GEN B)
+ZXQX_resultant_bound_i(GEN nf, GEN A, GEN B, GEN (*f)(GEN,GEN,long))
 {
   pari_sp av = avma;
-  GEN r, M = L2_bound(nf, NULL, &r);
+  GEN r, M = nf_L2_bound(nf, NULL, &r);
   long v = nf_get_varn(nf), i, l = lg(r);
   GEN a = cgetg(l, t_COL);
   for (i = 1; i < l; i++)
-    gel(a, i) =  RgX_RgXY_ResBound(gsubst(A, v, gel(r,i)),
-                                   gsubst(B, v, gel(r,i)), DEFAULTPREC);
+    gel(a, i) = f(gsubst(A, v, gel(r,i)), gsubst(B, v, gel(r,i)), DEFAULTPREC);
   return gc_ulong(av, (ulong) dbllog2(gmul(M,RgC_fpnorml2(a, DEFAULTPREC))));
 }
+static ulong
+ZXQX_resultant_bound(GEN nf, GEN A, GEN B)
+{ return ZXQX_resultant_bound_i(nf, A, B, &RgX_RgXY_ResBound); }
 
-/* Compute Res(A, B/dB) in Z, assuming A,B in Z[X], dB in Z or NULL (= 1)
- * If B=NULL, take B = A' and assume deg A > 1 and 'bound' is set */
+/* Compute Res(A, B/dB) in Z[X]/T, assuming A,B in Z[X,Y], dB in Z or NULL (= 1)
+ * If B=NULL, take B = A' and assume deg A > 1 */
 static GEN
 ZXQX_resultant_all(GEN A, GEN B, GEN T, GEN dB, ulong bound)
 {
@@ -2221,9 +2280,9 @@ ZXQX_resultant_all(GEN A, GEN B, GEN T, GEN dB, ulong bound)
     if (a < 0 || b < 0) return gen_0;
     if (!a) return gpowgs(gel(A,2), b);
     if (!b) return gpowgs(gel(B,2), a);
-    if (!bound) bound = ZXQX_resultant_bound(nfinit(T, DEFAULTPREC), A, B);
   } else
-    if (!bound) bound = ZXQX_resultant_bound(nfinit(T, DEFAULTPREC), A, RgX_deriv(A));
+    if (!bound) B = RgX_deriv(A);
+  if (!bound) bound = ZXQX_resultant_bound(nfinit(T, DEFAULTPREC), A, B);
   worker = snm_closure(is_entry("_ZXQX_resultant_worker"),
                        mkvec4(A, B? B: gen_0, T, dB? dB: gen_0));
   init_modular_big(&S);
@@ -2336,8 +2395,7 @@ QXQ_inv_slice(GEN A, GEN B, GEN P, GEN *mod)
       *mod = gen_1; return pol_0(v);
     }
     H = gerepilecopy(av, Flx_to_ZX(U));
-    *mod = utoi(p);
-    return H;
+    *mod = utoipos(p); return H;
   }
   T = ZV_producttree(P);
   A = ZX_nv_mod_tree(A, P, T);
@@ -2358,9 +2416,7 @@ QXQ_inv_slice(GEN A, GEN B, GEN P, GEN *mod)
   }
   if (redo) T = ZV_producttree(P);
   H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P, T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -2441,8 +2497,7 @@ QXQ_div_slice(GEN A, GEN B, GEN C, GEN P, GEN *mod)
     }
     U = Flxq_mul(a, bi, c, p);
     H = gerepilecopy(av, Flx_to_ZX(U));
-    *mod = utoi(p);
-    return H;
+    *mod = utoipos(p); return H;
   }
   T = ZV_producttree(P);
   A = ZX_nv_mod_tree(A, P, T);
@@ -2464,9 +2519,7 @@ QXQ_div_slice(GEN A, GEN B, GEN C, GEN P, GEN *mod)
   }
   if (redo) T = ZV_producttree(P);
   H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P, T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -2551,10 +2604,8 @@ ZXQ_minpoly_slice(GEN A, GEN B, long d, GEN P, GEN *mod)
     GEN a = ZX_to_Flx(A, p), b = ZX_to_Flx(B, p);
     GEN Hp = Flxq_minpoly(a, b, p);
     if (degpol(Hp) != d) { p = 1; Hp = pol0_Flx(v); }
-    H = Flx_to_ZX(Hp);
-    *mod = utoi(p);
-    gerepileall(av, 2, &H, mod);
-    return H;
+    H = gerepileupto(av, Flx_to_ZX(Hp));
+    *mod = utoipos(p); return H;
   }
   T = ZV_producttree(P);
   A = ZX_nv_mod_tree(A, P, T);
@@ -2569,9 +2620,7 @@ ZXQ_minpoly_slice(GEN A, GEN B, long d, GEN P, GEN *mod)
     gel(H, i) = m;
   }
   H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P, T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -2644,10 +2693,8 @@ ZX_ZXY_resultant_slice(GEN A, GEN B, GEN dB, long degA, long degB, long dres,
     ulong dp = dB ? umodiu(dB, p): 1;
     GEN a = ZX_to_Flx(A, p), b = ZXX_to_FlxX(B, p, vY);
     GEN Hp = ZX_ZXY_resultant_prime(a, b, dp, p, degA, degB, dres, sX);
-    H = Flx_to_ZX(Hp);
-    *mod = utoi(p);
-    gerepileall(av, 2, &H, mod);
-    return H;
+    H = gerepileupto(av, Flx_to_ZX(Hp));
+    *mod = utoipos(p); return H;
   }
   T = ZV_producttree(P);
   A = ZX_nv_mod_tree(A, P, T);
@@ -2662,9 +2709,7 @@ ZX_ZXY_resultant_slice(GEN A, GEN B, GEN dB, long degA, long degB, long dres,
     gel(H,i) = ZX_ZXY_resultant_prime(a, b, dp, p, degA, degB, dres, sX);
   }
   H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P, T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -2766,8 +2811,7 @@ ZX_direct_compositum_slice(GEN A, GEN B, GEN P, GEN *mod)
     GEN a = ZX_to_Flx(A, p), b = ZX_to_Flx(B, p);
     GEN Hp = Flx_direct_compositum(a, b, p);
     H = gerepileupto(av, Flx_to_ZX(Hp));
-    *mod = utoi(p);
-    return H;
+    *mod = utoipos(p); return H;
   }
   T = ZV_producttree(P);
   A = ZX_nv_mod_tree(A, P, T);
@@ -2780,9 +2824,7 @@ ZX_direct_compositum_slice(GEN A, GEN B, GEN P, GEN *mod)
     gel(H,i) = Flx_direct_compositum(a, b, p);
   }
   H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P, T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -2802,7 +2844,8 @@ ZX_direct_compositum(GEN A, GEN B, GEN lead)
   forprime_t S;
   ulong bound;
   GEN H, worker, mod;
-  bound = ZX_ZXY_ResBound(A, poleval(B,deg1pol(gen_1,pol_x(1),0)), NULL);
+  if (degpol(A) < degpol(B)) swap(A, B);
+  bound = ZX_ZXY_ResBound_1(A, B);
   worker = snm_closure(is_entry("_ZX_direct_compositum_worker"), mkvec2(A,B));
   init_modular_big(&S);
   H = gen_crt("ZX_direct_compositum", worker, &S, lead, bound, 0, &mod,
@@ -2860,8 +2903,7 @@ ZXQX_direct_compositum_slice(GEN A, GEN B, GEN C, GEN P, GEN *mod)
     GEN c = ZX_to_Flx(C, p);
     GEN Hp = FlxX_to_Flm(FlxqX_direct_compositum(a, b, c, p), dC);
     H = gerepileupto(av, Flm_to_ZM(Hp));
-    *mod = utoi(p);
-    return H;
+    *mod = utoipos(p); return H;
   }
   T = ZV_producttree(P);
   A = ZXX_nv_mod_tree(A, P, T, v);
@@ -2875,9 +2917,7 @@ ZXQX_direct_compositum_slice(GEN A, GEN B, GEN C, GEN P, GEN *mod)
     gel(H,i) = FlxX_to_Flm(FlxqX_direct_compositum(a, b, c, p), dC);
   }
   H = nmV_chinese_center_tree_seq(H, P, T, ZV_chinesetree(P, T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -2908,17 +2948,7 @@ ZXQX_direct_compositum(GEN A, GEN B, GEN T, ulong bound)
 
 static long
 ZXQX_direct_compositum_bound(GEN nf, GEN A, GEN B)
-{
-  pari_sp av = avma;
-  GEN r, M = L2_bound(nf, NULL, &r);
-  long v = nf_get_varn(nf), i, l = lg(r);
-  GEN a = cgetg(l, t_COL);
-  for (i = 1; i < l; i++)
-    gel(a, i) =  RgX_RgXY_ResBound(gsubst(A, v, gel(r,i)),
-                 poleval(gsubst(B, v, gel(r,i)),
-                         deg1pol(gen_1, pol_x(1), 0)), DEFAULTPREC);
-  return gc_long(av, (long) dbllog2(gmul(M,RgC_fpnorml2(a, DEFAULTPREC))));
-}
+{ return ZXQX_resultant_bound_i(nf, A, B, &RgX_RgXY_ResBound_1); }
 
 GEN
 nf_direct_compositum(GEN nf, GEN A, GEN B)
@@ -3125,7 +3155,7 @@ GEN
 ffsumnbirred(GEN p, long n)
 {
   pari_sp av = avma, av2;
-  GEN q, t = p, v = vecfactoru(1, n);
+  GEN q, t = p, v = vecfactoru_i(1, n);
   long i;
   q = cgetg(n+1,t_VEC); gel(q,1) = p;
   for (i=2; i<=n; i++) gel(q,i) = mulii(gel(q,i-1), p);
