@@ -21,6 +21,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_trans
+
 GEN
 trans_fix_arg(long *prec, GEN *s0, GEN *sig, GEN *tau, pari_sp *av, GEN *res)
 {
@@ -765,6 +767,60 @@ mpatanh(GEN x)
   shiftr_inplace(z, -1); return gerepileuptoleaf(av, z);
 }
 
+/* atanh(u/v) using binary splitting, 0 < u < v */
+GEN
+atanhuu(ulong u, ulong v, long prec)
+{
+  long i, nmax;
+  GEN u2 = sqru(u), v2 = sqru(v);
+  double d = ((double)v) / u;
+  struct abpq_res R;
+  struct abpq A;
+  /* satisfies (2n+1) (v/u)^2n > 2^bitprec */
+  nmax = (long)ceil(prec2nbits(prec) / (2*log2(d)));
+  abpq_init(&A, nmax);
+  A.a[0] = A.b[0] = gen_1;
+  A.p[0] = utoipos(u);
+  A.q[0] = utoipos(v);
+  for (i = 1; i <= nmax; i++)
+  {
+    A.a[i] = gen_1;
+    A.b[i] = utoipos((i<<1)+1);
+    A.p[i] = u2;
+    A.q[i] = v2;
+  }
+  abpq_sum(&R, 0, nmax, &A);
+  return rdivii(R.T, mulii(R.B,R.Q),prec);
+}
+/* atanh(u/v) using binary splitting, 0 < u < v */
+GEN
+atanhui(ulong u, GEN v, long prec)
+{
+  long i, nmax;
+  GEN u2 = sqru(u), v2 = sqri(v);
+  double d = gtodouble(v) / u;
+  struct abpq_res R;
+  struct abpq A;
+  /* satisfies (2n+1) (v/u)^2n > 2^bitprec */
+  nmax = (long)ceil(prec2nbits(prec) / (2*log2(d)));
+  abpq_init(&A, nmax);
+  A.a[0] = A.b[0] = gen_1;
+  A.p[0] = utoipos(u);
+  A.q[0] = v;
+  for (i = 1; i <= nmax; i++)
+  {
+    A.a[i] = gen_1;
+    A.b[i] = utoipos((i<<1)+1);
+    A.p[i] = u2;
+    A.q[i] = v2;
+  }
+  abpq_sum(&R, 0, nmax, &A);
+  return rdivii(R.T, mulii(R.B,R.Q),prec);
+}
+
+static void
+err_atanh(GEN x, GEN bad) { pari_err_DOMAIN("atanh", "x", "=", bad, x); }
+
 GEN
 gatanh(GEN x, long prec)
 {
@@ -774,6 +830,45 @@ gatanh(GEN x, long prec)
 
   switch(typ(x))
   {
+    case t_INT:
+      sx = signe(x);
+      if (!sx) return real_0(prec);
+      z = cgetg(3, t_COMPLEX); av = avma;
+      if (lgefint(x) == 3)
+      {
+        if (x[2] == 1) err_atanh(x, sx == 1? gen_1: gen_m1);
+        a = atanhuu(1, x[2], prec);
+      }
+      else
+        a = atanhui(1, x, prec);
+      gel(z,1) = gerepileuptoleaf(av, a);
+      gel(z,2) = Pi2n(-1, prec);
+      togglesign(sx > 0? gel(z,2): gel(z,1));
+      return z;
+    case t_FRAC:
+    {
+      long ly, lz;
+
+      y = gel(x,1); ly = lgefint(y);
+      z = gel(x,2); lz = lgefint(z); if (ly > 3 && lz > 3) break;
+      if (abscmpii(y, z) > 0) /* |y| > z; lz = 3 */
+      {
+        ulong u = z[2];
+        z = cgetg(3, t_COMPLEX); av = avma;
+        a = ly == 3? atanhuu(u, y[2], prec): atanhui(u, y, prec);
+        gel(z,1) = gerepileuptoleaf(av, a);
+        gel(z,2) = Pi2n(-1, prec);
+        togglesign(signe(y) > 0? gel(z,2): gel(z,1));
+      }
+      else
+      { /* |y| < z; ly = 3 */
+        av = avma;
+        a = lz == 3? atanhuu(y[2], z[2], prec): atanhui(y[2], z, prec);
+        z = gerepileuptoleaf(av, a);
+        if (signe(y) < 0) togglesign(z);
+      }
+      return z;
+    }
     case t_REAL:
       sx = signe(x);
       if (!sx) return real_0_bit(expo(x));
@@ -782,10 +877,10 @@ gatanh(GEN x, long prec)
       y = cgetg(3,t_COMPLEX);
       av = avma;
       z = subrs(x,1);
-      if (!signe(z)) pari_err_DOMAIN("atanh", "argument", "=", gen_1, x);
+      if (!signe(z)) err_atanh(x, gen_1);
       z = invr(z); shiftr_inplace(z, 1); /* 2/(x-1)*/
       z = addrs(z,1);
-      if (!signe(z)) pari_err_DOMAIN("atanh", "argument", "=", gen_m1, x);
+      if (!signe(z)) err_atanh(x, gen_m1);
       z = logr_abs(z);
       shiftr_inplace(z, -1); /* (1/2)log((1+x)/(x-1)) */
       gel(y,1) = gerepileuptoleaf(av, z);
@@ -852,36 +947,13 @@ mulu_interval_step_prec(long l, long m, long s, long prec)
 
 /* x * (i*(i+1)) */
 static GEN
-muliunu(GEN x, ulong i)
+muliunextu(GEN x, ulong i)
 {
   if (i & HIGHMASK) /* i(i+1) >= 2^BITS_IN_LONG*/
-    return muliu(muliu(x, i), i+1);
+    return mulii(x, muluu(i, i+1));
   else
     return muliu(x, i*(i+1));
 }
-/* x / (i*(i+1)) */
-GEN
-divrunu(GEN x, ulong i)
-{
-  if (i & HIGHMASK) /* i(i+1) >= 2^BITS_IN_LONG*/
-    return divru(divru(x, i), i+1);
-  else
-    return divru(x, i*(i+1));
-}
-/* x / (i*(i+1)) */
-GEN
-divgunu(GEN x, ulong i)
-{
-#ifdef LONG_IS_64BIT
-  if (i < 3037000500L) /* i(i+1) < 2^63 */
-#else
-  if (i < 46341L) /* i(i+1) < 2^31 */
-#endif
-    return gdivgs(x, i*(i+1));
-  else
-    return gdivgs(gdivgs(x, i), i+1);
-}
-
 /* arg(s+it) */
 double
 darg(double s, double t)
@@ -922,7 +994,7 @@ red_mod_2z(GEN x, GEN z)
 static GEN
 negeuler(long prec) { GEN g = mpeuler(prec); setsigne(g, -1); return g; }
 /* lngamma(1+z) = -Euler*z + sum_{i > 1} zeta(i)/i (-z)^i
- * at relative precision prec, |z| < 1 is small */
+ * at relative precision prec, |z| <= 1/2 is small */
 static GEN
 lngamma1(GEN z, long prec)
 { /* sum_{i > l} |z|^(i-1) = |z|^l / (1-|z|) < 2^-B
@@ -942,8 +1014,8 @@ lngamma1(GEN z, long prec)
 }
 /* B_i / (i(i-1)), i even. Sometimes NOT reduced (but gadd/gmul won't care)!*/
 static GEN
-bern_unu(long i)
-{ GEN B = bernfrac(i); return mkfrac(gel(B,1), muliunu(gel(B,2), i-1)); }
+bern_unextu(long i)
+{ GEN B = bernfrac(i); return mkfrac(gel(B,1), muliunextu(gel(B,2), i-1)); }
 /* B_i / i, i even. Sometimes NOT reduced (but gadd/gmul won't care)!*/
 static GEN
 bern_u(long i)
@@ -953,11 +1025,11 @@ static GEN
 lngamma_sum(GEN a, long N)
 {
   pari_sp av = avma;
-  GEN S = bern_unu(2*N);
+  GEN S = bern_unextu(2*N);
   long i;
   for (i = 2*N-2; i > 0; i -= 2)
   {
-    S = gadd(bern_unu(i), gmul(a,S));
+    S = gadd(bern_unextu(i), gmul(a,S));
     if (gc_needed(av,3))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"gamma: i = %ld", i);
@@ -984,11 +1056,62 @@ psi_sum(GEN a, long N)
   }
   return gmul(a,S);
 }
+static void
+gamma_optim(double ssig, double st, long prec, long *plim, long *pN)
+{
+  double la, l,l2,u,v, rlogs, ilogs;
+  long N = 1, lim;
+  dcxlog(ssig,st, &rlogs,&ilogs);
+  /* Re (s - 1/2) log(s) */
+  u = (ssig - 0.5)*rlogs - st * ilogs;
+  /* Im (s - 1/2) log(s) */
+  v = (ssig - 0.5)*ilogs + st * rlogs;
+  /* l2 = | (s - 1/2) log(s) - s + log(2Pi)/2 |^2 ~ |lngamma(s))|^2 */
+  u = u - ssig + log(2.*M_PI)/2;
+  v = v - st;
+  l2 = u*u + v*v;
+  if (l2 < 0.000001) l2 = 0.000001;
+  l = (prec2nbits_mul(prec, M_LN2) - log(l2)/2) / 2.;
+  if (l < 0) l = 0.;
+
+  if (st > 1 && l > 0)
+  {
+    double t = st * M_PI / l;
+    la = t * log(t);
+    if (la < 4.) la = 4.;
+    if (la > 150) la = t;
+  }
+  else
+    la = 4.; /* heuristic */
+  lim = (long)ceil(l / (1.+ log(la)));
+  if (lim == 0) lim = 1;
+
+  u = (lim-0.5) * la / M_PI;
+  l2 = u*u - st*st;
+  if (l2 > 0)
+  {
+    double t = ceil(sqrt(l2) - ssig);
+    if (t > 1) N = (long)t;
+  }
+  *plim = lim; *pN = N;
+}
+/* do we use lngamma1 instead of Euler-Maclaurin ? */
+static int
+gamma_use_1(double s, double t, long prec, long *plim, long *pN)
+{
+  double a = s-1, d = fabs(a) + fabs(t);
+  long k;
+  if (d < 1e-16) return 1;
+  gamma_optim(s, t, prec, plim, pN);
+  if (d >= 0.5) return 0;
+  k = bit_accuracy(prec) / -log2(dnorm(a, t)); /* 2k = lngamma1 bound */
+  return (t ? k: 1.5*k) < *plim + *pN;
+}
 static GEN
 cxgamma(GEN s0, int dolog, long prec)
 {
-  GEN s, a, y, res, sig, tau, p1, nnx, pi, pi2, sqrtpi2;
-  long i, lim, N, esig, et;
+  GEN s, a, y, res, sig, tau, B, nnx, pi, pi2;
+  long i, esig, et, lim, N = 1;
   pari_sp av, av2;
   int funeq = 0;
   pari_timer T;
@@ -1036,69 +1159,31 @@ cxgamma(GEN s0, int dolog, long prec)
     if (signe(l2) > 0)
     {
       l2 = gsub(gsqrt(l2,3), sig);
-      if (signe(l2) > 0) N = itos( gceil(l2) ); else N = 1;
+      if (signe(l2) > 0) N = itos( gceil(l2) );
     }
-    else
-      N = 1;
   }
   else
   { /* |s| is moderate. Use floats  */
     double ssig = rtodbl(sig);
     double st = typ(s) == t_REAL? 0.0: rtodbl(imag_i(s));
-    double la, l,l2,u,v, rlogs, ilogs;
 
-    if (fabs(ssig-1) + fabs(st) < 1e-16)
+    if (gamma_use_1(ssig, st, prec, &lim, &N))
     { /* s ~ 1: loggamma(1+u) ~ - Euler * u, cancellation */
       if (funeq) /* s0 ~ 0: use lngamma(s0)+log(s0) = lngamma(s0+1) */
-      {
-        if (dolog)
-          y = gsub(lngamma1(s0,prec), glog(s0,prec));
-        else
-          y = gdiv(gexp(lngamma1(s0,prec), prec), s0);
-      }
+        y = dolog? gsub(lngamma1(s0,prec), glog(s0,prec))
+                 : gdiv(gexp(lngamma1(s0,prec), prec), s0);
       else
       {
-        if (isint1(s0)) { set_avma(av); return dolog? real_0(prec): real_1(prec); }
+        if (isint1(s0))
+        {
+          set_avma(av);
+          return dolog? real_0(prec): real_1(prec);
+        }
         y = lngamma1(gsubgs(s0,1),prec);
         if (!dolog) y = gexp(y,prec);
       }
       set_avma(av); return affc_fixlg(y, res);
     }
-    dcxlog(ssig,st, &rlogs,&ilogs);
-    /* Re (s - 1/2) log(s) */
-    u = (ssig - 0.5)*rlogs - st * ilogs;
-    /* Im (s - 1/2) log(s) */
-    v = (ssig - 0.5)*ilogs + st * rlogs;
-    /* l2 = | (s - 1/2) log(s) - s + log(2Pi)/2 |^2 ~ |lngamma(s))|^2 */
-    u = u - ssig + log(2.*M_PI)/2;
-    v = v - st;
-    l2 = u*u + v*v;
-    if (l2 < 0.000001) l2 = 0.000001;
-    l = (prec2nbits_mul(prec, M_LN2) - log(l2)/2) / 2.;
-    if (l < 0) l = 0.;
-
-    if (st > 1 && l > 0)
-    {
-      double t = st * M_PI / l;
-      la = t * log(t);
-      if (la < 4.) la = 4.;
-      if (la > 150) la = t;
-    }
-    else
-      la = 4.; /* heuristic */
-    lim = (long)ceil(l / (1.+ log(la)));
-    if (lim == 0) lim = 1;
-
-    u = (lim-0.5) * la / M_PI;
-    l2 = u*u - st*st;
-    if (l2 > 0)
-    {
-      double t = ceil(sqrt(l2) - ssig);
-      N = (t < 1)? 1: (long)t;
-      if (N < 1) N = 1;
-    }
-    else
-      N = 1;
   }
   if (DEBUGLEVEL>5) err_printf("lim, N: [%ld, %ld]\n",lim,N);
   incrprec(prec);
@@ -1132,7 +1217,6 @@ cxgamma(GEN s0, int dolog, long prec)
         }
       }
     }
-    if (dolog) y = logr_abs(y);
   }
   else
   { /* Compute lngamma mod 2 I Pi */
@@ -1148,82 +1232,76 @@ cxgamma(GEN s0, int dolog, long prec)
       }
     }
     if (!odd(N)) y = gmul(y, gaddsg(N - 1, s));
-    if (dolog)
-    {
-      if (typ(s) == t_REAL) y = logr_abs(y);
-      else
-      { /* fix imaginary part */
-        long prec0 = LOWDEFAULTPREC;
-        GEN s0 = gprec_w(s, prec0), y0 = s0, k;
-        y0 = garg(y0, prec0); /* Im log(s) at low accuracy */
-        for (i=1; i < N; i++) y0 = gadd(y0, garg(gaddgs(s0,i), prec0));
-        y = glog(y, prec);
-        k = ground( gdiv(gsub(y0, imag_i(y)), Pi2n(1,prec0)) );
-        if (signe(k)) y = gadd(y, mulcxI(mulir(k, Pi2n(1, prec))));
-      }
-    }
   }
   if (DEBUGLEVEL>5) timer_printf(&T,"product from 0 to N-1");
   constbern(lim);
   nnx = gaddgs(s, N); a = ginv(nnx);
-  p1 = gsub(gmul(gsub(nnx, ghalf), glog(nnx,prec)), nnx);
-  p1 = gadd(p1, gmul(a, lngamma_sum(gsqr(a), lim)));
+  B = gadd(gsub(gmul(gsub(nnx, ghalf), glog(nnx,prec)), nnx),
+           gmul(a, lngamma_sum(gsqr(a), lim)));
   if (DEBUGLEVEL>5) timer_printf(&T,"Bernoulli sum");
 
-  pi = mppi(prec); pi2 = shiftr(pi, 1); sqrtpi2 = sqrtr(pi2);
-
+  pi = mppi(prec); pi2 = shiftr(pi, 1);
   if (dolog)
   {
-    if (funeq)
-    { /* recall that s = 1 - s0 */
-      GEN T = shiftr(sqrtpi2,-1); /* sqrt(2Pi)/2 */
-      if (typ(s) != t_REAL)
+    if (typ(s) == t_REAL)
+    {
+      if (!funeq) y = logr_abs(divrr(sqrtr(pi2), y));
+      else
       {
-        /* We compute log(sin(Pi s0)) so that it has branch cuts along
-        * (-oo, 0] and [1, oo).  To do this in a numerically stable way
+        GEN T = shiftr(sqrtr(pi2),-1); /* sqrt(Pi/2) */
+        /* s0 < 0, step (*) simplifies: imag(lngamma(s0)) = - Pi * floor(s0) */
+        y = logr_abs(divrr(mulrr(y, T), mpsin(gmul(pi,s0))));
+        y = mkcomplex(y, mulri(pi, gfloor(s0)));
+        B = gneg(B);
+      }
+    }
+    else
+    { /* log(y), fixing imaginary part */
+      long prec2 = LOWDEFAULTPREC;
+      GEN k, s2 = gprec_w(s, prec2), y2 = garg(s2, prec2); /* ~ Im log(s) */
+      for (i=1; i < N; i++) y2 = gadd(y2, garg(gaddgs(s2,i), prec2));
+      y = glog(y, prec);
+      k = ground( gdiv(gsub(y2, imag_i(y)), Pi2n(1,prec2)) );
+      if (signe(k)) y = gadd(y, mulcxI(mulir(k, Pi2n(1, prec))));
+      if (!funeq) y = gsub(shiftr(logr_abs(pi2),-1), y); /* y -> sqrt(2Pi)/y */
+      else
+      { /* recall that s = 1 - s0 */
+        GEN T = shiftr(sqrtr(pi2),-1); /* sqrt(Pi/2) */
+        /* (*) Compute log(sin(Pi s0)) so that it has branch cuts along
+        * (-oo, 0] and [1, oo). To do this in a numerically stable way
         * we must compute the log first then mangle its imaginary part.
         * The rounding operation below is stable because we're rounding
         * a number which is already within 1/4 of an integer. */
 
-        /* z = log( sin(Pi s0) / (sqrt(2Pi)/2) ) */
+        /* z = log(sin(Pi s0) / sqrt(Pi/2)) */
         GEN z = glog(gdiv(gsin(gmul(pi,s0),prec), T), prec);
-        /* b = (2 Re(s) - 1) / 4 */
-        GEN b = shiftr(subrs(shiftr(sig, 1), 1), -2);
+        GEN b = shiftr(subrs(shiftr(sig, 1), 1), -2); /* (2 Re(s)-1) / 4 */
         y = gsub(y, z);
         if (gsigne(imag_i(s)) > 0) togglesign(b);
-        /* z = 2Pi round( Im(z)/2Pi - b ) */
-        z = gmul(roundr(gsub(gdiv(imag_i(z), pi2), b)), pi2);
+        z = roundr(gsub(gdiv(imag_i(z), pi2), b)); /* round( Im(z)/2Pi - b ) */
         if (signe(z)) { /* y += I*z, z a t_REAL */
-          if (typ(y) == t_COMPLEX)
-            gel(y,2) = gadd(gel(y,2), z);
-          else
-            y = mkcomplex(y, z);
+          z = mulir(z, pi2);
+          if (typ(y) == t_COMPLEX) gel(y,2) = gadd(gel(y,2), z);
+          else y = mkcomplex(y, z);
         }
+        B = gneg(B);
       }
-      else
-      { /* s0 < 0, formula simplifies: imag(lngamma(s0)) = - Pi * floor(s0) */
-        GEN z = logr_abs(divrr(mpsin(gmul(pi,s0)), T));
-        y = gsub(y, z);
-        y = mkcomplex(y, mulri(pi, gfloor(s0)));
-      }
-      p1 = gneg(p1);
     }
-    else /* y --> sqrt(2Pi) / y */
-      y = gsub(logr_abs(sqrtpi2), y);
-    y = gadd(p1, y);
+    y = gadd(B, y);
   }
   else
   {
+    GEN sqrtpi2 = sqrtr(pi2);
     if (funeq)
     { /* y --> y Pi/(sin(Pi s) * sqrt(2Pi)) = y sqrt(Pi/2)/sin(Pi s) */
       y = gdiv(gmul(shiftr(sqrtpi2,-1),y), gsin(gmul(pi,s0), prec));
       /* don't use s above: sin(pi s0) = sin(pi s) and the former is
        * more accurate, esp. if s0 ~ 0 */
-      p1 = gneg(p1);
+      B = gneg(B);
     }
     else /* y --> sqrt(2Pi) / y */
       y = gdiv(sqrtpi2, y);
-    y = gmul(gexp(p1, prec), y);
+    y = gmul(gexp(B, prec), y);
   }
   set_avma(av); return affc_fixlg(y, res);
 }
@@ -1322,13 +1400,13 @@ gadw(GEN x, long p)
   gel(u, 1) = s;
   gel(u, 2) = s;
   for (j = 2; j < p; ++j)
-    gel(u, j+1) = gdivgs(gel(u, j), j);
+    gel(u, j+1) = gdivgu(gel(u, j), j);
   for (k = 1, kp = p; k < n; ++k, kp += p) /* kp = k*p */
   {
     GEN c;
-    gel(u, 1) = gdivgs(gadd(gel(u, 1), gel(u, p)), kp);
+    gel(u, 1) = gdivgu(gadd(gel(u, 1), gel(u, p)), kp);
     for (j = 1; j < p; ++j)
-      gel(u, j+1) = gdivgs(gadd(gel(u, j), gel(u, j+1)), kp + j);
+      gel(u, j+1) = gdivgu(gadd(gel(u, j), gel(u, j+1)), kp + j);
 
     t = gmul(t, gaddgs(x, k-1));
     c = leafcopy(gel(u,1)); setvalp(c, valp(c) + k); /* c = u[1] * p^k */
@@ -1358,12 +1436,12 @@ Qp_gamma_Dwork(GEN x, long p)
   if (k)
   {
     GEN x_k = gsubgs(x,k);
-    x = gdivgs(x_k, p);
+    x = gdivgu(x_k, p);
     p1 = gadw(x, p); if (!odd(k)) p1 = gneg(p1);
     for (j = 1; j < k; ++j) p1 = gmul(p1, gaddgs(x_k, j));
   }
   else
-    p1 = gneg(gadw(gdivgs(x, p), p));
+    p1 = gneg(gadw(gdivgu(x, p), p));
   return gerepileupto(ltop, p1);
 }
 
@@ -1379,7 +1457,7 @@ Qp_gamma_Morita(long n, GEN p, long e)
   for (i = 2; i < n; i++)
     if (!pp || i%pp)
     {
-      p2 = gmulgs(p2, i);
+      p2 = gmulgu(p2, i);
       if ((i&0xFL) == 0xFL)
         p2 = gerepileupto(ltop, p2);
     }
@@ -1415,6 +1493,22 @@ Qp_gamma(GEN x)
   return Qp_gamma_Dwork(x, itos(p));
 }
 
+static GEN
+Qp_lngamma(GEN x)
+{
+  GEN s, y, Y;
+  long v = valp(x), e, k, K;
+  if (v >= 0) return Qp_log(Qp_gamma(x));
+  e = precp(x) + v; K = (2 + (e + 4) / (-v)) >> 1;
+  s = gen_0; Y = y = ginv(x); y = gsqr(y); constbern(K);
+  for (k = 1; k <= K; k++)
+  {
+    s = gadd(s, gmul(gdivgunextu(bernfrac(2*k), 2*k-1), Y));
+    if (k < K) Y = gmul(Y, y); /* x^(1-2k) */
+  }
+  return gadd(s, gsub(gmul(gsub(x, ghalf), Qp_log(x)), x));
+}
+
 /* gamma(1+x) - 1, |x| < 1 is "small" */
 GEN
 ggamma1m1(GEN x, long prec) { return gexpm1(lngamma1(x, prec), prec); }
@@ -1673,13 +1767,7 @@ ggamma(GEN x, long prec)
       GEN a = gel(x,1), b = gel(x,2), c = gammafrac24(a, b, prec);
       if (c) return c;
       av = avma; c = subii(a,b);
-      if (expi(c) - expi(b) < -50)
-      { /* x = 1 + c/b is close to 1 */
-        x = mkfrac(c,b);
-        if (lgefint(b) >= prec) x = fractor(x,prec);
-        y = mpexp(lngamma1(x, prec));
-      }
-      else if (signe(a) < 0 || cmpii(shifti(a,1), b) < 0)
+      if (signe(a) < 0)
       { /* gamma will use functional equation x -> z = 1-x = -c/b >= 1/2.
          * Gamma(x) = Pi / (sin(Pi z) * Gamma(z)) */
         GEN z = mkfrac(negi(c), b), q = ground(z), r = gsub(z,q);
@@ -1687,12 +1775,27 @@ ggamma(GEN x, long prec)
         z = fractor(z, prec+EXTRAPRECWORD);
         y = divrr(pi, mulrr(mpsin(gmul(pi, r)), cxgamma(z, 0, prec)));
         if (mpodd(q)) togglesign(y);
+        return gerepileupto(av, y);
+      }
+      if (cmpii(shifti(a,1), b) < 0)
+      { /* 0 < x < 1/2 gamma would use funeq: adding 1 is cheaper. */
+        if (expi(a) - expi(b) < -3) /* close to 0 */
+        {
+          if (lgefint(b) >= prec) x = fractor(x,prec);
+          y = mpexp(lngamma1(x, prec));
+        }
+        else
+          y = cxgamma(fractor(mkfrac(addii(a,b), b), prec), 0, prec);
+        return gerepileupto(av, gdiv(y, x));
+      }
+      if (expi(c) - expi(b) < -3)
+      { /* x = 1 + c/b is close to 1 */
+        x = mkfrac(c,b);
+        if (lgefint(b) >= prec) x = fractor(x,prec);
+        y = mpexp(lngamma1(x, prec));
       }
       else
-      {
-        x = fractor(x, prec);
-        y = cxgamma(x, 0, prec);
-      }
+        y = cxgamma(fractor(x, prec), 0, prec);
       return gerepileupto(av, y);
     }
 
@@ -1806,20 +1909,32 @@ glngamma(GEN x, long prec)
       long e;
       if (c) return glog(c, prec);
       c = subii(a,b); e = expi(b) - expi(c);
-      if (e > 50)
-      {
-        x = mkfrac(c,b);
-        if (lgefint(b) >= prec) x = fractor(x,prec + nbits2nlong(e));
-        y = lngamma1(x, prec);
-      }
-      else if (signe(a) < 0 || cmpii(shifti(a,1), b) < 0)
+      if (signe(a) < 0)
       { /* gamma will use functional equation x -> z = 1-x = -c/b >= 1/2.
          * lngamma(x) = log |Pi / (sin(Pi z) * Gamma(z))| + I*Pi * floor(x) */
         GEN z = mkfrac(negi(c), b), q = ground(z), r = gsub(z,q);
         GEN pi = mppi(prec); /* |r| <= 1/2 */
         z = fractor(z, prec+EXTRAPRECWORD);
-        y = subrr(logr_abs(divrr(pi, mpsin(gmul(pi, r)))), cxgamma(z, 1, prec));
-        if (signe(a) < 0) y = gadd(y, mkcomplex(gen_0, mulri(pi, gfloor(x))));
+        y = subrr(logr_abs(divrr(pi, mpsin(gmul(pi,r)))), cxgamma(z, 1, prec));
+        y = gadd(y, mkcomplex(gen_0, mulri(pi, gfloor(x))));
+        return gerepileupto(av, y);
+      }
+      if (cmpii(shifti(a,1), b) < 0)
+      { /* 0 < x < 1/2 gamma would use funeq: adding 1 is cheaper. */
+        if (expi(a) - expi(b) < -3) /* close to 0 */
+        {
+          if (lgefint(b) >= prec) x = fractor(x,prec);
+          y = lngamma1(x, prec);
+        }
+        else
+          y = cxgamma(fractor(mkfrac(addii(a,b), b), prec), 1, prec);
+        return gerepileupto(av, gsub(y, glog(x, prec)));
+      }
+      if (e > 3)
+      {
+        x = mkfrac(c,b);
+        if (lgefint(b) >= prec) x = fractor(x,prec + nbits2nlong(e));
+        y = lngamma1(x, prec);
       }
       else
       {
@@ -1842,7 +1957,7 @@ glngamma(GEN x, long prec)
         t = gadd(t, glngamma(y0,prec));
       return gerepileupto(av, t);
 
-    case t_PADIC: return gerepileupto(av, Qp_log(Qp_gamma(x)));
+    case t_PADIC: return gerepileupto(av, Qp_lngamma(x));
   }
   return trans_eval("lngamma",glngamma,x,prec);
 }
@@ -1857,6 +1972,21 @@ err_psi(GEN s)
   pari_err_DOMAIN("psi","argument", "=",
                   strtoGENstr("nonpositive integer"), s);
 }
+/* L ~ |log s|^2 */
+static long
+psi_lim(double L, double la, long prec)
+{
+  double d = (prec2nbits_mul(prec, 2*M_LN2) - log(L)) / (4*(1+log(la)));
+  return (d < 2)? 2: 2 + (long)ceil(d);
+}
+/* max(|log (s + it - Euler)|, 1e-6) */
+static double
+dlogE(double s, double t)
+{
+  double rlog, ilog;
+  dcxlog(s - 0.57721566, t, &rlog,&ilog);
+  return maxdd(dnorm(rlog,ilog), 1e-6);
+}
 static GEN
 cxpsi(GEN s0, long prec)
 {
@@ -1876,42 +2006,24 @@ cxpsi(GEN s0, long prec)
   { /* |s| is HUGE. Play safe */
     GEN L, S = gprec_w(s,LOWDEFAULTPREC), rS = real_i(S), iS = imag_i(S);
     double l;
-
-    l = rtodbl( gnorm(glog(S, 3)) );
-    l = log(l) / 2.;
-    lim = 2 + (long)ceil((prec2nbits_mul(prec, M_LN2) - l) / (2*(1+log((double)la))));
-    if (lim < 2) lim = 2;
-
+    lim = psi_lim(rtodbl(gnorm(glog(S,LOWDEFAULTPREC))), la, prec);
     l = (2*lim-1)*la / (2.*M_PI);
     L = gsub(dbltor(l*l), gsqr(iS));
     if (signe(L) < 0) L = gen_0;
-
-    L = gsub(gsqrt(L, 3), rS);
+    L = gsub(gsqrt(L, LOWDEFAULTPREC), rS);
     if (signe(L) > 0) nn = (long)ceil(rtodbl(L)); else nn = 1;
-    if (DEBUGLEVEL>2) err_printf("lim, nn: [%ld, %ld]\n",lim,nn);
   }
   else
   {
-    double ssig = rtodbl(sig);
-    double st = typ(s) == t_REAL? 0.0: rtodbl(imag_i(s));
-    double l;
-    {
-      double rlog, ilog; /* log (s - Euler) */
-      dcxlog(ssig - 0.57721566, st, &rlog,&ilog);
-      l = dnorm(rlog,ilog);
-    }
-    if (l < 0.000001) l = 0.000001;
-    l = log(l) / 2.;
-    lim = 2 + (long)ceil((prec2nbits_mul(prec, M_LN2) - l) / (2*(1+log((double)la))));
-    if (lim < 2) lim = 2;
-
+    double l, rS = rtodbl(sig), iS = typ(s) == t_REAL? 0.0: rtodbl(imag_i(s));
+    lim = psi_lim(dlogE(rS, iS), la, prec);
     l = (2*lim-1)*la / (2.*M_PI);
-    l = l*l - st*st;
+    l = l*l - iS*iS;
     if (l < 0.) l = 0.;
-    nn = (long)ceil( sqrt(l) - ssig );
+    nn = (long)ceil( sqrt(l) - rS );
     if (nn < 1) nn = 1;
-    if (DEBUGLEVEL>2) err_printf("lim, nn: [%ld, %ld]\n",lim,nn);
   }
+  if (DEBUGLEVEL>2) err_printf("lim, nn: [%ld, %ld]\n",lim,nn);
   incrprec(prec); unr = real_1(prec); /* one extra word of precision */
   s2 = gmul2n(s, 1); sq = gsqr(s);
   a = gdiv(unr, gaddgs(s, nn)); /* 1 / (s+n) */
@@ -1971,11 +2083,11 @@ serpsiz0(GEN z0, long L, long v, long prec)
   /* Start from n = 3; in Luke's notation, A2 := A_{n-2}, A1 := A_{n-1},
    * A := A_n. Same for B */
   av = avma;
-  A2= gdivgs(mkpoln(2, gen_1, utoipos(6)), 2);
+  A2= gdivgu(mkpoln(2, gen_1, utoipos(6)), 2);
   B2 = scalarpol_shallow(utoipos(4), 0);
-  A1= gdivgs(mkpoln(3, gen_1, utoipos(82), utoipos(96)), 6);
+  A1= gdivgu(mkpoln(3, gen_1, utoipos(82), utoipos(96)), 6);
   B1 = mkpoln(2, utoipos(8), utoipos(28));
-  A = gdivgs(mkpoln(4, gen_1, utoipos(387), utoipos(2906), utoipos(1920)), 12);
+  A = gdivgu(mkpoln(4, gen_1, utoipos(387), utoipos(2906), utoipos(1920)), 12);
   B = mkpoln(3, utoipos(14), utoipos(204), utoipos(310));
   A2= tr(A2,z0, L);
   B2= tr(B2,z0, L);
@@ -2101,21 +2213,6 @@ serpsi(GEN y, long prec)
   return Q;
 }
 
-static GEN
-hrec(ulong a, long b)
-{
-  long m;
-  switch(b - a)
-  {
-    case 1: return mkfrac(gen_1, utoipos(a));
-    case 2: return mkfrac(utoipos(2*a + 1), muluu(a, a+1));
-  }
-  m = (a + b) >> 1;
-  return gadd(hrec(a, m), hrec(m, b));
-}
-/* exact Harmonic number H_n */
-static GEN
-harmonic(ulong n) { return hrec(1, n+1); }
 static ulong
 psi_n(ulong b)
 {
