@@ -19,6 +19,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_zetamult
+
 /********************************************************************/
 /**                           CONVERSIONS                          **/
 /********************************************************************/
@@ -283,7 +285,7 @@ zetamult_interpolate2_i(GEN avec, GEN t, long prec)
       pari_sp av = avma;
       GEN S = gel(b, j + 1);
       S = gsub(S, gcoeff(a, i+1, 1)); /* u = 0 */
-      if (j > 2) S = gsub(S, gmulgs(gcoeff(a, i+1, 2), j)); /* u = 1 */
+      if (j > 2) S = gsub(S, gmulgu(gcoeff(a, i+1, 2), j)); /* u = 1 */
       if (j >= 4)
       {
         GEN C = utoipos(j*(j-1) / 2);
@@ -296,14 +298,14 @@ zetamult_interpolate2_i(GEN avec, GEN t, long prec)
         }
         if (!odd(j)) S = gsub(S, gmul(C, gcoeff(a,i+1, j/2+1)));
       }
-      gcoeff(a, i+1, j) = gerepileupto(av, gdivgs(S, j));
+      gcoeff(a, i+1, j) = gerepileupto(av, gdivgu(S, j));
     }
   }
   _1 = real_1(prec + EXTRAPRECWORD); av = avma;
   ze = cgetg(n+1, t_VEC);
   for (i = 1; i <= n; i++)
   {
-    GEN S = gdivgs(gcoeff(a, n+2-i, 1), s), sj = utoipos(s);
+    GEN S = gdivgu(gcoeff(a, n+2-i, 1), s), sj = utoipos(s);
     for (j = 2; j <= l; j++)
     {
       sj = muliu(sj, s); /* = s^j */
@@ -429,7 +431,7 @@ filllg1(GEN ibin1, GEN r1, GEN y, long N, long prec)
 }
 static GEN
 fillrec(hashtable *H, GEN evec, long _0, long _1, GEN X, GEN pab, GEN r1,
-        long N, long prec)
+        long N)
 {
   long n, a, b, s, x0;
   GEN xy1, x, y, r, wmid, wini, wfin, mid, ini, fin;
@@ -439,9 +441,9 @@ fillrec(hashtable *H, GEN evec, long _0, long _1, GEN X, GEN pab, GEN r1,
   findabvgen(evec, _0, _1, &wmid, &wini, &wfin, &a, &b);
   x = gel(X, evec[1]); s = lg(evec)-1; /* > 1 */
   y = gel(X, evec[s]);
-  mid = fillrec(H, wmid, _0, _1, X, pab, r1, N, prec);
-  ini = fillrec(H, wini, _0, _1, X, pab, r1, N, prec);
-  fin = fillrec(H, wfin, _0, _1, X, pab, r1, N, prec);
+  mid = fillrec(H, wmid, _0, _1, X, pab, r1, N);
+  ini = fillrec(H, wini, _0, _1, X, pab, r1, N);
+  fin = fillrec(H, wfin, _0, _1, X, pab, r1, N);
   if (evec[1] == _0) { x0 = 1; xy1 = gdiv(r1, y); }
   else { x0 = 0; xy1 = gdiv(r1, gmul(gsubsg(1, x), y)); }
   r = cgetg(N+2, t_VEC); gel(r, N+1) = gen_0;
@@ -591,17 +593,41 @@ zetamult_Akhilesh(GEN e, long bit, long prec)
   if (DEBUGLEVEL) err_printf("polylogmult: k = %ld, %ld nodes\n", k, H->nb);
   return gprec_wtrunc(gel(r,1), prec);
 }
+
+/* lump together close entries + round to 1 entries that are ~ 1 */
+static GEN
+vec_round(GEN V, long b)
+{
+  GEN v = shallowcopy(V), w = shallowcopy(v);
+  long i, j, l = lg(v);
+  for (i = 1; i < l; i++)
+  {
+    long e;
+    if (!gel(v,i)) continue;
+    if (gexpo(gsubgs(gel(v,i), 1)) < b) gel(w,i) = gel(v,i) = gen_1;
+    e = gexpo(gel(v,i));
+    for (j = i+1; j < l; j++)
+      if (gel(v,j) && gexpo(gsub(gel(v,i), gel(v,j))) - e < b)
+      {
+        gel(v,j) = NULL;
+        gel(w,j) = gel(w,i);
+      }
+  }
+  return w;
+}
+
 /* evec t_VEC */
 static GEN
 zetamultevec(GEN evec, long prec)
 {
   pari_sp av = avma;
   double *x, *y, z = 0;
-  long i, j, l, lH, bitprec, prec2, N, _0 = 0, _1 = 0, k = lg(evec) - 1;
+  long b, i, j, l, lH, bitprec, prec2, N, _0 = 0, _1 = 0, k = lg(evec) - 1;
   GEN r1, r, pab, ibin, ibin1, X, Evec, v;
   hashtable *H;
 
   if (k == 0) return gen_1;
+  evec = vec_round(evec, 3 - prec2nbits(prec));
   v = vec_equiv(evec); l = lg(v);
   Evec = cgetg(k+1, t_VECSMALL);
   X = cgetg(l + 2, t_VEC); /* our alphabet */
@@ -629,12 +655,19 @@ zetamultevec(GEN evec, long prec)
   }
   for (i = 1; i < l; i++)
     for (j = i+1; j < l; j++) z = maxdd(z, x[i] + y[j]);
+  b = 0;
+  for (i = 1; i < l; i++)
+  {
+    GEN t = real_i(gel(X,i));
+    long e = -gexpo(gsubgs(gmul2n(t,1), 1));
+    b = maxss(b, e);
+  }
   set_avma(av);
   if (z >= 2) pari_err_IMPL("polylogmult in this range");
   bitprec = prec2nbits(prec) + 64*(1 + (k >> 5));
-  N = 1 + bitprec/ (2 - z);
+  N = 1 + bitprec / (2 - z);
   bitprec += z * N;
-  prec2 = nbits2prec(bitprec);
+  prec2 = nbits2prec(bitprec + b);
   X = gprec_wensure(X, prec2);
   get_ibin(&ibin, &ibin1, N, prec2);
   pab = get_pab(N, k);
@@ -643,7 +676,7 @@ zetamultevec(GEN evec, long prec)
   for (i = 1; i < l; i++)
     if (i != _0 && i != _1)
       hash_insert(H, mkvecsmall(i), filllg1(ibin1, r1, gel(X,i), N, prec2));
-  r = fillrec(H, Evec, _0, _1, X, pab, r1, N, prec2);
+  r = fillrec(H, Evec, _0, _1, X, pab, r1, N);
   if (DEBUGLEVEL) err_printf("polylogmult: k = %ld, %ld nodes\n", k, H->nb);
   return gprec_wtrunc(gel(r,1), prec);
 }
