@@ -22,6 +22,11 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #  include <windows.h>
 #  include "../systems/mingw/mingw.h"
 #endif
+#ifdef DEBUG_FLOATS
+#  undef  _GNU_SOURCE
+#  define _GNU_SOURCE
+#  include <fenv.h>
+#endif
 #include "pari.h"
 #include "paripriv.h"
 #include "gp.h"
@@ -111,7 +116,7 @@ gp_head(void)
 {
   pari_print_version();
   pari_putc('\n');
-  pari_center("Copyright (C) 2000-2020 The PARI Group");
+  pari_center("Copyright (C) 2000-2022 The PARI Group");
   pari_putc('\n');
   print_text("PARI/GP is free software, covered by the GNU General Public \
 License, and comes WITHOUT ANY WARRANTY WHATSOEVER.");
@@ -561,6 +566,9 @@ main(int argc, char **argv)
     puts("### Errors on startup, exiting...\n\n");
     exit(1);
   }
+#ifdef DEBUG_FLOATS
+  feenableexcept(FE_INVALID);
+#endif
 #ifdef __CYGWIN32__
   cyg_environment(argc, argv);
 #endif
