@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -83,7 +84,7 @@ static void DrawString(void *data, long x, long y, char *text, long numtext)
 static void
 draw(PARI_plot *T, GEN w, GEN x, GEN y)
 {
-  char tmppath[MAX_PATH], fname[MAX_PATH];
+  char tmppath[MAX_PATH], fname[2 * MAX_PATH];
   struct plot_eng plotWin32;
   HDC hEmf;
   int r, g, b;
