@@ -70,6 +70,24 @@ extern const long CATCH_ALL;
     return _y;\
   }
 
+#define pari_APPLY_pol(EXPR)\
+  { \
+    long i, _l; \
+    GEN _y = cgetg_copy(x, &_l); _y[1] = x[1];\
+    if (_l == 2) return _y;\
+    for (i=2; i<_l; i++) gel(_y,i) = EXPR;\
+    return normalizepol_lg(_y, _l);\
+  }
+
+#define pari_APPLY_ser(EXPR)\
+  { \
+    long i, _l; \
+    GEN _y = cgetg_copy(x, &_l); _y[1] = x[1];\
+    if (_l == 2) return _y;\
+    for (i=2; i<_l; i++) gel(_y,i) = EXPR;\
+    return normalizeser(_y);\
+  }
+
 #define pari_APPLY_type(TYPE, EXPR)\
   { \
     long i, _l = lg(x); \
@@ -101,13 +119,16 @@ extern const double LOG10_2, LOG2_10;
 #ifndef  M_LN2
 #  define M_LN2 0.693147180559945309417232121458176568
 #endif
+#ifndef M_E
+#define M_E 2.7182818284590452354
+#endif
 
 /* Common global variables: */
 extern int new_galois_format, factor_add_primes, factor_proven;
-extern ulong DEBUGFILES, DEBUGLEVEL, DEBUGMEM, precdl;
+extern ulong DEBUGLEVEL, DEBUGMEM, precdl;
 extern long DEBUGVAR;
 extern ulong pari_mt_nbthreads;
-extern THREAD GEN  zetazone, bernzone, primetab;
+extern THREAD GEN  zetazone, bernzone, eulerzone, primetab;
 extern GEN gen_m1,gen_1,gen_2,gen_m2,ghalf,gen_0,gnil,err_e_STACK;
 extern THREAD VOLATILE int PARI_SIGINT_block, PARI_SIGINT_pending;
 
