@@ -5,7 +5,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -30,7 +31,7 @@ invmod(GEN a, GEN b, GEN *res)
     long l, su, sa = signe(a), lb,lna;
     mp_size_t lu;
     GEN na;
-    if (!sa) { avma = av; *res = absi(b); return 0; }
+    if (!sa) { set_avma(av); *res = absi(b); return 0; }
     if (signe(b) < 0) b = negi(b);
     if (abscmpii(a, b) < 0)
       na = sa > 0? addii(a, b): subii(a, b);
@@ -48,11 +49,11 @@ invmod(GEN a, GEN b, GEN *res)
     /* na >= b */
     l = mpn_gcdext(LIMBS(d), LIMBS(u), &lu, LIMBS(ca), NLIMBS(ca), LIMBS(cb), NLIMBS(cb));
     d[1] = evalsigne(1)|evallgefint(l+2);
-    if (!is_pm1(d)) {avma=av; *res=icopy(d); return 0;}
+    if (!is_pm1(d)) {set_avma(av); *res=icopy(d); return 0;}
     su = lu?((sa ^ lu) < 0)? -1: 1: 0;
     u[1] = evalsigne(su) | evallgefint(labs(lu)+2);
     if (su < 0) u = addii(u, b);
-    avma=av; *res=icopy(u); return 1;
+    set_avma(av); *res=icopy(u); return 1;
   }
 }
 
@@ -159,7 +160,7 @@ bezout(GEN a, GEN b, GEN *pu, GEN *pv)
     d[1] = evalsigne(1)|evallgefint(l+2);
     u[1] = evalsigne(su)|evallgefint(lu+2);
     if (pv) v=diviiexact(subii(d,mulii(u,a)),b);
-    avma = av;
+    set_avma(av);
     if (pu) *pu=icopy(u);
     if (pv) *pv=icopy(v);
     return icopy(d);
