@@ -5,7 +5,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -258,7 +259,7 @@ sqrz_i(GEN z, GEN x, long lz)
     pari_sp av = avma;
     GEN hi = sqrispec_mirror(x+2, lz-2);
     mulrrz_end(z, hi, lz, 1, ez, hi[lz]);
-    avma = av; return;
+    set_avma(av); return;
   }
   if (lz == 3)
   {
@@ -314,7 +315,7 @@ mulrrz_int(GEN z, GEN x, GEN y, long lz, long flag, long sz)
   pari_sp av = avma;
   GEN hi = muliispec_mirror(y+2, x+2, lz+flag-2, lz-2);
   mulrrz_end(z, hi, lz, sz, expo(x)+expo(y), hi[lz]);
-  avma = av;
+  set_avma(av);
 }
 
 /* lz = 3 */
@@ -442,7 +443,7 @@ mulir(GEN x, GEN y)
     }
     else /* dubious: complete x with 0s and call mulrr */
       mulrrz_i(z, itor(x,lz), y, lz, 0, sx);
-    avma = av; return z;
+    set_avma(av); return z;
   }
 }
 
@@ -458,7 +459,7 @@ addmulii_gen(GEN x, GEN y, GEN z, long lz)
   if (ly == 2) return icopy(x); /* y = 0, wasteful copy */
   av = avma; (void)new_chunk(lx+ly+lz); /*HACK*/
   t = mulii(z, y);
-  avma = av; return addii(t,x);
+  set_avma(av); return addii(t,x);
 }
 /* x + y*z, lgefint(z) == 3 */
 static GEN
@@ -483,7 +484,7 @@ addmulii_lg3(GEN x, GEN y, GEN z)
   t = muluispec(w, y+2, ly-2);
   if (signe(y) < 0) s = -s;
   setsigne(t, s);
-  avma = av; return addii(x,t);
+  set_avma(av); return addii(x,t);
 }
 /* x + y*z */
 GEN
@@ -566,7 +567,7 @@ divir(GEN x, GEN y)
   }
   z = cgetr(ly); av = avma;
   affrr(divrr(itor(x, ly+1), y), z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 
 GEN
@@ -585,7 +586,7 @@ divur(ulong x, GEN y)
   }
   z = cgetr(ly); av = avma;
   affrr(divrr(utor(x,ly+1), y), z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 
 GEN
@@ -605,7 +606,7 @@ divsr(long x, GEN y)
   }
   z = cgetr(ly); av = avma;
   affrr(divrr(stor(x,ly+1), y), z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 
 /* returns 1/y, assume y != 0 */
@@ -616,7 +617,7 @@ invr_basecase(GEN y)
   GEN z = cgetr(ly);
   pari_sp av = avma;
   affrr(divrr(real_1(ly+1), y), z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 /* returns 1/b, Newton iteration */
 GEN
@@ -645,10 +646,10 @@ invr(GEN b)
     /* TODO: mulrr(a,x) should be a half product (the higher half is known).
      * mulrr(x, ) already is */
     affrr(addrr(x, mulrr(x, subsr(1, mulrr(a,x)))), x);
-    avma = (pari_sp)a;
+    set_avma((pari_sp)a);
   }
   x[1] = (b[1] & SIGNBITS) | evalexpo(expo(x)-expo(b));
-  avma = (pari_sp)x; return x;
+  set_avma((pari_sp)x); return x;
 }
 
 GEN
@@ -662,7 +663,7 @@ modii(GEN x, GEN y)
     {
       pari_sp av = avma;
       (void)new_chunk(lgefint(y));
-      x = remii(x,y); avma=av;
+      x = remii(x,y); set_avma(av);
       if (x==gen_0) return x;
       return subiispec(y+2,x+2,lgefint(y)-2,lgefint(x)-2);
     }
@@ -673,7 +674,7 @@ void
 modiiz(GEN x, GEN y, GEN z)
 {
   const pari_sp av = avma;
-  affii(modii(x,y),z); avma=av;
+  affii(modii(x,y),z); set_avma(av);
 }
 
 GEN
