@@ -439,24 +439,21 @@ INLINE void
 restore_vars(long nbmvar, long nblvar, long nblock)
 {
   long j;
-  for(j=1;j<=nbmvar;j++)
-    freelex();
-  for(j=1;j<=nblvar;j++)
-    { s_lvars.n--; pop_val(lvars[s_lvars.n]); }
-  for(j=1;j<=nblock;j++)
-    { s_locks.n--; gunclone(locks[s_locks.n]); }
+  for(j=1; j<=nbmvar; j++) freelex();
+  for(j=1; j<=nblvar; j++) { s_lvars.n--; pop_val(lvars[s_lvars.n]); }
+  for(j=1; j<=nblock; j++) { s_locks.n--; gunclone(locks[s_locks.n]); }
 }
 
 INLINE void
 restore_trace(long nbtrace)
 {
   long j;
-  for(j=1;j<=nbtrace;j++)
+  for(j=1; j<=nbtrace; j++)
   {
     GEN C = trace[s_trace.n-j].closure;
     clone_unlock(C);
   }
-  s_trace.n-=nbtrace;
+  s_trace.n -= nbtrace;
 }
 
 INLINE long
@@ -503,7 +500,7 @@ pop_lex(long n)
   long j;
   for(j=1; j<=n; j++)
     freelex();
-  s_trace.n--;
+  restore_trace(1);
 }
 
 static THREAD pari_stack s_relocs;
@@ -1345,42 +1342,43 @@ closure_eval(GEN C)
         break;
       }
 
-#define EVAL_f(f) \
+#define EVAL_f(f, type, resEQ) \
   switch (ep->arity) \
   { \
-    case 0: f(); break; \
-    case 1: sp--; f(st[sp]); break; \
-    case 2: sp-=2; f(st[sp],st[sp+1]); break; \
-    case 3: sp-=3; f(st[sp],st[sp+1],st[sp+2]); break; \
-    case 4: sp-=4; f(st[sp],st[sp+1],st[sp+2],st[sp+3]); break; \
-    case 5: sp-=5; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4]); break; \
-    case 6: sp-=6; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5]); break; \
-    case 7: sp-=7; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6]); break; \
-    case 8: sp-=8; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7]); break; \
-    case 9: sp-=9; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8]); break; \
-    case 10: sp-=10; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9]); break; \
-    case 11: sp-=11; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10]); break; \
-    case 12: sp-=12; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11]); break; \
-    case 13: sp-=13; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12]); break; \
-    case 14: sp-=14; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13]); break; \
-    case 15: sp-=15; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14]); break; \
-    case 16: sp-=16; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15]); break; \
-    case 17: sp-=17; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15],st[sp+16]); break; \
-    case 18: sp-=18; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15],st[sp+16],st[sp+17]); break; \
-    case 19: sp-=19; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15],st[sp+16],st[sp+17],st[sp+18]); break; \
-    case 20: sp-=20; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15],st[sp+16],st[sp+17],st[sp+18],st[sp+19]); break; \
+    case 0: resEQ ((type (*)(void))f)(); break; \
+    case 1: sp--;  resEQ ((type (*)(long))f)(st[sp]); break; \
+    case 2: sp-=2; resEQ((type (*)(long,long))f)(st[sp],st[sp+1]); break; \
+    case 3: sp-=3; resEQ((type (*)(long,long,long))f)(st[sp],st[sp+1],st[sp+2]); break; \
+    case 4: sp-=4; resEQ((type (*)(long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3]); break; \
+    case 5: sp-=5; resEQ((type (*)(long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4]); break; \
+    case 6: sp-=6; resEQ((type (*)(long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5]); break; \
+    case 7: sp-=7; resEQ((type (*)(long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6]); break; \
+    case 8: sp-=8; resEQ((type (*)(long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7]); break; \
+    case 9: sp-=9; resEQ((type (*)(long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8]); break; \
+    case 10: sp-=10; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9]); break; \
+    case 11: sp-=11; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10]); break; \
+    case 12: sp-=12; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11]); break; \
+    case 13: sp-=13; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12]); break; \
+    case 14: sp-=14; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13]); break; \
+    case 15: sp-=15; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14]); break; \
+    case 16: sp-=16; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15]); break; \
+    case 17: sp-=17; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15],st[sp+16]); break; \
+    case 18: sp-=18; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15],st[sp+16],st[sp+17]); break; \
+    case 19: sp-=19; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15],st[sp+16],st[sp+17],st[sp+18]); break; \
+    case 20: sp-=20; resEQ((type (*)(long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long,long))f)(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15],st[sp+16],st[sp+17],st[sp+18],st[sp+19]); break; \
     default: \
       pari_err_IMPL("functions with more than 20 parameters");\
       goto endeval; /*LCOV_EXCL_LINE*/ \
   }
 
+
     case OCcallgen:
       {
         entree *ep = (entree *)operand;
         GEN res;
         /* Macro Madness : evaluate function ep->value on arguments
          * st[sp-ep->arity .. sp]. Set res = result. */
-        EVAL_f(res = ((GEN (*)(ANYARG))ep->value));
+        EVAL_f(ep->value, GEN, res=);
         if (br_status) goto endeval;
         gel(st,sp++)=res;
         break;
@@ -1399,7 +1397,7 @@ closure_eval(GEN C)
       {
         entree *ep = (entree *)operand;
         long res;
-        EVAL_f(res = ((long (*)(ANYARG))ep->value));
+        EVAL_f(ep->value, long, res=);
         if (br_status) goto endeval;
         st[sp++] = res;
         break;
@@ -1408,7 +1406,7 @@ closure_eval(GEN C)
       {
         entree *ep = (entree *)operand;
         long res;
-        EVAL_f(res = ((int (*)(ANYARG))ep->value));
+        EVAL_f(ep->value, int, res=);
         if (br_status) goto endeval;
         st[sp++] = res;
         break;
@@ -1416,7 +1414,7 @@ closure_eval(GEN C)
     case OCcallvoid:
       {
         entree *ep = (entree *)operand;
-        EVAL_f(((void (*)(ANYARG))ep->value));
+        EVAL_f(ep->value, void,(void));
         if (br_status) goto endeval;
         break;
       }
@@ -1664,11 +1662,8 @@ void
 evalstate_reset(void)
 {
   mtstate_reset();
-  sp = 0;
-  rp = 0;
-  dbg_level = 0;
   restore_vars(s_var.n, s_lvars.n, s_locks.n);
-  s_trace.n = 0;
+  sp = rp = dbg_level = s_trace.n = 0;
   reset_break();
   compilestate_reset();
   parsestate_reset();
@@ -1814,7 +1809,7 @@ parsum(GEN a, GEN b, GEN code)
   worker = snm_closure(is_entry("_parapply_slice_worker"), mkvec(code));
   b = gfloor(b);
   N = addiu(subii(b, a), 1);
-  m = itou(sqrti(N)); if (cmpiu(N, m) < 0) m = itou(N);
+  m = itou(sqrti(N));
   mt_queue_start_lim(&pt, worker, m);
   L = cgetg(m + 2, t_VEC); v = mkvec(L);
   s = gen_0; a = setloop(a); pending = 0; av2 = avma;
@@ -2175,6 +2170,7 @@ parforeach(GEN x, GEN code, void *E, long call(void*, GEN, GEN))
   }
   set_avma(av2);
   mt_queue_end(&pt);
+  clone_unlock_deep(x);
   br_status = status;
   set_avma(av);
 }
