@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -79,9 +80,17 @@ hash_create(ulong minsize, ulong (*hash)(void*), int (*eq)(void*,void*),
   h->eq   = eq;
   setlen(h, len); return h;
 }
+static ulong
+hash_id(void *x) { return (ulong)x; }
+static int
+eq_id(void *x, void *y) { return x == y; }
+hashtable *
+hash_create_ulong(ulong s, long stack)
+{ return hash_create(s, &hash_id, &eq_id, stack); }
 
 void
-hash_init_GEN(hashtable *h, ulong minsize, int (*eq)(GEN,GEN), int use_stack)
+hash_init(hashtable *h, ulong minsize, ulong (*hash)(void*),
+                                       int (*eq)(void*,void*), int use_stack)
 {
   int i = get_prime_index(minsize);
   ulong len = hashprimes[i];
@@ -92,11 +101,21 @@ hash_init_GEN(hashtable *h, ulong minsize, int (*eq)(GEN,GEN), int use_stack)
   h->use_stack = use_stack;
   h->pindex = i;
   h->nb = 0;
-  h->hash = (ulong (*)(void*)) hash_GEN;
-  h->eq   = (int (*)(void*,void*)) eq;
+  h->hash = hash;
+  h->eq   = eq;
   setlen(h, len);
 }
 
+void
+hash_init_GEN(hashtable *h, ulong minsize, int (*eq)(GEN,GEN), int use_stack)
+{ hash_init(h, minsize,(ulong (*)(void*)) hash_GEN,
+                       (int (*)(void*,void*)) eq, use_stack);
+}
+
+void
+hash_init_ulong(hashtable *h, ulong minsize, int use_stack)
+{ hash_init(h, minsize,hash_id, eq_id, use_stack); }
+
 void
 hash_insert2(hashtable *h, void *k, void *v, ulong hash)
 {
@@ -209,6 +228,13 @@ hash_haskey_long(hashtable *h, void *k, long *v)
   else return 0;
 }
 
+GEN
+hash_haskey_GEN(hashtable *h, void *k)
+{
+  hashentry * e = hash_search(h, k);
+  return e ? (GEN) e->val: NULL;
+}
+
 hashentry *
 hash_remove_select(hashtable *h, void *k, void *E,
   int (*select)(void*,hashentry*))
@@ -255,13 +281,6 @@ hash_destroy(hashtable *h)
   }
   pari_free(h->table); pari_free(h);
 }
-static ulong
-hash_id(void *x) { return (ulong)x; }
-static int
-eq_id(void *x, void *y) { return x == y; }
-hashtable *
-hash_create_ulong(ulong s, long stack)
-{ return hash_create(s, &hash_id, &eq_id, stack); }
 
 static
 int strequal(void *a, void *b) { return !strcmp((char*)a,(char*)b); }
@@ -331,22 +350,13 @@ hash_GEN(GEN x)
       return h;
   }
 }
-
-/* djb's hash */
 ulong
-hash_str(const char *str)
+hash_zv(GEN x)
 {
-  ulong hash = 5381, c;
-  while ( (c = (ulong)*str++) )
-    hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
-  return hash;
-}
-
-/* hashvalue's underlying hash function */
-ulong
-hash_str2(const char *s)
-{
-  ulong n = 0, c;
-  while ( (c = (ulong)*s++) ) n = (n<<1) ^ c;
-  return n;
+  long i, lx = lg(x);
+  ulong h;
+  if (lx == 1) return 0;
+  h = x[1];
+  for (i = 1; i < lx; i++) h = glue(h, uel(x,i));
+  return h;
 }
