@@ -15,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_intnum
+
 static GEN
 intlin(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec);
 
@@ -129,7 +131,7 @@ qrom2(void *E, GEN (*eval)(void *, GEN), GEN a, GEN b, long bit)
       sum = gadd(sum, eval(E, x)); x = addrr(x,del);
       if ((j1 & 0x1ff) == 0) gerepileall(av2, 2, &sum,&x);
     }
-    sum = gmul(sum,del); p1 = gdivgs(gel(s,j-1),3);
+    sum = gmul(sum,del); p1 = gdivgu(gel(s,j-1),3);
     gel(s,j) = gerepileupto(av, gadd(p1,sum));
     if (j >= KLOC && (ss = interp(h, s, j, bit-(3*j/2)+3, KLOC)))
       return gmulsg(sig, ss);
@@ -1265,7 +1267,7 @@ intnum(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
     long v = gvar(gel(a,1));
     if (v != NO_VARIABLE) {
       na = cgetg(3,t_VEC);
-      gel(na,1) = polcoef(gel(a,1),0,v);
+      gel(na,1) = polcoef_i(gel(a,1),0,v);
       gel(na,2) = gel(a,2);
     }
     a = gel(a,1);
@@ -1274,7 +1276,7 @@ intnum(void *E, GEN (*eval)(void*, GEN), GEN a, GEN b, GEN tab, long prec)
     long v = gvar(gel(b,1));
     if (v != NO_VARIABLE) {
       nb = cgetg(3,t_VEC);
-      gel(nb,1) = polcoef(gel(b,1),0,v);
+      gel(nb,1) = polcoef_i(gel(b,1),0,v);
       gel(nb,2) = gel(b,2);
     }
     b = gel(b,1);
@@ -1462,14 +1464,13 @@ contfrac_Euler(GEN C)
 static GEN
 QD(GEN M, long lim)
 {
-  pari_sp av;
+  pari_sp av = avma;
+  long lim2 = lim / 2, j, k;
   GEN e, q, c;
-  long lim2, j, k;
   e = zerovec(lim);
   c = zerovec(lim+1); gel(c, 1) = gel(M, 1);
   q = cgetg(lim+1, t_VEC);
   for (k = 1; k <= lim; ++k) gel(q, k) = gdiv(gel(M, k+1), gel(M, k));
-  lim2 = lim/2; av = avma;
   for (j = 1; j <= lim2; ++j)
   {
     long l = lim - 2*j;
@@ -1640,16 +1641,10 @@ Pade(GEN M, GEN *pP, GEN *pQ)
 static GEN
 veczetaprime(GEN a, GEN b, long N, long prec)
 {
-  long newprec, fpr = prec2nbits(prec), pr = (long)ceil(fpr * 1.5);
-  long l = nbits2prec(pr), e = fpr / 2;
-  GEN eps, A, B;
-  newprec = nbits2prec(pr + BITS_IN_LONG);
-  a = gprec_wensure(a, newprec);
-  b = gprec_wensure(b, newprec);
-  eps = real2n(-e, l);
-  A = veczeta(a, gsub(b, eps), N, newprec);
-  B = veczeta(a, gadd(b, eps), N, newprec);
-  return gmul2n(RgV_sub(B, A), e-1);
+  long B = prec2nbits(prec) / 2;
+  GEN v, h = mkcomplex(gen_0, real2n(-B, prec));
+  v = veczeta(a, gadd(b, h), N, prec);
+  return gmul2n(imag_i(v), B);
 }
 
 struct mon_w {
@@ -1677,7 +1672,7 @@ wrapmonw2(void* E, GEN x)
 {
   struct mon_w *W = (struct mon_w*)E;
   GEN wnx = closure_callgen1prec(W->w, x, W->prec);
-  return gdiv(wnx, gpow(x, gadd(gmulgs(W->a, W->j), W->b), W->prec));
+  return gdiv(wnx, gpow(x, gadd(gmulgu(W->a, W->j), W->b), W->prec));
 }
 static GEN
 M_from_wrapmon(struct mon_w *S, GEN wfast, GEN n0)
@@ -1865,7 +1860,7 @@ sumnuminit(GEN fast, long prec)
 
   gel(res, 1) = d = mkfrac(gen_1, utoipos(4)); /* 1/4 */
   av = avma;
-  w = gtodouble(glambertW(ginv(d), LOWDEFAULTPREC));
+  w = gtodouble(glambertW(ginv(d), 0, LOWDEFAULTPREC));
   N = (long)ceil(M_LN2*bitprec/(w*(1+w))+5);
   k = (long)ceil(N*w); if (k&1) k--;
 
@@ -1987,7 +1982,7 @@ intnumgauexpinit(long prec)
   Pade(V, &P, &Q);
   N = RgX_recip(gsub(P, Q));
   E = RgX_recip(Q);
-  R = gdivgs(gdiv(N, RgX_deriv(E)), 2);
+  R = gdivgu(gdiv(N, RgX_deriv(E)), 2);
   vabs = RX_realroots(E,prec2);
   l = lg(vabs); settyp(vabs, t_VEC);
   vwt = cgetg(l, t_VEC);
@@ -2130,10 +2125,10 @@ intnumainfrat(GEN F, long N, double r, long prec)
   ser = gmul(F, real_1(prec + EXTRAPREC64));
   ser = rfracrecip_to_ser_absolute(ser, lim+2);
   v = valp(ser);
-  S = gdivgs(sercoeff(ser,lim+1), lim*N);
+  S = gdivgu(sercoeff(ser,lim+1), lim*N);
   /* goes down to 2, but coeffs are 0 in degree < v */
   for (k = lim; k >= v; k--) /* S <- (S + coeff(ser,k)/(k-1)) / N */
-    S = gdivgs(gadd(S, gdivgs(sercoeff(ser,k), k-1)), N);
+    S = gdivgu(gadd(S, gdivgu(sercoeff(ser,k), k-1)), N);
   if (v-2) S = gdiv(S, powuu(N, v-2));
   return gerepilecopy(av, gprec_wtrunc(S, prec));
 }
@@ -2142,7 +2137,7 @@ static GEN
 rfrac_eval0(GEN R)
 {
   GEN N, n, D = gel(R,2), d = constant_coeff(D);
-  if (gcmp0(d)) return NULL;
+  if (gequal0(d)) return NULL;
   N = gel(R,1);
   n = typ(N)==t_POL? constant_coeff(N): N;
   return gdiv(n, d);
@@ -2151,7 +2146,7 @@ static GEN
 rfrac_eval(GEN R, GEN a)
 {
   GEN D = gel(R,2), d = poleval(D,a);
-  return gcmp0(d)? NULL: gdiv(poleval(gel(R,1),a), d);
+  return gequal0(d)? NULL: gdiv(poleval(gel(R,1),a), d);
 }
 /* R = \sum_i vR[i], eval at a omitting poles */
 static GEN
@@ -2211,10 +2206,10 @@ sumnumrat_i(GEN F, GEN F0, GEN vF, long prec)
   S1 = add_sumrfrac(S1, F, vF, N-1);
   if (F0) S1 = gadd(S1, F0);
   S = gmul(_1, gsubst(F, vx, gaddgs(pol_x(vx), N)));
-  S = rfrac_to_ser(S, k + 2);
+  S = rfrac_to_ser_i(S, k + 2);
   S2 = gen_0;
   for (j = 2; j <= k; j += 2)
-    S2 = gadd(S2, gmul(gdivgs(bernfrac(j),j), sercoeff(S, j-1)));
+    S2 = gadd(S2, gmul(gdivgu(bernfrac(j),j), sercoeff(S, j-1)));
   return gadd(intf, gsub(S1, S2));
 }
 /* sum_{n >= a} F(n) */
@@ -2293,12 +2288,12 @@ prodnumrat(GEN F, long a, long prec)
   intf = intnumainfrat(gmul(pol_x(vx),G), N, r, prec);
   intf = gneg(gadd(intf, gmulsg(N, glog(gsubst(F, vx, stoi(N)), prec))));
   S = gmul(real_1(prec), gsubst(G, vx, gaddgs(pol_x(vx), N)));
-  S = rfrac_to_ser(S, k + 2);
+  S = rfrac_to_ser_i(S, k + 2);
   S1 = gsqrt(gsubst(F, vx, utoipos(N)), prec);
   for (m = 0; m < N; m++) S1 = gmul(S1, gsubst(F, vx, utoi(m)));
   S2 = gen_0;
   for (j = 2; j <= k; j += 2)
-    S2 = gadd(S2, gmul(gdivgs(bernfrac(j),j*(j-1)), sercoeff(S, j-2)));
+    S2 = gadd(S2, gmul(gdivgu(bernfrac(j),j*(j-1)), sercoeff(S, j-2)));
   return gerepileupto(ltop, gmul(S1, gexp(gsub(intf, S2), prec)));
 }
 
@@ -2397,8 +2392,9 @@ sumeulerrat(GEN F, GEN s, long a, long prec)
   }
   /* F t_RFRAC */
   if (a < 2) a = 2;
-  vF = -poldegree(F, -1);
   rs = gtodouble(real_i(s));
+  vF = -poldegree(F, -1);
+  if (vF <= 0) pari_err(e_MISC, "sum diverges in sumeulerrat");
   r = polmax(gel(F,2));
   N = maxss(30, a); lN = log2((double)N);
   RS = maxdd(1./vF, log2(r) / lN);
@@ -2417,10 +2413,14 @@ sumeulerrat(GEN F, GEN s, long a, long prec)
 static GEN
 rfrac_logderiv(GEN N, GEN D)
 {
-  if (typ(N) != t_POL || varn(N) != varn(D)) return gdiv(gneg(RgX_deriv(D)), D);
-  if (!degpol(D)) return gdiv(RgX_deriv(N), N);
-  return gdiv(RgX_sub(RgX_mul(RgX_deriv(N), D), RgX_mul(RgX_deriv(D), N)),
-              RgX_mul(N, D));
+  GEN a;
+  if (typ(N) != t_POL || varn(N) != varn(D) || !degpol(N))
+    return gdiv(gneg(RgX_deriv(D)), D);
+  if (!degpol(D))
+    return gdiv(RgX_deriv(N), N);
+  a = RgX_sub(RgX_mul(RgX_deriv(N), D), RgX_mul(RgX_deriv(D), N));
+  if (lg(a) > 3) gel(a,2) = gen_0;
+  return gdiv(a, RgX_mul(N, D));
 }
 
 /* prod_{p prime, p >= a} F(p^s), F rational function */
@@ -2454,7 +2454,7 @@ prodeulerrat(GEN F, GEN s, long a, long prec)
   (void)rfracrecip(&NF, &DF); /* returned value is 0 */
   if (!RgX_is_ZX(DF) || !is_pm1(gel(DF,2))
       || lim * log2(r) > 4 * B) NF = gmul(NF, real_1(prec2));
-  ser = integser(rfrac_to_ser(rfrac_logderiv(NF,DF), lim+3));
+  ser = integser(rfrac_to_ser_i(rfrac_logderiv(NF,DF), lim+3));
   /* ser = log f, f = F(1/x) + O(x^(lim+1)) */
   P = primes_interval(gen_2, utoipos(N));
   z = gexp(sumlogzeta(ser, s, P, rs, lN, vF, lim, prec), prec);
@@ -2476,7 +2476,7 @@ sumnumlagrange1init(GEN c1, long flag, long prec)
   c1d = c1 ? gtodouble(c1) : 0.332;
   N = (ulong)ceil(c1d*B); if ((N&1L) == 0) N++;
   prec2 = nbits2prec(B+(long)ceil(1.8444*N) + 16);
-  W = vecbinome(N);
+  W = vecbinomial(N);
   T = vecpowuu(N, N);
   V = cgetg(N+1, t_VEC); gel(V,N) = gel(T,N);
   for (n = N-1; n >= 1; n--)
@@ -2502,7 +2502,7 @@ sumnumlagrange2init(GEN c1, long flag, long prec)
 
   N = (ulong)ceil(c1d*B); if ((N&1L) == 0) N++;
   prec2 = nbits2prec(B+(long)ceil(1.18696*N) + 16);
-  W = vecbinome(2*N);
+  W = vecbinomial(2*N);
   T = vecpowuu(N, 2*N);
   V = cgetg(N+1, t_VEC); gel(V, N) = told = gel(T,N);
   for (n = N-1; n >= 1; n--)
@@ -2675,3 +2675,118 @@ sumnumlagrange(void *E, GEN (*eval)(void*,GEN,long), GEN a, GEN tab, long prec)
 GEN
 sumnumlagrange0(GEN a, GEN code, GEN tab, long prec)
 { EXPR_WRAP(code, sumnumlagrange(EXPR_ARGPREC, a, tab, prec)); }
+
+/********************************************************************/
+/*                          SIDI type programs                      */
+/********************************************************************/
+
+GEN
+sumnumsidi(void *E, GEN (*f)(void*, GEN, long), GEN a, double mu, long prec)
+{
+  pari_sp av;
+  GEN M, N, Wkeep = gen_0, W = gen_0, _1, S, t, Wp;
+  long bit = prec2nbits(prec), newbit = (long)(mu * bit) + 33;
+  long n, s, fail = 0, BIG = LONG_MAX, ekeep = LONG_MAX;
+
+  prec = nbits2prec(newbit); _1 = real_1(prec);
+  av = avma; S = real_0(prec); t = Wp = f(E, a, prec);
+  M = N = cgetg(1, t_VEC);
+  for (n = 1;; n++)
+  {
+    long e = BIG;
+    GEN c;
+    S = gadd(S, t); t = f(E, gaddsg(n, a), prec);
+    if (gequal0(t))
+      c = divru(real2n(bit, prec), n);
+    else
+      c = gdiv(_1, gmulsg(n, t));
+    /* Sidi's W algorithm */
+    M = vec_append(M, gmul(S, c));
+    N = vec_append(N, c); if (n == 1) continue;
+    for (s = n - 1; s >= 1; s--)
+    {
+      GEN d = sstoQ(s * n, n - s);
+      gel(M, s) = gmul(d, gsub(gel(M, s), gel(M, s + 1)));
+      gel(N, s) = gmul(d, gsub(gel(N, s), gel(N, s + 1)));
+    }
+    if (!gequal0(gel(N, 1)))  /* if N[1] = 0, count as failure */
+    {
+      W = gdiv(gel(M, 1), gel(N, 1));
+      e = gexpo(gsub(W, Wp));
+      if (e < -bit) break;
+    }
+    if (++fail >= 10)
+    {
+      if (DEBUGLEVEL)
+        err_printf("sumnumsidi: reached accuracy of %ld bits.", -ekeep);
+      bit = -ekeep; W = Wkeep; break;
+    }
+    if (e < ekeep) { fail = 0; ekeep = e; Wkeep = W; }
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"sumnumsidi");
+      gerepileall(av, 6, &W, &Wkeep, &S, &t, &M, &N);
+    }
+    Wp = W;
+  }
+  return gprec_w(W, nbits2prec(bit));
+}
+
+GEN
+sumnumsidi0(GEN a, GEN code, long safe, long prec)
+{ EXPR_WRAP(code, sumnumsidi(EXPR_ARGPREC, a, safe ? 1.56 : 1., prec)); }
+
+struct _osc_wrap
+{
+  void *E;
+  GEN (*f)(void*, GEN);
+  GEN a, H, tab;
+  long prec;
+};
+
+static GEN
+_int_eval(void *E, GEN (*f)(void*, GEN), GEN a, GEN n, GEN H, GEN T, long prec)
+{
+  GEN u = gmul(n, H);
+  if (a) u = gadd(a, u);
+  return intnumgauss(E, f, u, gadd(u, H), T, prec);
+}
+static GEN
+osc_wrap(void* E, GEN n)
+{
+  struct _osc_wrap *D = (struct _osc_wrap*)E;
+  return _int_eval(D->E, D->f, D->a, n, D->H, D->tab, D->prec);
+}
+static GEN
+osc_wrap_prec(void* E, GEN n, long prec)
+{
+  struct _osc_wrap *D = (struct _osc_wrap*)E;
+  return _int_eval(D->E, D->f, D->a, n, D->H, D->tab, prec);
+}
+
+GEN
+intnumosc(void *E, GEN (*f)(void*, GEN), GEN a, GEN H, long flag, GEN tab,
+          long prec)
+{
+  pari_sp av = avma;
+  struct _osc_wrap D;
+  GEN S;
+  if (flag < 0 || flag > 4) pari_err_FLAG("intnumosc");
+  if (!tab) tab = intnumgaussinit(0, prec + (flag == 0? (prec>>1): 0));
+  if (gequal0(a)) a = NULL;
+  D.E = E; D.f = f; D.a = a; D.H = H; D.tab = tab; D.prec = prec;
+  switch(flag)
+  {
+    case 0: S = sumnumsidi((void*)&D, osc_wrap_prec, gen_0, 1.56, prec); break;
+    case 1: S = sumnumsidi((void*)&D, osc_wrap_prec, gen_0, 1.0, prec); break;
+    case 2: S = sumalt((void*)&D, osc_wrap, gen_0, prec); break;
+    case 3: S = sumnumlagrange((void*)&D, osc_wrap_prec, gen_0, NULL, prec);
+            break;
+    default: S = sumpos((void*)&D, osc_wrap, gen_0, prec); break;
+  }
+  return gerepilecopy(av, S);
+}
+
+GEN
+intnumosc0(GEN a, GEN code, GEN H, long flag, GEN tab, long prec)
+{ EXPR_WRAP(code, intnumosc(EXPR_ARG, a, H, flag, tab, prec)); }
