@@ -14,6 +14,8 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_genus2red
+
 /********************************************************************/
 /**                                                                **/
 /**                       IGUSA INVARIANTS                         **/
@@ -265,8 +267,8 @@ discpart(GEN H, GEN p, long prec)
 static long
 theta_j(GEN B, GEN p, long j)
 {
-  long i, t = 60*myval(RgX_coeff(B,5-j), p);
-  for(i = 2+j; i <= 6; i++)
+  long i, t = VERYBIG;
+  for(i = 1+j; i <= 6; i++)
     t = minss(t, myval(RgX_coeff(B,6-i), p) * (60 / (i-j)));
   return t;
 }
@@ -299,6 +301,15 @@ factmz(GEN Q, GEN p, long *maxord)
   *maxord = m;
   return (m >= 3)? FpX_oneroot(gel(z,m), p): NULL;
 }
+static long
+get_lambda(GEN H, GEN p)
+{
+  if (!dvdii(RgX_coeff(H,3), p)) return 3;
+  if (!dvdii(RgX_coeff(H,4), p)) return 2;
+  if (!dvdii(RgX_coeff(H,5), p)) return 1;
+  if (!dvdii(RgX_coeff(H,6), p)) return 0;
+  return -1;
+}
 
 /* H integral ZX of degree 5 or 6, p > 2. Modify until
  *   y^2 = p^alpha H is minimal over Z_p, alpha = 0,1
@@ -312,65 +323,56 @@ factmz(GEN Q, GEN p, long *maxord)
 static GEN
 polymini(GEN H, GEN p)
 {
-  GEN a0, a1, a2, a3, Hp, rac;
-  long t60, alpha, lambda, maxord, quad = 0, beta = 0;
+  long t60, alpha, lambda, quad = 0, beta = 0;
+  GEN Hp;
 
   alpha = ZX_pvalrem(H, p, &H) & 1;
-  RgX_to_03(H, &a0,&a1,&a2,&a3);
-  if (dvdii(a0,p) && dvdii(a1,p) && dvdii(a2,p) && dvdii(a3,p))
-  {
-    H = RgX_recip6(H);
-    RgX_to_03(H, &a0,&a1,&a2,&a3);
-  }
-  if (!dvdii(a3,p)) lambda = 3;
-  else if (!dvdii(a2,p)) lambda = 2;
-  else if (!dvdii(a1,p)) lambda = 1;
-  else lambda = 0;
-
+  lambda = get_lambda(H, p);
+  if (lambda < 0) { H = RgX_recip6(H); lambda = get_lambda(H, p); }
   for(;;)
-  { /* lambda <= 3, t60 = 60*theta */
-    long e;
-    t60 = theta_j(H,p,lambda); e = t60 / 60;
-    if (e)
-    {
-      GEN pe = powiu(p,e);
-      /* H <- H(p^e X) / p^(e(6-lambda)) */
-      H = ZX_Z_divexact(ZX_unscale_div(H,pe), powiu(pe,5-lambda));
-      alpha = (alpha + lambda*e)&1;
-      beta += e;
-      t60 -= 60*e;
-    }
-    /* 0 <= t < 60 */
-    Hp = FpX_red(H, p);
-    if (t60) break;
+  {
+    for(;;)
+    { /* lambda <= 3, t60 = 60*theta */
+      GEN rac;
+      long e, maxord;
+      t60 = theta_j(H,p,lambda); e = t60 / 60;
+      if (e)
+      {
+        GEN pe = powiu(p,e);
+        /* H <- H(p^e X) / p^(e(6-lambda)) */
+        H = ZX_unscale_divpow(H, pe, 6-lambda);
+        alpha = (alpha + lambda*e)&1;
+        beta += e;
+        t60 -= 60*e;
+      }
+      /* 0 <= t < 60 */
+      Hp = FpX_red(H, p); if (t60) break;
 
-    rac = factmz(Hp,p, &maxord);
-    if (maxord <= 2)
-    {
-      if (degpol(Hp) <= 3) break;
-      goto end;
-    }
-    else
-    { /* maxord >= 3 */
+      rac = factmz(Hp,p, &maxord);
+      if (maxord <= 2)
+      {
+        if (degpol(Hp) <= 3) break;
+        goto end;
+      }
+      /* maxord >= 3 */
       if (!rac) { quad = 1; goto end; }
       if (signe(rac)) H = ZX_translate(H, rac);
-      lambda = 6-maxord;
+      lambda = 6 - maxord;
     }
-  }
-
-  if (lambda <= 2)
-  {
-    if (myval(RgX_coeff(H,2),p) > 1-alpha &&
-        myval(RgX_coeff(H,1),p) > 2-alpha &&
-        myval(RgX_coeff(H,0),p) > 3-alpha)
+    if (lambda <= 2)
     {
-      H = ZX_unscale(H, p);
-      if (alpha) H = ZX_Z_mul(H, p);
-      return polymini(H, p);
+      if (myval(RgX_coeff(H,2),p) > 1-alpha &&
+          myval(RgX_coeff(H,1),p) > 2-alpha &&
+          myval(RgX_coeff(H,0),p) > 3-alpha)
+      {
+        H = ZX_unscale(H, p);
+        if (alpha) H = ZX_Z_mul(H, p);
+        return polymini(H, p);
+      }
+      break;
     }
-  }
-  else if (lambda == 3 && alpha == 1)
-  {
+    /* lambda = 3 */
+    if (alpha == 0) break;
     if (degpol(Hp) == 3)
     {
       if (myval(RgX_coeff(H,6),p) >= 3 &&
@@ -381,28 +383,27 @@ polymini(GEN H, GEN p)
         H = ZX_Z_divexact(H, powiu(p, degpol(H)-3)); /* H(x/p)p^3 */
         t60 += 60; alpha = 0; beta--;
       }
+      break;
     }
-    else if (degpol(Hp) == 6 && t60)
+    if (degpol(Hp) != 6) break;
+    if (t60)
     {
-      rac = factmz(RgX_mulXn(Hp, -3), p, &maxord);
-      if (maxord == 3)
+      long m, maxord;
+      GEN v, T, rac = factmz(RgX_mulXn(Hp, -3), p, &maxord);
+      if (maxord <= 2) break;
+      T = ZX_affine(H, p, rac); /* H(rac + px) */
+      if (ZX_pval(T,p) < 3) break;
+
+      H = ZX_Z_divexact(T, powiu(p,3));
+      alpha = 0; beta++;
+      v = FpX_factor_squarefree(FpX_red(H,p), p);
+      m = lg(v)-1; /* maximal multiplicity */
+      if (m > 1)
       {
-        GEN T = ZX_unscale(ZX_translate(H,rac),p); /* H(rac + px) */
-        if (ZX_pval(T,p)>= 3)
-        {
-          H = RgX_Rg_div(T, powiu(p,3));
-          alpha = 0; beta++;
-          t60 = theta_j(H,p,3);
-          if (!t60)
-          {
-            Hp = FpX_red(H, p);
-            if (!signe(FpX_disc(Hp,p)))
-            {
-              rac = FpX_oneroot(Hp, p);
-              t60 = theta_j(ZX_translate(H,rac),p,3);
-            }
-          }
-        }
+        rac = FpX_oneroot(gel(v,m), p); /* v[m] is linear */
+        H = ZX_translate(H,rac);
+        t60 = theta_j(H,p,3);
+        if (t60 < 60) break;
       }
     }
   }
@@ -661,8 +662,7 @@ get_red(struct red *S, struct igusa_p *Ip, GEN polh, GEN p, long alpha, long r)
       S->g = cyclic(indice);
       return indice ? indice: 1;
     case 6:
-      if (alpha == 0) /* H(px) /p^3 */
-        polh = ZX_Z_divexact(ZX_unscale_div(polh,p), sqri(p));
+      if (alpha == 0) polh = ZX_unscale_divpow(polh, p, 3); /* H(px) /p^3 */
       indice = FpX_is_squarefree(FpX_red(polh,p), p)
                ? 0
                : val[6] - val[7] + val[8]/Ip->eps;
@@ -1652,8 +1652,7 @@ litredtp(long alpha, long alpha1, long t60, long t60_1, GEN polh, GEN polh1,
     if (Ip->tt == 6)
     {
       d = val[6] - val[7] + val[8]/Ip->eps;
-      if (Ip->r1 && alpha1 == 0) /* H(px) / p^3 */
-        polh1 = ZX_Z_divexact(ZX_unscale_div(polh1,p), sqri(p));
+      if (Ip->r1 && alpha1 == 0) polh1 = ZX_unscale_divpow(polh1, p, 3);
       if (FpX_is_squarefree(FpX_red(polh1,p),p))
       { indice = 0; condp = 3-Ip->r2/6; }
       else
@@ -1663,8 +1662,7 @@ litredtp(long alpha, long alpha1, long t60, long t60_1, GEN polh, GEN polh1,
     { /* Ip->tt == 7 */
       long d1;
       d = val[6] - 3*val[3] + val[8]/Ip->eps;
-      if (t60_1 == 60) /* H(px) / p^3 */
-        polh1 = ZX_Z_divexact(ZX_unscale_div(polh1,p), sqri(p));
+      if (t60_1 == 60) polh1 = ZX_unscale_divpow(polh1, p, 3);
       d1 = minss(val[7]-3*val[3],d/2);
       if (d == 2*d1) indice = d1;
       else
@@ -2052,15 +2050,17 @@ genus2localred(struct igusa *I, struct igusa_p *Ip, GEN p, GEN polmini)
   return condp;
 }
 
-static long
-chk_pol(GEN P) {
-  switch(typ(P))
-  {
-    case t_INT: break;
-    case t_POL: RgX_check_ZX(P,"genus2red"); return varn(P); break;
-    default: pari_err_TYPE("genus2red", P);
-  }
-  return -1;
+static GEN
+hyperellintegralmodel(GEN PQ)
+{
+  GEN D;
+  PQ = Q_remove_denom(PQ, &D);
+  if (!D) return PQ;
+  if (typ(PQ)==t_POL) return gmul(PQ,D);
+  if (typ(PQ) == t_VEC && lg(PQ) == 3)
+    return mkvec2(gmul(gel(PQ,1),D), gel(PQ,2));
+  pari_err_TYPE("hyperellintegralmodel",PQ);
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* P,Q are ZX, study Y^2 + Q(X) Y = P(X) */
@@ -2069,32 +2069,13 @@ genus2red(GEN PQ, GEN p)
 {
   pari_sp av = avma;
   struct igusa I;
-  GEN P, Q, D;
+  GEN P, Q;
   GEN j22, j42, j2j6, a0,a1,a2,a3,a4,a5,a6, V,polr,facto,factp, vecmini, cond;
-  long i, l, dd, vP,vQ;
-
-  PQ = Q_remove_denom(PQ, &D);
-  if (typ(PQ) == t_VEC && lg(PQ) == 3)
-  {
-    P = gel(PQ,1);
-    Q = gel(PQ,2);
-  }
-  else
-  {
-    P = PQ;
-    Q = gen_0;
-  }
-
-  vP = chk_pol(P);
-  vQ = chk_pol(Q);
-  if (vP < 0)
-  {
-    if (vQ < 0) pari_err_TYPE("genus2red",mkvec2(P,Q));
-    P = scalarpol(P,vQ);
-  }
-  else if (vQ < 0) Q = scalarpol(Q,vP);
+  long i, l, dd;
+  PQ = hyperellminimalmodel(hyperellintegralmodel(PQ), NULL, p ? mkvec(p): p);
+  P = gel(PQ,1);
+  Q = gel(PQ,2);
   if (p && typ(p) != t_INT) pari_err_TYPE("genus2red", p);
-  if (D) P = ZX_Z_mul(P,D);
 
   polr = ZX_add(ZX_sqr(Q), gmul2n(P,2)); /* ZX */
   switch(degpol(polr))
@@ -2180,5 +2161,5 @@ genus2red(GEN PQ, GEN p)
   cond = factorback(facto);
   /* remove denominator 2 coming from f = -1 in genuslocalred(, p = 2) */
   if (typ(cond) != t_INT) cond = gel(cond,1);
-  return gerepilecopy(av, mkvec4(cond, facto, polr, V));
+  return gerepilecopy(av, mkvec4(cond, facto, PQ, V));
 }
