@@ -12,12 +12,12 @@ Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
-/* This file is based on ratpoints-2.1.3 by Michael Stoll, see
+/* This file is based on ratpoints-2.2.1 by Michael Stoll, see
  * http://www.mathe2.uni-bayreuth.de/stoll/programs/
  * Original copyright / license: */
 /***********************************************************************
- * ratpoints-2.1.3                                                     *
- * Copyright (C) 2008, 2009  Michael Stoll                             *
+ * ratpoints-2.2.1                                                     *
+ * Copyright (C) 2008, 2009, 2022  Michael Stoll                       *
  *  - A program to find rational points on hyperelliptic curves        *
  *                                                                     *
  * This program is free software: you can redistribute it and/or       *
@@ -27,7 +27,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
  ***********************************************************************/
 
 #include "paricfg.h"
-#ifdef HAS_SSE2
+#ifdef HAS_AVX
+#include <immintrin.h>
+#elif defined(HAS_SSE2)
 #include <emmintrin.h>
 #endif
 
@@ -37,9 +39,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #define pel(a,b)  gel((a),(b)+2)
 
 #define RATPOINTS_ARRAY_SIZE 256           /* Array size in longs */
-#define RATPOINTS_DEFAULT_SP1 9            /* Default value for sp1 */
-#define RATPOINTS_DEFAULT_SP2 16           /* Default value for sp2 */
-#define RATPOINTS_DEFAULT_NUM_PRIMES 28    /* Default value for num_primes */
+#define RATPOINTS_DEFAULT_SP1 11           /* Default value for sp1 */
+#define RATPOINTS_DEFAULT_SP2 19           /* Default value for sp2 */
+#define RATPOINTS_DEFAULT_NUM_PRIMES 30    /* Default value for num_primes */
 #define RATPOINTS_DEFAULT_BIT_PRIMES 7     /* Default value for bit_primes */
 #define RATPOINTS_DEFAULT_MAX_FORBIDDEN 30 /* Default value for max_forbidden */
 
@@ -60,37 +62,118 @@ typedef struct {double low; double up;} ratpoints_interval;
 
 #define CEIL(a,b) (((a) <= 0) ? -(-(a) / (b)) : 1 + ((a)-1) / (b))
 
-/* Some Macros for working with SSE registers */
+/* define RBA_USE_VX provisionnaly */
+#define RBA_USE_VX
+#ifdef HAS_AVX512
+ /* Use AVX512 512 bit registers for the bit arrays */
+typedef ulong ratpoints_bit_array __attribute__ ((vector_size (64)));
+
+#define EXT0(a) ((ulong)a[0])
+#define EXT(a,i) ((ulong)a[i])
+#define TEST(a) (  EXT0(a)  || EXT(a,1) || EXT(a,2) ||EXT(a,3)\
+                || EXT(a,4) || EXT(a,5) || EXT(a,6) ||EXT(a,7) )
+
+#define RBA(a) ((ratpoints_bit_array) {((ulong) a), ((ulong) a), ((ulong) a), ((ulong) a)\
+                                     , ((ulong) a), ((ulong) a), ((ulong) a), ((ulong) a) })
+#define RBA_SHIFT (9)
+#define MASKL(a,s) { unsigned long *survl = (unsigned long *)(a); long sh = (s); \
+                     long l, qsh = sh>>TWOPOTBITS_IN_LONG, rsh = sh & (BITS_IN_LONG-1); \
+                     for(l = 0; l < qsh; l++) { *survl++ = 0UL; }; *survl &= (~0UL)<<rsh; }
+#define MASKU(a,s) { unsigned long *survl = (unsigned long *)(a); long sh = (s); \
+                     long l, qsh = RBA_PACK-1 - (sh>>TWOPOTBITS_IN_LONG), rsh = sh & (BITS_IN_LONG-1); \
+                     survl += qsh; *survl++ &= (~0UL)>>rsh; \
+                     for(l = qsh+1; l < RBA_PACK; l++) { *survl++ = 0UL; } }
+
+#elif defined(HAS_AVX)
+ /* Use AVX 256 bit registers for the bit arrays */
+typedef ulong ratpoints_bit_array __attribute__ ((vector_size (32)));
+
+#define EXT0(a) ((ulong)a[0])
+#define EXT(a,i) ((ulong)a[i])
+
+#ifdef __AVX2__
+#define TEST(a) ( _mm256_movemask_epi8(_mm256_cmpeq_epi8((__m256i)(a), (__m256i)RBA(0))) != 0xffffffffL )
+#elif defined(__AVX__)
+#define TEST(a) ( !_mm256_testz_si256((__m256i)(a), (__m256i)(a)) )
+#else
+#define TEST(a) (EXT(a,0) || EXT(a,1) || EXT(a,2) || EXT(a,3))
+#endif
+
+#define RBA(a) ((ratpoints_bit_array){((ulong) a), ((ulong) a), ((ulong) a), ((ulong) a)})
+#define RBA_SHIFT (8)
+#define MASKL(a,s) { unsigned long *survl = (unsigned long *)(a); long sh = (s); \
+                     if(sh >= 2*BITS_IN_LONG) \
+                     { sh -= 2*BITS_IN_LONG; survl[0] = 0UL; survl[1] = 0UL; \
+                       if(sh >= BITS_IN_LONG) \
+                       { survl[2] = 0UL; survl[3] &= (~0UL)<<(sh - BITS_IN_LONG); } \
+                       else { survl[2] &= ~(0UL)<<sh; } } \
+                     else if(sh >= BITS_IN_LONG) { survl[0] = 0UL; survl[1] &= (~0UL)<<(sh - BITS_IN_LONG); } \
+                     else { survl[0] &= ~(0UL)<<sh; } }
+#define MASKU(a,s) { unsigned long *survl = (unsigned long *)(a); long sh = (s); \
+                     if(sh >= 2*BITS_IN_LONG) \
+                     { sh -= 2*BITS_IN_LONG; survl[3] = 0UL; survl[2] = 0UL; \
+                       if(sh >= BITS_IN_LONG) \
+                       { survl[0] &= ~(0UL)>>(sh - BITS_IN_LONG); survl[1] = 0UL; } \
+                       else { survl[1] &= ~(0UL)>>sh; } } \
+                     else if(sh >= BITS_IN_LONG) { survl[2] &= ~(0UL)>>(sh - BITS_IN_LONG); survl[3] = 0UL; } \
+                     else { survl[3] &= ~(0UL)>>sh; } }
+#elif defined(HAS_SSE2) || defined(HAS_NEON)
+
 #ifdef HAS_SSE2
-
-#define AND(a,b) ((a)&(b))
-#define EXT0(a) ((ulong)__builtin_ia32_vec_ext_v2di((__v2di)(a), 0))
-#define EXT1(a) ((ulong)__builtin_ia32_vec_ext_v2di((__v2di)(a), 1))
-#define TEST(a) (EXT0(a) || EXT1(a))
-#define RBA(a) ((__v2di){((long) a), ((long) a)})
-
 /* Use SSE 128 bit registers for the bit arrays */
 typedef __v2di ratpoints_bit_array;
-
-#define RBA_LENGTH (128)
-#define RBA_SHIFT (7)
-#define RBA_ALIGN  (sizeof(ratpoints_bit_array))
-
+#define EXT0(a) ((ulong)__builtin_ia32_vec_ext_v2di((__v2di)(a), 0))
+#define EXT(a,i) ((ulong)__builtin_ia32_vec_ext_v2di((__v2di)(a), 1))
 #else
+typedef ulong ratpoints_bit_array __attribute__ ((vector_size (16)));
+#define EXT0(a) ((ulong)a[0])
+#define EXT(a,i) ((ulong)a[i])
+#endif
+
+#define TEST(a) (EXT0(a) || EXT(a,1))
+#define RBA(a) ((ratpoints_bit_array){((long) a), ((long) a)})
+#define RBA_SHIFT (7)
+#define MASKL(a,s) { unsigned long *survl = (unsigned long *)(a); long sh = (s); \
+                     if(sh >= BITS_IN_LONG) { survl[0] = 0UL; survl[1] &= (~0UL)<<(sh - BITS_IN_LONG); } \
+                     else { survl[0] &= ~(0UL)<<sh; } }
+#define MASKU(a,s) { unsigned long *survl = (unsigned long *)(a); long sh = (s); \
+                     if(sh >= BITS_IN_LONG) { survl[0] &= ~(0UL)>>(sh - BITS_IN_LONG); survl[1] = 0UL; } \
+                     else { survl[1] &= ~(0UL)>>sh; } }
+#else
+
 /* Use ulong for the bit arrays */
 typedef ulong ratpoints_bit_array;
-#define AND(a,b) ((a)&(b))
+
 #define EXT0(a) (a)
 #define TEST(a) (a)
 #define RBA(a) (a)
-
-#define RBA_LENGTH BITS_IN_LONG
 #define RBA_SHIFT TWOPOTBITS_IN_LONG
-#define RBA_ALIGN  (sizeof(long))
-
+#define MASKL(a,s) { *(a) &= ~(0UL)<<(s); }
+#define MASKU(a,s) { *(a) &= ~(0UL)>>(s); }
+#undef RBA_USE_VX
 #endif
 
-typedef struct { long p; long offset; ratpoints_bit_array *ptr; } sieve_spec;
+#define RBA_SIZE  (sizeof(ratpoints_bit_array))
+#define RBA_LENGTH  (RBA_SIZE<<3)
+#define RBA_PACK  (RBA_LENGTH>>TWOPOTBITS_IN_LONG)
+
+#ifdef RBA_USE_VX
+#define RATPOINTS_CHUNK 16
+#define CODE_INIT_SIEVE_COPY \
+{ ulong k; \
+      for (a = 0; a < p; a++) \
+        for(k = 1; k < RBA_PACK; k++) \
+          si[a+k*p] = si[a]; \
+      for(a = 0; (ulong)a < (RATPOINTS_CHUNK-1)*RBA_PACK; a++) \
+         si[a+p*RBA_PACK] = si[a];\
+}
+#else
+#define RATPOINTS_CHUNK 1
+#define CODE_INIT_SIEVE_COPY
+#endif
+
+typedef struct { long p; long offset; ratpoints_bit_array *ptr;
+                 ratpoints_bit_array *start; ratpoints_bit_array *end; } sieve_spec;
 
 typedef enum { num_all, num_even, num_odd, num_none } bit_selection;
 
@@ -125,33 +208,24 @@ typedef struct {
   ulong **sieves0;
 } ratpoints_args;
 
-#ifdef HAS_SSE2
-#define CODE_INIT_SIEVE_COPY for (a = 0; a < p; a++) si[a+p] = si[a];
-#else
-#define CODE_INIT_SIEVE_COPY
-#endif
-
 static ratpoints_bit_array *
-#if defined(__GNUC__)
-__attribute__((noinline))
-#endif
 sieve_init1(long p, ratpoints_sieve_entry *se1, long b1, ratpoints_args *args1)
 {
   ratpoints_sieve_entry *se = se1;
   ratpoints_args *args = args1;
-  register int *isfs = se->is_f_square;
-  register long b = b1;
+  int *isfs = se->is_f_square;
+  long b = b1;
   long lmp = BITS_IN_LONG % p;
   long ldp = BITS_IN_LONG / p;
   long p1 = (ldp + 1) * p;
   long diff_shift = p1 & LONG_MASK;
   long diff = BITS_IN_LONG - diff_shift;
-  register ulong help0;
-  register long a;
-  register long d = se->inverses[b];
-  register long ab = 0; /* a/b mod p */
-  register ulong test = 1UL;
-  register ulong he0 = 0UL;
+  ulong help0;
+  long a;
+  long d = se->inverses[b];
+  long ab = 0; /* a/b mod p */
+  ulong test = 1UL;
+  ulong he0 = 0UL;
   for (a = 0; a < p; a++)
   {
     if (isfs[ab]) he0 |= test;
@@ -161,10 +235,10 @@ sieve_init1(long p, ratpoints_sieve_entry *se1, long b1, ratpoints_args *args1)
   }
   help0 = he0;
   {
-    register ulong help1;
+    ulong help1;
      /* repeat bit pattern floor(BITS_IN_LONG/p) times */
-    register ulong pattern = help0;
-    register long i;
+    ulong pattern = help0;
+    long i;
     /* the p * (floor(BITS_IN_LONG/p) + 1) - BITS_IN_LONG
             = p - (BITS_IN_LONG mod p)
        upper bits into help[b][1] :
@@ -175,17 +249,17 @@ sieve_init1(long p, ratpoints_sieve_entry *se1, long b1, ratpoints_args *args1)
     { /* fill the bit pattern from help0/help1 into sieve[b][].
           sieve[b][a0] has the same semantics as help0/help1,
           but here, a0 runs from 0 to p-1 and all bits are filled. */
-      register long a;
+      long a;
       ulong *si = (ulong *)args->ba_next;
 
-      args->ba_next += p;
+      args->ba_next += p + RATPOINTS_CHUNK-1;
       /* copy the first chunk into sieve[b][] */
       si[0] = help0;
       /* now keep repeating the bit pattern,
          rotating it in help0/help1 */
       for (a = 1 ; a < p; a++)
       {
-        register ulong temp = help0 >> diff;
+        ulong temp = help0 >> diff;
         help0 = help1 | (help0 << diff_shift);
         si[a] = help0;
         help1 = temp;
@@ -200,15 +274,12 @@ sieve_init1(long p, ratpoints_sieve_entry *se1, long b1, ratpoints_args *args1)
 
 /* This is for p > BITS_IN_LONG */
 static ratpoints_bit_array *
-#if defined(__GNUC__)
-__attribute__((noinline))
-#endif
 sieve_init2(long p, ratpoints_sieve_entry *se1, long b1, ratpoints_args *args1)
 {
   ratpoints_sieve_entry *se = se1;
   ratpoints_args *args = args1;
-  register int *isfs = se->is_f_square;
-  register long b = b1;
+  int *isfs = se->is_f_square;
+  long b = b1;
   /* long ldp = 0;  = BITS_IN_LONG / p */
   /* long p1 = p; = (ldp + 1) * p; */
   long wp = p >> TWOPOTBITS_IN_LONG;
@@ -218,16 +289,16 @@ sieve_init2(long p, ratpoints_sieve_entry *se1, long b1, ratpoints_args *args1)
 
   /* initialize help */
   {
-    register ulong *he = &help[0];
-    register ulong *he1 = &he[(p>>TWOPOTBITS_IN_LONG) + 2];
+    ulong *he = &help[0];
+    ulong *he1 = &he[(p>>TWOPOTBITS_IN_LONG) + 2];
     while (he1 != he) { he1--; *he1 = 0UL; }
   }
-  { register ulong work = 0UL;
-    register long a;
-    register long ab = 0; /* a/b mod p */
-    register long d = se->inverses[b];
-    register long n = 0;
-    register ulong test = 1UL;
+  { ulong work = 0UL;
+    long a;
+    long ab = 0; /* a/b mod p */
+    long d = se->inverses[b];
+    long n = 0;
+    ulong test = 1UL;
     for (a = 0; a < p; )
     {
       if (isfs[ab]) work |= test;
@@ -244,17 +315,17 @@ sieve_init2(long p, ratpoints_sieve_entry *se1, long b1, ratpoints_args *args1)
   { /* fill the bit pattern from help[] into sieve[b][].
        sieve[b][a0] has the same semantics as help[b][a0],
        but here, a0 runs from 0 to p-1 and all bits are filled. */
-    register ulong *si = (ulong *)args->ba_next;
-    register long a1;
-    register long a;
+    ulong *si = (ulong *)args->ba_next;
+    long a1;
+    long a;
 
-    args->ba_next += p;
+    args->ba_next += p + RATPOINTS_CHUNK-1;
     /* copy the first chunk from help[] into sieve[num][b][] */
     for (a = 0; a < wp; a++) si[a] = help[a];
     /* now keep repeating the bit pattern, rotating it in help */
     for (a1 = a ; a < p; a++)
     {
-      register long t = (a1 == wp) ? 0 : a1+1;
+      long t = (a1 == wp) ? 0 : a1+1;
       help[a1] |= help[t]<<diff_shift;
       si[a] = help[a1];
       a1 = t;
@@ -318,18 +389,18 @@ gen_sieves0(GEN listprime)
 {
   long n;
   long nbprime = lg(listprime)-1;
-  ulong ** si = (ulong**) new_chunk(nbprime+1);
+  ulong ** w = (ulong**) new_chunk(nbprime+1);
   for (n = 1; n <= nbprime; n++)
   {
-    ulong i, p = uel(listprime,n);
-    ulong *w = (ulong *) stack_malloc_align(2*p*sizeof(ulong), RBA_ALIGN);
-    for (i = 0; i < p; i++) uel(w,i) = ~0UL;
-    for (i = 0; i < BITS_IN_LONG; i++)
-      uel(w,(p*i)>>TWOPOTBITS_IN_LONG) &= ~(1UL<<((p*i) & LONG_MASK));
-    for (i = 0; i < p; i++) uel(w,i+p) = uel(w,i);
-    si[n] = w;
+    ulong a, p = uel(listprime,n);
+    ulong *si = (ulong *) stack_malloc_align((p+RATPOINTS_CHUNK-1)*RBA_SIZE, RBA_SIZE);
+    for (a = 0; a < p; a++) si[a] = ~0UL;
+    for (a = 0; a < BITS_IN_LONG; a++)
+      uel(si,(p*a)>>TWOPOTBITS_IN_LONG) &= ~(1UL<<((p*a) & LONG_MASK));
+    CODE_INIT_SIEVE_COPY
+    w[n] = si;
   }
-  return si;
+  return w;
 }
 
 static void
@@ -346,9 +417,9 @@ ZX_positive_region(GEN P, long h, long bitprec)
 {
   long prec = nbits2prec(bitprec);
   GEN it = mkvec2(stoi(-h),stoi(h));
-  GEN R = ZX_Uspensky(P, it, 1, bitprec);
+  GEN R = realroots(P, it, prec);
   long nR = lg(R)-1;
-  long s = signe(poleval(P,gel(it,1)));
+  long s = signe(ZX_Z_eval(P,gel(it,1)));
   long i=1, j;
   GEN iv, st, en;
   if (s<0 && nR==0) return NULL;
@@ -484,55 +555,134 @@ _ratpoints_sift0(long b, long w_low, long w_high,
            ratpoints_bit_array *survivors, sieve_spec *sieves, int *quit,
            int process(long, long, GEN, void*, int*), void *info)
 {
-  long range = w_high - w_low;
-  long sp1 = args->sp1;
-  long sp2 = args->sp2;
-  long i, n, nb = 0, absb = labs(b);
+  long sp1 = args->sp1, sp2 = args->sp2;
+  long i, n, nb = 0, absb = labs(b), base = 0;
   ratpoints_bit_array *surv0;
 
   /* now do the sieving (fast!) */
+#if (RATPOINTS_CHUNK == 16)
+  long w_low_new;
+  ratpoints_bit_array *surv = survivors;
+
+  /* first set the start fields for the first and second phases of sieving */
+  for(n = 0; n < sp2; n++)
+    sieves[n].start = sieves[n].ptr + mod(w_low + sieves[n].offset, sieves[n].p);
+  /* Take RATPOINTS_CHUNK bit-arrays and apply phase 1 to them,
+   * then repeat with the next RATPOINTS_CHUNK bit-arrays. */
+  for(w_low_new = w_low; w_low_new < w_high; surv += RATPOINTS_CHUNK, w_low_new += RATPOINTS_CHUNK)
+  {
+    /* read data from memory into registers */
+    ratpoints_bit_array reg0 = surv[0];
+    ratpoints_bit_array reg1 = surv[1];
+    ratpoints_bit_array reg2 = surv[2];
+    ratpoints_bit_array reg3 = surv[3];
+    ratpoints_bit_array reg4 = surv[4];
+    ratpoints_bit_array reg5 = surv[5];
+    ratpoints_bit_array reg6 = surv[6];
+    ratpoints_bit_array reg7 = surv[7];
+    ratpoints_bit_array reg8 = surv[8];
+    ratpoints_bit_array reg9 = surv[9];
+    ratpoints_bit_array reg10 = surv[10];
+    ratpoints_bit_array reg11 = surv[11];
+    ratpoints_bit_array reg12 = surv[12];
+    ratpoints_bit_array reg13 = surv[13];
+    ratpoints_bit_array reg14 = surv[14];
+    ratpoints_bit_array reg15 = surv[15];
+
+    for(n = 0; n < sp1; n++)
+    { /* retrieve the pointer to the beginning of the relevant bits */
+      ratpoints_bit_array *siv1 = sieves[n].start;
+      reg0 &= *siv1++;
+      reg1 &= *siv1++;
+      reg2 &= *siv1++;
+      reg3 &= *siv1++;
+      reg4 &= *siv1++;
+      reg5 &= *siv1++;
+      reg6 &= *siv1++;
+      reg7 &= *siv1++;
+      reg8 &= *siv1++;
+      reg9 &= *siv1++;
+      reg10 &= *siv1++;
+      reg11 &= *siv1++;
+      reg12 &= *siv1++;
+      reg13 &= *siv1++;
+      reg14 &= *siv1++;
+      reg15 &= *siv1++;
+
+      /* update the pointer for the next round
+       * (RATPOINTS_CHUNK-1 bit-arrays after sieves[n].end) */
+      while(siv1 >= sieves[n].end) siv1 -= sieves[n].p;
+      sieves[n].start = siv1;
+    }
+    /* store the contents of the registers back into memory */
+    surv[0] = reg0;
+    surv[1] = reg1;
+    surv[2] = reg2;
+    surv[3] = reg3;
+    surv[4] = reg4;
+    surv[5] = reg5;
+    surv[6] = reg6;
+    surv[7] = reg7;
+    surv[8] = reg8;
+    surv[9] = reg9;
+    surv[10] = reg10;
+    surv[11] = reg11;
+    surv[12] = reg12;
+    surv[13] = reg13;
+    surv[14] = reg14;
+    surv[15] = reg15;
+  }
+#else /* RATPOINTS_CHUNK not between 2 and 16 */
+  long range = w_high - w_low;
   for (n = 0; n < sp1; n++)
   {
     ratpoints_bit_array *sieve_n = sieves[n].ptr;
-    register long p = sieves[n].p;
+    long p = sieves[n].p;
     long r = mod(-w_low-sieves[n].offset, p);
-    register ratpoints_bit_array *surv = survivors;
+    ratpoints_bit_array *surv = survivors;
 
     if (w_high < w_low + r)
     { /* if we get here, r > 0, since w_high >= w_low always */
-      register ratpoints_bit_array *siv1 = &sieve_n[p-r];
-      register ratpoints_bit_array *siv0 = siv1 + range;
+      ratpoints_bit_array *siv1 = &sieve_n[p-r];
+      ratpoints_bit_array *siv0 = siv1 + range;
 
-      while (siv1 != siv0) { *surv = AND(*surv, *siv1++); surv++; }
+      while (siv1 != siv0) { *surv &= *siv1++; surv++; }
     }
     else
     {
-      register ratpoints_bit_array *siv1 = &sieve_n[p-r];
-      register ratpoints_bit_array *surv_end = &survivors[range - p];
-      register long i;
-      for (i = r; i; i--) { *surv = AND(*surv, *siv1++); surv++; }
+      ratpoints_bit_array *siv1 = &sieve_n[p-r];
+      ratpoints_bit_array *surv_end = &survivors[range - p];
+      long i;
+      for (i = r; i; i--) { *surv &= *siv1++; surv++; }
       siv1 -= p;
       while (surv <= surv_end)
       {
-        for (i = p; i; i--) { *surv = AND(*surv, *siv1++); surv++; }
+        for (i = p; i; i--) { *surv &= *siv1++; surv++; }
         siv1 -= p;
       }
       surv_end += p;
-      while (surv < surv_end) { *surv = AND(*surv, *siv1++); surv++; }
+      while (surv < surv_end) { *surv &= *siv1++; surv++; }
     }
-  } /* for n */
+  }
+  /* initialize pointers in sieve for the second phase */
+  for(n = sp1; n < sp2; n++)
+    sieves[n].start = sieves[n].ptr + mod(w_low + sieves[n].offset, sieves[n].p);
+#endif /* RATPOINTS_CHUNK */
+
   /* 2nd phase of the sieve: test each surviving bit array with more primes */
   surv0 = &survivors[0];
-  for (i = w_low; i < w_high; i++)
+  for (i = w_low; i < w_high; i++, base++)
   {
-    register ratpoints_bit_array nums = *surv0++;
+    ratpoints_bit_array nums = *surv0++;
     sieve_spec *ssp = &sieves[sp1];
-    register long n;
+    long n;
 
     for (n = sp2-sp1; n && TEST(nums); n--)
     {
-      register long p = ssp->p;
-      nums = AND(nums, ssp->ptr[mod(i + ssp->offset, p)]);
+      ratpoints_bit_array *ptr = (ssp->start) + base;
+      long p = ssp->p;
+      while(ptr >= ssp->end) ptr -= p;
+      nums &= *ptr;
       ssp++;
     }
 
@@ -540,10 +690,8 @@ _ratpoints_sift0(long b, long w_low, long w_high,
     if (TEST(nums))
     {
       long a0, a, d;
-#ifdef HAS_SSE2
-      long da = (which_bits == num_all) ? RBA_LENGTH/2 : RBA_LENGTH;
-#endif
-           /* a will be the numerator corresponding to the selected bit */
+      ulong nums0 = EXT0(nums);
+      /* a will be the numerator corresponding to the selected bit */
       if (which_bits == num_all)
       {
         d = 1; a0 = i * RBA_LENGTH;
@@ -553,43 +701,39 @@ _ratpoints_sift0(long b, long w_low, long w_high,
         d = 2; a0 = i * 2 * RBA_LENGTH;
         if (which_bits == num_odd) a0++;
       }
-      {
-#ifdef HAS_SSE2
-        ulong nums0 = EXT0(nums);
-        ulong nums1 = EXT1(nums);
-#else
-        ulong nums0 = nums;
-#endif
-
-        for (a = a0; nums0; a += d, nums0 >>= 1)
-        { /* test one bit */
-          if (odd(nums0) && ugcd(labs(a), absb)==1)
-          {
-            if (!args->bc) set_bc(b, args);
-            nb += _ratpoints_check_point(a, b, args, quit, process, info);
-            if (*quit) return nb;
-          }
+      for (a = a0; nums0; a += d, nums0 >>= 1)
+      { /* test one bit */
+        if (odd(nums0) && ugcd(labs(a), absb)==1)
+        {
+          if (!args->bc) set_bc(b, args);
+          nb += _ratpoints_check_point(a, b, args, quit, process, info);
+          if (*quit) return nb;
         }
-#ifdef HAS_SSE2
-        for (a = a0 + da; nums1; a += d, nums1 >>= 1)
-        { /* test one bit */
-          if (odd(nums1) && ugcd(labs(a), absb)==1)
-          {
-            if (!args->bc) set_bc(b, args);
-            nb += _ratpoints_check_point(a, b, args, quit, process, info);
-            if (*quit) return nb;
-          }
-        }
-#endif
       }
+#ifdef RBA_USE_VX
+      {
+        ulong k, da = d<<TWOPOTBITS_IN_LONG;
+        for (k = 1; k < RBA_PACK; k++)
+        {
+          ulong nums1 = EXT(nums,k);
+          a0 += da;
+          for (a = a0; nums1; a += d, nums1 >>= 1)
+          { /* test one bit */
+            if (odd(nums1) && ugcd(labs(a), absb)==1)
+            {
+              if (!args->bc) set_bc(b, args);
+              nb += _ratpoints_check_point(a, b, args, quit, process, info);
+              if (*quit) return nb;
+            }
+          }
+        }
+      }
+#endif
     }
   }
   return nb;
 }
 
-#define MAX_DIVISORS 512
- /* Maximal length of array for squarefree divisors of leading coefficient */
-
 typedef struct { double r; ratpoints_sieve_entry *ssp; } entry;
 
 static const int squares16[16] = {1,1,0,0,1,0,0,0,0,1,0,0,0,0,0,0};
@@ -626,10 +770,10 @@ find_points_init(ratpoints_args *args, long bit_primes)
   for (n = 1; n <= nbprime; n++)
   {
     ulong p = args->listprime[n];
-    need += p*p;
+    need += p*(p + RATPOINTS_CHUNK-1);
   }
   args->ba_buffer = (ratpoints_bit_array*)
-     stack_malloc_align(need*sizeof(ratpoints_bit_array),RBA_ALIGN);
+     stack_malloc_align(need*RBA_SIZE,RBA_SIZE);
   args->ba_next = args->ba_buffer;
 
   /* allocate space for int_buffer */
@@ -780,8 +924,11 @@ get_2adic_info(ratpoints_args *args, ulong *den_bits,
       if (np2 > 0)       db |= 0x1010UL; /* v_2(den) = 2 */
       if (np3 > 0)       db |= 0x0100UL; /* v_2(den) = 3 */
       if (np4 > 0)       db |= 0x0001UL; /* v_2(den) >= 4 */
-      if (db == 0) { *den_bits = 0UL; return num_none; }
-
+      if (db == 0)
+      {
+        for (i = 0 ; i < 16; i++) num_bits[i] = RBA(0UL);
+        *den_bits = 0UL; return num_none;
+      }
       for (i = 16; i < BITS_IN_LONG; i <<= 1) db |= db << i;
       *den_bits = db;
     }
@@ -1140,6 +1287,9 @@ sift(long b, ratpoints_bit_array *survivors, ratpoints_args *args,
         /* copy if already initialized, else initialize */
         ssp[n].ptr = sptr ? sptr : (p<BITS_IN_LONG?sieve_init1(p, se, bp, args)
                                                   :sieve_init2(p, se, bp, args));
+        ssp[n].start = ssp[n].ptr;
+        ssp[n].end = ssp[n].ptr + p;
+
       }
     }
 
@@ -1171,39 +1321,14 @@ sift(long b, ratpoints_bit_array *survivors, ratpoints_args *args,
         for (i = range; i; i--) survivors[i-1] = bits16;
         /* boundary words */
         if (w_low0 == w_low)
-        {
-          long sh = low - RBA_LENGTH * w_low;
-          ulong *survl = (ulong *)survivors;
-
-#ifdef HAS_SSE2
-          if (sh >= BITS_IN_LONG)
-          {
-            survl[0] = 0UL;
-            survl[1] &= (~0UL)<<(sh - BITS_IN_LONG);
-          }
-          else
-            survl[0] &= ~(0UL)<<sh;
-#else
-          survl[0] &= ~(0UL)<<sh;
-#endif
-        }
+          MASKL(survivors,low - RBA_LENGTH * w_low)
         if (w_high0 == w_high)
-        {
-          long sh = RBA_LENGTH * w_high - high;
-          ulong *survl = (ulong *)&survivors[range-1];
+          MASKU(&survivors[range-1], RBA_LENGTH * w_high - high)
 
-#ifdef HAS_SSE2
-          if (sh >= BITS_IN_LONG)
-          {
-            survl[0] &= ~(0UL)>>(sh - BITS_IN_LONG);
-            survl[1] = 0UL;
-          }
-          else
-            survl[1] &= ~(0UL)>>sh;
-#else
-          survl[0] &= ~(0UL)>>sh;
+#if (RATPOINTS_CHUNK > 1)
+        while(range%RATPOINTS_CHUNK != 0)
+          { survivors[range] = RBA(0); range++; w_high0++; }
 #endif
-        }
         nb += _ratpoints_sift0(b, w_low0, w_high0, args, which_bits,
                          survivors, &ssp[0], quit, process, info);
         if (*quit) return;
@@ -1259,19 +1384,21 @@ find_points_work(ratpoints_args *args,
 
   if (args->b_low < 1)  args->b_low = 1;
   if (args->b_high < 1) args->b_high = height;
-  if (args->b_high > height) args->b_high = height;
   if (args->max_forbidden < 0)
     args->max_forbidden = RATPOINTS_DEFAULT_MAX_FORBIDDEN;
   if (args->max_forbidden > nbprime)
     args->max_forbidden = nbprime;
   if (args->array_size <= 0) args->array_size = RATPOINTS_ARRAY_SIZE;
   {
-    long s = 2*CEIL(height, BITS_IN_LONG);
+    long s = 2*maxss(1,CEIL(height, BITS_IN_LONG));
     if (args->array_size > s) args->array_size = s;
   }
+  /* make sure that array size is a multiple of RATPOINTS_CHUNK */
+  args->array_size = CEIL(args->array_size, RATPOINTS_CHUNK)*RATPOINTS_CHUNK;
+
   /* Don't reverse if intervals are specified or limits for the denominator
      are given */
-  if (args->num_inter > 0 || args->b_low > 1 || args->b_high < height)
+  if (args->num_inter > 0 || args->b_low > 1 || args->b_high != height)
     args->flags |= RATPOINTS_NO_REVERSE;
 
   /* Check if reversal of polynomial might be better:
@@ -1374,7 +1501,7 @@ find_points_work(ratpoints_args *args,
   /* now do the sieving */
   {
     ratpoints_bit_array *survivors = (ratpoints_bit_array *)
-      stack_malloc_align((args->array_size)*sizeof(ratpoints_bit_array), RBA_ALIGN);
+      stack_malloc_align((args->array_size)*RBA_SIZE, RBA_SIZE);
     if (args->flags & (RATPOINTS_USE_SQUARES | RATPOINTS_USE_SQUARES1))
     {
       if (args->flags & RATPOINTS_USE_SQUARES)
@@ -1632,8 +1759,6 @@ ZX_hyperellratpoints(GEN P, GEN h, long flag)
   struct points data;
   ulong flags = 0;
 
-  if (!ZX_is_squarefree(P))
-    pari_err_DOMAIN("hyperellratpoints","issquarefree(pol)","=",gen_0, P);
   if (!args_h(&args, h))
   {
     pari_err_TYPE("hyperellratpoints", h);
@@ -1670,14 +1795,15 @@ QX_hyperellratpoints(GEN P, GEN h, long flag, GEN *den)
 }
 
 static GEN
-ZX_homogenous_evalpow(GEN Q, GEN A, GEN B)
+QX_homogenous_evalpow(GEN P, GEN A, GEN B)
 {
   pari_sp av = avma;
+  GEN den, Q = Q_remove_denom(P, &den);
   long i, d = degpol(Q);
   GEN s = gel(Q, d+2);
   for (i = d-1; i >= 0; i--)
     s = addii(mulii(s, A), mulii(gel(B,d+1-i), gel(Q,i+2)));
-  return gerepileuptoint(av, s);
+  return gerepileupto(av, den ? gdiv(s,den): s);
 }
 
 static GEN
@@ -1734,7 +1860,7 @@ hyperellratpoints(GEN PQ, GEN h, long flag)
   {
     GEN Li = gel(L,i), x = gel(Li,1), y = gel(Li,2), z = gel(Li,3);
     GEN B = gpowers(z, dQ);
-    GEN Qx = gdiv(ZX_homogenous_evalpow(Q, x, B), gel(B, dQ+1));
+    GEN Qx = gdiv(QX_homogenous_evalpow(Q, x, B), gel(B, dQ+1));
     GEN zdy = powiu(z, dy);
     if (den) zdy = mulii(zdy, den);
     gel(L,i) = mkvec2(gdiv(x,z), gmul2n(gsub(gdiv(y,zdy),Qx),-1));
