@@ -17,13 +17,13 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "../src/language/anal.h"
 #include "mt.h"
 
-static THREAD int pari_MPI_size, pari_MPI_rank;
-static THREAD long nbreq = 0, mt_issingle = 0;
+#define DEBUGLEVEL DEBUGLEVEL_mt
 
-enum PMPI_cmd { PMPI_close, PMPI_worker, PMPI_work, PMPI_parisizemax,
-                PMPI_parisize, PMPI_precreal, PMPI_primetab,
-                PMPI_varpriority, PMPI_eval,
-                PMPI_exportadd, PMPI_exportdel};
+static THREAD int pari_MPI_size, pari_MPI_rank;
+static THREAD long nbreq = 0;
+
+enum PMPI_cmd { PMPI_close, PMPI_worker, PMPI_work,
+                PMPI_eval, PMPI_exportadd, PMPI_exportdel};
 
 struct mt_mstate
 {
@@ -45,35 +45,76 @@ send_long(long a, long dest)
 }
 
 static void
-send_vlong(long *a, long n, long dest)
+send_bcast_long(long a, MPI_Comm comm)
 {
   BLOCK_SIGINT_START
-  MPI_Send(a, n, MPI_LONG, dest, 0, MPI_COMM_WORLD);
+  MPI_Bcast(&a, 1, MPI_LONG, 0, comm);
   BLOCK_SIGINT_END
 }
 
+static long
+recv_bcast_long(MPI_Comm comm)
+{
+  long a;
+  BLOCK_SIGINT_START
+  MPI_Bcast(&a, 1, MPI_LONG, 0, comm);
+  BLOCK_SIGINT_END
+  return a;
+}
+
 static void
 send_request(enum PMPI_cmd ecmd, long dest)
 {
   send_long((long)ecmd, dest);
 }
 
+static void
+send_request_all(enum PMPI_cmd ecmd, long n)
+{
+  long i;
+  for (i=1; i<=n; i++)
+    send_long((long)ecmd, i);
+}
+
 static void
 send_GEN(GEN elt, int dest)
 {
   pari_sp av = avma;
-  int size;
   GEN reloc = copybin_unlink(elt);
   GENbin *buf = copy_bin_canon(mkvec2(elt,reloc));
-  size = sizeof(GENbin) + buf->len*sizeof(ulong);
+  long size = sizeof(GENbin) + buf->len*sizeof(ulong);
   {
     BLOCK_SIGINT_START
-    MPI_Send(buf, size, MPI_CHAR, dest, 0, MPI_COMM_WORLD);
+    MPI_Send(buf, size/sizeof(long), MPI_LONG, dest, 0, MPI_COMM_WORLD);
     BLOCK_SIGINT_END
   }
   pari_free(buf); set_avma(av);
 }
 
+static void
+send_bcast_GEN(GEN elt, MPI_Comm comm)
+{
+  pari_sp av = avma;
+  GEN reloc = copybin_unlink(elt);
+  GENbin *buf = copy_bin_canon(mkvec2(elt,reloc));
+  long size = sizeof(GENbin) + buf->len*sizeof(ulong);
+  {
+    send_bcast_long(size, comm);
+    BLOCK_SIGINT_START
+    MPI_Bcast(buf, size/sizeof(long), MPI_LONG, 0, comm);
+    BLOCK_SIGINT_END
+  }
+  pari_free(buf); set_avma(av);
+}
+
+static void
+send_bcast_vlong(long *a, long n, MPI_Comm comm)
+{
+  BLOCK_SIGINT_START
+  MPI_Bcast(a, n, MPI_LONG, 0, comm);
+  BLOCK_SIGINT_END
+}
+
 static void
 send_request_GEN(enum PMPI_cmd ecmd, GEN elt, int dest)
 {
@@ -81,20 +122,6 @@ send_request_GEN(enum PMPI_cmd ecmd, GEN elt, int dest)
   send_GEN(elt, dest);
 }
 
-static void
-send_request_long(enum PMPI_cmd ecmd, long elt, int dest)
-{
-  send_request(ecmd, dest);
-  send_long(elt, dest);
-}
-
-static void
-send_request_vlong(enum PMPI_cmd ecmd, long *a, long n, int dest)
-{
-  send_request(ecmd, dest);
-  send_vlong(a, n, dest);
-}
-
 static long
 recvfrom_long(int src)
 {
@@ -105,14 +132,6 @@ recvfrom_long(int src)
   return a;
 }
 
-static void
-recvfrom_vlong(long *a, long n, int src)
-{
-  BLOCK_SIGINT_START
-  MPI_Recv(a, n, MPI_LONG, src, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
-  BLOCK_SIGINT_END
-}
-
 static enum PMPI_cmd
 recvfrom_request(int src)
 {
@@ -126,9 +145,9 @@ recvstatus_buf(int source, MPI_Status *status)
   GENbin *buf;
   BLOCK_SIGINT_START
 
-  MPI_Get_count(status, MPI_CHAR, &size);
-  buf = (GENbin *)pari_malloc(size);
-  MPI_Recv(buf, size, MPI_CHAR, source, 0/* tag */,
+  MPI_Get_count(status, MPI_LONG, &size);
+  buf = (GENbin *)pari_malloc(size*sizeof(long));
+  MPI_Recv(buf, size, MPI_LONG, source, 0/* tag */,
           MPI_COMM_WORLD, MPI_STATUS_IGNORE);
   BLOCK_SIGINT_END
   return buf;
@@ -162,6 +181,34 @@ recvfrom_GEN(int src)
   return recvstatus_GEN(src, &status);
 }
 
+static GEN
+recv_bcast_GEN(MPI_Comm comm)
+{
+  GEN res;
+  GENbin *buf;
+  long size;
+
+  size = recv_bcast_long(comm);
+  buf = (GENbin *)pari_malloc(size);
+
+  BLOCK_SIGINT_START
+  MPI_Bcast(buf, size/sizeof(long), MPI_LONG, 0, comm);
+  BLOCK_SIGINT_END
+
+  buf->rebase = &shiftaddress_canon;
+  res = bin_copy(buf);
+  bincopy_relink(gel(res,1),gel(res,2));
+  return gel(res,1);
+}
+
+static void
+recv_bcast_vlong(long *a, long n, MPI_Comm comm)
+{
+  BLOCK_SIGINT_START
+  MPI_Bcast(a, n, MPI_LONG, 0, comm);
+  BLOCK_SIGINT_END
+}
+
 static GEN
 recvany_GEN(int *source)
 {
@@ -204,52 +251,53 @@ pari_MPI_child(void)
     switch (recvfrom_request(0))
     {
     case PMPI_worker:
-      paristack_setsize(rsize, vsize);
-      gp_context_save(&rec);
-      worker = recvfrom_GEN(0);
-      av = avma;
-      break;
+      {
+        MPI_Comm comm;
+        long n = recv_bcast_long(MPI_COMM_WORLD), status = pari_MPI_rank <= n;
+        MPI_Comm_split(MPI_COMM_WORLD, status, pari_MPI_rank, &comm);
+        if (status==0)
+        {
+          MPI_Comm_free(&comm);
+          break;
+        }
+        rsize = recv_bcast_long(comm);
+        vsize = recv_bcast_long(comm);
+        precreal = recv_bcast_long(comm);
+        recv_bcast_vlong(varpriority-1,MAXVARN+2,comm);
+        {
+          pari_sp ltop = avma;
+          GEN tab = recv_bcast_GEN(comm);
+          if (!gequal(tab, primetab))
+          {
+            long i, l = lg(tab);
+            GEN old = primetab, t = cgetg_block(l, t_VEC);
+            for (i = 1; i < l; i++) gel(t,i) = gclone(gel(tab,i));
+            primetab = t;
+            gunclone_deep(old);
+          }
+          set_avma(ltop);
+        }
+        paristack_setsize(rsize, vsize);
+        worker = recv_bcast_GEN(comm);
+        MPI_Comm_free(&comm);
+        gp_context_save(&rec);
+        av = avma;
+        break;
+      }
     case PMPI_work:
       work = recvfrom_GEN(0);
       done = closure_callgenvec(worker, work);
       send_GEN(done, 0);
       set_avma(av);
       break;
-    case PMPI_parisizemax:
-      vsize = recvfrom_long(0);
-      break;
-    case PMPI_parisize:
-      rsize = recvfrom_long(0);
-      break;
-    case PMPI_precreal:
-      precreal = recvfrom_long(0);
-      break;
-    case PMPI_primetab:
-      {
-        pari_sp ltop = avma;
-        GEN tab = recvfrom_GEN(0);
-        if (!gequal(tab, primetab))
-        {
-          long i, l = lg(tab);
-          GEN old = primetab, t = cgetg_block(l, t_VEC);
-          for (i = 1; i < l; i++) gel(t,i) = gclone(gel(tab,i));
-          primetab = t;
-          gunclone_deep(old);
-        }
-        set_avma(ltop);
-      }
-      break;
     case PMPI_eval:
-      (void) closure_evalgen(recvfrom_GEN(0));
+      (void) closure_evalgen(recv_bcast_GEN(MPI_COMM_WORLD));
       set_avma(av);
       break;
-    case PMPI_varpriority:
-      recvfrom_vlong(varpriority-1,MAXVARN+2,0);
-      break;
     case PMPI_exportadd:
     {
-      GEN str = recvfrom_GEN(0);
-      GEN val = recvfrom_GEN(0);
+      GEN str = recv_bcast_GEN(MPI_COMM_WORLD);
+      GEN val = recv_bcast_GEN(MPI_COMM_WORLD);
       entree *ep = fetch_entry(GSTR(str));
       export_add(ep->name, val);
       set_avma(av);
@@ -257,7 +305,7 @@ pari_MPI_child(void)
     }
     case PMPI_exportdel:
     {
-      GEN str = recvfrom_GEN(0);
+      GEN str = recv_bcast_GEN(MPI_COMM_WORLD);
       entree *ep = fetch_entry(GSTR(str));
       export_del(ep->name);
       set_avma(av);
@@ -275,11 +323,12 @@ void
 mt_err_recover(long er)
 {
   if (pari_MPI_rank) longjmp(child_env,er);
-  else if (mt_issingle) mtsingle_err_recover(er);
+  else mtsingle_err_recover(er);
 }
 void mt_sigint_block(void) { }
 void mt_sigint_unblock(void) { }
 void mt_sigint(void) {}
+void mt_thread_init(void) { }
 
 int
 mt_is_parallel(void)
@@ -290,7 +339,7 @@ mt_is_parallel(void)
 int
 mt_is_thread(void)
 {
-  return pari_MPI_rank ? 1 : mt_issingle ? mtsingle_is_thread(): 0;
+  return pari_MPI_rank ? 1 : mtsingle_is_thread();
 }
 
 long
@@ -303,18 +352,15 @@ void
 mt_export_add(const char *str, GEN val)
 {
   pari_sp av = avma;
-  long i, n = pari_MPI_size-1;
+  long n = pari_MPI_size-1;
   GEN s;
   if (pari_mt || pari_MPI_rank)
     pari_err(e_MISC,"export not allowed during parallel sections");
   export_add(str, val);
   s = strtoGENstr(str);
-  for (i=1; i <= n; i++)
-  {
-    send_request(PMPI_exportadd, i);
-    send_GEN(s, i);
-    send_GEN(val, i);
-  }
+  send_request_all(PMPI_exportadd, n);
+  send_bcast_GEN(s, MPI_COMM_WORLD);
+  send_bcast_GEN(val, MPI_COMM_WORLD);
   set_avma(av);
 }
 
@@ -322,24 +368,23 @@ void
 mt_export_del(const char *str)
 {
   pari_sp av = avma;
-  long i, n = pari_MPI_size-1;
-  GEN s;
+  long n = pari_MPI_size-1;
   if (pari_MPI_rank)
     pari_err(e_MISC,"unexport not allowed during parallel sections");
   export_del(str);
-  s = strtoGENstr(str);
-  for (i=1; i <= n; i++)
-    send_request_GEN(PMPI_exportdel, s, i);
+  send_request_all(PMPI_exportdel, n);
+  send_bcast_GEN(strtoGENstr(str), MPI_COMM_WORLD);
   set_avma(av);
 }
 
 void
 mt_broadcast(GEN code)
 {
-  long i;
   if (!pari_MPI_rank && !pari_mt)
-    for (i=1;i<pari_MPI_size;i++)
-      send_request_GEN(PMPI_eval, code, i);
+  {
+    send_request_all(PMPI_eval, pari_MPI_size-1);
+    send_bcast_GEN(code, MPI_COMM_WORLD);
+  }
 }
 
 void
@@ -427,29 +472,29 @@ mt_queue_start_lim(struct pari_mt *pt, GEN worker, long lim)
 {
   if (lim==0) lim = pari_mt_nbthreads;
   else        lim = minss(pari_mt_nbthreads, lim);
-  if (pari_mt || pari_MPI_rank || pari_MPI_size <= 2 || lim <= 1)
-  {
-    mt_issingle = 1;
+  if (pari_MPI_rank)
+    mtsequential_queue_start(pt, worker);
+  else if (pari_mt || pari_MPI_size <= 2 || lim <= 1)
     mtsingle_queue_start(pt, worker);
-  }
   else
   {
+    MPI_Comm comm;
     struct mt_mstate *mt = &pari_mt_data;
-    long i, n = minss(lim, pari_MPI_size-1);
+    long n = minss(lim, pari_MPI_size-1);
     long mtparisize = GP_DATA->threadsize? GP_DATA->threadsize: pari_mainstack->rsize;
     long mtparisizemax = GP_DATA->threadsizemax;
     pari_mt = mt;
-    mt_issingle = 0;
     mt->workid = (long*) pari_malloc(sizeof(long)*(n+1));
-    for (i=1; i <= n; i++)
-    {
-      send_request_long(PMPI_parisize, mtparisize, i);
-      send_request_long(PMPI_parisizemax, mtparisizemax, i);
-      send_request_long(PMPI_precreal, get_localbitprec(), i);
-      send_request_vlong(PMPI_varpriority,varpriority-1,MAXVARN+2, i);
-      send_request_GEN(PMPI_primetab, primetab, i);
-      send_request_GEN(PMPI_worker, worker, i);
-    }
+    send_request_all(PMPI_worker, pari_MPI_size-1);
+    send_bcast_long(n, MPI_COMM_WORLD);
+    MPI_Comm_split(MPI_COMM_WORLD, 1, 0, &comm);
+    send_bcast_long(mtparisize, comm);
+    send_bcast_long(mtparisizemax, comm);
+    send_bcast_long(get_localbitprec(), comm);
+    send_bcast_vlong(varpriority-1,MAXVARN+2, comm);
+    send_bcast_GEN(primetab, comm);
+    send_bcast_GEN(worker, comm);
+    MPI_Comm_free(&comm);
     mt->n = n;
     mt->nbint = 1;
     mt->source = 1;
